#ifndef CGUL_BDF_GLYPH_H
#define CGUL_BDF_GLYPH_H

/**
 * Each instance of this class is one glyph from a BDF font file.  It
 * is an integral part of the <code>cgul_bdf</code> class.
 *
 * @file
 * @author Paul Serice
 * @brief glyph for BDF fonts
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_libc.h"
#include "cgul_string.h"
#include "cgul_wchar.h"

CGUL_BEGIN_C

/** Opaque reference to a <code>cgul_bdf_glyph</code> instance. */
typedef struct cgul_bdf_glyph* cgul_bdf_glyph_t;

/**
 * This method creates a new <code>cgul_bdf_glyph</code> object.  The
 * caller is responsible for calling
 * <code>cgul_bdf_glyph__delete()</code> on the pointer returned.  If
 * memory cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 */
extern CGUL_EXPORT cgul_bdf_glyph_t
cgul_bdf_glyph__new(cgul_exception_t* cex);

/**
 * This method deletes the <code>glyph</code> object freeing all
 * internally allocated resources.  After calling this method,
 * <code>glyph</code> should no longer be dereferenced.
 *
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__delete(cgul_bdf_glyph_t glyph);

/**
 * This method does a deep copy of <code>src</code> to
 * <code>dst</code>.  If an error occurs, an exception is thrown, and
 * <code>dst</code> will be in an undefined state.
 *
 * @param[in,out] cex c-style exception
 * @param[in] dst destination glyph
 * @param[in] src source glyph
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__copy(cgul_exception_t* cex,
                     cgul_bdf_glyph_t dst,
                     cgul_bdf_glyph_t src);

/**
 * This method returns the descriptive name for the character.  This
 * is the name that belongs on the STARTCHAR line.  If no descriptive
 * name is available, <code>NULL</code> is returned.  Thus, it is
 * important that you always check the return value.  The pointer
 * returned belongs to this class and should not be altered or freed.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return descriptive name or <code>NULL</code>
 */
extern CGUL_EXPORT const char*
cgul_bdf_glyph__get_descriptive_name(cgul_exception_t* cex,
                                     cgul_bdf_glyph_t glyph);

/**
 * Get the encoding.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return encoding
 */
extern CGUL_EXPORT cgul_wchar_t
cgul_bdf_glyph__get_encoding(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph);

/**
 * Get the ascent for the glyph.  This is a value derived from the
 * glyph's bounding box.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return ascent of glyph
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_ascent(cgul_exception_t* cex,
                           cgul_bdf_glyph_t glyph);

/**
 * Get the descent for the glyph.  This is a value derived from the
 * glyph's bounding box.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return descent of glyph
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_descent(cgul_exception_t* cex,
                            cgul_bdf_glyph_t glyph);

/**
 * Get the scalable width in the x-direction.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return scalable width in the x-direction
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_swidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph);

/**
 * Get the scalable width in the y-direction.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return scalable width in the y-direction
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_swidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph);

/**
 * Get the device width in the x-direction.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return device width in the x-direction
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_dwidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph);

/**
 * Get the device width in the y-direction.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return device width in the y-direction
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_dwidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph);

/**
 * Get the width of the glyph's bounding box.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return width of the glyph's bounding box
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_bb_width(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph);

/**
 * Get the height of the glyph's bounding box.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return height of the glyph's bounding box
 */
extern CGUL_EXPORT unsigned int
cgul_bdf_glyph__get_bb_height(cgul_exception_t* cex,
                              cgul_bdf_glyph_t glyph);

/**
 * Get the x-offset of the lower-left corner of the glyph's bounding box.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return x-offset of the lower-left corner of the glyph's bounding box
 */
extern CGUL_EXPORT int
cgul_bdf_glyph__get_bb_x_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph);

/**
 * Get the y-offset of the lower-left corner of the glyph's bounding box.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return y-offset of the lower-left corner of the glyph's bounding box
 */
extern CGUL_EXPORT int
cgul_bdf_glyph__get_bb_y_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph);

/**
 * This method returns <code>true</code> if the dot at
 * <code>(x,y)</code> inside the bounding box should be drawn.  The x-
 * and y-coordinates are relative to the top-left corner of the
 * glyph's bounding box, not the top-left corner of the glyph itself.
 * Also note that <code>x</code> and <code>y</code> values start with
 * zero, not one. <p>
 *
 * NOTE: There is a <code>cgul_bdf_glyph__set_dot()</code> method in
 * the private header. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] x x-coordinate
 * @param[in] y y-coordinate
 * @return whether the dot should be drawn
 */
extern CGUL_EXPORT int
cgul_bdf_glyph__get_dot(cgul_exception_t* cex,
                        cgul_bdf_glyph_t glyph,
                        unsigned int x,
                        unsigned int y);

/**
 * Get the width of the glyph in bytes.  Remember that the data is
 * a bitmap so it is packed with 8 pixels per byte.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return width of the glyph in bytes
 */
extern CGUL_EXPORT size_t
cgul_bdf_glyph__get_bytes_wide(cgul_exception_t* cex,
                               cgul_bdf_glyph_t glyph);

/**
 * Get the file offset of the STARTCHAR line that describes this glyph
 * in the BDF.  This information is used to generate the *.bdfi index
 * file so that the glyph can be loaded on demand.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @return width of the glyph in bytes
 */
extern CGUL_EXPORT unsigned long
cgul_bdf_glyph__get_file_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph);

/**
 * This method appends the bitmap of <code>glyph</code> as a character
 * sequence of hex digits to the string <code>s</code>.  The appended
 * character sequence can be used as the data in the BITMAP section of
 * a BDF file.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] s string
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__append_bitmap_to_string(cgul_exception_t* cex,
                                        cgul_bdf_glyph_t glyph,
                                        cgul_string_t s);

/**
 * Write the string representation of <code>glyph</code> to
 * <code>f</code>.  This is different from
 * <code>cgul_bdf_glyph__fdraw()</code> in that it produces a valid
 * snippet of BDF that can be used as the entire <code>STARTCHAR
 * ... ENDCHAR</code> section.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] f file
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__write_file(cgul_exception_t* cex,
                           cgul_bdf_glyph_t glyph,
                           FILE* f);

/**
 * This method appends an ASCII image of the glyph to the file
 * <code>f</code>.  The image only shows the part of the glyph that is
 * in the bounding box.  Each active pixel is represented by the
 * character <code>c_active</code>.  Each inactive pixel is
 * represented by <code>c_inactive</code>.  If <code>prefix</code> is
 * not <code>NULL</code>, it will be prefixed to each scan line.  If
 * <code>suffix</code> is not <code>NULL</code>, it will be appended
 * to each scan line.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] f file
 * @param[in] prefix prefix for each scan line
 * @param[in] suffix suffix for each scan line
 * @param[in] c_active active character
 * @param[in] c_inactive inactive character
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__fdraw(cgul_exception_t* cex,
                      cgul_bdf_glyph_t glyph,
                      FILE* f,
                      const char* prefix,
                      const char* suffix,
                      char c_active,
                      char c_inactive);

/**
 * This method appends an ASCII image of the glyph to the string
 * <code>s</code>.  The image only shows the part of the glyph that is
 * in the bounding box.  Each active pixel is represented by the
 * character <code>c_active</code>.  Each inactive pixel is
 * represented by <code>c_inactive</code>.  If <code>prefix</code> is
 * not <code>NULL</code>, it will be prefixed to each scan line.  If
 * <code>suffix</code> is not <code>NULL</code>, it will be appended
 * to each scan line.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in,out] s string
 * @param[in] prefix prefix for each scan line
 * @param[in] suffix suffix for each scan line
 * @param[in] c_active active character
 * @param[in] c_inactive inactive character
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__sdraw(cgul_exception_t* cex,
                      cgul_bdf_glyph_t glyph,
                      cgul_string_t s,
                      const char* prefix,
                      const char* suffix,
                      char c_active,
                      char c_inactive);

CGUL_END_C

#endif  /*  CGUL_BDF_GLYPH_H  */
