#ifndef CGUL_BLOCK_READER_FILE_CXX_H
#define CGUL_BLOCK_READER_FILE_CXX_H

/**
 * C++ wrapper for <code>cgul_block_reader_file</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_block_reader_file</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_block_reader_cxx.h"
#include "cgul_block_reader_file.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_block_reader_file</code> objects.  The main purpose of
 * this class is to convert the C-style function calls and exception
 * handling in <code>cgul_block_reader_file</code> into C++-style
 * function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_block_reader_file</code>
 */
class cgul_block_reader_file_cxx : public cgul_block_reader_cxx {

public:

    /**
     * Create a new <code>cgul_block_reader_file</code> instance.
     * After this method returns, <code>open_fname()</code> or
     * <code>open_file()</code> should be called before calling
     * <code>read_next_block()</code>.  The client is responsible for
     * deleting the object returned.  If an error occurs, an exception
     * is thrown.
     */
    cgul_block_reader_file_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_block_reader_file__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * This method creates a new <code>cgul_block_reader_file</code>
     * instance and calls <code>open_fname()</code> passing it
     * <code>fname</code>.  The file will be closed when this instance
     * is deleted.  The client is responsible for deleting the object
     * returned.  If an error occurs, an exception is thrown.
     *
     * @param[in] fname file name
     */
    cgul_block_reader_file_cxx(const char* fname)
    {
        cgul_exception_t local = NULL;
        obj = cgul_block_reader_file__new_from_fname(&local, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * This method creates a new <code>cgul_block_reader_file</code>
     * instance and calls <code>open_file()</code> passing it
     * <code>f</code>.  The class does not take ownership of
     * <code>f</code>.  Thus, the client is still responsible for
     * calling <code>fclose()</code> on it.  The client is also
     * responsible for deleting the object returned.  If an error
     * occurs, an exception is thrown.
     *
     * @param[in] f file
     */
    cgul_block_reader_file_cxx(FILE* f)
    {
        cgul_exception_t local = NULL;
        obj = cgul_block_reader_file__new_from_file(&local, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * This virtual method deletes the instance freeing all internally
     * allocated resources.  This does not include closing the
     * underlying file if it was obtained from the client by a call to
     * <code>open_file()</code>.
     */
    virtual
    ~cgul_block_reader_file_cxx()
    {
        if (obj) {
            cgul_block_reader_file__delete(obj);
            obj = NULL;
        }
    }
    
    /**
     * Open the file with name <code>fname</code> and use it when reading
     * blocks of data.  If a data source is already open, it will be
     * closed before this method attempts to open the new data source.
     * The new file will be closed when this instance is deleted.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] fname file name
     */
    virtual void
    open_fname(const char* fname)
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_file__open_fname(&local, obj, fname);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Use the file <code>f</code> when reading blocks of data.  If a data
     * source is already open, it is closed before attempting to open the
     * new data source.  This class does not take ownership of
     * <code>f</code>.  Thus, the client is still responsible for calling
     * <code>fclose()</code> on <code>f</code>.  If an error occurs, an
     * exception is thrown.
     *
     * @param[in] f file
     */
    virtual void
    open_file(FILE* f)
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_file__open_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Close the file.  After this method returns,
     * <code>open_fname()</code> or <code>open_file()</code> should be
     * called before calling any other method.
     */
    virtual void
    close()
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_file__close(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Virtual method used to read one block of data <code>block</code> of
     * size <code>block_size</code> bytes from the underlying file.  The
     * user is responsible for making sure <code>block</code> points to at
     * least <code>block_size</code> contiguous bytes of memory.  This
     * method returns the number of bytes read or <code>0</code> if EOF is
     * reached without error.  If an error occurs, <code>0</code> is
     * returned, and an exception is thrown.
     *
     * @param[in] block block
     * @param[in] block_size block size in bytes
     * @return number of bytes read
     */
    virtual size_t
    read_next_block(void* block,
                    size_t block_size)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_file__read_next_block(&local,
                                                     obj,
                                                     block,
                                                     block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * Return the name of underlying file.  If
     * <code>open_file()</code> was used to initialize the object,
     * <code>NULL</code> is returned instead.  This method throws an
     * exception if no file is currently open.
     *
     * @return name of underlying file or <code>NULL</code>
     */
    virtual const char*
    get_fname() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_file__get_fname(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * Virtual method that returns the presentation name.
     *
     * @return presentation name
     */
    virtual const char*
    get_presentation_name() const
    {
        const char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_file__get_presentation_name(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * Virtual method that sets the presentation name.  The presentation
     * name is primarily used for error reporting and can be set to
     * anything, but it does need to be set again if the input file
     * changes.  If the presentation name is not set, <code>"FILE"</code>
     * is returned.  An exception is thrown if an attempt is made to set
     * the presentation name without first setting the input file.
     *
     * @param[in] presentation_name presentation name
     */
    virtual void
    set_presentation_name(const char* presentation_name)
    {
        cgul_exception_t local = NULL;
        cgul_block_reader_file__set_presentation_name(&local,
                                                      obj,
                                                      presentation_name);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
    
    /**
     * Return the currently opened file.  The client must not do anything
     * that would invalidate the pointer or corrupt the associated data
     * stream.  This method throws an exception if no file is currently
     * open.
     *
     * @return currently opened file or memory
     */
    virtual FILE*
    get_file() const
    {
        FILE* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_file__get_file(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the underlying <code>cgul_block_reader</code>
     * object. <p>
     *
     * @note <code>get_obj()</code> is similar to this method except
     * it returns a pointer to the most derived class; whereas, this
     * method returns the same pointer but cast to the base class. <p>
     *
     * @return underlying <code>cgul_block_reader</code> object
     */
    virtual cgul_block_reader_t
    as_block_reader() const
    {
        cgul_block_reader_t rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_block_reader_file__as_block_reader(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Get the underlying <code>cgul_block_reader_file</code> object.
     *
     * @return underlying object
     */
    virtual cgul_block_reader_file_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_block_reader_file</code> object.
     * This means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_block_reader_file__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_block_reader_file_t
    take_obj()
    {
        cgul_block_reader_file_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_block_reader_file_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_block_reader_file__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_block_reader_file_cxx(const cgul_block_reader_file_cxx&);
    cgul_block_reader_file_cxx& operator=(const cgul_block_reader_file_cxx&);

    /* Underlying <code>cgul_block_reader_file</code> object. */
    cgul_block_reader_file_t obj;

};

#endif  /*  CGUL_BLOCK_READER_FILE_CXX_H  */
