#ifndef CGUL_CACHE_H
#define CGUL_CACHE_H

/**
 * This class allows you to cache objects so that they can be reused
 * without have to reallocate and free the memory.  This improves
 * performance because malloc() and free() are relatively expensive.
 * It should also help reduce memory fragmentation. <p>
 *
 * This class can also be used to reserve objects for future use.
 * This can be used to allow synchronous error reporting in classes
 * like <code>cgul_event_router</code> which keep a pending event
 * queue.  Without reserving an object when
 * <code>cgul_event_router__add_listener()</code> has to put an event
 * on the pending event queue, that object would have to be allocated
 * later which could cause an asynchronous out-of-memory exception,
 * but by allocating and reserving that object immediately, the client
 * gets synchronous notification of out-of-memory exceptions. <p>
 *
 * Lastly, this class can be used as an abstraction layer for creating
 * objects.  For example, the constructors for
 * <code>cgul_rbtree</code> set up a <code>cgul_cache</code> to return
 * different nodes depending on whether the user wants to enable the
 * indexing feature of <code>cgul_rbtree</code>.  This allows the user
 * to save two <code>unsigned int</code> values per node which can
 * lead to a substantial savings for a large tree.  Most of the rest
 * of the code in <code>cgul_rbtree</code> is written without having
 * to know the exact type of node that is being used and simply calls
 * <code>cgul_cache__get()</code> and <code>cgul_cache__put()</code>
 * when appropriate. <p>
 *
 * This class is used internally by <code>cgul_list</code> and
 * <code>cgul_rbtree</code> to cache their nodes.  You can use these
 * classes to test this class by, for example, creating a
 * <code>cgul_list</code> and playing with
 * <code>cgul_list_set_cache_size()</code>.
 *
 * @file
 * @author Paul Serice
 * @brief cache objects so they can be reused without thrashing malloc()
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_cache</code> instance. */
typedef struct cgul_cache* cgul_cache_t;

/*
 * The next three typedef statements define the "cacheable" interface.
 * For any class of objects the client wants to cache, the "cacheable"
 * interface must be implemented.
 */

/**
 * This function pointer typedef is needed so that the
 * <code>cgul_cache</code> knows how to create new default instances
 * of the objects you want to cache.
 */
typedef void* (*cgul_cache__constructor_t)(cgul_exception_t* cex);

/**
 * This function pointer typedef is needed so that the
 * <code>cgul_cache</code> knows how to delete instances of the
 * objects you want to cache.  The client's implementation of this
 * function should not throw an exception.
 */
typedef void (*cgul_cache__destructor_t)(void* obj);

/**
 * This function pointer typedef is needed so that the
 * <code>cgul_cache</code> knows how take a fully constructed object
 * and return it to its initial default state.  This callback is
 * invoked when objects are put back on the cache to minimize the
 * memory footprint of objects on the cache.  If the restructor is
 * set to <code>NULL</code>, the client is responsible for resetting
 * the object before putting it back on the cgul_cache. */
typedef void (*cgul_cache__restructor_t)(void* obj);

/**
 * Static method used to return the <code>cgul_cache</code> instance
 * that is an adapter for <code>free()</code>.  Because the object
 * returned is statically allocated, the client should not attempt to
 * call <code>cgul_cache__delete()</code> on the pointer returned.  If
 * the client forgets, <code>cgul_cache__delete()</code> will ignore
 * the request. <p>
 *
 * The purpose of this static method is to allow clients to use
 * methods (like <code>cgul_rbtree__remove_range()</code>) that accept
 * <code>cgul_cache</code> objects for cleaning up keys or values even
 * though the objects are allocated directly using one of the
 * <code>malloc()</code> family of function calls (instead of through
 * the <code>cgul_cache</code> interface). <p>
 *
 * The <code>cgul_cache</code> object returned only implements the
 * <code>cgul_cache__destructor_t</code> part of the
 * <code>cgul_cache</code> interface.  Thus, any attempt to use the
 * object returned to construct a new object will result in
 * <code>NULL</code> being returned. <p>
 *
 * Because the <code>cgul_cache</code> object returned is only used in
 * the destruction context, the conventional
 * <code>cgul_exception</code> object required of most methods is not
 * required of this method. <p>
 *
 * @return <code>cgul_cache</code> adapter for <code>free()</code>
 */
extern CGUL_EXPORT cgul_cache_t
cgul_cache__get_freer();

/**
 * Create a new <code>cgul_cache</code> instance.  If
 * <code>restructor</code> is set to <code>NULL</code>, the client is
 * responsible for resetting the object before putting it back on the
 * <code>cgul_cache</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned and an exception is thrown. <p>
 *
 * By default, the maximum cache size is set to 0.  This means no
 * objects are cached by default.  When you want to enable caching,
 * you can change the maximum cache size by calling
 * <code>cgul_cache__set_size()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] constructor constructor for the cached objects
 * @param[in] destructor destructor for the cached objects
 * @param[in] restructor restructor for the cached objects
 * @return new <code>cgul_cache</code> instance
 */
extern CGUL_EXPORT cgul_cache_t
cgul_cache__new(cgul_exception_t* cex,
                cgul_cache__constructor_t constructor,
                cgul_cache__destructor_t destructor,
                cgul_cache__restructor_t restructor);

/**
 * This method first calls the destructor for each object that is
 * still cached.  It then frees all of its internally allocated
 * resources.  The client should not use the <code>cache</code>
 * instance after calling this method.
 *
 * @param[in] cache <code>cgul_cache</code> instance
 */
extern CGUL_EXPORT void
cgul_cache__delete(cgul_cache_t cache);

/**
 * Get the maximum number of objects the cache can hold when the
 * reserve is zero.  To get the total number of elements the cache can
 * hold, you must add the reserve to the size.
 *
 * @param[in] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @return maximum size of the cache
 */
extern CGUL_EXPORT unsigned long int
cgul_cache__get_size(cgul_exception_t* cex,
                     cgul_cache_t cache);

/**
 * Set the maximum number of objects the cache can hold when the
 * reserve is zero. <p>
 *
 * Calling this method with <code>size</code> less than the reserve
 * does not free any of the reserved objects; however, it may free
 * objects that have recently been "moved" from the reserve to the
 * main cache.  (See <code>cgul_cache__set_reserve()</code>.) <p>
 *
 * You can release all the cached objects and disable caching of
 * objects by setting the cache reserve to 0 and the cache size to 0
 * (in that order).  This is the default. <p>
 *
 * If an error occurs (while enlarging the cache), an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @param[in] size maximum size of the cache
 */
extern CGUL_EXPORT void
cgul_cache__set_size(cgul_exception_t* cex,
                     cgul_cache_t cache,
                     unsigned long int size);

/**
 * Get the reserve limit.  This is the number of objects that are
 * guaranteed to be held in reserve by the cache.  These objects will
 * not be returned by the cache until the reserve is reduced.  The
 * reserve is independent of the size of the cache.
 *
 * @param[in] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @return reserve
 */
extern CGUL_EXPORT unsigned long int
cgul_cache__get_reserve(cgul_exception_t* cex,
                        cgul_cache_t cache);

/**
 * Set the reserve limit.  This method causes <code>cache</code> to
 * reserve objects.  The reserved objects are held indefinitely until
 * the reserve is reduced.  The reserve is independent of the size of
 * the cache. <p>
 *
 * It is important to understand that if these reserved objects are
 * not already being held by the cache, the cache will instantiate
 * them.  This allows the client to immediately instantiate objects
 * without having to worry about out-of-memory exceptions when the
 * objects are used later. <p>
 *
 * Note that if you want to get a pointer to one of the reserved
 * object, you first call <code>cgul_cache__set_reserve()</code> to
 * reduce the number of the reserved objects.  This simply "moves" the
 * now unreserved objects to the main cache of objects.  You then
 * immediately call <code>cgul_cache__get()</code> to get an object
 * from the main cache. <p>
 *
 * It is done like this so that you can code strictly in terms of
 * <code>cgul_cache__get()</code>.  You can then reserve an object
 * now and trigger when it is used later just by calling this method.
 * The code that uses <code>cgul_cache__get()</code> does not need to
 * know anything about reserved objects.  It will just naturally use
 * them. <p>
 *
 * Objects that are "moved" from the reserve to the main cache are not
 * guaranteed to be available if <code>cgul_cache__set_size()</code>
 * has since been called.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @param[in] reserve reserve
 */
extern CGUL_EXPORT void
cgul_cache__set_reserve(cgul_exception_t* cex,
                        cgul_cache_t cache,
                        unsigned long int reserve);

/**
 * Try to return a re-initialized object from the cache; if the cache
 * is empty, a newly instantiated object will be returned instead.  If
 * an error occurs, <code>NULL</code> is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @return cached or new object
 */
extern CGUL_EXPORT void*
cgul_cache__get(cgul_exception_t* cex,
                cgul_cache_t cache);

/**
 * Try to put <code>obj</code> on the cache.  If the cache is full,
 * <code>obj</code> will be delete instead.
 *
 * @param[in] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @param[in] obj object to cache or delete
 */
extern CGUL_EXPORT void
cgul_cache__put(cgul_exception_t* cex,
                cgul_cache_t cache,
                void* obj);

CGUL_END_C

#endif  /*  CGUL_CACHE_H  */
