#ifndef CGUL_CRC16_H
#define CGUL_CRC16_H

/**
 * The <code>cgul_crc16</code> class is designed to handle any of the
 * different common permutations on the CRC-16 algorithm.  This class
 * is parameterized to allow different generator polynomials, initial
 * values of the shift register, final XOR values, whether to append
 * the size, whether to XOR the size before appending, and whether the
 * data stream is reflected.
 *
 * This class was first written in java, and the naming convention for
 * methods in this class follows what is established by
 * java.security.MessageDigest instead of what is established by
 * java.util.zip.CRC32.
 *
 * A simple example of how to use this class follows:
 *
 * <pre>
 *
 *     cgul_exception_t* cex = NULL;
 *     cgul_crc16_t crc =
 *         cgul_crc16__new(cex, 0x8408, 0, 0, 1);
 *     unsigned short result =
 *         cgul_crc16__crc_block(cex, crc, "123456789", 9);
 *     cgul_libc_printf(&cex, "%#06hx\n", result);
 *
 * </pre>
 *
 * The above code will calculate the CRC16-CCITT CRC for the ASCII
 * text string "123456789".  The CRC that is printed to the screen
 * will be 0x2189. <p>
 *
 * The following is a list of 16-bit CRCs and how to call
 * <code>cgul_crc16__new()</code> to create a new
 * <code>cgul_crc16</code> objects for each CRC: <p>
 *
 * <pre>
 *
 *  // CRC-16 (aka CRC-IBM, CRC-16/ARC)
 *  g = 0xa001;
 *  init = 0;
 *  final_xor = 0;
 *  reflected = 1;
 *
 *  // CRC-CCITT (aka kermit)
 *  g = 0x8408;
 *  init = 0;
 *  final_xor = 0;
 *  reflected = 1;
 *
 *  // X.25
 *  g = 0x8408;
 *  init = 0xffff;
 *  final_xor = 0xffff;
 *  reflected = 1;
 *
 *  // CRC-16/RFID
 *  g = 0x1021;
 *  init = 0xffff;
 *  final_xor = 0xffff;
 *  reflected = 0;
 *
 *  // CRC-16/USB
 *  g = 0xa001;
 *  init = 0xffff;
 *  final_xor = 0xffff;
 *  reflected = 1;
 *
 *  // xmodem (aka zmodem)
 *  g = 0x1021;
 *  init = 0;
 *  final_xor = 0;
 *  reflected = 0;
 *
 * </pre>
 *
 * @note If you use this class from multiple threads, you MUST provide
 * external synchronization to all methods so that you do not corrupt
 * internal data structures. <p>
 *
 * @file
 * @author Paul Serice
 * @brief CRC-16
 * @see cgul_crc16_incr
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to <code>cgul_crc16</code> instance. */
typedef struct cgul_crc16* cgul_crc16_t;

/**
 * Create a new <code>cgul_crc16</code> object.  Objects created with
 * this constructor do not append the size to the data stream that is
 * CRCed.  The caller is responsible for freeing the object by calling
 * <code>cgul_crc16__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] g generator polynomial
 * @param[in] init initial value loaded into the accumulator
 * @param[in] final_xor value to xor with the calculated crc
 * @param[in] reflected whether to use the reflected algorithm
 * @return new <code>cgul_crc16</code> instance
 */
extern CGUL_EXPORT cgul_crc16_t
cgul_crc16__new(cgul_exception_t* cex,
                unsigned short g,
                unsigned short init,
                unsigned short final_xor,
                int reflected);

/**
 * Create a new <code>cgul_crc16</code> object.  Objects created with
 * this constructor append the size to the data stream that is CRCed.
 * The caller is responsible for freeing the object by calling
 * cgul_crc16__delete().  If memory cannot be allocated,
 * <code>NULL</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] g polynomial
 * @param[in] init initial value loaded into the accumulator
 * @param[in] final_xor value to xor with the calculated crc
 * @param[in] reflected whether to use the reflected algorithm
 * @param[in] size_xor value to xor with the size before appending
 * @return new <code>cgul_crc16</code> instance
 */
extern CGUL_EXPORT cgul_crc16_t
cgul_crc16__new_appends_size(cgul_exception_t* cex,
                             unsigned short g,
                             unsigned short init,
                             unsigned short final_xor,
                             int reflected,
                             size_t size_xor);

/**
 * This method frees all internally allocated memory.  Do not try to
 * access <code>crc</code> after calling this method.
 *
 * @param[in] crc <code>cgul_crc16</code> instance
 */
extern CGUL_EXPORT void
cgul_crc16__delete(cgul_crc16_t crc);

/**
 * Assign the current accumulator and size of <code>rhs</code> to
 * <code>lhs</code>.  This operation probably only makes sense if
 * <code>lhs</code> and <code>rhs</code> have been constructed
 * identically.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_crc16__assign(cgul_exception_t* cex,
                   cgul_crc16_t lhs,
                   cgul_crc16_t rhs);

/**
 * This method completes the CRC calculation by performing the size
 * XOR and the final XOR.  You will generally call one of the update()
 * methods before calling this method.  When this method returns, the
 * internal state of the object is reset so that it is ready to
 * calculate the next CRC.  The return value is the final value for
 * the CRC.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @return final crc value
 */
extern CGUL_EXPORT unsigned short
cgul_crc16__crc(cgul_exception_t* cex,
                cgul_crc16_t crc);

/**
 * This is the method to call when you only have one block for which
 * you want to calculate a CRC.  You can also call this if you have
 * called one of the update() methods before, and you know that
 * <code>block</code> is your last block.  When this method returns,
 * the internal state of the object is reset so that it is ready to
 * calculate the next CRC.  <code>length</code> is the number of bytes
 * inside of <code>block</code> that you want to add to the CRC
 * calculation.  The return value is the final value for the CRC.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] length length of <code>block</code>
 * @return final crc value
 */
extern CGUL_EXPORT unsigned short
cgul_crc16__crc_block(cgul_exception_t* cex,
                      cgul_crc16_t crc,
                      const void* block,
                      size_t length);

/**
 * This is the method to call when you want to calculate the CRC from
 * a range of bytes from one block.  You can also call this if you
 * have called one of the update() methods before, and you know that
 * <code>block</code> is your last block.  When this method returns,
 * the internal state of the object is reset so that it is ready to
 * calculate the next CRC.  <code>offset</code> is the index into
 * <code>block</code> where you want to add <code>length</code> bytes
 * to the CRC calculation.  The return value is the final value for
 * the CRC.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] offset offset into <code>block</code>
 * @param[in] length length of <code>block</code>
 * @return final crc value
 */
extern CGUL_EXPORT unsigned short
cgul_crc16__crc_range_of_block(cgul_exception_t* cex,
                               cgul_crc16_t crc,
                               const void* block,
                               size_t offset,
                               size_t length);

/**
 * This is the method to call when you only have one file for which
 * you want to calculate a CRC.  You can also call this if you have
 * called one of the update() methods before (like
 * <code>cgul_crc16__update_file()</code>), and you know that
 * <code>f</code> is your last file.  When this method returns, the
 * internal state of the object is reset so that it is ready to
 * calculate the next CRC.  The return value is the final value for
 * the CRC.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] f input file
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc16__crc_file(cgul_exception_t* cex,
                     cgul_crc16_t crc,
                     FILE* f);

/**
 * This is the method to call when you only have one file for which
 * you want to calculate a CRC.  You can also call this if you have
 * called one of the update() methods before (like
 * <code>cgul_crc16__update_fname()</code>), and you know that
 * <code>fname</code> is your last file.  When this method returns,
 * the internal state of the object is reset so that it is ready to
 * calculate the next CRC.  The return value is the final value for
 * the CRC.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] fname input file name
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc16__crc_fname(cgul_exception_t* cex,
                      cgul_crc16_t crc,
                      const char* fname);

/**
 * This method updates the CRC by adding the byte <code>c</code> to
 * the CRC calculation.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] c byte to add to crc
 */
extern CGUL_EXPORT void
cgul_crc16__update_byte(cgul_exception_t* cex,
                        cgul_crc16_t crc,
                        unsigned char c);

/**
 * This method updates the CRC by adding all the bytes in
 * <code>block</code> to the calculation.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] length length of <code>block</code>
 */
extern CGUL_EXPORT void
cgul_crc16__update_block(cgul_exception_t* cex,
                         cgul_crc16_t crc,
                         const void* block,
                         size_t length);

/**
 * This method updates the CRC by adding a range of bytes in
 * <code>block</code> to the calculation.  The range starts at
 * <code>offset</code> and extends <code>length</code> bytes.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] offset offset into <code>block</code>
 * @param[in] length length of <code>block</code>
 */
extern CGUL_EXPORT void
cgul_crc16__update_range_of_block(cgul_exception_t* cex,
                                  cgul_crc16_t crc,
                                  const void* block,
                                  size_t offset,
                                  size_t length);

/**
 * This method updates the CRC by adding all the bytes (starting at
 * the current offset) in the file <code>f</code> to the calculation.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] f input file
 */
extern CGUL_EXPORT void
cgul_crc16__update_file(cgul_exception_t* cex,
                        cgul_crc16_t crc,
                        FILE* f);

/**
 * This method updates the CRC by adding all the bytes in the file
 * with name <code>fname</code> to the calculation.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @param[in] fname input file name
 */
extern CGUL_EXPORT void
cgul_crc16__update_fname(cgul_exception_t* cex,
                         cgul_crc16_t crc,
                         const char* fname);

/**
 * Reset the <code>cgul_crc16</code> object for further use.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 */
extern CGUL_EXPORT void
cgul_crc16__reset(cgul_exception_t* cex,
                  cgul_crc16_t crc);

/**
 * You can use this method to see the CRC lookup table that is being
 * used.  The returned value is the CRC lookup table for this object
 * as a string.  The caller is responsible for calling free() on the
 * string that is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc16</code> instance
 * @return crc lookup table as a string
 */
extern CGUL_EXPORT char*
cgul_crc16__get_table_as_string(cgul_exception_t* cex,
                                cgul_crc16_t crc);


CGUL_END_C

#endif  /*  CGUL_CRC16_H  */
