#ifndef CGUL_ENDIAN_H
#define CGUL_ENDIAN_H

/**
 * Cconvert between big-endian and little-endian formats.
 *
 * @file
 * @author Paul Serice
 * @brief convert between big-endian and little-endian formats
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * Return whether the host is big endian.  This is a fast run-time
 * test, but calling it before each swap will adversely affect
 * performance.
 *
 * @param[in] cex c-style exception
 * @return whether the host is big-endian
 */
extern CGUL_EXPORT int
cgul_endian__is_big_endian(cgul_exception_t* cex);

/**
 * Return whether the host is little endian.  This is a fast run-time
 * test, but calling it before each swap will adversely affect
 * performance.
 *
 * @param[in] cex c-style exception
 * @return whether the host is little endian
 */
extern CGUL_EXPORT int
cgul_endian__is_little_endian(cgul_exception_t* cex);

/**
 * Byte swap the first <code>value_size</code> bytes in the region of
 * memory pointed to by <code>value</code>.  If <code>value</code> is
 * in big-endian format, it will be converted to little-endian format
 * and vice versa. <p>
 *
 * The following example shows how to convert an unsigned, big-endian
 * 32-bit integer to the host format (assuming the host is either big
 * endian or little endian instead of one of the more exotic formats):
 *
 * <pre>
 *     cgul_uint32_t value;
 *     ...
 *
 *     if (cgul_endian__is_little_endian(cex)) {
 *         ...
 *         cgul_endian__swap(cex, (char*)&value, sizeof(value));
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in,out] value value
 * @param[in] value_size size of <code>value</code>
 * @see cgul_int
 */
extern CGUL_EXPORT void
cgul_endian__swap(cgul_exception_t* cex,
                  void* value,
                  size_t value_size);

CGUL_END_C

#endif  /*  CGUL_ENDIAN_H  */
