#ifndef CGUL_HEAP_SORT_H
#define CGUL_HEAP_SORT_H

/**
 * Heap sort an array.
 *
 * @file
 * @author Paul Serice
 * @brief heap sort
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_heap_primitives.h"

CGUL_BEGIN_C

/**
 * Generic in-place sort with the same parameter signature as stdlib's
 * qsort().  <code>v</code> is a pointer to the top of your array.
 * <code>v_count</code> is the number of elements in <code>v</code>.
 * <code>element_size</code> is the number of contiguous bytes that
 * comprise each element.  <code>cf</code> is the comparison function.
 * It will be passed pointers to two elements.  It must return less
 * than zero, zero, or greater than zero if the first element is less
 * than, equal to, or greater than the second element.  As explained
 * in the next paragraph, this function has to internally allocate
 * memory to perform its task.  If out-of-memory occurs, an exception
 * is thrown. <p>
 *
 * One of the main reasons for using heap sort is that it requires no
 * extra space, but this function creates an extra copy of
 * <code>v</code> and another array of pointers that gets passed to
 * <code>cgul_heap_sort_pointers()</code>.  So, if the reason for
 * using heap sort is to conserve space, client code should be
 * organized around arrays of pointers that can be passed directly to
 * <code>cgul_heap_sort_pointers()</code>. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code> (in bytes)
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_heap_sort(cgul_exception_t* cex,
               void* v,
               size_t v_count,
               size_t element_size,
               cgul_heap_primitives__compare_t cf);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be sorted using
 * <code>cf</code> the same as described for
 * <code>cgul_heap_sort()</code> except this function operates without
 * using any extra space. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 */
extern CGUL_EXPORT void
cgul_heap_sort_pointers(cgul_exception_t* cex,
                        void** v,
                        size_t v_count,
                        cgul_heap_primitives__compare_t cf);

CGUL_END_C

#endif  /*  CGUL_HEAP_SORT_H  */
