#ifndef CGUL_HEAP_SORT_CXX_H
#define CGUL_HEAP_SORT_CXX_H

/**
 * C++ wrapper for <code>cgul_heap_sort</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_heap_sort</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_heap_primitives_cxx.h"
#include "cgul_heap_sort.h"

/**
 * This class provides the C++ bindings for
 * <code>cgul_heap_sort</code>.  The main purpose of this class is to
 * convert the C-style function calls and exception handling into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_heap_sort</code>
 */
class cgul_heap_sort_cxx {

public:

    /**
     * Generic in-place sort with the same parameter signature as stdlib's
     * qsort().  <code>v</code> is a pointer to the top of your array.
     * <code>v_count</code> is the number of elements in <code>v</code>.
     * <code>element_size</code> is the number of contiguous bytes that
     * comprise each element.  <code>cf</code> is the comparison function.
     * It will be passed pointers to two elements.  It must return less
     * than zero, zero, or greater than zero if the first element is less
     * than, equal to, or greater than the second element.  As explained
     * in the next paragraph, this function has to internally allocate
     * memory to perform its task.  If out-of-memory occurs, an exception
     * is thrown. <p>
     *
     * One of the main reasons for using heap sort is that it requires no
     * extra space, but this function creates an extra copy of
     * <code>v</code> and another array of pointers that gets passed to
     * <code>cgul_heap_sort_pointers()</code>.  So, if the reason for
     * using heap sort is to conserve space, client code should be
     * organized around arrays of pointers that can be passed directly to
     * <code>cgul_heap_sort_pointers()</code>. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
     * better solutions. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] element_size size of each element in <code>v</code> (in bytes)
     * @param[in] cf comparison function
     */
    static void
    sort(void* v,
         size_t v_count,
         size_t element_size,
         cgul_heap_primitives_cxx::compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_heap_sort(&local, v, v_count, element_size, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * If your array is already composed of pointers, you can directly
     * call this function.  <code>v</code> will be sorted using
     * <code>cf</code> the same as described for
     * <code>cgul_heap_sort()</code> except this function operates without
     * require any extra space (other than O(log n) stack
     * consumption). <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
     * better solutions. <p>
     *
     * @param[in] v array of pointers to elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] cf comparison function
     */
    static void
    sort_pointers(void** v,
                  size_t v_count,
                  cgul_heap_primitives_cxx::compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_heap_sort_pointers(&local, v, v_count, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

};

#endif  /*  CGUL_HEAP_SORT_CXX_H  */
