#ifndef CGUL_HEX_DUMP_H
#define CGUL_HEX_DUMP_H

/**
 * Create a hex dump of arbitrary binary data.
 *
 * @file
 * @author Paul Serice
 * @brief hex dump
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_string.h"

CGUL_BEGIN_C

typedef struct cgul_hex_dump* cgul_hex_dump_t;

/**
 * Create a new <code>cgul_hex_dump</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_hex_dump__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_hex_dump</code> instance
 */
extern CGUL_EXPORT cgul_hex_dump_t
cgul_hex_dump__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  Do not try to
 * use <code>hd</code> after calling this method.
 *
 * @param[in] hd <code>cgul_hex_dump</code> instance
 */
extern CGUL_EXPORT void
cgul_hex_dump__delete(cgul_hex_dump_t hd);

/**
 * Bytes from <code>buffer</code> to <code>buffer +
 * buffer_length</code> will be append to the conversion.  You can
 * call this function as often as you like.  For efficiency, a new
 * cgul_string is internally allocated for each call you make to
 * <code>cgul_hex_dump__append()</code>.  The new cgul_string is then
 * put on a linked list.  When you call
 * <code>cgul_hex_dump__get_value()</code>, all the cgul_string
 * objects on the list are concatenated.  If this function fails to
 * allocate memory, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] hd <code>cgul_hex_dump</code> instance
 * @param[in] buffer buffer
 * @param[in] buffer_length buffer length
 */
extern CGUL_EXPORT void
cgul_hex_dump__append(cgul_exception_t* cex,
                      cgul_hex_dump_t hd,
                      const void* buffer,
                      size_t buffer_length);

/**
 * Call this method at any point to get a <b>copy</b> of all the
 * strings appended together that comprise elements on the internal
 * linked list.  The caller is responsible for freeing the copy by
 * calling <code>cgul_string_delete()</code>.  If this function fails
 * to allocate memory, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] hd <code>cgul_hex_dump</code> instance
 * @return copy of current hex dump as a single string
 */
extern CGUL_EXPORT cgul_string_t
cgul_hex_dump__get_value(cgul_exception_t* cex,
                         cgul_hex_dump_t hd);

/**
 * The method lets you reuse a <code>cgul_hex_dump</code> object.  If
 * <code>reset_byte_count</code> is <code>true</code>, the object will
 * be reset to the same state as any new <code>cgul_hex_dump</code>
 * object would have.  If <code>reset_byte_count</code> is
 * <code>false</code>, the object will have the same state as any new
 * <code>cgul_hex_dump</code> object except the byte count that runs
 * along the left edge of the output will not be reset.
 *
 * @param[in] cex c-style exception
 * @param[in] hd <code>cgul_hex_dump</code> instance
 * @param[in] reset_byte_count whether to reset the byte count
 */
extern CGUL_EXPORT void
cgul_hex_dump__clear(cgul_exception_t* cex,
                     cgul_hex_dump_t hd,
                     int reset_byte_count);

CGUL_END_C

#endif  /*  CGUL_HEX_DUMP_H  */
