#ifndef CGUL_LINE_WRAPPER_CXX_H
#define CGUL_LINE_WRAPPER_CXX_H

/**
 * C++ wrapper for <code>cgul_line_wrapper</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_line_wrapper</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_string_cxx.h"
#include "cgul_wstring_cxx.h"
#include "cgul_line_wrapper.h"

/**
 * This class provides the C++ bindings for C <code>cgul_line_wrapper</code>
 * objects.  The main purpose of this class is to convert the C-style
 * function calls and exception handling in <code>cgul_line_wrapper</code>
 * into C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_line_wrapper</code>
 */
class cgul_line_wrapper_cxx {

public:

    /**
     * This function wraps the UTF-8 string <code>lines</code> at
     * <code>column</code> and returns the result in <code>lines</code>.
     * This is used for formatting text destined for the console.  It
     * assumes all characters have the same, unspecified width.  So the
     * wrapping is based on the maximum number of characters per line. <p>
     *
     * It is permissible for <code>lines</code> to have embedded new-line
     * characters before calling this method.  When an embedded new-line
     * character is encountered, this method will honor it and force the
     * start of a new line.
     *
     * If an error occurs, an exception is thrown. <p>
     *
     * @param[in,out] lines lines to wrap
     * @param[in] column column at which to wrap
     */
    static void
    wrap_utf8_at_column(cgul_string_cxx& lines,
                        unsigned int column)
    {
        cgul_exception_t local = NULL;
        cgul_line_wrapper__wrap_utf8_at_column(&local,
                                               lines.get_obj(),
                                               column);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This function wraps the UTF-32 string <code>wlines</code> at
     * <code>column</code> and returns the result in <code>wlines</code>.
     * This is used for formatting text destined for the console.  It
     * assumes all characters have the same, unspecified width.  So the
     * wrapping is based on the maximum number of wide characters per
     * line. <p>
     *
     * It is permissible for <code>wlines</code> to have embedded new-line
     * wide characters before calling this method.  When an embedded
     * new-line wide character is encountered, this method will honor it
     * and force the start of a new line.
     *
     * If an error occurs, an exception is thrown. <p>
     *
     * @param[in,out] wlines lines to wrap
     * @param[in] column column at which to wrap
     */
    static void
    wrap_utf32_at_column(cgul_wstring_cxx& wlines,
                         unsigned int column)
    {
        cgul_exception_t local = NULL;
        cgul_line_wrapper__wrap_utf32_at_column(&local,
                                                wlines.get_obj(),
                                                column);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This function wraps the UTF-8 string <code>lines</code> at
     * <code>width</code> and returns the result in <code>lines</code>.
     * This is used for formatting text destined to be drawn on a
     * graphical surface.  The amount of space each line consumes is
     * calculated using the BDF font <code>bdf</code> by calling
     * <code>cgul_bdf__get_utf8_extents()</code>.  Thus, in order for
     * this function to be accurate, your code must draw text as explained
     * in the comments to <code>cgul_bdf__get_utf8_extents()</code> and
     * the value you pass in here for <code>substitute</code> must
     * match. <p>
     *
     * It is permissible for <code>lines</code> to have embedded new-line
     * characters before calling this method.  When an embedded new-line
     * character is encountered, this method will honor it and force the
     * start of a new line. <p>
     *
     * If <code>substitute</code> is set, this method assumes the code
     * that draws the text will substitute missing glyphs with the default
     * glyph for the font, e.g., an empty box.  If <code>substitute</code>
     * is not set, this method will throw an exception if it cannot find
     * the glyph for one of the characters in <code>lines</code>. <p>
     *
     * If you just need to calculate the width of a string in pixels, you
     * can use <code>cgul_bdf__get_utf8_extents()</code> instead. <p>
     *
     * If an error occurs, an exception is thrown. <p>
     *
     * @param[in,out] lines lines to wrap
     * @param[in] bdf BDF font
     * @param[in] substitute whether to substitute missing glyphs with
     * the default
     * @param[in] max_line_width maximum line width before wrapping occurs
     */
    static void
    wrap_utf8_at_width(cgul_string_cxx& lines,
                       cgul_bdf_cxx& bdf,
                       int substitute,
                       unsigned int max_line_width)
    {
        cgul_exception_t local = NULL;
        cgul_line_wrapper__wrap_utf8_at_width(&local,
                                              lines.get_obj(),
                                              bdf.get_obj(),
                                              substitute,
                                              max_line_width);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This function wraps the UTF-32 string <code>wlines</code> at
     * <code>width</code> and returns the result in <code>wlines</code>.
     * This is used for formatting text destined to be drawn on a
     * graphical surface.  The amount of space each line consumes is
     * calculated using the BDF font <code>bdf</code> by calling
     * <code>cgul_bdf__get_utf32_extents()</code>.  Thus, in order for
     * this function to be accurate, your code must draw text as explained
     * in the comments to <code>cgul_bdf__get_utf32_extents()</code> and
     * the value you pass in here for <code>substitute</code> must
     * match. <p>
     *
     * It is permissible for <code>wlines</code> to have embedded new-line
     * characters before calling this method.  When an embedded new-line
     * character is encountered, this method will honor it and force the
     * start of a new line. <p>
     *
     * If <code>substitute</code> is set, this method assumes the code
     * that draws the text will substitute missing glyphs with the default
     * glyph for the font, e.g., an empty box.  If <code>substitute</code>
     * is not set, this method will throw an exception if it cannot find
     * the glyph for one of the characters in <code>lines</code>. <p>
     *
     * If you just need to calculate the width of a string in pixels, you
     * can use <code>cgul_bdf__get_utf32_extents()</code> instead. <p>
     *
     * If an error occurs, an exception is thrown. <p>
     *
     * @param[in,out] wlines lines to wrap
     * @param[in] bdf BDF font
     * @param[in] substitute whether to substitute missing glyphs with
     * the default
     * @param[in] max_line_width maximum line width before wrapping occurs
     */
    static void
    wrap_utf32_at_width(cgul_wstring_cxx& wlines,
                        cgul_bdf_cxx& bdf,
                        int substitute,
                        unsigned int max_line_width)
    {
        cgul_exception_t local = NULL;
        cgul_line_wrapper__wrap_utf32_at_width(&local,
                                               wlines.get_obj(),
                                               bdf.get_obj(),
                                               substitute,
                                               max_line_width);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

};

#endif  /*  CGUL_LINE_WRAPPER_CXX_H  */
