#ifndef CGUL_MERGE_SORT_CXX_H
#define CGUL_MERGE_SORT_CXX_H

/**
 * C++ wrapper for <code>cgul_merge_sort</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_merge_sort</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_merge_sort.h"

/**
 * This class provides the C++ bindings for
 * <code>cgul_merge_sort</code>.  The main purpose of this class is to
 * convert the C-style function calls and exception handling into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_merge_sort</code>
 */
class cgul_merge_sort_cxx {

public:

    /**
     * This typedef is the interface the client must define in order
     * for <code>sort()</code> or <code>sort_pointers()</code> to sort
     * an array.
     */
    typedef cgul_merge_sort__compare_t compare_t;

    /**
     * Generic sort with the same parameter signature as stdlib's
     * qsort().  <code>v</code> is a pointer to the top of your array.
     * <code>v_count</code> is the number of elements in
     * <code>v</code>.  <code>element_size</code> is the number of
     * contiguous bytes that comprise each element.  <code>cf</code>
     * is the comparison function.  It will be passed pointers to two
     * elements.  It must return less than zero, zero, or greater than
     * zero if the first element is less than, equal to, or greater
     * than the second element.  This function has to internally
     * allocate memory to perform its task.  If out-of-memory occurs,
     * an exception is thrown, and the array is returned partially
     * sorted. <p>
     *
     * The main draw back to using merge sort when sorting arrays is
     * that it needs to create a copy of the original array in order
     * to work efficiently.  This particular implementation actually
     * requires 1.5 copies of the original array at peak memory usage.
     * If you are short on memory, you can either use
     * cgul_merge_sort_iteratively() which only requires 1 copy or
     * <code>cgul_list_cxx::sort()</code> which requires no extra
     * copies because it operates on linked lists which have efficient
     * removal and insertion operations. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree_cxx</code> or <code>cgul_multimap_cxx</code>
     * are usually better solutions. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] element_size size of each element in <code>v</code>
     * @param[in] cf comparison function
     */
    static void
    sort(void* v,
         size_t v_count,
         size_t element_size,
         cgul_merge_sort__compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_merge_sort(&local, v, v_count, element_size, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Generic sort with the same parameter signature as stdlib's
     * qsort().  <code>v</code> is a pointer to the top of your array.
     * <code>v_count</code> is the number of elements in
     * <code>v</code>.  <code>element_size</code> is the number of
     * contiguous bytes that comprise each element.  <code>cf</code>
     * is the comparison function.  It will be passed pointers to two
     * elements.  It must return less than zero, zero, or greater than
     * zero if the first element is less than, equal to, or greater
     * than the second element.  This function has to internally
     * allocate memory to perform its task.  If out-of-memory occurs,
     * an exception is thrown, and the array is returned partially
     * sorted. <p>
     *
     * There are three main differences between this function and
     * <code>sort()</code>.  First, <code>sort()</code> is about 30%
     * faster than this function.  Second, <code>sort()</code> is a
     * recursive implementation that consumes more stack space than
     * this function which is an iterative implementation.  Third,
     * <code>sort()</code> uses 50% more memory than this
     * implementation because it requires 1.5 extra copies of
     * <code>v</code>; whereas, this implementation only requires 1
     * extra copy.  If in doubt, use <code>sort()</code> instead of
     * this function. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree_cxx</code> or <code>cgul_multimap_cxx</code>
     * are usually better solutions. <p>
     *
     * @param[in] v array of elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] element_size size of each element in <code>v</code>
     * @param[in] cf comparison function
     */
    static void
    sort_iteratively(void* v,
                     size_t v_count,
                     size_t element_size,
                     cgul_merge_sort__compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_merge_sort_iteratively(&local, v, v_count, element_size, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * If your array is already composed of pointers, you can directly
     * call this function.  <code>v</code> will be sorted using
     * <code>cf</code> the same as described for <code>sort()</code>.
     * If out-of-memory occurs, an exception is thrown, and the array
     * is returned partially sorted. <p>
     *
     * The main draw back to using merge sort when sorting arrays is
     * that it needs to create a copy of the original array in order
     * to work efficiently.  This particular implementation actually
     * requires 1.5 copies of the original array at peak memory usage.
     * If you are short on memory, you can either use
     * <code>sort_pointers_iteratively()</code> which only requires 1
     * copy or <code>cgul_list_cxx::sort()</code> which requires no
     * extra copies because it operates on linked list which have
     * efficient removal and insertion operations. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree_cxx</code> or <code>cgul_multimap_cxx</code>
     * are usually better solutions. <p>
     *
     * @param[in] v array of pointers to elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] cf comparison function
     */
    static void
    sort_pointers(void** v,
                  size_t v_count,
                  cgul_merge_sort__compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_merge_sort_pointers(&local, v, v_count, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * If your array is already composed of pointers, you can directly
     * call this function.  <code>v</code> will be sorted using
     * <code>cf</code> the same as described for
     * <code>cgul_merge_sort_iteratively()</code>.  If out-of-memory
     * occurs, an exception is thrown, and the array is returned
     * partially sorted. <p>
     *
     * There are three main differences between this function and
     * <code>sort_pointers()</code>.  First,
     * <code>sort_pointers()</code> is about 30% faster than this
     * function.  Second, <code>sort_pointers()</code> is a recursive
     * implementation that consumes more stack space than this
     * function which is an iterative implementation.  Third,
     * <code>sort_pointers()</code> uses 50% more memory than this
     * implementation because it requires 1.5 extra copies of
     * <code>v</code>; whereas, this implementation only requires 1
     * extra copy.  If in doubt, use <code>sort_pointers()</code>
     * instead of this function. <p>
     *
     * If you need to constantly maintain a sorted list,
     * <code>cgul_rbtree_cxx</code> or <code>cgul_multimap_cxx</code>
     * are usually better solutions. <p>
     *
     * @param[in] v array of pointers to elements to sort
     * @param[in] v_count number of elements in <code>v</code>
     * @param[in] cf comparison function
     */
    static void
    sort_pointers_iteratively(void** v,
                              size_t v_count,
                              cgul_merge_sort__compare_t cf)
    {
        cgul_exception_t local = NULL;
        cgul_merge_sort_pointers_iteratively(&local, v, v_count, cf);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }
};

#endif  /*  CGUL_MERGE_SORT_CXX_H  */
