#ifndef CGUL_MICROXML_H
#define CGUL_MICROXML_H

/**
 * Parser for MicroXML as specified by the W3C MicroXML Community
 * Group at the following URLs: <p>
 *
 * <pre>
 *     https://www.w3.org/community/microxml/
 *     https://dvcs.w3.org/hg/microxml/raw-file/tip/spec/microxml.html
 * </pre>
 *
 * By default, white space and comments in the prolog and epilog are
 * not tokenized.  Also, comments (but not white space) under the root
 * element are also not tokenized.  To receive all the tokens that are
 * skipped by default, the following methods can be used: <p>
 *
 * <code>
 *     cgul_microxml__set_skip_prolog_or_epilog_tokens(&local, uxml, 0);
 *     cgul_microxml__set_skip_comments(&local, uxml, 0);
 * </code>
 *
 * White space under the root element can be skipped as follows: <p>
 *
 * <code>
 *     cgul_microxml__set_skip_white_space(&local, uxml, 1);
 * </code>
 *
 * @file
 * @author Paul Serice
 * @brief MicroXML Parser
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_hash.h"
#include "cgul_stream_reader.h"
#include "cgul_string.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_microxml_t</code> instance. */
typedef struct cgul_microxml* cgul_microxml_t;

/**
 * Static method used to escape the special MicroXML characters '&',
 * '<', '>', '"', and "'" in the string <code>s</code>.  The client is
 * repsonsible for calling <code>free()</code> on the pointer
 * returned.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s UTF-8 input string
 * @return UTF-8 escaped output string
 */
extern CGUL_EXPORT char*
cgul_microxml__escape(cgul_exception_t* cex,
                      const char* s);

/**
 * Return a new <code>cgul_microxml</code> instance.  After this
 * method returns, <code>cgul_microxml__set_stream_reader()</code>
 * should be called before calling
 * <code>cgul_microxml__read_next_token()</code>.  The client is
 * responsible for calling <code>cgul_microxml__delete()</code> on the
 * pointer returned.  If an error occurs, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_microxml</code> instance
 * @see cgul_microxml__set_maximum_text_length()
 * @see cgul_microxml__set_skip_comments()
 * @see cgul_microxml__set_skip_white_space()
 */
extern CGUL_EXPORT cgul_microxml_t
cgul_microxml__new(cgul_exception_t* cex);

/**
 * This method creates a new <code>cgul_microxml</code> instance and
 * calls <code>cgul_microxml__open_fname()</code> passing it
 * <code>fname</code>.  The file will be closed when this instance is
 * deleted.  The client is responsible for calling
 * <code>cgul_microxml__delete()</code> on the object returned.  If an
 * error occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name
 * @return new <code>cgul_microxml</code> instance
 * @see cgul_microxml__set_text_size()
 * @see cgul_microxml__set_comment_size()
 * @see cgul_microxml__set_skip_comments()
 * @see cgul_microxml__set_skip_white_space()
 */
extern CGUL_EXPORT cgul_microxml_t
cgul_microxml__new_from_fname(cgul_exception_t* cex,
                              const char* fname);

/**
 * This method creates a new <code>cgul_microxml</code> instance and
 * calls <code>cgul_microxml__open_file()</code> passing it
 * <code>f</code>.  The class does not take ownership of
 * <code>f</code>.  Thus, the client is still responsible for calling
 * <code>fclose()</code> on it.  The client is also responsible for
 * calling <code>cgul_microxml__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @return new <code>cgul_microxml</code> instance
 * @see cgul_microxml__set_text_size()
 * @see cgul_microxml__set_comment_size()
 * @see cgul_microxml__set_skip_comments()
 * @see cgul_microxml__set_skip_white_space()
 */
extern CGUL_EXPORT cgul_microxml_t
cgul_microxml__new_from_file(cgul_exception_t* cex,
                             FILE* f);

/**
 * This method creates a new <code>cgul_microxml</code> instance and
 * calls <code>cgul_microxml__open_memory()</code> passing it
 * <code>buffer</code> and <code>buffer_size</code>.  This class does
 * not take ownership of <code>buffer</code> so the client is still
 * responsible for freeing <code>buffer</code> if necessary.  The
 * client is also responsible for calling
 * <code>cgul_microxml__delete()</code> on the object returned.  If an
 * error occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] buffer memory buffer
 * @param[in] buffer_size size of <code>buffer</code> in bytes
 * @return new <code>cgul_microxml</code> instance
 * @see cgul_microxml__set_text_size()
 * @see cgul_microxml__set_comment_size()
 * @see cgul_microxml__set_skip_comments()
 * @see cgul_microxml__set_skip_white_space()
 */
extern CGUL_EXPORT cgul_microxml_t
cgul_microxml__new_from_memory(cgul_exception_t* cex,
                               const char* buffer,
                               size_t buffer_size);

/**
 * This method creates a new <code>cgul_microxml</code> instance and
 * calls <code>cgul_microxml__set_stream_reader()</code> passing it
 * <code>sr</code>.  The client is responsible for calling
 * <code>cgul_microxml__delete()</code> on the object returned.  If an
 * error occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr stream reader
 * @return new <code>cgul_microxml</code> instance
 * @see cgul_microxml__set_maximum_text_length()
 * @see cgul_microxml__set_skip_comments()
 * @see cgul_microxml__set_skip_white_space()
 */
extern CGUL_EXPORT cgul_microxml_t
cgul_microxml__new_from_stream_reader(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr);

/**
 * Delete the MicroXML instance <code>uxml</code> freeing all
 * internally allocated resources.  The client must not attempt to use
 * <code>uxml</code> after calling this method.
 *
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
extern CGUL_EXPORT void
cgul_microxml__delete(cgul_microxml_t uxml);

/**
 * Open the MicroXML file with file name <code>fname</code>.  If a
 * file or memory buffer is already open, it is closed before
 * attempting to open the new file.  The new file will be closed when
 * this instance is deleted.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] fname file name
 */
extern CGUL_EXPORT void
cgul_microxml__open_fname(cgul_exception_t* cex,
                          cgul_microxml_t uxml,
                          const char* fname);

/**
 * Open the MicroXML file <code>f</code>.  If a file or memory buffer
 * is already open, it is closed before attempting to open the new
 * file.  This class does not take ownership of <code>f</code>.  Thus,
 * the client is still responsible for calling <code>fclose()</code>
 * on <code>f</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] f input file name
 */
extern CGUL_EXPORT void
cgul_microxml__open_file(cgul_exception_t* cex,
                         cgul_microxml_t uxml,
                         FILE* f);

/**
 * Open the MicroXML file contained in the memory buffer
 * <code>buffer</code> holding size <code>buffer_size</code> bytes
 * (not Unicode characters).  If a file or memory buffer is already
 * open, it is closed before attempting to open the new memory buffer.
 * This class does not take ownership of <code>buffer</code>.  Thus,
 * the client is still responsible for freeing the buffer if
 * necessary.  This method does not alter <code>buffer</code>, but
 * <code>buffer</code> should not be changed while it is being used by
 * this class.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] buffer memory buffer
 * @param[in] buffer_size buffer size in bytes (not Unicode characters)
 */
extern CGUL_EXPORT void
cgul_microxml__open_memory(cgul_exception_t* cex,
                           cgul_microxml_t uxml,
                           const char* buffer,
                           size_t buffer_size);

/**
 * Close the open MicroXML file (if any).  If an error occurs, an
 * exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
extern CGUL_EXPORT void
cgul_microxml__close(cgul_exception_t* cex,
                     cgul_microxml_t uxml);

/**
 * Return the stream reader.  If no stream reader is currently
 * selected, <code>NULL</code> is returned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return stream reader
 */
extern CGUL_EXPORT cgul_stream_reader_t
cgul_microxml__get_stream_reader(cgul_exception_t* cex,
                                 cgul_microxml_t uxml);

/**
 * Set the stream reader <code>sr</code>.  This class does not take
 * ownership of <code>sr</code>.  Thus, the client is still
 * responsible for deleting it.  If <code>sr</code> is
 * <code>NULL</code>, the old stream reader will no longer be used
 * making it safe to delete the old stream reader without having to
 * set a new one.  If an error occurs, an exception is thrown. <p>
 *
 * The only constraint on <code>sr</code> is that it must yield
 * Unicode characters.  If the stream is UTF-8 or starts with a
 * Unicode Byte-Order Mark (BOM), <code>cgul_stream_reader</code>
 * should automatically detect the correct decoder to use; otherwise,
 * manually set <code>sr</code> to use one of the following decoders: <p>
 *
 * <code>
 *     cgul_stream_reader__decode_utf8
 *     cgul_stream_reader__decode_utf16be
 *     cgul_stream_reader__decode_utf16le
 *     cgul_stream_reader__decode_utf32be
 *     cgul_stream_reader__decode_utf32le
 * </code>
 *
 * If you need to parse MicroXML fragments, multiple block readers can
 * be added to the stream reader at any time using
 * <code>cgul_stream_reader__add_block_reader()</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] sr stream reader
 * @see cgul_stream_reader__add_block_reader()
 */
extern CGUL_EXPORT void
cgul_microxml__set_stream_reader(cgul_exception_t* cex,
                                 cgul_microxml_t uxml,
                                 cgul_stream_reader_t sr);

/**
 * Return the presentation name.  If an input source is not currently
 * set, <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return presentation name
 */
extern CGUL_EXPORT const char*
cgul_microxml__get_presentation_name(cgul_exception_t* cex,
                                     cgul_microxml_t uxml);

/**
 * Set the presentation name.  The presentation name is primarily used
 * for error reporting and can be set to anything, but it does need to
 * be set again if the input file changes.  If the presentation name
 * is not set, it defaults to the file name that was opened or to
 * <code>"FILE"</code> or <code>"MEMORY"</code> if using a
 * <code>FILE*</code> or memory buffer respectively.  An exception is
 * thrown if an attempt is made to set the presentation name without
 * first setting the input file.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] presentation_name presentation name
 */
extern CGUL_EXPORT void
cgul_microxml__set_presentation_name(cgul_exception_t* cex,
                                     cgul_microxml_t uxml,
                                     const char* presentation_name);

/**
 * Return the maximum entity length.  If set to <code>0</code>, the
 * length is limited only by available memory.  The default is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return maximum entity length
 */
extern CGUL_EXPORT size_t
cgul_microxml__get_maximum_entity_length(cgul_exception_t* cex,
                                         cgul_microxml_t uxml);

/**
 * Set the maximum entity length.  If set to <code>0</code>, the
 * length is limited only by available memory.  The default is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] entity_length_max maximum entity length
 */
extern CGUL_EXPORT void
cgul_microxml__set_maximum_entity_length(cgul_exception_t* cex,
                                         cgul_microxml_t uxml,
                                         size_t entity_length_max);

/**
 * Return the starting line number for the current token.  If no input
 * source is set, <code>0</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return current line number
 */
extern CGUL_EXPORT size_t
cgul_microxml__get_line_number(cgul_exception_t* cex,
                               cgul_microxml_t uxml);

/**
 * Return the starting column number for the current token.  If no
 * input source is set, <code>0</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return current column number
 */
extern CGUL_EXPORT size_t
cgul_microxml__get_column_number(cgul_exception_t* cex,
                                 cgul_microxml_t uxml);

/**
 * Read the next token from the MicroXML input source.  This method
 * can be used to implement a pull parser.  If an error occurs, an
 * exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @see cgul_microxml__is_document_start()
 * @see cgul_microxml__is_document_end()
 * @see cgul_microxml__is_element_start()
 * @see cgul_microxml__is_element_end()
 * @see cgul_microxml__is_text()
 * @see cgul_microxml__is_comment()
 */
extern CGUL_EXPORT void
cgul_microxml__read_next_token(cgul_exception_t* cex,
                               cgul_microxml_t uxml);

/**
 * Return whether the current token indicates the start of the
 * document.  This is always the first token.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether the current token indicates the start of the document
 */
extern CGUL_EXPORT int
cgul_microxml__is_document_start(cgul_exception_t* cex,
                                 cgul_microxml_t uxml);

/**
 * Return whether the current token indicates the end of the document.
 * This is always the last token if no errors occurr.  Attempting to
 * get the next token after this one results in an error.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether the current token indicates the end of the document
 */
extern CGUL_EXPORT int
cgul_microxml__is_document_end(cgul_exception_t* cex,
                               cgul_microxml_t uxml);

/**
 * Return whether the start tag for an element has been read.  Use
 * <code>cgul_microxml__get_element_name()</code> and
 * <code>cgul_microxml__get_attributes()</code> to get the name and
 * attributes for the element.  Use
 * <code>cgul_microxml__is_empty_element()</code> to determine if the
 * element is an empty element (e.g., <code>&lt;foo/&gt;</code>).
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether the start tag for an element has been read
 */
extern CGUL_EXPORT int
cgul_microxml__is_element_start(cgul_exception_t* cex,
                                cgul_microxml_t uxml);

/**
 * Return whether the end tag for an element has been read or, if an
 * empty element was parsed earlier, whether an end tag is being
 * simulated on behalf of the earlier empty element.  This way the
 * following two MicroXML samples generate the same token stream: <p>
 *
 * <pre>
 *     &lt;foo&gt;&lt;/foo&gt;
 *     &lt;foo/&gt;
 * </pre>
 *
 * Use <code>cgul_microxml__get_element_name()</code> to get the name
 * of the element.  Use <code>cgul_microxml__is_empty_element()</code>
 * to determine if the end tag is being simulated because the matching
 * start tag is an empty element. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether the end tag for an element has been read
 */
extern CGUL_EXPORT int
cgul_microxml__is_element_end(cgul_exception_t* cex,
                              cgul_microxml_t uxml);

/**
 * Return whether text has been read.  Use
 * <code>cgul_microxml__get_text()</code> to get the text.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether text has been read
 */
extern CGUL_EXPORT int
cgul_microxml__is_text(cgul_exception_t* cex,
                       cgul_microxml_t uxml);

/**
 * Return whether a comment has been read.  Use
 * <code>cgul_microxml__get_comment()</code> to get the comment.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether comment has been read
 */
extern CGUL_EXPORT int
cgul_microxml__is_comment(cgul_exception_t* cex,
                          cgul_microxml_t uxml);

/**
 * Return the name of the current element.  The string is owned by
 * this class so the client must not attempt to free it.  This method
 * is valid when processing the start or end tag of an element;
 * otherwise, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * Element names are interned by the <code>cgul_microxml</code>
 * instance making them valid until the <code>cgul_microxml</code>
 * instance is deleted or reset.  It is common for elements to be
 * repeated.  In this case, this method returns the same pointer for
 * the repeated elements which allows DOMs to be created in a memory
 * efficient manner.  Because the element names are shared, it is
 * important that the client treat them as immutable which should be
 * natural in most contexts. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return name associated with token
 */
extern CGUL_EXPORT const char*
cgul_microxml__get_element_name(cgul_exception_t* cex,
                                cgul_microxml_t uxml);

/**
 * Return the value for the attribute name
 * <code>attribute_name</code>.  If no attribute has that name,
 * <code>NULL</code> is returned.  Ownership of the attribute value
 * remains with the <code>cgul_microxml</code> instance so the client
 * must not attempt to delete it.  This method is valid when
 * processing the start or end tag of an element; otherwise,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * Unlike element names and attribute names, attribute values are only
 * valid until the next token is read unless
 * <code>cgul_string__take_value()</code> is called on the returned
 * attribute value in which case the client is responsible for calling
 * <code>free()</code> on the value taken. <p>
 *
 * The attribute value can be converted to numeric types using methods
 * from <code>cgul_string</code> like
 * <code>cgul_string__to_int()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] attribute_name attribute name
 * @return value for the given attribute name or <code>NULL</code>
 * @see cgul_microxml__get_attributes()
 * @see cgul_microxml__take_attributes()
 */
extern CGUL_EXPORT cgul_string_t
cgul_microxml__get_attribute_value(cgul_exception_t* cex,
                                   cgul_microxml_t uxml,
                                   const char* attribute_name);

/**
 * Return the entire attribute map associated with the current
 * element.  The keys and values stored in the attributes map are
 * owned by this class so the client must not attempt to free or
 * delete them.  This method is valid when processing the start tag of
 * an element; otherwise, <code>NULL</code> is returned, and an
 * exception is thrown. <p>
 *
 * The keys in the attributes map correspond to the names of the
 * attributes and are C-style strings.  The keys are valid until this
 * <code>cgul_microxml</code> instance is deleted or reset. <p>
 *
 * The values in the attributes map correspond to the attribute values
 * and are <code>cgul_string</code> instances (not C-style strings).
 * Unlike the keys, the values are only valid until the next token is
 * read unless <code>cgul_string__take_value()</code> is called on the
 * the attribute values in which case the client is responsible for
 * calling <code>free()</code> on each value taken. <p>
 *
 * The attribute values can be converted to numeric types using
 * methods from <code>cgul_string</code> like
 * <code>cgul_string__to_int()</code>. <p>
 *
 * You can iterate over the attribute names and values as follows: <p>
 *
 * <code>
 *     cgul_hash_node_t attr = NULL;
 *     attr = cgul_hash__get_front(cex, attrs);
 *     for ( ; attr ; attr = cgul_hash_node__get_next(cex, attr)) {
 *      // Get the key/value pair for this attribute.
 *      const char* key = (const char*)cgul_hash_node__get_key(cex, attr);
 *      cgul_string_t tmp = (cgul_string_t)cgul_hash_node__get_value(cex, attr);
 *      const char* value = cgul_string__get_value(cex, tmp);
 *      ...
 *     }
 * </code>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return map of attributes associated with token
 * @see cgul_microxml__take_attributes()
 * @see cgul_microxml__get_attribute_value()
 */
extern CGUL_EXPORT cgul_hash_t
cgul_microxml__get_attributes(cgul_exception_t* cex,
                              cgul_microxml_t uxml);

/**
 * Return the attributes map associated with the current element.
 * This method transfers ownership of the attributes map requiring the
 * client to delete the map by calling
 * <code>cgul_microxml__delete_attributes()</code>.  The element
 * attributes can only be taken once per element.  This method is
 * valid when processing the start tag of an element; otherwise,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * The keys in the attributes map correspond to the names of the
 * attributes and are C-style strings.  The keys are only valid until
 * <code>cgul_microxml__delete_attributes()</code> is called or until
 * this <code>cgul_microxml</code> instance is deleted or reset.  So
 * do *not* try use the keys after deleting <code>uxml</code>!  (The
 * reason for this is that attribute names are interned in a symbol
 * table owned by <code>uxml</code> so that multiple occurrences of
 * the same attribute name can share the same symbol without requiring
 * extra memory for a duplicate string.) <p>
 *
 * The values in the attributes map correspond to the attribute values
 * and are <code>cgul_string</code> instances (not C-style strings).
 * Unlike the keys, the values are valid until
 * <code>cgul_microxml__delete_attributes()</code> is called even if
 * <code>uxml</code> has been deleted or reset. <p>
 *
 * The attribute values can be converted to numeric types using
 * methods from <code>cgul_string</code> like
 * <code>cgul_string__to_int()</code>. <p>
 *
 * You can iterate over the attribute names and values as follows: <p>
 *
 * <code>
 *     cgul_hash_node_t attr = NULL;
 *     attr = cgul_hash__get_front(cex, attrs);
 *     for ( ; attr ; attr = cgul_hash_node__get_next(cex, attr)) {
 *      // Get the key/value pair for this attribute.
 *      const char* key = (const char*)cgul_hash_node__get_key(cex, attr);
 *      cgul_string_t tmp = (cgul_string_t)cgul_hash_node__get_value(cex, attr);
 *      const char* value = cgul_string__get_value(cex, tmp);
 *      ...
 *     }
 * </code>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return map of attributes associated with token
 * @see cgul_microxml__get_attributes()
 * @see cgul_microxml__get_attribute_value()
 */
extern CGUL_EXPORT cgul_hash_t
cgul_microxml__take_attributes(cgul_exception_t* cex,
                               cgul_microxml_t uxml);

/**
 * Convenience method used to delete the attributes map
 * <code>attributes</code> returned by the function
 * <code>cgul_microxml__take_attributes()</code>.  The client must not
 * use <code>attributes</code> after this method returns. <p>
 *
 * @note Do not use this function for attribute maps returned by
 * <code>cgul_microxml__get_attributes()</code>.  Only use it on
 * attributes maps returned by
 * <code>cgul_microxml__take_attributes()</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] attributes attributes map
 */
extern CGUL_EXPORT void
cgul_microxml__delete_attributes(cgul_exception_t* cex,
                                 cgul_microxml_t uxml,
                                 cgul_hash_t attributes);

/**
 * Return whether the element is an empty element (e.g.,
 * <code>&lt;foo/&gt;</code>).  This method is valid when processing
 * the start or end tag of an element; otherwise, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether the empty-element syntax was used to close the start tag
 */
extern CGUL_EXPORT int
cgul_microxml__is_empty_element(cgul_exception_t* cex,
                                cgul_microxml_t uxml);

/**
 * Return the current text.  The returned value is owned by this class
 * so the client must not attempt to delete it.  This method is valid
 * when processing text; otherwise, <code>NULL</code> is returned, and
 * an exception is thrown. <p>
 *
 * The text returned is only valid until the next token is read unless
 * <code>cgul_string__take_value()</code> is called on the returned
 * value in which case the client is responsible for calling
 * <code>free()</code> on the text taken. <p>
 *
 * The text can be converted to numeric types using methods from
 * <code>cgul_string</code> like
 * <code>cgul_string__to_int()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return text associated with token
 * @see cgul_microxml__set_maximum_text_length()
 * @see cgul_microxml__set_skip_white_space()
 */
extern CGUL_EXPORT cgul_string_t
cgul_microxml__get_text(cgul_exception_t* cex,
                        cgul_microxml_t uxml);

/**
 * Return whether the current text is comprised of only white space
 * characters.  The MicroXML specification defines white space as tab,
 * new line, or space characters.  This method is valid when
 * processing text; otherwise, <code>0</code> is returned, and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether text is pure white space
 */
extern CGUL_EXPORT int
cgul_microxml__is_white_space(cgul_exception_t* cex,
                              cgul_microxml_t uxml);

/**
 * Return whether to skip text under the root element that is pure
 * white space.  By default, this white space is not skipped.  For
 * white space in the prolog and epilog, see
 * <code>cgul_microxml__get_skip_prolog_or_epilog_tokens()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether to skip text that is pure white space
 * @see cgul_microxml__get_skip_prolog_or_epilog_tokens()
 */
extern CGUL_EXPORT int
cgul_microxml__get_skip_white_space(cgul_exception_t* cex,
                                    cgul_microxml_t uxml);

/**
 * Set whether to skip text under the root element that is pure white
 * space.  This is convenient when the document being parsed is
 * formatted with line breaks and indentation that are not
 * significant.  By default, this white space is not skipped.  For
 * white space in the prolog and epilog, see
 * <code>cgul_microxml__set_skip_prolog_or_epilog_tokens()</code>. <p>
 *
 * If enabled, <code>cgul_microxml__set_maximum_text_length()</code>
 * should be used to increase the maximum size of each block of text
 * (if necessary) so that at least one character that is not white
 * space is always included in each block of significant white
 * space. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] skip whether to skip text that is pure white space
 * @see cgul_microxml__set_maximum_text_length()
 * @see cgul_microxml__set_skip_prolog_or_epilog_tokens()
 */
extern CGUL_EXPORT void
cgul_microxml__set_skip_white_space(cgul_exception_t* cex,
                                    cgul_microxml_t uxml,
                                    int skip);

/**
 * Return the maximum size in Unicode characters of text returned by
 * <code>cgul_microxml__get_text()</code>.  If set to <code>0</code>,
 * the length is limited only by available memory.  The default is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return maximum text length in Unicode characters
 */
extern CGUL_EXPORT size_t
cgul_microxml__get_maximum_text_length(cgul_exception_t* cex,
                                       cgul_microxml_t uxml);

/**
 * Set the maximum size in Unicode characters of text returned by
 * <code>cgul_microxml__get_text()</code>.  If set to <code>0</code>,
 * the length is limited only by available memory.  The default is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] text_length_max maximum text length in Unicode characters
 */
extern CGUL_EXPORT void
cgul_microxml__set_maximum_text_length(cgul_exception_t* cex,
                                       cgul_microxml_t uxml,
                                       size_t text_length_max);

/**
 * Return the current comment.  The returned value is owned by this
 * class so the client must not attempt to delete it.  This method is
 * valid when processing comments; otherwise, <code>NULL</code> is
 * returned, and an exception is thrown. <p>
 *
 * The comment returned is only valid until the next token is read
 * unless <code>cgul_string__take_value()</code> is called on the
 * returned value in which case the client is responsible for calling
 * <code>free()</code> on the comment taken. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return comment associated with token
 * @see cgul_microxml__set_skip_comments()
 */
extern CGUL_EXPORT cgul_string_t
cgul_microxml__get_comment(cgul_exception_t* cex,
                           cgul_microxml_t uxml);

/**
 * Return whether to skip comments under the root element.  By
 * default, these comments are skipped.  For comments in the prolog
 * and epilog, see
 * <code>cgul_microxml__get_skip_prolog_or_epilog_tokens()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether to skip text that is pure white space
 * @see cgul_microxml__get_skip_prolog_or_epilog_tokens()
 */
extern CGUL_EXPORT int
cgul_microxml__get_skip_comments(cgul_exception_t* cex,
                                 cgul_microxml_t uxml);

/**
 * Set whether to skip comments under the root element.  By default,
 * these comments are skipped.  For comments in the prolog and epilog,
 * see <code>cgul_microxml__set_skip_prolog_or_epilog_tokens()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] skip whether to skip text that is pure white space
 * @see cgul_microxml__set_skip_prolog_or_epilog_tokens()
 */
extern CGUL_EXPORT void
cgul_microxml__set_skip_comments(cgul_exception_t* cex,
                                 cgul_microxml_t uxml,
                                 int skip);

/**
 * Return the maximum comment length.  If set to <code>0</code>, the
 * length is limited only by available memory.  The default is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return maximum comment length
 */
extern CGUL_EXPORT size_t
cgul_microxml__get_maximum_comment_length(cgul_exception_t* cex,
                                          cgul_microxml_t uxml);

/**
 * Set the maximum comment length.  If set to <code>0</code>, the
 * length is limited only by available memory.  The default is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] comment_length_max maximum comment length
 */
extern CGUL_EXPORT void
cgul_microxml__set_maximum_comment_length(cgul_exception_t* cex,
                                          cgul_microxml_t uxml,
                                          size_t comment_length_max);

/**
 * Return whether to skip tokens in the prolog or epilog.  These are
 * the white space and comment tokens that surround the root element.
 * By default, these tokens are skipped.  For white space and comments
 * under the root element, see
 * <code>cgul_microxml__get_skip_white_space()</code> and
 * <code>cgul_microxml__get_skip_comments()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @return whether tokens in the prolog or epilog
 * @see cgul_microxml__get_skip_white_space()
 * @see cgul_microxml__get_skip_comments()
 */
extern CGUL_EXPORT int
cgul_microxml__get_skip_prolog_or_epilog_tokens(cgul_exception_t* cex,
                                                cgul_microxml_t uxml);

/**
 * Set whether to skip tokens in the prolog or epilog.  These are the
 * white space and comment tokens that surround the root element.  By
 * default, these tokens are skipped.  For white space and comments
 * under the root element, see
 * <code>cgul_microxml__set_skip_white_space()</code> and
 * <code>cgul_microxml__set_skip_comments()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] skip whether tokens in the prolog or epilog
 * @see cgul_microxml__set_skip_white_space()
 * @see cgul_microxml__set_skip_comments()
 */
extern CGUL_EXPORT void
cgul_microxml__set_skip_prolog_or_epilog_tokens(cgul_exception_t* cex,
                                                cgul_microxml_t uxml,
                                                int skip);

CGUL_END_C

#endif  /*  CGUL_MICROXML_H  */
