#ifndef CGUL_MIME_CXX_H
#define CGUL_MIME_CXX_H

/**
 * C++ wrapper for <code>cgul_mime</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_mime</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_mime.h"

/**
 * This class provides the C++ bindings for <code>cgul_mime</code>.
 * The main purpose of this class is to convert the C-style function
 * calls and exception handling in <code>cgul_mime</code> into
 * C++-style function calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_mime</code>
 */
class cgul_mime_cxx {

public:

    /**
     * The sections of a multipart MIME file.
     */
    typedef cgul_mime__section_t section_t;

    /**
     * This typedef is the interface for the callback function invoked
     * by <code>parse()</code> to produce blocks from the original
     * MIME source so that the blocks can be parsed. <p>
     *
     * Functions that implement this interface should read bytes into
     * the block <code>block</code> that extends for
     * <code>block_size</code> bytes and return the number of bytes
     * read into <code>block</code>.  If and only if EOF is reached,
     * this function should return <code>0</code>.  It is permissible
     * to return a short read count greater than zero.  If the parser
     * still needs more data, it will just invoke the callback again.
     * If an error occurs, this function should throw a c-style
     * exception (and only a c-style exception) which will be
     * translated into a <code>cgul_exception_cxx</code>
     * exception. <p>
     *
     * The value of <code>data</code> is the same as what was initially
     * passed into <code>parse()</code> when the callback was
     * registered allowing the client to pass arbitrary data to the
     * producer. <p>
     *
     * @param[in,out] cex c-style exception
     * @param[in] block block
     * @param[in] block_size block size
     * @param[in] data client data
     * @return number of bytes read into <code>block</code> or zero if EOF
     */
    typedef size_t
    (*produce_t)(cgul_exception_t* cex,
                 char* block,
                 size_t block_size,
                 void* data);

    /**
     * This typedef is the interface for the callback function invoked by
     * the following functions to consume the blocks generated by the
     * parser: <p>
     *
     * <pre>
     *     parse()
     *     parse_file()
     *     parse_fname()
     *     parse_memory()
     * </pre>
     *
     * Functions that implement this interface will be passed each block
     * <code>block</code> and its size in bytes <code>block_size</code>
     * from the MIME file as it is parsed along with the MIME section type
     * <code>section</code> from which the data originated.  The function
     * will also be passed the current part <code>part</code> of the
     * multipart MIME file that is being parsed.  The value for
     * <code>part</code> will be <code>0</code> for the preamble and the
     * epilogue, <code>1</code> for the first part, <code>2</code> for the
     * second part, and so on. <p>
     *
     * The value of <code>data</code> is the same as what was
     * initially passed in when the callback was registered allowing
     * the client to pass arbitrary data to the consumer.  If an error
     * occurs, this function should throw a c-style exception (and
     * only a c-style exception) which will be translated into a
     * <code>cgul_exception_cxx</code> exception. <p>
     *
     * @param[in,out] cex c-style exception
     * @param[in] part MIME part of the multipart message
     * @param[in] section MIME section where the block originates
     * @param[in] block block
     * @param[in] block_size block size
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int
    (*consume_t)(cgul_exception_t* cex,
                 unsigned long int part,
                 section_t section,
                 char* block,
                 size_t block_size,
                 void* data);

    /**
     * This is the generic parsing function that works by parsing blocks
     * of data supplied by the <code>produce</code> function.  For each
     * block supplied by <code>produce</code>, it invokes the client's
     * callback function <code>consume</code> at least once passing in a
     * block that holds a subset of bytes from the current MIME message,
     * the size of the block, the MIME section type, and the client data
     * <code>consume_data</code>.  This function returns the total number
     * of MIME parts.  If an error occurs, <code>0</code> is returned, and
     * an exception is thrown. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the first 16K of the file.  If
     * it cannot automatically detect the EOL sequence, this function will
     * use "\r\n" as the EOL sequence. <p>
     *
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @param[in] produce callback function to produce MIME blocks
     * @param[in] produce_data client data for <code>produce</code>
     * @param[in] consume callback function to consume parsed blocks
     * @param[in] consume_data client data for <code>consume</code>
     * @return total number of parts
     * @see parse_memory()
     * @see parse_file()
     */
    static unsigned long int
    parse(const char* boundary,
          const char* eol,
          produce_t produce,
          void* produce_data,
          consume_t consume,
          void* consume_data)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__parse(&local,
                              boundary,
                              eol,
                              (cgul_mime__produce_t)produce,
                              produce_data,
                              (cgul_mime__consume_t)consume,
                              consume_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is an adapter for <code>parse()</code> for
     * parsing multipart MIME messages already resident in memory.  It
     * registers a special "producer" that allows the parser to directly
     * access the MIME message in memory rather than having to copy it
     * block-by-block into the parser's internal buffer. <p>
     *
     * This function works by parsing a multipart MIME message that is in
     * memory at <code>block</code> and extends for
     * <code>block_size</code> bytes.  The client's callback function
     * <code>consume</code> is invoked once for each MIME section passing
     * in a block that holds all the bytes from the current MIME section,
     * the size of the block, the MIME section type, and the client data
     * <code>consume_data</code>.  This function returns the total number
     * of MIME parts.  If an error occurs, <code>0</code> is returned, and
     * an exception is thrown. <p>
     *
     * Even though this function passes all the bytes from each MIME
     * section to the <code>consume</code> callback, it may be worthwhile
     * to write <code>consume</code> so that it can also be used with
     * <code>parse_file()</code> which is similar but generally
     * requires multiple calls to <code>consume</code> in order to pass in
     * the same amount of data. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the first 16K of the file.  If
     * it cannot automatically detect the EOL sequence, this function will
     * use "\r\n" as the EOL sequence. <p>
     *
     * @param[in] block block
     * @param[in] block_size block size
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @param[in] consume callback function to consume parsed blocks
     * @param[in] consume_data client data for <code>consume</code>
     * @return total number of parts
     * @see split_memory()
     */
    static unsigned long int
    parse_memory(char* block,
                 size_t block_size,
                 const char* boundary,
                 const char* eol,
                 consume_t consume,
                 void* consume_data)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__parse_memory(&local,
                                     block,
                                     block_size,
                                     boundary,
                                     eol,
                                     (cgul_mime__consume_t)consume,
                                     consume_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is an adapter for <code>parse()</code> for
     * parsing multipart MIME messages from a file.  It registers a
     * "producer" that copies the MIME message from file block-by-block
     * into the parser's internal buffer. <p>
     *
     * This function works by parsing a multipart MIME message in the file
     * <code>fin</code>.  For each block read from the file, it invokes
     * the client's callback function <code>consume</code> at least once
     * passing in a block that holds a subset of bytes from the current
     * MIME section, the size of the block, the MIME section type, and the
     * client data <code>consume_data</code>.  This function returns the
     * total number of MIME parts.  If an error occurs, <code>0</code> is
     * returned, and an exception is thrown. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the first 16K of the file.  If
     * it cannot automatically detect the EOL sequence, this function will
     * use "\r\n" as the EOL sequence. <p>
     *
     * @param[in] fin input file
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @param[in] consume callback function to consume parsed blocks
     * @param[in] consume_data client data for <code>consume</code>
     * @return total number of parts
     * @see split_file()
     */
    static unsigned long int
    parse_file(FILE* fin,
               const char* boundary,
               const char* eol,
               consume_t consume,
               void* consume_data)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__parse_file(&local,
                                   fin,
                                   boundary,
                                   eol,
                                   (cgul_mime__consume_t)consume,
                                   consume_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is an adapter for <code>parse()</code> for
     * parsing multipart MIME messages from a file.  It registers a
     * "producer" that copies the MIME message from file block-by-block
     * into the parser's internal buffer. <p>
     *
     * This function works by parsing a multipart MIME message in the file
     * with name <code>fname</code>.  For each block read from the file,
     * it invokes the client's callback function <code>consume</code> at
     * least once passing in a block that holds a subset of bytes from the
     * current MIME section, the size of the block, the MIME section type,
     * and the client data <code>consume_data</code>.  This function
     * returns the total number of MIME parts.  If an error occurs,
     * <code>0</code> is returned, and an exception is thrown. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the first 16K of the file.  If
     * it cannot automatically detect the EOL sequence, this function will
     * use "\r\n" as the EOL sequence. <p>
     *
     * @param[in] fname name of the input file
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @param[in] consume callback function to consume parsed blocks
     * @param[in] consume_data client data for <code>consume</code>
     * @return total number of parts
     * @see split_fname()
     */
    static unsigned long int
    parse_fname(const char* fname,
                const char* boundary,
                const char* eol,
                consume_t consume,
                void* consume_data)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__parse_fname(&local,
                                    fname,
                                    boundary,
                                    eol,
                                    (cgul_mime__consume_t)consume,
                                    consume_data);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is an adapter for <code>parse_memory()</code>
     * that splits a multipart MIME message in memory into multiple
     * files on the host file system. <p>
     *
     * Split the multipart MIME message in memory starting at
     * <code>block</code> and extending for <code>block_size</code> bytes
     * into preamble, header, body, and epilogue sections.  The directory
     * where the files should be written is given by <code>dname</code>.
     * The name of the preamble is given by <code>pname</code>.  The names
     * for the header files will have the form <code>hname-%lu</code>.
     * The names for the body files will have the form
     * <code>bname-%lu</code> where "%lu" will be replaced by the part
     * number (starting with 1).  If any of <code>pname</code>,
     * <code>hname</code>, <code>bname</code>, or <code>ename</code> are
     * <code>NULL</code>, those files will not be written.  The name of
     * the epilogue file is given by <code>ename</code>.  This function
     * returns the total number of MIME parts.  If an error occurs,
     * <code>0</code> is returned, and an exception is thrown. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the MIME message.  If it cannot
     * automatically detect the EOL sequence, this function will use
     * "\r\n" as the EOL sequence. <p>
     *
     * @param[in] block block
     * @param[in] block_size block size
     * @param[in] dname directory name where output files will be created
     * @param[in] pname name for the preamble file
     * @param[in] hname base name for header files
     * @param[in] bname base name for body files
     * @param[in] ename name for the epilogue file
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @return total number of parts
     * @see parse_memory()
     */
    static unsigned long int
    split_memory(char* block,
                 size_t block_size,
                 const char* dname,
                 const char* pname,
                 const char* hname,
                 const char* bname,
                 const char* ename,
                 const char* boundary,
                 const char* eol)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__split_memory(&local,
                                     block,
                                     block_size,
                                     dname,
                                     pname,
                                     hname,
                                     bname,
                                     ename,
                                     boundary,
                                     eol);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is an adapter for <code>parse_file()</code> that
     * splits a multipart MIME message in a file into multiple files
     * on the host file system. <p>
     *
     * Split the multipart MIME message in the file <code>fin</code> into
     * preamble, header, body, and epilogue sections.  The directory where
     * the files should be written is given by <code>dname</code>.  The
     * name of the preamble is given by <code>pname</code>.  The names for
     * the header files will have the form <code>hname-%lu</code>.  The
     * names for the body files will have the form <code>bname-%lu</code>
     * where "%lu" will be replaced by the part number (starting with 1).
     * If any of <code>pname</code>, <code>hname</code>,
     * <code>bname</code>, or <code>ename</code> are <code>NULL</code>,
     * those files will not be written.  The name of the epilogue file is
     * given by <code>ename</code>.  This function returns the total
     * number of MIME parts.  If an error occurs, <code>0</code> is
     * returned, and an exception is thrown. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the first 16K of the file.  If
     * it cannot automatically detect the EOL sequence, this function will
     * use "\r\n" as the EOL sequence. <p>
     *
     * @param[in] fin MIME input file
     * @param[in] dname directory name where output files will be created
     * @param[in] pname name for the preamble file
     * @param[in] hname base name for header files
     * @param[in] bname base name for body files
     * @param[in] ename name for the epilogue file
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @return total number of parts
     * @see parse_file()
     */
    static unsigned long int
    split_file(FILE* fin,
               const char* dname,
               const char* pname,
               const char* hname,
               const char* bname,
               const char* ename,
               const char* boundary,
               const char* eol)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__split_file(&local,
                                   fin,
                                   dname,
                                   pname,
                                   hname,
                                   bname,
                                   ename,
                                   boundary,
                                   eol);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function is an adapter for <code>parse_fname()</code> that
     * splits a multipart MIME message in a file into multiple files
     * on the host file system. <p>
     *
     * Split the multipart MIME message in the file with name
     * <code>fname</code> into preamble, header, body, and epilogue
     * sections.  The directory where the files should be written is given
     * by <code>dname</code>.  The name of the preamble is given by
     * <code>pname</code>.  The names for the header files will have the
     * form <code>hname-%lu</code>.  The names for the body files will
     * have the form <code>bname-%lu</code> where "%lu" will be replaced
     * by the part number (starting with 1).  If any of
     * <code>pname</code>, <code>hname</code>, <code>bname</code>, or
     * <code>ename</code> are <code>NULL</code>, those files will not be
     * written.  The name of the epilogue file is given by
     * <code>ename</code>.  This function returns the total number of MIME
     * parts.  If an error occurs, <code>0</code> is returned, and an
     * exception is thrown. <p>
     *
     * The MIME parts are separated from each other by the MIME boundary
     * <code>boundary</code>.  Unfortunately, the boundary separators are
     * required to start with "--", but the "Content-Type" header
     * specifies the boundary without the implicit "--" prefix.  To make
     * it possible to directly pass the boundary read from the
     * "Content-Type" header into this function, <code>boundary</code>
     * must not include the implicit "--" prefix; instead, the "--" prefix
     * will automatically be inserted by this function. <p>
     *
     * According to the MIME standard, all EOL sequences should be "\r\n",
     * but as a practical matter, MIME messages often use the native EOL
     * sequence instead.  As a result, the client needs to pass in the
     * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
     * <code>NULL</code> this function will attempt to automatically
     * detect the EOL sequence by scanning the first 16K of the file.  If
     * it cannot automatically detect the EOL sequence, this function will
     * use "\r\n" as the EOL sequence. <p>
     *
     * @param[in] fname name of MIME input file
     * @param[in] dname directory name where output files will be created
     * @param[in] pname name for the preamble file
     * @param[in] hname base name for header files
     * @param[in] bname base name for body files
     * @param[in] ename name for the epilogue file
     * @param[in] boundary MIME boundary
     * @param[in] eol EOL sequence
     * @return total number of parts
     * @see parse_fname()
     */
    static unsigned long int
    split_fname(const char* fname,
                const char* dname,
                const char* pname,
                const char* hname,
                const char* bname,
                const char* ename,
                const char* boundary,
                const char* eol)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_mime__split_fname(&local,
                                    fname,
                                    dname,
                                    pname,
                                    hname,
                                    bname,
                                    ename,
                                    boundary,
                                    eol);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};

#endif  /*  CGUL_MIME_CXX_H  */
