#ifndef CGUL_RBTREE_CXX_H
#define CGUL_RBTREE_CXX_H

/**
 * C++ wrapper for <code>cgul_rbtree</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_rbtree</code>
 */

#include "cgul_cache_cxx.h"
#include "cgul_exception_cxx.h"
#include "cgul_rbtree.h"
#include "cgul_rbtree_node_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_rbtree</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_rbtree</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_rbtree</code>
 * @see cgul_rbtree_t
 */
class cgul_rbtree_cxx {

public:

    /**
     * This typedef is the interface the client must define in order
     * for <code>cgul_rbtree_cxx</code> to sort your keys as they are
     * inserted. <p>
     */
    typedef cgul_rbtree__compare_t compare_t;

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_keys()</code>
     *     <code>foldr_keys()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_key_t)(const void* key,
                              void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_values()</code>
     *     <code>foldr_values()</code>
     * </pre>
     *
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_value_t)(void* value,
                                void* data);

    /**
     * This typedef is the interface for the combining function used
     * by the following methods: <p>
     *
     * <pre>
     *     <code>foldl_pairs()</code>
     *     <code>foldr_pairs()</code>
     * </pre>
     *
     * @param[in] key key
     * @param[in] value value
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*fold_pair_t)(const void* key,
                               void* value,
                               void* data);

    /**
     * This typedef is the interface for the callback function used by
     * the following methods: <p>
     *
     * <pre>
     *     <code>traverse()</code>
     *     <code>traverse_range()</code>
     * </pre>
     *
     * @param[in] t tree
     * @param[in] n node
     * @param[in] data client data
     * @return whether to continue
     */
    typedef int (*traverse_t)(cgul_rbtree_cxx* t,
                              cgul_rbtree_node_cxx* n,
                              void* data);

    /**
     * Create a new <code>cgul_rbtree_cxx</code> object.
     * <code>compare</code> is the comparison function that allows the
     * tree to maintain its sorted order.  The caller is responsible
     * for arranging for <code>delete</code> to be called on the
     * object.  If memory cannot be allocated, an exception is
     * thrown. <p>
     *
     * Be default <code>with_indexing</code> is false which means the
     * ability to index into the tree as though it were an array will
     * not be enabled.  If <code>with_indexing</code> is set to true,
     * each node will be larger containing two extra <code>unsigned
     * long</code> values and will run about 15% slower. <p>
     *
     * This class does not take ownership of the inserted keys/value
     * pairs.  Instead, the client is responsible for freeing the
     * key/value pairs only after their nodes have been permanently
     * removed from the <code>cgul_rbtree</code>.  The
     * <code>traverse()</code> method can be used to remove each node
     * safely before calling <code>delete</code> on the object created. <p>
     *
     * @param[in] compare comparison function
     * @param[in] with_indexing whether to enable indexing
     */
    cgul_rbtree_cxx(compare_t compare,
                    int with_indexing = 0)
    {
        cgul_exception_t local = NULL;
        if (with_indexing) {
            obj = cgul_rbtree__new_with_indexing(&local, compare);
        } else {
            obj = cgul_rbtree__new(&local, compare);
        }
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Create a new <code>cgul_rbtree_cxx</code> object by wrapping
     * an existing <code>cgul_rbtree</code> object.
     *
     * @param[in] rhs right-hand side
     */
    cgul_rbtree_cxx(cgul_rbtree_t rhs)
    {
        obj = rhs;
    }

    /**
     * This method frees all internally allocated memory.  This does
     * not include the key/value pairs stored in the tree.  The client
     * is responsible for freeing the key/value pairs when the client
     * thinks it is convenient and safe to do so; however, the client
     * must understand that freeing a key before removing its node
     * invalidates the data structure. <p>
     *
     * As a convenience, you may want to call <code>clear()</code>
     * before calling this method in order to properly put the keys
     * and values back on their respective <code>cgul_cache_cxx</code>
     * objects (if <code>cgul_cache_cxx</code> objects are being used)
     * before you delete the tree.  If <code>cgul_cache_cxx</code>
     * objects are not being used, the client needs to arrange some
     * other mechanism to free the keys or values. <p>
     *
     * @see free_keys()
     * @see free_values()
     */
    virtual
    ~cgul_rbtree_cxx()
    {
        if (obj) {
            cgul_rbtree__delete(obj);
            obj = NULL;
        }
    }

    /**
     * This method calls <code>free()</code> on all the keys in the
     * tree.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free keys.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the tree.
     */
    virtual void
    free_keys()
    {
        cgul_rbtree__free_keys(obj);
    }

    /**
     * This method calls <code>free()</code> on all the values in the
     * tree.  Because this is such a common operation, it is an
     * exception to the rule that cgul containers never free values.
     * This method should only ever be called immediately before
     * calling <code>delete</code> because it otherwise invalidates
     * the tree.
     */
    virtual void
    free_values()
    {
        cgul_rbtree__free_values(obj);
    }

    /**
     * Get the size of the cache of nodes.
     *
     * @return size of the cache of nodes
     */
    virtual unsigned long int
    get_cache_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__get_cache_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the size of the cache of nodes. <p>
     *
     * For efficiency, the <code>cgul_rbtree_cxx</code> object can
     * keep a cache of nodes that can be reused.  In many situations,
     * this can greatly reduce the number of calls to
     * <code>malloc()</code> which can greatly improve performance and
     * reduce memory fragmentation. <p>
     *
     * You can release all the cached nodes and disable caching of
     * nodes by setting the cache reserve to 0 and the cache size to 0
     * (in that order).  This is the default. <p>
     *
     * If an error occurs (while enlarging the cache), an exception is
     * thrown. <p>
     *
     * @param[in] size of the cache of nodes
     */
    virtual void
    set_cache_size(unsigned long int size)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__set_cache_size(&local, obj, size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the reserve limit of the cache of nodes.
     *
     * @return reserve limit of the cache of nodes
     */
    virtual unsigned long int
    get_cache_reserve() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__get_cache_reserve(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the reserve limit of the cache of nodes and guarantee that
     * at least that many nodes are allocated and waiting in
     * reserve. <p>
     *
     * When you use this method to increase the reserve, you are
     * guaranteed that at least <code>reserve</code> count of nodes
     * will be allocated and reserved for future use. <p>
     *
     * This is useful because <code>insert()</code> can throw an
     * exception if it cannot allocate a node to hold the key/value
     * pair.  Thus, by allocating and reserving a node now, we know
     * that <code>insert()</code> will be able to run without error
     * later when the node is "unreserved".  Normally, this is not an
     * issue because you can almost always just insert at the moment
     * when it is requested, but if the <code>cgul_rbtree_cxx</code>
     * is busy at that moment (perhaps it is iterating over its
     * nodes), you'll have to queue insert (and remove) requests for
     * later execution. <p>
     *
     * To "unreserve" a node so that it can be used, you simply call this
     * method again with a value for <code>reserve</code> that is less
     * than the current reserve limit. <p>
     *
     * If an error occurs while allocating the reserved nodes, an
     * exception is thrown. <p>
     *
     * WARNING: If you reserve nodes without "unreserving" them, you will
     * introduce a memory leak that is difficult to track down because
     * this class is careful to free all of the cached nodes (including
     * the reserved nodes) when it is deleted, but the reserved nodes will
     * just sit around in memory until that time.  Thus, after reserving
     * nodes by calling this method, you must remember to timely set the
     * reserve level back to zero in order to make those nodes
     * available. <p>
     *
     * @param[in] reserve of the cache of nodes
     * @see cgul_cache_cxx::set_reserve()
     */
    virtual void
    set_cache_reserve(unsigned long int reserve)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__set_cache_reserve(&local, obj, reserve);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return 1 if the tree is empty; otherwise, return 0.
     *
     * @return whether the tree is empty
     */
    virtual int
    is_empty() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__is_empty(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Insert <code>key</code> into the tree.  If <code>key</code>
     * already exists, another node with the same key is not created.
     * The new or pre-existing node is returned in <code>*n</code> if
     * <code>n</code> is not <code>NULL</code>. <p>
     *
     * If a matching node is not found, <code>1</code> is returned because
     * a new node can and will be inserted.  If a matching node is found,
     * <code>0</code> is returned because the insertion failed because the
     * key already exists.<p>
     *
     * If an error occurs, 0 is returned, <code>*n</code> is set to
     * <code>NULL</code> (if <code>n</code> is not <code>NULL</code>), and
     * an exception is thrown. <p>
     *
     * This method operates as it does so that you only have to search the
     * tree once when doing an insert.  Logically, you would usually like
     * to do a find to see if the tree holds your key.  If it does, you
     * use the value associated with that key in some way.  If it does
     * not, you then want to insert a new key/value pair into the tree.
     * The problem with divorcing the find from the insert is that it then
     * requires you to search the tree twice.  Once to see if the key
     * exists and once to do the insertion. <p>
     *
     * To prevent having to search the tree twice, you can just call this
     * method to handle both inserting and updating nodes.  It will search
     * the tree once, and it will always return a pointer to the correct
     * node.  You can then use <code>cgul_rbtree_node_cxx::get_value()</code>
     * or <code>cgul_rbtree_node_cxx::set_value()</code> to get or set the
     * value indexed by "key". <p>
     *
     * TIP: You might find it convenient to use <code>cgul_string_cxx</code>
     * to build up a string that you use as a key.  You can then use
     * <code>cgul_string_cxx::get_value()</code> [yes, get_value()] and
     * pass the string to <code>insert()</code>.  If the insert
     * succeeds, you can inform the <code>cgul_string_cxx</code> object
     * that it is no longer the owner of the underlying C-style string
     * by calling <code>cgul_string_cxx::take_value()</code> [yes,
     * take_value()].  An example is as follows: <p>
     *
     * <pre>
     *
     *     cgul_string_cxx word;
     *     cgul_rbtree_node_cxx* n = NULL;
     *
     *     if (t->insert(word.get_value(), &n)) {
     *         word.take_value();
     *         . . .
     *     } else {
     *         . . .
     *     }
     *
     * </pre>
     *
     * NOTE: Inserting a key/value pair does not cause this class to take
     * ownership of <code>key</code> or <code>value</code>.  The client
     * must make sure the values pointed to by <code>key</code> and
     * <code>value</code> are not invalidated during the life of the newly
     * created node.  If the client needs to delete the <code>key</code>
     * or the <code>value</code> or both, the client needs to manually
     * arrange for this. <p>
     *
     * @param[in] key key
     * @param[out] n node that already existed or was newly created
     * @return 1 if a new node was inserted; 0 if a node with the same key
     * already exists
     */
    virtual int
    insert(const void* key,
           cgul_rbtree_node_cxx** n)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__insert(&local, obj, key, (cgul_rbtree_node_t*)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node indexed by <code>key</code>.  If the key does not
     * exist, <code>NULL</code> is returned; otherwise, the node
     * associated with the key is returned.
     *
     * @param[in] key key
     * @return node associated with the key or <code>NULL</code>
     */
    virtual cgul_rbtree_node_cxx*
    find(const void* key)
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__find(&local, obj, key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node indexed by position.  This lets you use a
     * <code>cgul_rbtree_cxx</code> instances like you would an
     * array.  Indexing into <code>cgul_rbtree_cxx</code> is a
     * logarithmic function which will be slower than an array, but
     * you get all the advantages of a balanced binary tree including
     * fast insertion and deletion.  So, if you have ever feel the
     * need to shift the elements of an array, this method is for you. <p>
     *
     * If <code>index</code> is out of range, <code>NULL</code> is
     * returned; otherwise, the node returned is the same node that
     * would be returned if you had an array <code>t</code> that held
     * all the nodes in the tree in sorted order, and you asked for
     * <code>t[index]</code>.
     *
     * This method throws an exception if the tree was not created
     * with indexing enabled. <p>
     *
     * @param[in] index index into sorted keys of node to return
     * @return node associated with <code>index</code>
     * @see find_rank()
     */
    virtual cgul_rbtree_node_cxx*
    find_at(unsigned long int index)
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__find_at(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return whether the rank of the key <code>key</code> could be
     * determined.  If the rank could be determined it is returned in
     * <code>*rank</code> (if <code>rank</code> is not
     * <code>NULL</code>). <p>
     *
     * The rank is the number of keys less than <code>key</code>.  It
     * can be used as the index parameter to <code>find_at()</code> in
     * order to lookup <code>key</code> by index. <p>
     *
     * This method throws an exception if <code>t</code> was not created
     * with indexing enabled. <p>
     *
     * @param[in] key key
     * @param[out] rank rank of key
     * @return whether the key exists in the tree
     * @see find_at()
     */
    virtual int
    find_rank(const void* key,
              unsigned long int* rank)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__find_rank(&local, obj, key, rank);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node having the largest key less than or equal to the
     * search key <code>search_key</code> and return it.  If such a
     * key does not exist, <code>NULL</code> is returned instead.
     *
     * @param[in] search_key search key
     * @return floor node or <code>NULL</code>
     */
    virtual cgul_rbtree_node_cxx*
    find_floor(const void* search_key)
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__find_floor(
                 &local, obj, search_key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Find the node having the smallest key greater than or equal to
     * the search key <code>search_key</code> and return it.  If such
     * a key does not exist, <code>NULL</code> is returned instead.
     *
     * @param[in] search_key search key
     * @return ceiling node or <code>NULL</code>
     */
    virtual cgul_rbtree_node_cxx*
    find_ceiling(const void* search_key)
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__find_ceiling(
                 &local, obj, search_key);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

#if 0
    /**
     * This function lets you search the tree using arbitrary client data
     * rather than a specific key.  The comparison function
     * <code>compare</code> will be called with <code>key1</code> set to
     * the current node which is of type <code>cgul_rbtree_node_t</code>
     * and <code>key2</code> set to <code>data</code>.  The comparison
     * function should return less than one if you want to move to the
     * current node's left child.  It should return greater than one if
     * you want to move to the current node's right child.  It should
     * return zero, if you have found what you are looking for.  If your
     * comparison function returns zero, that is the node that will be
     * returned; otherwise, <code>NULL</code> is returned.
     *
     * @param[in] compare comparison function
     * @param[in] data data passed to <code>compare</code> in lieu of
     * a second key
     * @return node that was found or <code>NULL</code> if no node
     * could be found
     */
    virtual cgul_rbtree_node_cxx*
    find_special(compare_t compare,
                 const void* data);
#endif

    /**
     * Return the first node according to sort order.  This operation is
     * O(1).  If the tree is empty, <code>NULL</code> is returned.  This
     * class does not have to search the tree in order to find the front
     * node because it efficiently keeps a direct pointer to the front
     * node up to date.
     *
     * The following example shows how to iterate over the entire tree in
     * sort order:
     *
     * <pre>
     *     cgul_rbtree_node_cxx* n = t->get_front();
     *     for ( ; n ; n = n->get_next()) {
     *         ...
     *     }
     * </pre>
     *
     * @return front node
     */
    virtual cgul_rbtree_node_cxx*
    get_front() const
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__get_front(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the last node according to sort order.  This operation is
     * O(1).  If the tree is empty, <code>NULL</code> is returned.  This
     * class does not have to search the tree in order to find the back
     * node because it efficiently keeps a direct pointer to the back node
     * up to date.
     *
     * The following example shows how to iterate over the entire tree in
     * reverse sort order:
     *
     * <pre>
     *     cgul_rbtree_node_cxx* n = t->get_back();
     *     for ( ; n ; n = n->get_prev()) {
     *         ...
     *     }
     * </pre>
     *
     * @return back node
     */
    virtual cgul_rbtree_node_cxx*
    get_back() const
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__get_back(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the oldest node according to chronological order (i.e., the
     * order in which the nodes are inserted).  This operation is O(1).
     * If the tree is empty, <code>NULL</code> is returned.  This class
     * does not have to search the tree in order to find the oldest node
     * because it efficiently keeps a direct pointer to the oldest node up
     * to date. <p>
     *
     * The following example shows how to iterate over the entire tree in
     * chronological order:
     *
     * <pre>
     *     cgul_rbtree_node_cxx* n = t->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return oldest node
     * @see cgul_rbtree_node_cxx::get_younger()
     */
    virtual cgul_rbtree_node_cxx*
    get_oldest() const
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__get_oldest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the oldest node in the tree to be <code>n</code>.  Calling this
     * method has the potential to confuse iterators and should be handled
     * with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires the
     * oldest node in the tree, and you want to force <code>n</code> to be
     * the next node to expire.
     *
     * @param[in] n node
     * @see set_youngest()
     */
    virtual void
    set_oldest(cgul_rbtree_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__set_oldest(&local, obj, (cgul_rbtree_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the youngest node according to chronological order (i.e.,
     * the order in which the nodes are inserted).  This operation is
     * O(1).  If the tree is empty, <code>NULL</code> is returned.  This
     * class does not have to search the tree in order to find the
     * youngest node because it efficiently keeps a direct pointer to the
     * youngest node up to date. <p>
     *
     * The following example shows how to iterate over the entire tree in
     * reverse chronological order:
     *
     * <pre>
     *     cgul_rbtree_node_cxx* n = t->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return youngest node
     * @see cgul_rbtree_node_cxx::get_older()
     */
    virtual cgul_rbtree_node_cxx*
    get_youngest() const
    {
        cgul_rbtree_node_cxx* rv = 0;
        cgul_exception_t local = NULL;
        rv = (cgul_rbtree_node_cxx*)cgul_rbtree__get_youngest(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the youngest node in the tree to be <code>n</code>.  Calling
     * this method has the potential to confuse iterators and should be
     * handled with roughly the same level of caution as calling
     * <code>remove_node()</code>. <p>
     *
     * This method could be used, for example, if your code expires
     * the least-recently used (LRU) node.  By calling
     * <code>set_youngest()</code> each time a node is used, the
     * least-recently used node will be the oldest node in the tree.
     *
     * @param[in] n node
     * @see set_oldest()
     */
    virtual void
    set_youngest(cgul_rbtree_node_cxx* n)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__set_youngest(&local, obj, (cgul_rbtree_node_t)n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Remove the node associated with <code>key</code>.
     * The key and value pointers stored in the node that is to be removed
     * will be returned in <code>key_out</code> and <code>value_out</code>
     * if you pass in pointers that are not <code>NULL</code>.  This
     * method returns 1 if the node was removed; it returns 0
     * otherwise. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the tree.  See <code>remove_node()</code> for
     * details. <p>
     *
     * @param[in] key_in key associated with the node to be removed
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove(const void* key_in,
           void** key_out,
           void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__remove(&local, obj, key_in, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Remove <code>node</code> from the tree.  The key and
     * value pointers stored in the node that is to be removed will be
     * returned in <code>key_out</code> and <code>value_out</code> if you
     * pass in pointers that are not <code>NULL</code>. <p>
     *
     * There is no speed advantage to using this method instead of
     * <code>remove()</code>.  The reason is that removing a
     * node has the potential to unbalance the tree requiring the code to
     * descend the tree even though we already have the node in hand. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the tree.  The problem is that calling this
     * method invalidates the node making it impossible to call
     * <code>node->get_next()</code> afterward.  The solution is
     * simple.  Just call <code>node->get_next()</code> before calling
     * this method:
     *
     * <pre>
     *
     *     for (curr = t->find(key) ; curr ; curr = next) {
     *         next = curr->get_next();
     *         t->remove_node(curr, &key_out, &value_out);
     *         free(key_out);
     *         free(value_out);
     *     }
     *
     * </pre>
     *
     * Alternatively, you can use <code>remove_range()</code>,
     * <code>traverse()</code>, or <code>traverse_range()</code>. <p>
     *
     * If you use one of the traversal methods, it is safe to call this
     * method naively.  For example, if your tree stores
     * <code>cgul_string_cxx*</code> objects for both its keys and values,
     * another way to remove nodes is as follows: <p>
     *
     * <pre>
     *
     *     static int
     *     remove_node_cb(cgul_rbtree_cxx* t,
     *                    cgul_rbtree_node_cxx* node,
     *                    void* data)
     *     {
     *         cgul_string_cxx* key = NULL;
     *         cgul_string_cxx* value = NULL;
     *         t->remove_node(node, (void**)&key, (void**)&value);
     *         delete key;
     *         delete value;
     *         return 1;
     *     }
     *
     *     cgul_rbtree_node_cxx* first = t->find(key1);
     *     cgul_rbtree_node_cxx* last = t->find(key2);
     *     if (first && last) {
     *         t->traverse_range(first, last, &remove_node_cb, NULL);
     *     }
     *
     * </pre>
     *
     * @param[in] n node to be removed
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     */
    virtual void
    remove_node(cgul_rbtree_node_cxx* n,
                void** key_out,
                void** value_out)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__remove_node(&local,
                                 obj,
                                 (cgul_rbtree_node_t)n,
                                 key_out,
                                 value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method lets you remove a node based on its sorted order
     * <code>index</code>.  If the index is out of bounds, 0 is returned;
     * otherwise 1 is returned.  If you are interested in getting your
     * hands on the key and value pointers stored in the node that is to
     * be removed, they will be returned in <code>key_out</code> and
     * <code>value_out</code> if you pass in pointers that are not
     * <code>NULL</code>.  <code>index</code> is zero-based.
     *
     * This method throws an exception if the tree was not
     * created with indexing enabled. <p>
     *
     * It is almost always a mistake to naively call this method while
     * iterating over the tree.  See <code>remove_node()</code> for
     * details. <p>
     *
     * @param[in] index index of the sorted key associated with the node
     * to be removed
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_at(unsigned long int index,
              void** key_out,
              void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__remove_at(&local, obj, index, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * As with all remove operations,
     * <code>remove_front()</code> has to descend the tree
     * making sure the tree is balanced for the removal.  The performance
     * advantage to using this method is that it does not have to call the
     * comparison function to decide whether it should take the right or
     * left child of each node.  It knows to always take the left child. <p>
     *
     * If you are interested in getting your hands on the key and value
     * pointers stored in the node that is to be removed, they will be
     * returned in <code>key_out</code> and <code>value_out</code> if you
     * pass in pointers that are not <code>NULL</code>.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_front(void** key_out,
                 void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__remove_front(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * As with all remove operations,
     * <code>remove_back()</code> has to descend the tree
     * making sure the tree is balanced for the removal.  The performance
     * advantage to using this method is that it does not have to call the
     * comparison function to decide whether it should take the right or
     * left child of each node.  It knows to always take the right
     * child. <p>
     *
     * If you are interested in getting your hands on the key and value
     * pointers stored in the node that is to be removed, they will be
     * returned in <code>key_out</code> and <code>value_out</code> if you
     * pass in pointers that are not <code>NULL</code>.
     *
     * @param[out] key_out pointer for the key stored in the node
     * @param[out] value_out pointer for the value stored in the node
     * @return whether the node was removed
     */
    virtual int
    remove_back(void** key_out,
                void** value_out)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__remove_back(&local, obj, key_out, value_out);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method removes nodes in the range <code>first</code>
     * (inclusive) to <code>last</code> (inclusive).  Strictly as a
     * convenience, this method is an exception to the rule that cgul
     * containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches. <p>
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @return whether any nodes were removed.
     * @see cgul_cache_cxx::get_freer()
     */
    virtual int
    remove_range(cgul_rbtree_node_cxx* first,
                 cgul_rbtree_node_cxx* last,
                 cgul_cache_cxx* keys_cache = NULL,
                 cgul_cache_cxx* values_cache = NULL)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__remove_range(
                 &local,
                 obj,
                 (cgul_rbtree_node_t)first,
                 (cgul_rbtree_node_t)last,
                 (keys_cache ? keys_cache->get_obj() : NULL),
                 (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method clears the tree by removing each node individually.
     * Strictly as a convenience, this method is an exception to the rule
     * that cgul containers never free keys or values.  If you pass in
     * <code>keys_cache</code> or <code>values_cache</code> instances that
     * are not <code>NULL</code>, the keys or values will be put back on
     * their respective caches. <p>
     *
     * @param[in] keys_cache keys cache
     * @param[in] values_cache values cache
     * @see remove_range()
     * @see cgul_cache_cxx::get_freer()
     */
    virtual void
    clear(cgul_cache_cxx* keys_cache,
          cgul_cache_cxx* values_cache)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__clear(&local,
                           obj,
                           (keys_cache ? keys_cache->get_obj() : NULL),
                           (values_cache ? values_cache->get_obj() : NULL));
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the total number of elements stored in the tree.
     *
     * @return size of the tree
     */
    virtual unsigned long int
    get_size() const
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__get_size(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function performs a basic sanity check on the tree and returns
     * 1 if the tree passes and 0 if the tree fails.  For now, it checks
     * to see that there are never two red nodes without at least one
     * intervening black node.  It also checks to make sure that the black
     * depth of all nodes are within one level of each other.  There are
     * many unbalanced trees that satisfy this criteria however.
     *
     * @return 1 if passed basic sanity check; otherwise, 0
     */
    virtual int
    is_ok() const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_rbtree__is_ok(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large trees, this should be much faster than trying to do the
     * same thing using removes and inserts.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_rbtree_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_rbtree__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method performs a left fold of the tree with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each key in the tree starting at the front of the tree and
     * iterating forward to the end of the tree. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_keys(fold_key_t f,
               void* data)
    {
        const void* key = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__foldl_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_rbtree_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_rbtree_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the tree with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key in the tree starting at the back of the tree
     * and iterating backward to the front of the tree. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_keys(fold_key_t f,
               void* data)
    {
        const void* key = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__foldr_keys()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_rbtree_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_rbtree_node__get_key(&local, curr);
            if (!(*f)(key, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the tree with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each value in the tree starting at the front of the tree and
     * iterating forward to the end of the tree. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_values(fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__foldl_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_rbtree_node__get_next(&local, curr);
            /* Call back. */
            value = cgul_rbtree_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the tree with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each value in the tree starting at the back of the
     * tree and iterating backward to the front of the tree. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * value.  The second parameter passed into <code>f</code> is the
     * client data which is where the result of the fold should be
     * accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_values(fold_value_t f,
                 void* data)
    {
        void* value = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__foldr_values()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_rbtree_node__get_prev(&local, curr);
            /* Call back. */
            value = cgul_rbtree_node__get_value(&local, curr);
            if (!(*f)(value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a left fold of the tree with the combining
     * function <code>f</code>.  <code>f</code> is called once for
     * each key/value pair in the tree starting at the front of the
     * tree and iterating forward to the end of the tree. <p>
     * 
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldl_pairs(fold_pair_t f,
                void* data)
    {
        const void* key = NULL;
        void* value = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__foldl_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_front(&local, obj);
        for ( ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = cgul_rbtree_node__get_next(&local, curr);
            /* Call back. */
            key = cgul_rbtree_node__get_key(&local, curr);
            value = cgul_rbtree_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * This method performs a right fold of the tree with the
     * combining function <code>f</code>.  <code>f</code> is called
     * once for each key/value pair in the tree starting at the back
     * of the tree and iterating backward to the front of the
     * tree. <p>
     *
     * The first parameter passed into <code>f</code> is the current
     * key.  The second parameter passed into <code>f</code> is the
     * current value.  The third parameter passed into <code>f</code>
     * is the client data which is where the result of the fold should
     * be accumulated. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * iteration to continue. <p>
     *
     * @param[in] f combining function
     * @param[in] data client data passed to <code>f</code>
     */
    virtual void
    foldr_pairs(fold_pair_t f,
                void* data)
    {
        const void* key = NULL;
        void* value = NULL;
        cgul_rbtree_node_t curr = NULL;
        cgul_rbtree_node_t prev = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__foldr_pairs()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */

        /* Iterate over the nodes. */
        curr = cgul_rbtree__get_back(&local, obj);
        for ( ; curr ; curr = prev) {
            /* Get prev now in case f deletes curr. */
            prev = cgul_rbtree_node__get_prev(&local, curr);
            /* Call back. */
            key = cgul_rbtree_node__get_key(&local, curr);
            value = cgul_rbtree_node__get_value(&local, curr);
            if (!(*f)(key, value, data)) {
                break;
            }
        }
    }

    /**
     * Traverse all nodes passing each node to the function
     * <code>f</code>. <p>
     *
     * The first parameter passed into <code>f</code> is the tree
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the tree elements.  In fact, I would
     * recommend that you use
     * <code>cgul_rbtree_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_rbtree_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the rbtree, but you do have to be careful.
     *
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse_range()
     */
    virtual void
    traverse(traverse_t f,
             void* data)
    {
        traverse_range(NULL, NULL, f, data);
    }

    /**
     * Traverse all nodes starting with <code>first</code> (inclusive) and
     * ending with <code>last</code> (inclusive) passing each node to the
     * function <code>f</code>.  If you know the first node, but are not
     * sure of the last node, just use <code>NULL</code> as the last node.
     * This will cause this method to iterate until it reaches the end of
     * the tree.  You can than have <code>f</code> return 0 when it
     * determines that the last node has been reached. <p>
     *
     * The first parameter passed into <code>f</code> is the tree
     * <code>this</code>.  The second paramenter passed into
     * <code>f</code> is the node for this iteration.  The third
     * parameter passed into <code>f</code> is the client data
     * <code>data</code>. <p>
     *
     * <code>f</code> is provided with a safe context in which it can
     * remove the node that is passed into <code>f</code> by calling
     * <code>remove_node()</code>. <p>
     *
     * <code>f</code> must return true after each iteration in order for
     * the traversal to continue. <p>
     *
     * If <code>first</code> is <code>NULL</code>, iteration starts at the
     * beginning of the tree.  If <code>last</code> is <code>NULL</code>,
     * iteration stops at the end of the tree. <p>
     *
     * NOTE: It is not strictly necessary that you use
     * <code>traverse()</code> or <code>traverse_range()</code> in
     * order to iterate over the tree elements.  In fact, I would
     * recommend that you use
     * <code>cgul_rbtree_node_cxx::get_next()</code> for most of your
     * iteration needs.  If you need to remove nodes though, you
     * should probably use this method.  If you look at this method's
     * source however, you will see that the
     * <code>cgul_rbtree_node_cxx</code> class provides all the public
     * methods you need to safely remove nodes while you are iterating
     * over the tree, but you do have to be careful.
     *
     * @param[in] first first node in range
     * @param[in] last last node in range
     * @param[in] f traversal callback function
     * @param[in] data client data passed to <code>f</code>
     * @see traverse()
     */
    virtual void
    traverse_range(cgul_rbtree_node_cxx* first,
                   cgul_rbtree_node_cxx* last,
                   traverse_t f,
                   void* data)
    {
        cgul_rbtree_node_cxx* curr = NULL;
        cgul_rbtree_node_cxx* next = NULL;
        cgul_exception_t local = NULL;

        /*
         * NOTE: We do NOT forward the call to cgul_rbtree__traverse_range()
         * because it causes problems if f throws a C++ exception.
         * Instead, we do all of this from C++ code so that C++
         * exceptions can unwind naturally.
         */
        
        if (first == NULL) {
            first = (cgul_rbtree_node_cxx*)cgul_rbtree__get_front(&local, obj);
        }
        if (last == NULL) {
            last = (cgul_rbtree_node_cxx*)cgul_rbtree__get_back(&local, obj);
        }

        /* Iterate over the nodes. */
        for (curr = first ; curr ; curr = next) {
            /* Get next now in case f deletes curr. */
            next = (cgul_rbtree_node_cxx*)
                       cgul_rbtree_node__get_next(&local,
                                                  (cgul_rbtree_node_t)curr);
            /* Call back. */
            if (!(*f)(this, curr, data)) {
                break;
            }
            /* We are done if we just processed the last node. */
            if (curr == last) {
                break;
            }
        }
    }

    /**
     * Get the underlying <code>cgul_rbtree</code> object.
     *
     * @return underlying object
     */
    virtual cgul_rbtree_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_rbtree</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_rbtree__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_rbtree_t
    take_obj()
    {
        cgul_rbtree_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_rbtree_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_rbtree__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_rbtree_cxx();
    cgul_rbtree_cxx(const cgul_rbtree_cxx&);
    cgul_rbtree_cxx& operator=(const cgul_rbtree_cxx&);

    /** Underlying <code>cgul_rbtree</code> object. */
    cgul_rbtree_t obj;

};


#endif  /*  CGUL_RBTREE_CXX_H  */
