#ifndef CGUL_STREAM_READER_H
#define CGUL_STREAM_READER_H

/**
 * The <code>cgul_stream_reader</code> class is an adaptor for one or
 * more <code>cgul_block_reader</code> instances.  It returns a stream
 * of characters read from each each block reader in turn.  If a
 * decoder is not explicitly set, <code>cgul_stream_reader</code> will
 * attempt to determine the correct Unicode decoder by looking for the
 * Unicode Byte-Order Mark (BOM).  If the stream is not Unicode, an
 * ASCII decoder is provided or the user can write their own.  If a
 * decoder is not provided and not explicitly set, UTF-8 is assumed.
 * For automatic detection to work correctly, the underlying block
 * must have a size of at least 4 bytes.  The default block size is
 * 16K. <p>
 *
 * @file
 * @author Paul Serice
 * @brief stream reader abstraction
 * @see cgul_block_reader
 * @see cgul_crlf_file
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_block_reader.h"
#include "cgul_wchar.h"
#include "cgul_wstring.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_stream_reader</code> instance. */
typedef struct cgul_stream_reader* cgul_stream_reader_t;

/**
 * Type definition for the callback function used by
 * <code>cgul_stream_reader</code> to decode and return the next block
 * of characters.  The function should decode up to
 * <code>size_encoded</code> bytes from <code>block_encoded</code>
 * starting with the <code>index_encoded</code> byte and storing the
 * index of the first byte that could not be decoded in
 * <code>index_encoded</code>.  If the entire block is decoded, then
 * <code>index_encoded</code> should be set to
 * <code>size_encoded</code> on return.  Up to
 * <code>size_decoded</code> wide characters should be placed in order
 * in <code>block_decoded</code>, and <code>count_decoded</code>
 * should be set to the number of characters that were decoded.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 */
typedef void
(*cgul_stream_reader__decode_t)(cgul_exception_t* cex,
                                unsigned char* block_encoded,
                                size_t size_encoded,
                                size_t* index_encoded,
                                cgul_wchar_t* block_decoded,
                                size_t size_decoded,
                                size_t* count_decoded);

/**
 * This function implements the
 * <code>cgul_stream_reader__decode_t</code> callback and can be used
 * to configure a <code>cgul_stream_reader</code> instance to adapt a
 * <code>cgul_block_reader</code> to return an ASCII character stream.
 * Clients do not typically call this function directly.  Instead,
 * they usually only have to pass it into a
 * <code>cgul_stream_reader</code> instance so the instance can call
 * this function.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__decode_ascii(cgul_exception_t* cex,
                                 unsigned char* block_encoded,
                                 size_t size_encoded,
                                 size_t* index_encoded,
                                 cgul_wchar_t* block_decoded,
                                 size_t size_decoded,
                                 size_t* count_decoded);

/**
 * This function implements the
 * <code>cgul_stream_reader__decode_t</code> callback and can be used
 * to configure a <code>cgul_stream_reader</code> instance to adapt a
 * <code>cgul_block_reader</code> to return a UTF-8 character stream.
 * Clients do not typically call this function directly.  Instead,
 * they usually only have to pass it into a
 * <code>cgul_stream_reader</code> instance so the instance can call
 * this function. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__decode_utf8(cgul_exception_t* cex,
                                unsigned char* block_encoded,
                                size_t size_encoded,
                                size_t* index_encoded,
                                cgul_wchar_t* block_decoded,
                                size_t size_decoded,
                                size_t* count_decoded);

/**
 * This function implements the
 * <code>cgul_stream_reader__decode_t</code> callback and can be used
 * to configure a <code>cgul_stream_reader</code> instance to adapt a
 * <code>cgul_block_reader</code> to return a UTF-16BE character
 * stream.  Clients do not typically call this function directly.
 * Instead, they usually only have to pass it into a
 * <code>cgul_stream_reader</code> instance so the instance can call
 * this function. <p>
 *
 * @note For UTF-16 streams (having a Unicode Byte-Order Mark (BOM)),
 * do not set an explicit decoder.  The
 * <code>cgul_stream_reader</code> instance will automatically detect
 * the BOM and set the correct decoder.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__decode_utf16be(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded);

/**
 * This function implements the
 * <code>cgul_stream_reader__decode_t</code> callback and can be used
 * to configure a <code>cgul_stream_reader</code> instance to adapt a
 * <code>cgul_block_reader</code> to return a UTF-16LE character
 * stream.  Clients do not typically call this function directly.
 * Instead, they usually only have to pass it into a
 * <code>cgul_stream_reader</code> instance so the instance can call
 * this function. <p>
 *
 * @note For UTF-16 streams (having a Unicode Byte-Order Mark (BOM)),
 * do not set an explicit decoder.  The
 * <code>cgul_stream_reader</code> instance will automatically detect
 * the BOM and set the correct decoder.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__decode_utf16le(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded);

/**
 * This function implements the
 * <code>cgul_stream_reader__decode_t</code> callback and can be used
 * to configure a <code>cgul_stream_reader</code> instance to adapt a
 * <code>cgul_block_reader</code> to return a UTF-32BE character
 * stream.  Clients do not typically call this function directly.
 * Instead, they usually only have to pass it into a
 * <code>cgul_stream_reader</code> instance so the instance can call
 * this function. <p>
 *
 * @note For UTF-32 streams (having a Unicode Byte-Order Mark (BOM)),
 * do not set an explicit decoder.  The
 * <code>cgul_stream_reader</code> instance will automatically detect
 * the BOM and set the correct decoder.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__decode_utf32be(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded);

/**
 * This function implements the
 * <code>cgul_stream_reader__decode_t</code> callback and can be used
 * to configure a <code>cgul_stream_reader</code> instance to adapt a
 * <code>cgul_block_reader</code> to return a UTF-32LE character
 * stream.  Clients do not typically call this function directly.
 * Instead, they usually only have to pass it into a
 * <code>cgul_stream_reader</code> instance so the instance can call
 * this function. <p>
 *
 * @note For UTF-32 streams (having a Unicode Byte-Order Mark (BOM)),
 * do not set an explicit decoder.  The
 * <code>cgul_stream_reader</code> instance will automatically detect
 * the BOM and set the correct decoder.
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__decode_utf32le(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded);

/**
 * Create a new <code>cgul_stream_reader</code> instance.  After this
 * method returns, <code>cgul_stream_reader__add_block_reader()</code>
 * should be called and <code>cgul_stream_reader__set_decoder()</code>
 * may be called before calling
 * <code>cgul_stream_reader__read()</code>.  The client is responsible
 * for calling <code>cgul_stream_reader__delete()</code> on the
 * pointer returned.  If an error occurs, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @return new <code>cgul_stream_reader</code> instance
 * @see cgul_stream_reader__set_decoder()
 */
extern CGUL_EXPORT cgul_stream_reader_t
cgul_stream_reader__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_stream_reader</code> instance.  After this
 * method returns, <code>cgul_stream_reader__set_decoder()</code> may
 * be called before calling <code>cgul_stream_reader__read()</code>.
 * This class does not take ownership of the block reader
 * <code>br</code> so the client is responsible for deleting it when
 * this class is no longer using it.  The client is responsible for
 * calling <code>cgul_stream_reader__delete()</code> on the pointer
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] br block reader to add
 * @return new <code>cgul_stream_reader</code> instance
 * @see cgul_block_reader
 * @see cgul_block_reader_file
 * @see cgul_block_reader_memory
 * @see cgul_stream_reader__set_decoder()
 */
extern CGUL_EXPORT cgul_stream_reader_t
cgul_stream_reader__new_from_block_reader(cgul_exception_t* cex,
                                          cgul_block_reader_t br);

/**
 * This method deletes the instance <code>sr</code> freeing all
 * internally allocated resources.  Clients must not attempt to use
 * <code>sr</code> after this method returns.
 *
 * @param[in] sr <code>cgul_stream_reader</code> instance
 */
extern CGUL_EXPORT void
cgul_stream_reader__delete(cgul_stream_reader_t sr);

/**
 * Reset the <code>cgul_stream_reader</code> instance <code>sr</code>
 * so it can be used with a new set of block readers.  This clears the
 * end of file (EOF), resets the line and column numbers to
 * <code>1</code>, clears the queue of block readers, and clears the
 * decoder reverting back to automatic detection of UTF-8, UTF-16, and
 * UTF-32 input.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 */
extern CGUL_EXPORT void
cgul_stream_reader__reset(cgul_exception_t* cex,
                          cgul_stream_reader_t sr);

/**
 * Read the next encoded character from the stream and return the
 * result as <code>*next_char</code> provided <code>next_char</code>
 * is not <code>NULL</code>.  This methods advances the character
 * stream.  If successful, <code>1</code> is returned.  If the end of
 * file is reached, <code>0</code> is returned and
 * <code>*next_char</code> is set to <code>CGUL_WCHAR__NUL</code>.  If
 * an error occurs, <code>0</code> is returned,
 * <code>*next_char</code> is set to <code>CGUL_WCHAR__NUL</code>, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[out] next_char next character
 * @return whether the read completed successfully
 */
extern CGUL_EXPORT int
cgul_stream_reader__read(cgul_exception_t* cex,
                         cgul_stream_reader_t sr,
                         cgul_wchar_t* next_char);

/**
 * Return whether the end of file (EOF) has been reached.  Calling
 * <code>cgul_stream_reader__reset()</code> will clear the EOF flag.
 * Putting a character back will also clear the EOF flag.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return whether the end of file has been reached
 * @see cgul_stream_reader__peek()
 */
extern CGUL_EXPORT int
cgul_stream_reader__eof(cgul_exception_t* cex,
                        cgul_stream_reader_t sr);

/**
 * Return whether line and column number tracking is enabled.  By
 * default, line and column number tracking is enabled.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return whether line and column number tracking is enabled
 */
extern CGUL_EXPORT int
cgul_stream_reader__get_line_and_column_tracking(cgul_exception_t* cex,
                                                 cgul_stream_reader_t sr);

/**
 * Set whether line and column number tracking is enabled.  By
 * default, line and column number tracking is enabled.  Disabling
 * line and column number tracking results in about a 10% performance
 * increase.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] is_tracking_enabled whether line and column number
 * tracking is enabled
 */
extern CGUL_EXPORT void
cgul_stream_reader__set_line_and_column_tracking(cgul_exception_t* cex,
                                                 cgul_stream_reader_t sr,
                                                 int is_tracking_enabled);

/**
 * Return the line number for the current character (i.e., the
 * character that will be returned by the next call to
 * <code>cgul_stream_reader__read()</code>).  The line number is
 * 1-based (i.e., the first line is assigned the number 1). <p>
 *
 * @note If line and column number tracking has been disabled,
 * <code>0</code> is returned.  (See see
 * cgul_stream_reader__set_line_and_column_tracking().)
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return current line number
 * @see cgul_stream_reader__get_line_number_previous()
 * @see cgul_stream_reader__set_line_and_column_tracking()
 */
extern CGUL_EXPORT size_t
cgul_stream_reader__get_line_number_current(cgul_exception_t* cex,
                                            cgul_stream_reader_t sr);

/**
 * Return the column number for the current character (i.e., the
 * character that will be returned by the next call to
 * <code>cgul_stream_reader__read()</code>).  The column number is
 * 1-based (i.e., the first column is assigned the number 1).
 *
 * @note If line and column number tracking has been disabled,
 * <code>0</code> is returned.  (See see
 * cgul_stream_reader__set_line_and_column_tracking().)
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return current column number or <code>0</code>
 * @see cgul_stream_reader__get_column_number_previous()
 * @see cgul_stream_reader__set_line_and_column_tracking()
 */
extern CGUL_EXPORT size_t
cgul_stream_reader__get_column_number_current(cgul_exception_t* cex,
                                              cgul_stream_reader_t sr);

/**
 * Return the line number for the previous character (i.e., the
 * character that was returned by the previous call to
 * <code>cgul_stream_reader__read()</code>).  The line number is
 * 1-based (i.e., the first line is assigned the number 1).  If
 * <code>0</code> is returned, it means the first character has not
 * yet been read. <p>
 *
 * @note If line and column number tracking has been disabled,
 * <code>0</code> is returned.  (See see
 * cgul_stream_reader__set_line_and_column_tracking().)
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return current line number
 * @see cgul_stream_reader__get_line_number_current()
 * @see cgul_stream_reader__set_line_and_column_tracking()
 */
extern CGUL_EXPORT size_t
cgul_stream_reader__get_line_number_previous(cgul_exception_t* cex,
                                             cgul_stream_reader_t sr);

/**
 * Return the column number for the previous character (i.e., the
 * character that was returned by the previous call to
 * <code>cgul_stream_reader__read()</code>) or <code>0</code> if no
 * character has been read yet.  The column number is 1-based (i.e.,
 * the first column is assigned the number 1).
 *
 * @note If line and column number tracking has been disabled,
 * <code>0</code> is returned.  (See see
 * cgul_stream_reader__set_line_and_column_tracking().)
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return current column number or <code>0</code>
 * @see cgul_stream_reader__get_column_number_current()
 * @see cgul_stream_reader__set_line_and_column_tracking()
 */
extern CGUL_EXPORT size_t
cgul_stream_reader__get_column_number_previous(cgul_exception_t* cex,
                                               cgul_stream_reader_t sr);

/**
 * Return whether end-of-line (EOL) character sequences are
 * normalized.  If normalization is enabled, "\r" and "\r\n" sequences
 * are normalized to "\n".  All other characters pass though without
 * replacement.  By default EOL normalization is enabled.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return whether EOLs are normalized
 */
extern CGUL_EXPORT int
cgul_stream_reader__get_normalize_eol(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr);

/**
 * Set whether end-of-line (EOL) character sequences are normalized.
 * If normalization is enabled, "\r" and "\r\n" sequences are
 * normalized to "\n".  All other characters passed though without
 * replacement.  By default EOL normalization is enabled.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] normalize whether to normalize EOLs
 */
extern CGUL_EXPORT void
cgul_stream_reader__set_normalize_eol(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr,
                                      int normalize);

/**
 * Add the block reader <code>br</code> to the queue of block readers.
 * The block reader must end on an encoded character boundary (i.e.,
 * encoded characters must not span across block readers).  Characters
 * are streamed from the front block reader on the queue.  When end of
 * file is reach for the front block reader, it is removed from the
 * queue, and characters are then streamed from the next block reader
 * in the queue.  Once a block reader is added to the queue, it should
 * not be modified externally.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] br block reader to add
 * @see cgul_block_reader
 * @see cgul_block_reader_file
 * @see cgul_block_reader_memory
 */
extern CGUL_EXPORT void
cgul_stream_reader__add_block_reader(cgul_exception_t* cex,
                                     cgul_stream_reader_t sr,
                                     cgul_block_reader_t br);

/**
 * Return the size of the block used to read from the block reader.
 * The default block size is 16K.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return size of block used to read from the block reader
 */
extern CGUL_EXPORT size_t
cgul_stream_reader__get_block_size(cgul_exception_t* cex,
                                   cgul_stream_reader_t sr);

/**
 * Set the block size <code>block_size</code> used to read from the
 * block reader.  The default block size is 16K.  If a block reader
 * has been added, the block size can only be increased.  If the block
 * size is not large enough to decode any character in the input
 * stream, the decoder will throw an exception when
 * <code>cgul_stream_reader__read()</code>,
 * <code>cgul_stream_reader__peek()</code>, or
 * <code>cgul_stream_reader__accept()</code> (or related methods) is
 * called.  If successful, <code>1</code> is returned; otherwise,
 * <code>0</code> is returned.  If an error occurs, <code>0</code> is
 * returned, and an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] block_size block size
 * @return whether the block size was set
 */
extern CGUL_EXPORT int
cgul_stream_reader__set_block_size(cgul_exception_t* cex,
                                   cgul_stream_reader_t sr,
                                   size_t block_size);

/**
 * Return the decoder used to decode characters from the block reader.
 * If a decoder has not been set, <code>NULL</code> is returned.  If a
 * decoder is not explicitly set, <code>cgul_stream_reader</code> will
 * attempt to determine the correct Unicode decoder by looking for the
 * Unicode Byte-Order Mark (BOM).  If the stream is not Unicode, an
 * ASCII decoder is provided or the user can write their own.  If a
 * decoder is not provided and not explicitly set, UTF-8 is assumed.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return decoder
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT cgul_stream_reader__decode_t
cgul_stream_reader__get_decoder(cgul_exception_t* cex,
                                cgul_stream_reader_t sr);

/**
 * Set the decoder <code>decoder</code> used to decode characters from
 * the block reader.  If a decoder is not explicitly set,
 * <code>cgul_stream_reader</code> will attempt to determine the
 * correct Unicode decoder by looking for the Unicode Byte-Order Mark
 * (BOM).  If the stream is not Unicode, an ASCII decoder is provided
 * or the user can write their own.  If a decoder is not provided and
 * not explicitly set, UTF-8 is assumed.  For automatic detection to
 * work, the underlying block must have a size of at least 4 bytes.
 * The default block size is 16K. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] decoder decoder
 * @see cgul_stream_reader__decode_t
 */
extern CGUL_EXPORT void
cgul_stream_reader__set_decoder(cgul_exception_t* cex,
                                cgul_stream_reader_t sr,
                                cgul_stream_reader__decode_t decoder);

/**
 * Return the presentation name for the stream reader.  If a
 * presentation name has been explicitly set using
 * <code>cgul_stream_reader__set_presentation_name()</code>, it is
 * returned.  If no presentation name has been explicitly set and only
 * one block reader has been added, the presentation name for the
 * block reader is returned (and should not be cached by the client).
 * If no presentation name has been explicitly set and more than one
 * block reader has been added, <code>"MULTIPART"</code> is returned.
 * If no block reader has been set, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @return presentation name
 */
extern CGUL_EXPORT const char*
cgul_stream_reader__get_presentation_name(cgul_exception_t* cex,
                                          cgul_stream_reader_t sr);

/**
 * Set the presentation name for the stream reader. If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] presentation_name presentation name
 */
extern CGUL_EXPORT void
cgul_stream_reader__set_presentation_name(cgul_exception_t* cex,
                                          cgul_stream_reader_t sr,
                                          const char* presentation_name);

CGUL_END_C

#endif  /*  CGUL_STREAM_READER_H  */
