#ifndef CGUL_WCHAR_H
#define CGUL_WCHAR_H

/**
 * This file defines <code>cgul_wchar_t</code> which is a 32-bit,
 * wide-character type.  It also defines the basic set of functions
 * that manipulate raw, 32-bit, wide-character strings. <p>
 *
 * @file
 * @author Paul Serice
 * @brief cgul 32-bit wide-character support
 * @see cgul_wstring
 * @see cgul_unicode
 * @see cgul_hchar
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/**
 * The <code>cgul_wchar_t</code> typedef always defines a 32-bit wide
 * character which means it is always large enough to hold any Unicode
 * value even on machines that have a native <code>wchar_t</code> that
 * is only 16-bits wide. <p>
 *
 * If you need to convert between <code>cgul_wchar_t</code> and
 * <code>wchar_t</code>, there are routines in
 * <code>cgul_unicode</code> that convert between UTF-32 and UTF-8.
 * You can then use an operating system dependent mechanism for
 * converting from UTF-8 to <code>wchar_t</code>.  Alternatively, it
 * may just be easier to write simple loop that copies each character
 * from a <code>cgul_wchar_t</code> string to a <code>wchar_t</code>
 * string or vice versa. <p>
 */
typedef cgul_uint32_t cgul_wchar_t;

/**
 * The NUL character in a <code>cgul_wchar_t</code> string.
 */
#define CGUL_WCHAR__NUL ((cgul_wchar_t)0)

/**
 * The Unicode Byte-Order Mark (BOM) character.
 */
#define CGUL_WCHAR__BOM ((cgul_wchar_t)0xfeff)

/**
 * The maximum number of bytes needed to hold any UTF-8 multi-byte
 * sequence that represents one Unicode character.
 */
#define CGUL_WCHAR__MB_LEN_MAX (4)

/**
 * The most annoying thing about using <code>cgul_wchar_t</code>
 * instead of <code>wchar_t</code> is that we lose support for the
 * compiler automatically generating wide-character strings by simply
 * prepending 'L' to the string literal.  Because we often need the
 * empty string, this class provides one as a convenience.
 */
extern CGUL_EXPORT const cgul_wchar_t CGUL_WCHAR__EMPTY_STRING[1];

/**
 * This function returns the number of 32-bit wide-characters in
 * <code>ws</code>.  The return value is determined by counting all
 * the wide characters that come before the trailing NUL wide
 * character.  The caller is responsible for insuring that
 * <code>ws</code> is NUL-terminated.
 *
 * @param[in] cex c-style exception
 * @param[in] ws 32-bit wide-character string
 * @return number of 32-bit wide characters in <code>ws</code>
 */
extern CGUL_EXPORT size_t
cgul_wchar__wcslen(cgul_exception_t* cex,
                   const cgul_wchar_t* ws);

/**
 * This function copies the 32-bit wide characters in <code>src</code>
 * to <code>dst</code>.  The caller is responsible for making sure
 * that <code>src</code> is NUL-terminated and that <code>dst</code>
 * is large enough to hold all the characters in <code>src</code>
 * including the trailing NUL character.
 *
 * @param[in] cex c-style exception
 * @param[in] dst destination 32-bit wide-character string
 * @param[in] src source 32-bit wide-character string
 */
extern CGUL_EXPORT void
cgul_wchar__wcscpy(cgul_exception_t* cex,
                   cgul_wchar_t* dst,
                   const cgul_wchar_t* src);

/**
 * Perform a case-significant, 32-bit, wide-character string
 * comparison of <code>ws1</code> and <code>ws2</code>.  Return -1, 0,
 * or 1 if <code>ws1</code> is less than, equal to, or greater than
 * <code>ws2</code> respectively. <p>
 *
 * For simplicity, this function compares strings stricly by the
 * ordinal value of each character which should be sufficient, for
 * example, when comparing strings when inserting them as keys into a
 * <code>cgul_rbtree</code>; however, the comparison results are not
 * likely to be the same as what <code>strcoll()</code> would return
 * if used on the UTF-8 version of the strings. <p>
 *
 * @note This function does not take a <code>cgul_exception</code>
 * object as its first parameter in order to make it easier to use the
 * function as the comparison function of <code>cgul_rbtree</code>
 * objects. <p>
 *
 * @param[in] ws1 left-hand side
 * @param[in] ws2 right-hand side
 * @return -1, 0, or 1 if <code>ws1</code> is less than, equal to, or
 * greater than <code>ws2</code> respectively
 * @see cgul_wstring__compare()
 */
extern CGUL_EXPORT int
cgul_wchar__wcscmp(const cgul_wchar_t* ws1,
                   const cgul_wchar_t* ws2);

/**
 * Return whether the 32-bit, wide character <code>wc</code> is
 * considered to be white-space.  To avoid locale dependencies,
 * white-space is defined as any character from the following list:
 * ' ', '\\t', '\\n', '\\r', '\\f', '\\v'
 *
 * @param[in] cex c-style exception
 * @param[in] wc 32-bit, wide character
 * @return whether the character is white-space
 */
extern CGUL_EXPORT int
cgul_wchar__isspace(cgul_exception_t* cex,
                    cgul_wchar_t wc);

CGUL_END_C

#endif  /*  CGUL_WCHAR_H  */
