#ifndef CGUL_WSTRING_H
#define CGUL_WSTRING_H

/**
 * Simple 32-bit wide-character string class.  Unlike the
 * <code>cgul_string</code> class, this class and
 * <code>cgul_hstring</code> co-opt the role of <code>wchar_t</code>
 * strings in C because, as a practical matter, <code>wchar_t</code>
 * is not portable. <p>
 *
 * As a result, this class defines <code>cgul_wchar_t</code> to
 * always be an unsigned 32-bit value.  This means you cannot portably
 * use any of the standard C functions with these strings, but you can
 * convert these strings to UTF-8 and use the standard
 * <code>char*</code> functions. <p>
 *
 * If you need to read these strings from file, I recommend that the
 * file be UTF-8 encoded.  You can then read the strings from file
 * using <code>cgul_crlf_file</code>.  You can then convert them from
 * UTF-8 using <code>cgul_unicode__mbstowcs()</code>. <p>
 *
 * If you need to write these strings to file, I recommend that you
 * convert the string to UTF-8 using
 * <code>cgul_unicode__wcstombs()</code>.  You can then write the
 * resulting <code>char*</code> string to file using
 * <code>cgul_fprintf()</code>. <p>
 *
 * @file
 * @author Paul Serice
 * @brief 32-bit wide-character string class
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_wchar.h"

CGUL_BEGIN_C

/** Opaque pointer to <code>cgul_wstring</code> instance. */
typedef struct cgul_wstring* cgul_wstring_t;

/**
 * Create a new <code>cgul_wstring</code> object that holds an empty
 * string.  The caller is responsible for freeing the object by
 * calling <code>cgul_wstring__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * The following three methods implement the "cacheable" interface
 * defined by <code>cgul_cache</code>:
 *
 * <pre>
 *
 *     cgul_wstring__new()
 *     cgul_wstring__delete()
 *     cgul_wstring__clear()
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT cgul_wstring_t
cgul_wstring__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_wstring</code> object that is initialized
 * with a copy of <code>value</code>.  The caller is responsible for
 * freeing the object by calling <code>cgul_wstring__delete()</code>.
 * If memory cannot be allocated, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] value initial value
 * @return new <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT cgul_wstring_t
cgul_wstring__new_from_wcstring(cgul_exception_t* cex,
                                const cgul_wchar_t* value);

/**
 * cgul_wstring__delete() frees all internally allocated memory.  Do
 * not try to use <code>ws</code> after calling this method.
 *
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__delete(cgul_wstring_t ws);

/**
 * This method returns a pointer to the internal wide-character
 * string.  You are allowed to manually set individual characters in
 * the string.  Just remember to call
 * <code>cgul_wstring__externally_adjusted()</code> afterward.
 * Compare this method to <code>cgul_wstring__take_value()</code>.
 * 
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @return underlying wide-character string
 * @see cgul_wstring__externally_adjusted
 */
extern CGUL_EXPORT cgul_wchar_t*
cgul_wstring__get_value(cgul_exception_t* cex,
                        cgul_wstring_t ws);

/**
 * This method lets you take ownership of the underlying
 * wide-character string which is returned.  Immediately upon taking
 * ownership, the internal state of this class is reset including
 * resetting a new underlying wide-character string to be empty.  One
 * use of this function is to build up a string using
 * <code>cgul_wstring</code> that you then pass to
 * <code>cgul_hash</code> or <code>cgul_rbtree</code>.  Compare this
 * function with <code>cgul_wstring__get_value()</code>.
 * 
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @return underlying wide-character string
 */
extern CGUL_EXPORT cgul_wchar_t*
cgul_wstring__take_value(cgul_exception_t* cex,
                         cgul_wstring_t ws);

/**
 * Set the value of the underlying wide-character string by copying
 * characters from <code>value</code> to <code>s</code>.  The caller
 * is responsible for making sure <code>value</code> is terminated
 * with a <code>CGUL_WCHAR__NUL</code> character.  This method does
 * not uses <code>strcpy()</code> or <code>memcpy()</code>.  This
 * means that <code>value</code> and <code>s</code> can overlap.
 * 
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] value new value
 */
extern CGUL_EXPORT void
cgul_wstring__set_value(cgul_exception_t* cex,
                        cgul_wstring_t ws,
                        const cgul_wchar_t* value);

/**
 * This method returns the character in <code>ws</code> indicated by
 * <code>index</code> where <code>index</code> is zero-based.  If
 * <code>index</code> is out of bounds, <code>(cgul_wchar_t)0</code>
 * is returned and an exception is thrown.
 * 
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] index index into <code>ws</code>
 * @return wide character at <code>index</code>
 */
extern CGUL_EXPORT cgul_wchar_t
cgul_wstring__get_value_at(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           size_t index);

/**
 * This method sets the wide character in <code>ws</code> indicated by
 * <code>index</code> where <code>index</code> is zero-based to
 * <code>wc</code>.  If <code>index</code> is out of bounds, an
 * exception is thrown.
 * 
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] index index into <code>s</code>
 * @param[in] wc character
 */
extern CGUL_EXPORT void
cgul_wstring__set_value_at(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           size_t index,
                           cgul_wchar_t wc);

/**
 * This method is equivalent to calling
 * <code>cgul_wstring__set_value()</code> where <code>value</code> is
 * an empty string.  This method is designed to be used as the
 * <code>cgul_cache_restructor_t</code> method.  This means it is not
 * allowed to throw an exception.
 *
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__clear(cgul_wstring_t ws);

/**
 * This method converts the string to a <code>signed char</code>.
 * Typically, "base" will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading character in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 * @see cgul_wstring__get_value_at()
 */
extern CGUL_EXPORT signed char
cgul_wstring__to_schar(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       int base);

/**
 * This method converts the string to an <code>unsigned char</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 * @see cgul_wstring__get_value_at()
 */
extern CGUL_EXPORT unsigned char
cgul_wstring__to_uchar(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       int base);

/**
 * This method converts the string to a <code>short</code>.
 * Typically, "base" will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading character in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT short int
cgul_wstring__to_short(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       int base);

/**
 * This method converts the string to an <code>unsigned short</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT unsigned short int
cgul_wstring__to_ushort(cgul_exception_t* cex,
                        cgul_wstring_t ws,
                        int base);

/**
 * This method converts the string to an <code>int</code>.  Typically,
 * "base" will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading character in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT int
cgul_wstring__to_int(cgul_exception_t* cex,
                     cgul_wstring_t ws,
                     int base);

/**
 * This method converts the string to an <code>unsigned</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT unsigned int
cgul_wstring__to_unsigned(cgul_exception_t* cex,
                          cgul_wstring_t ws,
                          int base);

/**
 * This method converts the string to a <code>long</code>.  Typically,
 * <code>base</code> will be set to 0 to allow conversion for decimal,
 * hexadecimal, and octal based on the leading characters in the
 * string.  The string cannot have any extraneous characters.  If it
 * does, the conversion will fail.  On success, the converted value is
 * returned; otherwise, 0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT long int
cgul_wstring__to_long(cgul_exception_t* cex,
                      cgul_wstring_t ws,
                      int base);

/**
 * This method converts the string to an <code>unsigned long</code>.
 * Typically, <code>base</code> will be set to 0 to allow conversion
 * for decimal, hexadecimal, and octal based on the leading characters
 * in the string.  The string cannot have any extraneous characters.
 * If it does, the conversion will fail.  On success, the converted
 * value is returned; otherwise, 0 is returned and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] base base to use for the conversion
 * @return converted value
 */
extern CGUL_EXPORT unsigned long int
cgul_wstring__to_unsigned_long(cgul_exception_t* cex,
                               cgul_wstring_t ws,
                               int base);

/**
 * This method converts the string to a <code>double</code>.  The
 * string cannot have any extraneous characters.  If it does, the
 * conversion will fail.  On success, the converted value is returned;
 * otherwise, 0.0 is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @return converted value
 */
extern CGUL_EXPORT double
cgul_wstring__to_double(cgul_exception_t* cex,
                        cgul_wstring_t ws);

/**
 * Appends <code>value</code> to <code>ws</code>.  The implementation
 * of this method is such that it is possible for <code>value</code>
 * to be a substring of <code>ws</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] value value to append
 */
extern CGUL_EXPORT void
cgul_wstring__append(cgul_exception_t* cex,
                     cgul_wstring_t ws,
                     const cgul_wchar_t* value);

/**
 * Append a block of non-<code>CGUL_WCHAR__NUL</code> characters.  The
 * <code>block_size</code> is the number of characters (not bytes) in
 * the block.  The caller should verify that there are no embedded NUL
 * characters in the string.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] block block to append
 * @param[in] block_size size of block to append
 */
extern CGUL_EXPORT void
cgul_wstring__append_block(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           const void* block,
                           size_t block_size);

/**
 * Append <code>wc</code> to <code>ws</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] wc wide character to append
 */
extern CGUL_EXPORT void
cgul_wstring__append_char(cgul_exception_t* cex,
                          cgul_wstring_t ws,
                          cgul_wchar_t wc);


/**
 * Inserts <code>value</code> into <code>ws</code> at
 * <code>index</code> which can range from 0 to the length (inclusive)
 * of <code>ws</code>.  If <code>index</code> is out of bounds or if
 * memory cannot be allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] index index in <code>s</code> where <code>value</code> is inserted
 * @param[in] value value to insert
 * @param[in] value_length length of value
 */
extern CGUL_EXPORT void
cgul_wstring__insert(cgul_exception_t* cex,
                     cgul_wstring_t ws,
                     size_t index,
                     const cgul_wchar_t* value,
                     size_t value_length);

/**
 * Inserts <code>value</code> into <code>ws</code> at
 * <code>index</code> which can range from 0 to the length (inclusive)
 * of <code>ws</code>.  If <code>index</code> is out of bounds or if
 * memory cannot be allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] index index in <code>s</code> where <code>value</code> is inserted
 * @param[in] value value to insert
 */
extern CGUL_EXPORT void
cgul_wstring__insert_char(cgul_exception_t* cex,
                          cgul_wstring_t ws,
                          size_t index,
                          cgul_wchar_t value);

/**
 * Pad the current string in <code>ws</code> such that it is at least
 * <code>width</code> characters wide using the <code>pad</code>
 * character.  <code>ws</code> will be right aligned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] width padding width
 * @param[in] pad wide character to use for padding
 */
extern CGUL_EXPORT void
cgul_wstring__pad_right(cgul_exception_t* cex,
                        cgul_wstring_t ws,
                        size_t width,
                        cgul_wchar_t pad);

/**
 * Pad the current string in <code>ws</code> such that it is at least
 * <code>width</code> characters wide using the <code>pad</code>
 * character.  <code>ws</code> will be left aligned.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] width padding width
 * @param[in] pad wide character to use for padding
 */
extern CGUL_EXPORT void
cgul_wstring__pad_left(cgul_exception_t* cex,
                       cgul_wstring_t ws,
                       size_t width,
                       cgul_wchar_t pad);

/**
 * Trim the string by removing leading and trailing white-space where
 * white-space is definded by <code>cgul_wchar__isspace()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__trim(cgul_exception_t* cex,
                   cgul_wstring_t ws);

/**
 * Trim the string by removing leading white-space where white-space
 * is definded by <code>cgul_wchar__isspace()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__trim_left(cgul_exception_t* cex,
                        cgul_wstring_t ws);

/**
 * Trim the string by removing trailing white-space where white-space
 * is definded by <code>cgul_wchar__isspace()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__trim_right(cgul_exception_t* cex,
                         cgul_wstring_t ws);

/**
 * Reverse the order of the characaters in <code>ws</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__reverse(cgul_exception_t* cex,
                      cgul_wstring_t ws);

/**
 * This method returns the length of the string.  This is the number
 * of wide characters in the string, not the number of bytes.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @return length of the string
 */
extern CGUL_EXPORT size_t
cgul_wstring__get_length(cgul_exception_t* cex,
                         cgul_wstring_t ws);

/**
 * This method can be used to set the length of <code>ws</code>.  It
 * can only be used to make <code>ws</code> shorter.  If it is used in
 * an attempt to make <code>ws</code> longer, an exception is thrown
 * and zero is returned.  If the length of <code>ws</code> is
 * successfully changed, the previous length of the string is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] length new, shorter length of the string
 * @return previous length of the string
 */
extern CGUL_EXPORT size_t
cgul_wstring__set_length(cgul_exception_t* cex,
                         cgul_wstring_t ws,
                         size_t length);

/**
 * If you manually alter the underlying buffer by modifying the
 * <code>cgul_wchar_t*</code> string returned by
 * <code>cgul_wstring__get_value()</code>, you'll need to manually
 * have the class reset its internal state.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @see cgul_wstring__get_value
 */
extern CGUL_EXPORT void
cgul_wstring__externally_adjusted(cgul_exception_t* cex,
                                  cgul_wstring_t ws);

/**
 * Trim the string by returning excess capacity to the heap.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__trim_capacity(cgul_exception_t* cex,
                            cgul_wstring_t ws);

/**
 * Get the minimum capacity.
 *
 * @param[in] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @return minimum capacity
 */
extern CGUL_EXPORT size_t
cgul_wstring__get_minimum_capacity(cgul_exception_t* cex,
                                   cgul_wstring_t ws);

/**
 * Set the minimum capacity.  The default value is <code>128</code>
 * which means there is always capacity for 127 characters plus one
 * NUL-terminator.  Decreasing the capacity will decrease the amount
 * of space used and (typically) also decrease performance.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] minimum_capacity minimum capacity
 */
extern CGUL_EXPORT void
cgul_wstring__set_minimum_capacity(cgul_exception_t* cex,
                                   cgul_wstring_t ws,
                                   size_t minimum_capacity);

/**
 * Perform a case-significant wide-character string comparison of
 * <code>ws1</code> and <code>ws2</code>.  Return -1, 0, or 1 if
 * <code>ws1</code> is less than, equal to, or greater than
 * <code>ws2</code> respectively. <p>
 *
 * For simplicity, this function compares strings stricly by the
 * ordinal value of each character which should be sufficient, for
 * example, when inserting entire <code>cgul_wstring</code> objects as
 * keys into a <code>cgul_rbtree</code> object; however, the
 * comparison results are not likely to be the same as what
 * <code>strcoll()</code> would return if used on the UTF-8 version of
 * the strings. <p>
 *
 * @note Use <code>cgul_wchar__wcscmp()</code> instead if inserting
 * raw <code>cgul_wchar</code> strings as keys into a
 * <code>cgul_rbtree</code> object. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] ws1 left-hand side
 * @param[in] ws2 right-hand side
 * @return -1, 0, or 1 if <code>ws1</code> is less than, equal to, or
 * greater than <code>ws2</code> respectively
 * @see cgul_wchar__wcscmp()
 */
extern CGUL_EXPORT int
cgul_wstring__compare(cgul_exception_t* cex,
                      cgul_wstring_t ws1,
                      cgul_wstring_t ws2);

/**
 * Swap the underlying data for <code>ws1</code> and <code>ws2</code>.
 * For large strings, this should be much faster than trying to do the
 * same thing using <code>cgul_wstring__set_value()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] ws1 <code>cgul_wstring</code> instance
 * @param[in] ws2 <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__swap(cgul_exception_t* cex,
                   cgul_wstring_t ws1,
                   cgul_wstring_t ws2);


CGUL_END_C

#endif  /*  CGUL_WSTRING_H  */
