#include "cgul_bcd_p.h"
#include "cgul_big_integer_p.h"
#include "cgul_string_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"

/**
 * Macro used in several places in <code>cgul_bcd__divide()</code> to
 * lazily create a table of products that are the result of
 * multiplying a single <code>digit</code> by the
 * <code>divisor</code>.  <code>divisor</code> should be a
 * <code>cgul_bcd</code> and <code>digit</code> should be an integral
 * type that holds a single decimal digit than can be type cast to a
 * <code>char</code>.
 *
 * @param[in] divisor divisor
 * @param[in] digit digit
 */
#define CGUL_BCD__DIVIDE__LAZILY_CREATE_TABLE_OF_PRODUCTS(divisor, digit) \
do { \
    /* Lazily create the table of products. */ \
    if (table[(digit)] == NULL) { \
        table[(digit)] = cgul_bcd__new(cex); \
        cgul_bcd__multiply_by_single_digit(cex, \
                                           table[(digit)], \
                                           (divisor), \
                                           (char)(digit)); \
        if (*cex) { \
            goto out; \
        } \
    } \
} while (0)

/**
 * This method removes all leading zeros from <code>bcd</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] bcd <code>cgul_bcd</code> instance
 */
static void
cgul_bcd__remove_leading_zeros(cgul_exception_t* cex,
                               cgul_bcd_t bcd)
{
    size_t i = 0;
    size_t digits_len = 0;
    char* digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Get the array of digits and its length. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    /* Find the first leading digit that is not a zero except that we
     * allow "digits[0]" to be zero so that "bcd" can be zero. */
    for (i = digits_len - 1 ; i > 0 ; --i) {
        if (digits[i]) {
            /* Found a non-zero digit. */
            break;
        }
    }

    /* If we have leading zeros, trim them. */
    if (i != digits_len - 1) {
        cgul_string__set_length(cex, bcd->digits, i + 1);
    }

 out:
    return;
}

/**
 * This function converts a binary ascii character into a long and
 * returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_bcd__get_binary_digit(cgul_exception_t* cex,
                           char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * This function converts an octal ascii character into a long and
 * returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_bcd__get_octal_digit(cgul_exception_t* cex,
                          char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * This function converts a decimal ascii character into a long and
 * returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_bcd__get_decimal_digit(cgul_exception_t* cex,
                            char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch (c) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    case '8':
        rv = 8;
        break;
    case '9':
        rv = 9;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * This function converts a hexadecimal ascii character into a long
 * and returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_bcd__get_hexadecimal_digit(cgul_exception_t* cex,
                                char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    case '8':
        rv = 8;
        break;
    case '9':
        rv = 9;
        break;
    case 'a':
    case 'A':
        rv = 10;
        break;
    case 'b':
    case 'B':
        rv = 11;
        break;
    case 'c':
    case 'C':
        rv = 12;
        break;
    case 'd':
    case 'D':
        rv = 13;
        break;
    case 'e':
    case 'E':
        rv = 14;
        break;
    case 'f':
    case 'F':
        rv = 15;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * Pass in the full string in <code>value</code>.  Pass in the length
 * of <code>value</code> in <code>value_length</code>.  Pass in the
 * index in <code>*i</code> of the first character where you want this
 * function to start looking for a prefix that determines the base.
 * The index is to help you scan pass leading white space in
 * <code>value</code>.  This function returns 16 if the string at
 * <code>&value[*i]</code> starts with "0x" or "0X".  Returns 2 if it
 * starts with "0b" or "0B".  Returns 8 if it starts with "0" followed
 * by a digit between 0 and 7.  It otherwise returns 10.  Also, it
 * will increment <code>*i</code> to index one past the last character
 * in the prefix that determines the base.
 *
 * @param[in] cex c-style exception
 * @param[in] value value of big integer as a string
 * @param[in] value_length length of <code>value</code>
 * @param[in,out] index into <code>value</code>
 * @return base to use for <code>value</code>
 */
static long
cgul_bcd__determine_base(cgul_exception_t* cex,
                         const char* value,
                         size_t value_length,
                         size_t* i)
{
    long rv = 10;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(value);
    assert(i);

    if( value[*i] == '0' ) {
        size_t next = *i + 1;
        if( next < value_length ) {
            if( (value[next] == 'x') || (value[next] == 'X') ) {
                rv = 16;
                *i += 2;
            } else if( (value[next] == 'b') || (value[next] == 'B') ) {
                rv = 2;
                *i += 2;
            } else {
                long digit =
                    cgul_bcd__get_octal_digit(cex, value[next]);
                if (digit != -1) {
                    rv = 8;
                    ++*i;
                }
            }
        }
    }

 out:
    return rv;
}

/**
 * Skip leading white space.
 *
 * @param[in] cex c-style exception
 * @param[in] s string
 * @param[in] s_length length of <code>s</code>
 * @param[in,out] index into <code>s</code>
 */
static void
cgul_bcd__skip_white_space(cgul_exception_t* cex,
                           const char* s,
                           size_t s_length,
                           size_t* i)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(i);

    for ( ; *i < s_length ; ++*i) {
        if (!isspace((int)s[*i])) {
            break;
        }
    }

 out:
    return;
}

/**
 * Compare just the magnitudes of <code>lhs</code> and
 * <code>rhs</code> after effectively shifting <code>rhs</code> by
 * <code>shift_count</code> digits to the left.  For a normal
 * comparison of two numbers, <code>shift_count</code> should be
 * zero. <p>
 *
 * This function returns less than zero if lhs is less than rhs, zero
 * if lhs equals rhs, and greater than zero if lhs is greater than
 * rhs.  This is NOT typically how you want to compare
 * <code>cgul_bcd</code> types.  Use
 * <code>cgul_bcd__compare()</code> instead because it accounts for
 * the signs of <code>lhs</code> and <code>rhs</code>. <p>
 *
 * The <code>shift_count</code> parameter was added because shifting a
 * <code>cgul_bcd</code> object left is expensive (because the digits
 * are stored in an array which means shifting left requires copying
 * all digits), and <code>cgul_bcd__divide()</code> needs to
 * repeatedly shift left.  So rather than performing an expensive
 * shift left before performing the comparison, this method uses the
 * <code>shift_count</code> parameter to calculate the comparison
 * without actually having to shift left.
 *
 * @param[in] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 * @param[in] shift_count shift count
 * @return comparison of <code>lhs</code> and <code>rhs</code>
 */
static int
cgul_bcd__compare_magnitudes(cgul_exception_t* cex,
                             cgul_bcd_t lhs,
                             cgul_bcd_t rhs,
                             size_t shift_count)
{
    int rv = 0;
    size_t i = 0;
    size_t j = 0;
    size_t lhs_digits_len = 0;
    size_t rhs_digits_len = 0;
    char* lhs_digits = NULL;
    char* rhs_digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* Normalize the lengths so you can make the comparison. */
    cgul_bcd__remove_leading_zeros(cex, lhs);
    cgul_bcd__remove_leading_zeros(cex, rhs);

    /* Get the length of the arrays that hold the digits. */
    lhs_digits_len = cgul_string__get_length(cex, lhs->digits);
    rhs_digits_len = cgul_string__get_length(cex, rhs->digits);

    /* If one side has more digits, it has the greater magnitude. */
    if (lhs_digits_len != rhs_digits_len + shift_count) {
        if (lhs_digits_len > rhs_digits_len + shift_count) {
            rv = 1;
        } else {
            rv = -1;
        }
        goto out;
    }

    /* Get the arrays that hold the digits. */
    lhs_digits = cgul_string__get_value(cex, lhs->digits);
    rhs_digits = cgul_string__get_value(cex, rhs->digits);

    /* The magnitudes are roughly the same (lhs_digits_len ==
     * rhs_digits_len + shift_count).  We have to inspect each digit
     * individually at this point starting with the most-significant
     * digit. */
    i = lhs_digits_len;
    j = rhs_digits_len;
    for ( ; (i > 0) && (j > 0) ; --i, --j) {
        if (lhs_digits[i - 1] != rhs_digits[j - 1]) {
            if (lhs_digits[i - 1] > rhs_digits[j - 1]) {
                rv = 1;
            } else {
                rv = -1;
            }
            goto out;
        }
    }

    /* We have exhausted all the digits in rhs proper, but we need to
     * seamlessly shift zeros into rhs in order to satisfy
     * shift_count. */
    for ( ; i > 0 ; --i) {
        if (lhs_digits[i - 1] > 0) {
            rv = 1;
            goto out;
        }
    }

 out:
    return rv;
}

/**
 * Add the magnitudes of rhs1 and rhs2 after effectively shifting
 * <code>rhs2</code> by <code>shift_count</code> digits to the left.
 * Put the value in lhs without changing the sign of lhs.  It is safe
 * for <code>rhs1</code> or <code>rhs2</code> and <code>lhs</code> to
 * point to the same underlying <code>cgul_bcd</code> object.  For
 * normal addition, <code>shift_count</code> should be zero. <p>
 *
 * The <code>shift_count</code> parameter tells this method how far to
 * shift <code>rhs2</code> to the left before adding it with
 * <code>rhs1</code>.  This parameter is primarily for
 * <code>cgul_bcd__multiply()</code>.  The idea is that when you are
 * generating your rows that you add together during the multiply,
 * there are only 9 different rows provided you disregard the left
 * shifts.  So rather than make <code>cgul_bcd__multiply()</code>
 * generate those properly shifted numbers before passing them into
 * this method, we simply have it pass in
 * <code>shift_count</code>. <p>
 *
 * Note that an exception is thrown if this method is called with
 * <code>shift_count</code> greater than zero and <code>lhs</code> ==
 * </code>rhs2</code>.  The reason for this is explained in the
 * comments.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 * @param[in] shift_count shift count
 */
static void
cgul_bcd__add_magnitudes(cgul_exception_t* cex,
                         cgul_bcd_t lhs,
                         cgul_bcd_t rhs1,
                         cgul_bcd_t rhs2,
                         size_t shift_count)
{
    char d = 0;
    int carry = 0;
    size_t i = 0;
    size_t j = 0;
    size_t lhs_digits_len = 0;
    size_t rhs1_digits_len = 0;
    size_t rhs2_digits_len = 0;
    size_t rhs2_digits_len_after_shift = 0;
    size_t lhs_min_len = 0;
    size_t lhs_extra_digits_len = 0;
    char* lhs_digits = NULL;
    char* rhs1_digits = NULL;
    char* rhs2_digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* When (shift_count > 0) and (lhs == rhs2), rhs2 starts to be
     * overwritten by the digits in rhs1 when the shift is simulated.
     * Because this method is only called internally, we know that
     * this is not something we need.  So, instead of making this
     * method even more complicated, we just throw an exception. */
    if ((shift_count > 0) && (lhs == rhs2)) {
        *cex = cgul_exception__new("cgul_bcd__add_magnitudes: "
                                   "(shift_count > 0) and (lhs == rhs) "
                                   "is not permitted");
        goto out;
    }

    /* Get the lengths of the arrays of digits. */
    lhs_digits_len = cgul_string__get_length(cex, lhs->digits);
    rhs1_digits_len = cgul_string__get_length(cex, rhs1->digits);
    rhs2_digits_len = cgul_string__get_length(cex, rhs2->digits);

    /* Resize lhs (which may also resize rhs1 or rhs2 if they happen
     * to point to the same underlying object).  We add 1 to
     * lhs_min_len to account for a possible carry.  We do not need to
     * add 1 for a trailing null-terminator as that is automatically
     * done for us if the capacity is increased. */
    rhs2_digits_len_after_shift =
        cgul_defensive__add_size_type(cex, rhs2_digits_len, shift_count);
    if (*cex) { goto out; }
    lhs_min_len = CGUL_MAX(rhs1_digits_len, rhs2_digits_len_after_shift);
    lhs_min_len = cgul_defensive__add_size_type(cex, lhs_min_len, 1);
    if (*cex) { goto out; }
    if (lhs_digits_len < lhs_min_len) {
        lhs_extra_digits_len = lhs_min_len - lhs_digits_len;
        cgul_string__increase_capacity_if_necessary(cex,
                                                    lhs->digits,
                                                    lhs_extra_digits_len);
        if (*cex) {
            goto out;
        }
    }

    /* Because rhs1 or rhs2 could point to the same object as lhs, it
     * is not safe to get the arrays of digits until after lhs has
     * been resized above. */
    lhs_digits = cgul_string__get_value(cex, lhs->digits);
    rhs1_digits = cgul_string__get_value(cex, rhs1->digits);
    rhs2_digits = cgul_string__get_value(cex, rhs2->digits);

    /* If shift_count is specified, it tells us how far to shift rhs2
     * before doing the addition.  So, the digits in rhs1 in that
     * range are simply copied to lhs.  Note that you do not need to
     * copy digits from rhs1 if it points to the same object as
     * lhs. */
    if (lhs == rhs1) {
        /* Update i so we know where we are.  Because we cannot "copy"
         * more digits than already exist in lhs, i is set to
         * min(rhs1_digits_line, shift_count). */
        i = (rhs1_digits_len < shift_count) ? rhs1_digits_len : shift_count;
    } else {
        /* Copy digits from rhs1 to lhs. */
        for (i = 0 ; (i < shift_count) && (i < rhs1_digits_len) ; ++i) {
            lhs_digits[i] = rhs1_digits[i];
        }
    }

    /* If rhs1 is shorter than shift count, we need to complete the
     * above loop by adding zeros. */
    for ( ; i < shift_count ; ++i) {
        lhs_digits[i] = 0;
    }

    /*
     * Add the digits together while they line up.
     */
    for ( ; (i < rhs1_digits_len) && (j < rhs2_digits_len) ; ++i, ++j) {

        /* Add the two digits together and the carry. */
        d = rhs1_digits[i] + rhs2_digits[j] + carry;

        /* Check for a new carry. */
        if (d < 10) {
            carry = 0;
        } else {
            carry = 1;
            d -= 10;
        }

        /* Write the resulting digit. */
        lhs_digits[i] = d;

    }

    /* If rhs1 has more digits, then we just exhausted rhs2, and we
     * need to propagate the carry through the rest of rhs1. */
    for ( ; i < rhs1_digits_len ; ++i) {

        /* With carry = 0, this loop just copies digits from rhs1 to
         * lhs, but if lhs == rhs1, the digits are already present. */
        if ((carry == 0) && (lhs == rhs1)) {
            /* Update i so we know whether lhs needs to be truncated below. */
            i = rhs1_digits_len;
            break;
        }

        /* Add the digit and the carry. */
        d = rhs1_digits[i] + carry;

        /* Check for a new carry. */
        if (d < 10) {
            carry = 0;
        } else {
            carry = 1;
            d -= 10;
        }

        /* Write the resulting digit. */
        lhs_digits[i] = d;

    }

    /* If rhs2 has more digits, then we just exhausted rhs1, and we
     * need to propagate the carry through the rest of rhs2. */
    for ( ; j < rhs2_digits_len ; ++i, ++j) {

        /* With carry == 0, this loop just copies digits from rhs2 to
         * lhs, but if lhs == rhs2, the digits are already present. */
        if ((carry == 0) && (lhs == rhs2)) {
            /* Update i so we know the length of the string. */
            i = rhs2_digits_len + shift_count;
            break;
        }

        /* Add the digit and the carry. */
        d = rhs2_digits[j] + carry;

        /* Check for a new carry. */
        if (d < 10) {
            carry = 0;
        } else {
            carry = 1;
            d -= 10;
        }

        /* Write the resulting digit. */
        lhs_digits[i] = d;

    }

    /* There may still be a carry remaining. */
    if (carry) {
        /* Write the resulting digit. */
        lhs_digits[i] = 1;
        ++i;
    }

    /* Write a trailing null-terminator and adjust the length of the string. */
    lhs_digits[i] = '\0';
    lhs->digits->length = i;

    /* Make sure we didn't write off the end of the array above.
     * Remember that lhs_min_len is one less that the number of bytes
     * actually allocated. */
    assert(i <= lhs_min_len);

 out:
    return;
}

/**
 * Subtract the magnitude of <code>rhs2</code> from <code>rhs1</code>
 * after effectively shifting <code>rhs2</code> by
 * <code>shift_count</code> digits to the left.  Put the value in
 * <code>lhs</code> without changing the sign of <code>lhs</code>. <p>
 *
 * It is IMPORTANT that the caller arrange for <code>rhs1</code> to
 * ALWAYS have a larger magnitude than <code>rhs2</code>(after
 * shifting).  Remember, a larger magnitude does not necessarily mean
 * <code>rhs1</code> is greater than <code>rhs2</code>.  (Think about
 * two negative numbers.)  You can determine which has the greater
 * magnitude by calling <code>cgul_bcd__compare_magnitudes()</code>
 * (which also takes a <code>shift_count</code> parameter).  It is
 * safe for <code>rhs1</code> or <code>rhs2</code> and
 * <code>lhs</code> to point to the same underlying
 * <code>cgul_bcd</code> object.
 *
 * The <code>shift_count</code> parameter was added because shifting a
 * <code>cgul_bcd</code> object left is expensive (because the digits
 * are stored in an array which means shifting left requires copying
 * all digits), and <code>cgul_bcd__divide()</code> needs to
 * repeatedly shift left.  So rather than performing an expensive
 * shift left before performing the subtraction, this method uses the
 * <code>shift_count</code> parameter to calculate the subtraction
 * without actually having to shift left.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 * @param[in] shift_count shift count
 */
static void
cgul_bcd__subtract_magnitudes(cgul_exception_t* cex,
                              cgul_bcd_t lhs,
                              cgul_bcd_t rhs1,
                              cgul_bcd_t rhs2,
                              size_t shift_count)
{
    char d = 0;
    int borrow = 0;
    size_t i = 0;
    size_t j = 0;
    size_t lhs_digits_len = 0;
    size_t rhs1_digits_len = 0;
    size_t rhs2_digits_len = 0;
    size_t lhs_extra_digits_len = 0;
    char* lhs_digits = NULL;
    char* rhs1_digits = NULL;
    char* rhs2_digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* When (shift_count > 0) and (lhs == rhs2), rhs2 starts to be
     * overwritten by the digits in rhs1 when the shift is simulated.
     * Because this method is only called internally, we know that
     * this is not something we need.  So, instead of making this
     * method even more complicated, we just throw an exception. */
    if ((shift_count > 0) && (lhs == rhs2)) {
        *cex = cgul_exception__new("cgul_bcd__subtract_magnitudes: "
                                   "(shift_count > 0) and (lhs == rhs) "
                                   "is not permitted");
        goto out;
    }

    /* Get the length of the arrays of digits. */
    lhs_digits_len = cgul_string__get_length(cex, lhs->digits);
    rhs1_digits_len = cgul_string__get_length(cex, rhs1->digits);
    rhs2_digits_len = cgul_string__get_length(cex, rhs2->digits);

    /* Resize lhs to make sure it is large enough to hold the result.
     * We do this because profiling shows it is much more efficient to
     * write directly to the array than to write through
     * cgul_string__append_char(). */
    if (lhs_digits_len < rhs1_digits_len) {
        lhs_extra_digits_len = rhs1_digits_len - lhs_digits_len;
        cgul_string__increase_capacity_if_necessary(cex,
                                                    lhs->digits,
                                                    lhs_extra_digits_len);
        if (*cex) {
            goto out;
        }
    }

    /* Because rhs1 or rhs2 could point to the same object as lhs, it
     * is not safe to get the arrays of digits until after lhs has
     * been resized above. */
    lhs_digits = cgul_string__get_value(cex, lhs->digits);
    rhs1_digits = cgul_string__get_value(cex, rhs1->digits);
    rhs2_digits = cgul_string__get_value(cex, rhs2->digits);

    /* If shift_count is specified, it tells us how far to shift rhs2
     * before doing the subtraction.  So, the digits in rhs1 in that
     * range are simply copied to lhs.  Note that you do not need to
     * copy digits from rhs1 if it points to the same object as
     * lhs. */
    if (lhs == rhs1) {
        /* Update i so we know where we are.  Because we cannot "copy"
         * more digits than already exist in lhs, i is set to
         * min(rhs1_digits_line, shift_count). */
        i = shift_count;
    } else {
        /* Copy digits from rhs1 to lhs. */
        assert(shift_count <= rhs1_digits_len);
        for (i = 0 ; i < shift_count ; ++i) {
            lhs_digits[i] = rhs1_digits[i];
        }
    }

    /* Subtract where both numbers have aligned digits.  This exhausts
     * the digits of rhs2. */
    assert(shift_count + rhs2_digits_len <= rhs1_digits_len);
    for (j = 0 ; j < rhs2_digits_len ; ++i, ++j) {

        /* Get the next digit from rhs1. */
        d = rhs1_digits[i];

        /* If we have a pending borrow ... */
        if (borrow) {
            /* ... and this digit is not zero, we can satisfy and
             * clear the borrow by subtracting one from d. */
            if (d > 0) {
                --d;
                borrow = 0;
            } else {
                /* ... and this digit is zero, the borrow passes
                 * through causing this digit to be a 9. */
                d = 9;
            }
        }

        /* Determine if a borrow is needed. */
        if (d < rhs2_digits[j]) {
            d += 10;
            borrow = 1;
        }

        /* Subtract. */
        lhs_digits[i] = d - rhs2_digits[j];

    }

    /* Propagate any borrow that remains. */
    for ( ; i < rhs1_digits_len ; ++i) {

        /* With borrow == 0, this loop just copies digits from rhs1 to
         * lhs, but if lhs == rhs1, the digits are already present. */
        if ((borrow == 0) && (lhs == rhs1)) {
            /* Update i so we know the length of the string. */
            i = rhs1_digits_len;
            break;
        }

        /* Get the next digit from rhs1. */
        d = rhs1_digits[i];

        /* If we have a pending borrow ... */
        if (borrow) {
            /* ... and this digit is not zero, we can satisfy and
             * clear the borrow by subtracting one from d. */
            if (d > 0) {
                --d;
                borrow = 0;
            } else {
                /* ... and this digit is zero, the borrow passes
                 * through causing this digit to be a 9. */
                d = 9;
            }
        }

        /* Subtract (d - 0 = d). */
        lhs_digits[i] = d;
        
    }

    /* We have finished subtracting.  If we still have a borrow, it
     * means rhs1 had a smaller magnitude than rhs2 which is not
     * allowed. */
    assert(borrow == 0);

    /* Add a trailing null-terminator and set the length of the string. */
    lhs_digits[i] = '\0';
    lhs->digits->length = i;

    /* Remove leading zeros. */
    cgul_bcd__remove_leading_zeros(cex, lhs);

 out:
    return;
}

/**
 * This method is a helper for <code>cgul_bcd__multiply()</code>.  It
 * is designed to multiply <code>rhs1</code> by the single, positive
 * digit <code>rhs2</code> and write the result in <code>lhs</code>.
 * The result can then be used as one of the rows (after a proper
 * shift) that gets added together during the multiplication.  This
 * method can be called with <code>lhs</code> and <code>rhs1</code>
 * pointing to the same underlying object.
 *
 * @param[in,out] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
static void
cgul_bcd__multiply_by_single_digit(cgul_exception_t* cex,
                                   cgul_bcd_t lhs,
                                   cgul_bcd_t rhs1,
                                   unsigned char rhs2)
{
    int m = 0;
    int carry = 0;
    size_t i = 0;
    size_t lhs_digits_len = 0;
    size_t rhs1_digits_len = 0;
    size_t lhs_min_len = 0;
    size_t lhs_extra_digits_len = 0;
    char* rhs1_digits;
    char* lhs_digits;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2 < 10);

    /* Get the length of lhs and rhs1. */
    lhs_digits_len = cgul_string__get_length(cex, lhs->digits);
    rhs1_digits_len = cgul_string__get_length(cex, rhs1->digits);

    /* Adjust the capacity of lhs so we can write directly to the
     * underlying array of digits.  We add one to account for a new
     * digit produced by a possible final carry.  Profiling shows that
     * this helps. */
    lhs_min_len = cgul_defensive__add_size_type(cex, rhs1_digits_len, 1);
    if (*cex) { goto out; }
    if (lhs_digits_len < lhs_min_len) {
        lhs_extra_digits_len = lhs_min_len - lhs_digits_len;
        cgul_string__increase_capacity_if_necessary(cex,
                                                    lhs->digits,
                                                    lhs_extra_digits_len);
        if (*cex) {
            goto out;
        }
    }
    
    /* Wait until after resizing lhs to get the arrays of digits. */
    lhs_digits = cgul_string__get_value(cex, lhs->digits);
    rhs1_digits = cgul_string__get_value(cex, rhs1->digits);

    /* Multiply. */
    for (i = 0 ; i < rhs1_digits_len ; ++i) {
        m = rhs1_digits[i] * rhs2 + carry;
        lhs_digits[i] = m % 10;
        carry = m / 10;
    }

    /* Append the final carry if necessary. */
    if (carry) {
        lhs_digits[i] = carry;
        ++i;
    }

    /* Add a trailing null-terminator to lhs and set its new length. */
    lhs_digits[i] = '\0';
    lhs->digits->length = i;

    /* Assign the sign.  This is simplified because rhs2 is always positive. */
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs1->is_positive;
    }

    /* Make sure we didn't write off the end of the array above.
     * Remember that lhs_min_len is one less that the number of bytes
     * actually allocated. */
    assert(i <= lhs_min_len);

 out:
    return;
}

cgul_bcd_t
cgul_bcd__new(cgul_exception_t* cex)
{
    cgul_bcd_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the object itself. */
    rv = (cgul_bcd_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the string that holds the digits initializing it to zero. */
    rv->digits = cgul_string__new(cex);

    /* Set this object to zero. */
    cgul_bcd__set_to_zero(cex, rv);

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_bcd__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_bcd_t
cgul_bcd__new_from_bcd(cgul_exception_t* cex,
                       cgul_bcd_t value)
{
    cgul_bcd_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new object. */
    rv = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign from bcd. */
    cgul_bcd__assign(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_bcd__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_bcd_t
cgul_bcd__new_from_long(cgul_exception_t* cex,
                        long value)
{
    cgul_bcd_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new object. */
    rv = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign from long. */
    cgul_bcd__assign_from_long(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_bcd__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_bcd_t
cgul_bcd__new_from_unsigned_long(cgul_exception_t* cex,
                                 unsigned long value)
{
    cgul_bcd_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new object. */
    rv = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign from unsigned long. */
    cgul_bcd__assign_from_unsigned_long(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_bcd__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_bcd_t
cgul_bcd__new_from_size_type(cgul_exception_t* cex,
                             size_t value)
{
    cgul_bcd_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new object. */
    rv = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign from size type. */
    cgul_bcd__assign_from_size_type(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_bcd__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_bcd_t
cgul_bcd__new_from_string(cgul_exception_t* cex,
                          const char* s)
{
    cgul_bcd_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Allocate a new object. */
    rv = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign from string. */
    cgul_bcd__assign_from_string(cex, rv, s);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_bcd__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_bcd__delete(cgul_bcd_t bcd)
{
    if (bcd) {

        /* digits */
        if (bcd->digits) {
            cgul_string__delete(bcd->digits);
        }

        /* Free the object itself. */
        free(bcd);
    }
}

int
cgul_bcd__is_positive(cgul_exception_t* cex,
                      cgul_bcd_t bcd)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    rv = bcd->is_positive && !cgul_bcd__is_zero(cex, bcd);

 out:
    return rv;
}

int
cgul_bcd__is_negative(cgul_exception_t* cex,
                      cgul_bcd_t bcd)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    rv = !bcd->is_positive && !cgul_bcd__is_zero(cex, bcd);

 out:
    return rv;
}

int
cgul_bcd__is_zero(cgul_exception_t* cex,
                  cgul_bcd_t bcd)
{
    int rv = 0;
    size_t digits_len = 0;
    char* digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Remove leading zeros. */
    cgul_bcd__remove_leading_zeros(cex, bcd);

    /* Get the array of digits and the length of the array. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    assert(digits_len);

    /* Check for both positive and negative zero (essentially by
     * ignoring bcd->is_positive. */
    if ((digits_len == 1) && (digits[0] == 0)) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_bcd__set_to_zero(cgul_exception_t* cex,
                      cgul_bcd_t bcd)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    cgul_string__set_value(cex, bcd->digits, "");
    cgul_string__append_char(cex, bcd->digits, (char)0);
    bcd->is_positive = 1;

 out:
    return;
}

int
cgul_bcd__is_one(cgul_exception_t* cex,
                 cgul_bcd_t bcd)
{
    int rv = 0;
    size_t digits_len = 0;
    char* digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Remove leading zeros. */
    cgul_bcd__remove_leading_zeros(cex, bcd);

    /* Get the array of digits and the length of the array. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    assert(digits_len);

    /* Check if the value is one. */
    if (bcd->is_positive && (digits_len == 1) && (digits[0] == 1)) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_bcd__set_to_one(cgul_exception_t* cex,
                     cgul_bcd_t bcd)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    cgul_string__set_value(cex, bcd->digits, "");
    cgul_string__append_char(cex, bcd->digits, (char)1);
    bcd->is_positive = 1;

 out:
    return;
}

int
cgul_bcd__is_even(cgul_exception_t* cex,
                  cgul_bcd_t bcd)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(bcd->digits);
    assert(bcd->digits->value);

    if ((bcd->digits->value[0] & 0x01) == 0) {
        rv = 1;
    }

 out:
    return rv;
}

int
cgul_bcd__is_odd(cgul_exception_t* cex,
                 cgul_bcd_t bcd)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(bcd->digits);
    assert(bcd->digits->value);

    if ((bcd->digits->value[0] & 0x1) == 0x1) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_bcd__assign(cgul_exception_t* cex,
                 cgul_bcd_t lhs,
                 cgul_bcd_t rhs)
{
    size_t rhs_digits_len = 0;
    char* rhs_digits = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We do not have anything to do if both sides are already the same. */
    if (lhs == rhs) {
        goto out;
    }

    /* Clear lhs. */
    cgul_string__set_value(cex, lhs->digits, "");

    /* Get the rhs array of digits and its length. */
    rhs_digits = cgul_string__get_value(cex, rhs->digits);
    rhs_digits_len = cgul_string__get_length(cex, rhs->digits);

    assert(rhs_digits_len);

    /* Copy the digits from rhs to lhs being mindful that any '0'
     * digit in rhs will look like a null terminator for a string. */
    cgul_string__append_block(cex, lhs->digits, rhs_digits, rhs_digits_len);

    /* Remember the sign. */
    if (cgul_bcd__is_zero(cex, rhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive;
    }

 out:
    return;
}

void
cgul_bcd__assign_from_long(cgul_exception_t* cex,
                           cgul_bcd_t lhs,
                           long rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    if (rhs < 0) {
        cgul_bcd__assign_from_unsigned_long(cex, lhs, -rhs);
        lhs->is_positive = 0;
    } else {
        cgul_bcd__assign_from_unsigned_long(cex, lhs, rhs);
    }

 out:
    return;
}

void
cgul_bcd__assign_from_unsigned_long(cgul_exception_t* cex,
                                    cgul_bcd_t lhs,
                                    unsigned long rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    /* Clear lhs. */
    cgul_string__set_value(cex, lhs->digits, "");

    /* Peel off the digits in reverse order using integer division. */
    while (rhs) {
        char d = (char)(rhs % 10);
        cgul_string__append_char(cex, lhs->digits, d);
        rhs /= 10;
        if (*cex) {
            goto out;
        }
    }

    /* If no digits were copied to lhs, the value is zero. */
    if (cgul_string__get_length(cex, lhs->digits) == 0) {
        cgul_string__append_char(cex, lhs->digits, (char)0);
    }

    lhs->is_positive = 1;

 out:
    return;
}

void
cgul_bcd__assign_from_size_type(cgul_exception_t* cex,
                                cgul_bcd_t lhs,
                                size_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    /* Clear lhs. */
    cgul_string__set_value(cex, lhs->digits, "");

    /* Peel off the digits in reverse order using integer division. */
    while (rhs) {
        char d = (char)(rhs % 10);
        cgul_string__append_char(cex, lhs->digits, d);
        rhs /= 10;
        if (*cex) {
            goto out;
        }
    }

    /* If no digits were copied to lhs, the value is zero. */
    if (cgul_string__get_length(cex, lhs->digits) == 0) {
        cgul_string__append_char(cex, lhs->digits, (char)0);
    }

    lhs->is_positive = 1;

 out:
    return;
}

void
cgul_bcd__assign_from_cbi_element(cgul_exception_t* cex,
                                  cgul_bcd_t lhs,
                                  cbi_element_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    /* Clear lhs. */
    cgul_string__set_value(cex, lhs->digits, "");

    /* Peel off the digits in reverse order using integer division. */
    while (rhs) {
        char d = (char)(rhs % 10);
        cgul_string__append_char(cex, lhs->digits, d);
        rhs /= 10;
        if (*cex) {
            goto out;
        }
    }

    /* If no digits were copied to lhs, the value is zero. */
    if (cgul_string__get_length(cex, lhs->digits) == 0) {
        cgul_string__append_char(cex, lhs->digits, (char)0);
    }

    lhs->is_positive = 1;

 out:
    return;
}

void
cgul_bcd__assign_from_string(cgul_exception_t* cex,
                             cgul_bcd_t lhs,
                             const char* rhs)
{
    /* Error message if conversion fails. */
    static const char* errmsg = "cgul_bcd__assign_from_string: "
                                "invalid conversion";

    /* We need to know what the current locale considers to be the
     * thousands separator. */
    struct lconv* locale_info = NULL;

    /* thousands separator */
    char thousands_separator = '\0';

    /* Determine the base for the digits in the "rhs" string.  This is
     * either 16 (hexadecimal), 10 (decimal), 8 (octal), or 2
     * (binary). */
    long rhs_base = 10;

    /* Function used to determine if the digit in "rhs" is valid for
     * the base of "rhs". */
    long (*local_get_digit)(cgul_exception_t*, char) = NULL;

    /* Iterator for traversing "rhs". */
    size_t i = 0;

    /* String length of "rhs". */
    size_t rhs_length = 0;

    /* Save the sign.  You need to set this last because multiplying
     * by zero will set the sign back to positive. */
    int is_positive_saved = 1;

    /* Next digit to convert. */
    long digit = 0;

    /* "digit" as a cgul_bcd. */
    cgul_bcd_t bcd_digit = NULL;

    /* The base as a cgul_bcd. */
    cgul_bcd_t bcd_rhs_base =  NULL;

    /* Store the converstion in "tmp" and swap with "lhs" only if the
     * conversion completes successfully. */
    cgul_bcd_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    rhs_length = strlen(rhs);
    locale_info = localeconv();
    thousands_separator = *locale_info->thousands_sep;

    /*
     * Search for leading positive or negative sign.
     */

    cgul_bcd__skip_white_space(cex, rhs, rhs_length, &i);

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    /* See if there is a leading positive or negative sign.  I believe
     * all locales uses '+' and '-' for LC_NUMERICAL positive and
     * negative signs (as opposed to LC_MONETARY which accounts for
     * special monetary versions of positive and negative). */
    if( rhs[i] == '+' ) {
        is_positive_saved = 1;
        ++i;
    } else if( rhs[i] == '-' ) {
        is_positive_saved = 0;
        ++i;
    }

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    cgul_bcd__skip_white_space(cex, rhs, rhs_length, &i);

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    /* Determine the base of "rhs". */
    rhs_base = cgul_bcd__determine_base(cex, rhs, rhs_length, &i);
 
    /* Set up function that determines whether the digit in "rhs" is
     * valid for the particular base.  */
    if( rhs_base == 10 ) {
        local_get_digit = &cgul_bcd__get_decimal_digit;
    } else if( rhs_base == 16 ) {
        local_get_digit = &cgul_bcd__get_hexadecimal_digit;
    } else if( rhs_base == 8 ) {
        local_get_digit = &cgul_bcd__get_octal_digit;
    } else if( rhs_base == 2 ) {
        local_get_digit = &cgul_bcd__get_binary_digit;
    }

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    /* If the string is not in decimal, prepare to do the math. */
    if (rhs_base != 10) {
        bcd_digit = cgul_bcd__new(cex);
        bcd_rhs_base = cgul_bcd__new_from_long(cex, rhs_base);
    }

    /* Allocate the cgul_bcd object that holds the conversion. */
    tmp = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* If the string is already in decimal, we are going to avoid the
     * arithmetic and simply shift the digits into place instead.
     * This means we must start with tmp being empty. */
    if (rhs_base == 10) {
        cgul_string__set_value(cex, tmp->digits, "");
    }

    /* Iterate over the digits in "rhs". */
    for( ; i < rhs_length ; ++i ) {
        if( rhs[i] == thousands_separator ) {
            continue;
        }
        if( (digit = local_get_digit(cex, rhs[i])) == -1 ) {
            *cex = cgul_exception__new("%s", errmsg);
            goto out;
        }
        /* It is more efficient to simply shift digits in if the
         * string is in decimal. */
        if (rhs_base == 10) {
            cgul_string__append_char(cex, tmp->digits, (char)digit);
        } else {
            /* For other bases, we have to do the math. */
            cgul_bcd__multiply(cex, tmp, tmp, bcd_rhs_base);
            cgul_bcd__assign_from_long(cex, bcd_digit, digit);
            cgul_bcd__add_magnitudes(cex, tmp, tmp, bcd_digit, 0);
        }
        if (*cex) {
            goto out;
        }
    }

    /* We read digits out of rhs starting with the most-significant
     * digit, but if this was a base 10 converstion we simply appended
     * them to tmp->digits resulting in the digits being in reverse
     * order. */
    if (rhs_base == 10) {

        /* Reverse the string of digits so that they are in the
         * correct order with the least-significant digit at
         * tmp->digits->value[0]. */
        cgul_string__reverse(cex, tmp->digits);

        /* Also, if the string was already in decimal, we tried to
         * just shift digits into place after clearing tmp->digits.
         * It is possible that no digits were shifted in.  In this
         * case, we need to manually set tmp to zero. */
        if (cgul_string__get_length(cex, tmp->digits) == 0) {
            cgul_string__append_char(cex, tmp->digits, 0);
        }
    }

    if (*cex) {
        goto out;
    }

    /* Success! */
    cgul_bcd__swap(cex, lhs, tmp);
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = is_positive_saved;
    }

 out:

    if (tmp) {
        cgul_bcd__delete(tmp);
    }
    if (bcd_rhs_base) {
        cgul_bcd__delete(bcd_rhs_base);
    }
    if (bcd_digit) {
        cgul_bcd__delete(bcd_digit);
    }

    return;
}

void
cgul_bcd__negate(cgul_exception_t* cex,
                 cgul_bcd_t lhs,
                 cgul_bcd_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    cgul_bcd__assign(cex, lhs, rhs);
    if (*cex) {
        goto out;
    }

    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive ? 0 : 1;
    }

 out:
    return;
}

void
cgul_bcd__abs(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    cgul_bcd__assign(cex, lhs, rhs);
    lhs->is_positive = 1;

 out:
    return;
}

void
cgul_bcd__add(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs1,
              cgul_bcd_t rhs2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    if( rhs1->is_positive == rhs2->is_positive ) {
        /* rhs1 and rhs2 have the same sign.  Adjust lhs's sign and
         * add the magnitudes. */
        cgul_bcd__add_magnitudes(cex, lhs, rhs1, rhs2, 0 /* shift_count */);
        lhs->is_positive = rhs1->is_positive;
    } else {
        /* rhs1 and rhs2 have different signs.  Adjust lhs's sign and
         * subtract the magnitudes. */
        int cmp = cgul_bcd__compare_magnitudes(cex,
                                               rhs1,
                                               rhs2,
                                               0 /* shift_count */);
        if( cmp < 0 ) {
            cgul_bcd__subtract_magnitudes(cex,
                                          lhs,
                                          rhs2,
                                          rhs1,
                                          0 /* shift_count */);
            lhs->is_positive = rhs2->is_positive;
        } else if( cmp > 0 ) {
            cgul_bcd__subtract_magnitudes(cex,
                                          lhs,
                                          rhs1,
                                          rhs2,
                                          0 /* shift_count */);
            lhs->is_positive = rhs1->is_positive;
        } else {
            /* The numbers have opposite signs and the same magnitude
             * so an addition will result in a value of zero. */
            cgul_bcd__set_to_zero(cex, lhs);
        }
    }
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    }

 out:
    return;
}

void
cgul_bcd__subtract(cgul_exception_t* cex,
                   cgul_bcd_t lhs,
                   cgul_bcd_t rhs1,
                   cgul_bcd_t rhs2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    if( rhs1->is_positive != rhs2->is_positive ) {
        /* rhs1 and rhs2 have opposite signs.  Adjust lhs's sign and
         * add the magnitudes. */
        cgul_bcd__add_magnitudes(cex, lhs, rhs1, rhs2, 0 /* shift_count */);
        lhs->is_positive = rhs1->is_positive;
    } else {
        /* rhs1 and rhs2 have the same signs.  Adjust lhs's sign and
         * subtract the magnitudes. */
        int cmp = cgul_bcd__compare_magnitudes(cex,
                                               rhs1,
                                               rhs2,
                                               0 /* shift_count */);
        if( cmp < 0 ) {
            /* The answer has the opposite sign. */
            cgul_bcd__subtract_magnitudes(cex,
                                          lhs,
                                          rhs2,
                                          rhs1,
                                          0 /* shift_count */);
            lhs->is_positive = rhs1->is_positive ? 0 : 1;
        } else if( cmp > 0 ) {
            /* The answer has the same sign. */
            cgul_bcd__subtract_magnitudes(cex,
                                          lhs,
                                          rhs1,
                                          rhs2,
                                          0 /* shift_count */);
            lhs->is_positive = rhs1->is_positive;
        } else {
            /* The numbers have the same sign and the same magnitude
             * so a subtraction will result in a value of zero. */
            cgul_bcd__set_to_zero(cex, lhs);
        }
    }
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    }

 out:
    return;
}

/*
 * This method performs "long multiplication" which is O(n^2).  Do not
 * let the O(n^2) fool you.  On a machine with a fast integer
 * multiplier, this method actually runs faster than the other
 * multiplication methods up to around 72 digits.  All the
 * divide-and-conquer algorithms eventually produce factors that are
 * small enough for this method to handle efficiently.
 */
void
cgul_bcd__multiply_long(cgul_exception_t* cex,
                        cgul_bcd_t lhs,
                        cgul_bcd_t rhs1,
                        cgul_bcd_t rhs2)
{
    char d = 0;
    int result_is_positive = 0;
    size_t i = 0;
    size_t rhs1_digits_len = 0;
    size_t rhs2_digits_len = 0;
    char* rhs2_digits = NULL;
    cgul_bcd_t accumulator = NULL;
    cgul_bcd_t table[10];

    /* Initialize table.  This must be done before testing if an
     * exception is being thrown so that the clean up code in "out:"
     * will know whether it should delete elements of "table". */
    for (i = 0 ; i < 10 ; ++i) {
        table[i] = NULL;
    }

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /*
     * The algorithm used is almost identical to the pencil-and-paper
     * method.  There is one slight difference in that we add each row
     * immediately to the accumulator rather than determining all rows
     * before starting to add.
     */

    /* Determine the sign of the result. */
    if (rhs1->is_positive == rhs2->is_positive) {
        result_is_positive = 1;
    } else {
        result_is_positive = 0;
    }

    /* Get the number of digits in rhs1 and rhs2. */
    rhs1_digits_len = cgul_string__get_length(cex, rhs1->digits);
    rhs2_digits_len = cgul_string__get_length(cex, rhs2->digits);

    /* To possibly reduce the amount of looping, we swap the rhs1 and
     * rhs2 pointers if rhs1 is shorter than rhs2. */
    if (rhs1_digits_len < rhs2_digits_len) {
        size_t tmp_len = rhs1_digits_len;
        cgul_bcd_t tmp_ptr = rhs1;
        rhs1 = rhs2;
        rhs2 = tmp_ptr;
        rhs1_digits_len = rhs2_digits_len;
        rhs2_digits_len = tmp_len;
    }

    /*
     * At this point in the code, the rhs2 pointer points to is the
     * cgul_bcd object that has the fewest digits.
     */

    /* Get the array of digits for rhs2. */
    rhs2_digits = cgul_string__get_value(cex, rhs2->digits);

    /* Allocate the accumulator. */
    accumulator = cgul_bcd__new(cex);

    if (*cex) {
        goto out;
    }

    /* For each digit in rhs2, multiply it with rhs1 and shift the
     * result the correct number of places.  Then add that to the
     * accumultor. */
    for (i = 0 ; i < rhs2_digits_len ; ++i) {

        /* Get the next digit. */
        d = rhs2_digits[i];

        /* zero has no effect on the result. */
        if (d == 0) {
            continue;
        }

        /* There are only 10 decimal digits.  So there are really only
         * 10 different rows if you disregard shifting and only 9 if
         * you disregard zero as well.  To avoid repeatedly doing the
         * same 9 multiplications only to get the same 9 results, we
         * lazily generate and store the result in "table". */
        if (table[(int)d] == NULL) {
            table[(int)d] = cgul_bcd__new(cex);
            cgul_bcd__multiply_by_single_digit(cex, table[(int)d], rhs1, d);
            if (*cex) {
                goto out;
            }
        }

        /*
         * Add this row to the accumulator.  The call below to
         * cgul_bcd__add_magnitudes() is equivalent to the following:
         *
         *     accumulator = accumulator + table[d] * (10 ^ i)
         *
         * Basically, we are shifting table[d] by i digits to the left
         * before we add it to accumulator.
         */
        cgul_bcd__add_magnitudes(cex,
                                 accumulator,
                                 accumulator,
                                 table[(int)d],
                                 i);
        if (*cex) {
            goto out;
        }
    }

    /* Swap the result into lhs. */
    cgul_bcd__swap(cex, lhs, accumulator);
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = result_is_positive;
    }

 out:

    if (accumulator) {
        cgul_bcd__delete(accumulator);
    }
    for (i = 0 ; i < 10 ; ++i) {
        if (table[i]) {
            cgul_bcd__delete(table[i]);
        }
    }

    return;
}

static void
cgul_bcd__multiply_karatsuba(cgul_exception_t* cex,
                             cgul_bcd_t lhs,
                             cgul_bcd_t rhs1,
                             cgul_bcd_t rhs2)
{
    size_t m = 0;
    cgul_bcd_t x1 = NULL;
    cgul_bcd_t x2 = NULL;
    cgul_bcd_t y1 = NULL;
    cgul_bcd_t y2 = NULL;
    cgul_bcd_t a = NULL;
    cgul_bcd_t b = NULL;
    cgul_bcd_t c = NULL;
    cgul_bcd_t one = NULL;
    cgul_bcd_t tmp1 = NULL;
    cgul_bcd_t tmp2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Sanity check.  If the smaller factor is too small to split use
     * the "long multiplication" algorithm instead. */
    m = CGUL_MIN(rhs1->digits->length, rhs2->digits->length);
    if (m <= 1) {
        cgul_bcd__multiply_long(cex, lhs, rhs1, rhs2);
        goto out;
    }

    /* Calculate the number of digits in the lower-order parts.  For
     * Karatsuba to work, both factors have to be split the same.  We
     * determine this simply by halving the size of the smaller
     * factor. */
    m >>= 1;

    /* Allocate. */
    x1 = cgul_bcd__new(cex);
    x2 = cgul_bcd__new(cex);
    y1 = cgul_bcd__new(cex);
    y2 = cgul_bcd__new(cex);
    a = cgul_bcd__new(cex);
    b = cgul_bcd__new(cex);
    c = cgul_bcd__new(cex);
    one = cgul_bcd__new_from_long(cex, 1);
    tmp1 = cgul_bcd__new(cex);
    tmp2 = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Initialize x1 and x2.  Note that x2 is just the low-order
     * digits which are extracted from rhs1 by shifting x1 back to the
     * right.  This results in a value similar to rhs1 except all the
     * lower-order digits are zero.  The low-order digits can now be
     * extracted by simply subtracting from rhs1. */
    cgul_bcd__shift_right(cex, x1, rhs1, m);
    cgul_bcd__shift_left(cex, x2, x1, m);
    cgul_bcd__subtract(cex, x2, rhs1, x2);

    /* Initialize y1 and y2.  Note that y2 is just the low-order
     * digits which are extracted from rhs2 by shifting y1 back to the
     * right as explained above for x2. */
    cgul_bcd__shift_right(cex, y1, rhs2, m);
    cgul_bcd__shift_left(cex, y2, y1, m);
    cgul_bcd__subtract(cex, y2, rhs2, y2);

    /* Initialize a and b. */
    cgul_bcd__multiply(cex, a, x1, y1);
    cgul_bcd__multiply(cex, b, x2, y2);

    /* Initialize c. */
    cgul_bcd__add(cex, tmp1, x1, x2);
    cgul_bcd__add(cex, tmp2, y1, y2);
    cgul_bcd__multiply(cex, c, tmp1, tmp2);
    cgul_bcd__subtract(cex, c, c, a);
    cgul_bcd__subtract(cex, c, c, b);

    /* Calculate the final product. */
    cgul_bcd__shift_left(cex, lhs, a, (m << 1));
    cgul_bcd__shift_left(cex, c, c, m);
    cgul_bcd__add(cex, lhs, lhs, c);
    cgul_bcd__add(cex, lhs, lhs, b);

 out:

    /* Clean up. */
    if (tmp1) {
        cgul_bcd__delete(tmp1);
    }
    if (tmp2) {
        cgul_bcd__delete(tmp2);
    }
    if (one) {
        cgul_bcd__delete(one);
    }
    if (c) {
        cgul_bcd__delete(c);
    }
    if (b) {
        cgul_bcd__delete(b);
    }
    if (a) {
        cgul_bcd__delete(a);
    }
    if (y2) {
        cgul_bcd__delete(y2);
    }
    if (y1) {
        cgul_bcd__delete(y1);
    }
    if (x2) {
        cgul_bcd__delete(x2);
    }
    if (x1) {
        cgul_bcd__delete(x1);
    }

    return;
}

/*
 * This method just redirects the multiplication operation to the
 * correct algorithm based on the size of the factors rhs1 and rhs2.
 */
void
cgul_bcd__multiply(cgul_exception_t* cex,
                   cgul_bcd_t lhs,
                   cgul_bcd_t rhs1,
                   cgul_bcd_t rhs2)
{
    size_t min_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Get the size of the smaller factor. */
    min_size = CGUL_MIN(rhs1->digits->length, rhs2->digits->length);

    /* Pick a multiplication algorithm based on min_size. */
    if (min_size <= 72) {
        cgul_bcd__multiply_long(cex, lhs, rhs1, rhs2);
    } else {
        cgul_bcd__multiply_karatsuba(cex, lhs, rhs1, rhs2);
    }

 out:
    return;
}

void
cgul_bcd__divide(cgul_exception_t* cex,
                 cgul_bcd_t quotient,
                 cgul_bcd_t remainder,
                 cgul_bcd_t dividend,
                 cgul_bcd_t divisor)
{
    int compare_rv = 0;
    int is_guess_bad = 0;
    int is_quotient_positive = 0;
    int is_remainder_positive = 0;
    int is_quotient_temporary = 0;
    unsigned int guess_divisor = 0;
    unsigned int guess_dividend = 0;
    unsigned int digit = 0;
    size_t i = 0;
    size_t quotient_size = 0;
    size_t shift_count = 0;
    size_t shift_count_index = 0;
    size_t dividend_digits_len = 0;
    size_t divisor_digits_len = 0;
    size_t accumulator_digits_len = 0;
    size_t working_quotient_digits_len = 0;
    size_t extra_size = 0;
    char* working_quotient_digits = NULL;
    cgul_bcd_t accumulator = NULL;
    cgul_bcd_t working_quotient = NULL;
    cgul_bcd_t table[10];

    /* Initialize table.  This must be done before testing if an
     * exception is being thrown so that the clean up code in "out:"
     * will know whether it should delete elements of "table". */
    for (i = 0 ; i < 10 ; ++i) {
        table[i] = NULL;
    }

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(quotient);
    assert(remainder);
    assert(dividend);
    assert(divisor);

    /* Check for division by zero. */
    if (cgul_bcd__is_zero(cex, divisor)) {
        *cex = cgul_exception__new("cgul_bcd__divide: division by zero");
        goto out;
    }

    /* Now check if the dividend is zero. */
    if (cgul_bcd__is_zero(cex, dividend)) {
        cgul_bcd__set_to_zero(cex, quotient);
        cgul_bcd__set_to_zero(cex, remainder);
        goto out;
    }

    /* Now check if the divisor is one. */
    if (cgul_bcd__is_one(cex, divisor)) {
        /* Assign the dividend to the quotient before setting the
         * remainder to zero.  This is necessary because the remainder
         * is allowed to point to the same object as the dividend. */
        cgul_bcd__assign(cex, quotient, dividend);
        cgul_bcd__set_to_zero(cex, remainder);
        goto out;
    }

    /* Determine the sign of the quotient. */
    is_quotient_positive =
        (dividend->is_positive == divisor->is_positive) ? 1 : 0;

    /* Determine the sign of the remainder. */
    is_remainder_positive = dividend->is_positive;

    /* If the magnitude of the dividend is less than the magnitude of
     * the divisor, the quotient is zero and the dividend is the
     * remainder.  By performing this test, we guarantee that the
     * quotient will always have at least one digit if we continue
     * into the main loop. */
    if (cgul_bcd__compare_magnitudes(cex, dividend, divisor, 0) < 0) {
        /* Assign the dividend to the remainder before setting the
         * quotient to zero.  This is necessary because the quotient
         * is allowed to point to the same object as the dividend. */
        cgul_bcd__assign(cex, remainder, dividend);
        remainder->is_positive = is_remainder_positive;
        cgul_bcd__set_to_zero(cex, quotient);
        goto out;
    }

    /* Determine the number of digits in the arrays. */
    dividend_digits_len = cgul_string__get_length(cex, dividend->digits);
    divisor_digits_len = cgul_string__get_length(cex, divisor->digits);

    /* Determine the size of the quotient when we are done.  Note that
     * we just finished comparing the magnitudes of the dividend and
     * divisor.  So we know that the size of the dividend is at least
     * as large as the size of the divisor.  So, it is safe to
     * subtract these unsigned values here. */
    quotient_size = dividend_digits_len - divisor_digits_len + 1;

    /* Determine whether you can directly write to the quotient or
     * whether you need a temporary quotient. */
    if( (quotient == dividend) || (quotient == divisor) ) {
        working_quotient = cgul_bcd__new(cex);
        if (*cex) {
            goto out;
        }
        is_quotient_temporary = 1;
    } else {
        /* We can gain a little speed by overwriting "quotient" directly. */
        working_quotient = quotient;
    }

    /* Make sure the working quotient has enough capacity. */
    working_quotient_digits_len =
        cgul_string__get_length(cex, working_quotient->digits);
    if (working_quotient_digits_len < quotient_size) {
        extra_size = quotient_size - working_quotient_digits_len;
        cgul_string__increase_capacity_if_necessary(cex,
                                                    working_quotient->digits,
                                                    extra_size);
        if (*cex) {
            goto out;
        }
    }

    /* We always need an accumulator. */
    accumulator = cgul_bcd__new(cex);
    cgul_bcd__assign(cex, accumulator, dividend);
    if (*cex) {
        goto out;
    }

    /* Now that everything has been resized, get the array of digits. */
    working_quotient_digits =
        cgul_string__get_value(cex, working_quotient->digits);

    /* Generate the divisor that we use to make our guesses.  We need
     * the two most significant digits of "divisor." */
    if (divisor_digits_len >= 1) {
        guess_divisor =
            divisor->digits->value[divisor_digits_len - 1];
    }
    if (divisor_digits_len >= 2) {
        guess_divisor =
            (10 * guess_divisor)
            + divisor->digits->value[divisor_digits_len - 2];
    }

    /* We divide by guess_divisor so make sure it is not zero. */
    if (guess_divisor == 0) {
        *cex = cgul_exception__new(
                   "cgul_bcd__divide: "
                   "Internal Error: "
                   "guess_divisor is zero!");
        goto out;
    }

    /*
     * Chinese Remainder Theorem.
     *
     * The main for() loop is controlled by "shift_count".  The
     * "shift_count" variable is used to control how far over we are
     * shifting "divisor" before multiplying it by the next digit in
     * the quotient and subtracting the multiple from the accumulator.
     * This necessarily means that "shift_count" also tells us exactly
     * where in the quotient the next digit belongs.
     */

    for (shift_count_index = quotient_size ;
         shift_count_index > 0 ;
         --shift_count_index)
    {

        /* shift_count_index is "one beyond" the real shift count. */
        shift_count = shift_count_index - 1;

        /* Reset the digit. */
        digit = 0;

        /* This value is used to prevent an infinite loop below when
         * the guess is bad. */
        is_guess_bad = 0;

        /* If the accumulator is already less than the divisor, this
         * digit is a zero so we can skip the subtraction and
         * immediately append a zero.  */
        compare_rv = cgul_bcd__compare_magnitudes(cex,
                                                  accumulator,
                                                  divisor,
                                                  shift_count);
        if (compare_rv < 0) {
            goto append_digit;
        }

        /* Get the length of the accumulator. */
        accumulator_digits_len = accumulator->digits->length;

        /* Get the 3 most-significant digits of the accumulator. */
        guess_dividend = 0;
        if (accumulator_digits_len >= 1) {
            guess_dividend =
                accumulator->digits->value[accumulator_digits_len - 1];
        }
        if (accumulator_digits_len >= 2) {
            guess_dividend =
                (10 * guess_dividend)
                + accumulator->digits->value[accumulator_digits_len - 2];
        }
        if (accumulator_digits_len >= 3) {
            guess_dividend =
                (10 * guess_dividend)
                + accumulator->digits->value[accumulator_digits_len - 3];
        }

        /* Guess the next digit in the quotient.  First, we try to
         * divide the guess_dividend by the guess_divisor.  Our first
         * guess might be way off, for example, if guess_divisor is 1.
         * In this case, you could easily end up with a "digit" that
         * is two digits long.  In this case, the right thing to do is
         * to just use the most significant digit of the accumulator
         * instead of the two most significant bits.  This is what we
         * are testing for and what we are doing inside the while()
         * immediately below.
         */
        digit = guess_dividend / guess_divisor;
        while (digit >= 10) {
            digit = digit / 10;
        }

        /* This guessing business is not an exact science.  Actually,
         * I suppose it is an exact science, but it is hard to predict
         * how good the guess is.  So, we iterate _and_ test to make
         * sure we have found the correct digit. */
        for (;;) {

            /* Lazily create the table of products. */
            CGUL_BCD__DIVIDE__LAZILY_CREATE_TABLE_OF_PRODUCTS(divisor,
                                                              digit);

            /* Compare the magnitudes of the accumulator and the
             * shifted product. */
            compare_rv = cgul_bcd__compare_magnitudes(cex,
                                                      accumulator,
                                                      table[digit],
                                                      shift_count);

            /* If the magnitude of the accumulator is less than the
             * magnitude of the shifted product, the digit is too
             * large.  Carefully reduce it by one. */    
            if (compare_rv < 0) {
                /* First, make sure "digit" is not going to underflow. */
                if (digit == 0) {
                    *cex = cgul_exception__new(
                               "cgul_bcd__divide: "
                               "Internal Error: "
                               "Underflow detected when guessing digit!");
                    goto out;
                }

                /* Reduce our guess by one to account for inaccuracies
                 * in our guess (e.g., "81000 / 901") */
                --digit;

                /* Try this digit. */
                continue;
            }

            /* If you reach here, the magnitude of the accumulator is
             * equal to or greater than the magnitude of the shifted
             * product.  So we try "digit" by subtracting the shifted
             * product from the accumulator. */
            cgul_bcd__subtract_magnitudes(cex,
                                          accumulator,
                                          accumulator,
                                          table[digit],
                                          shift_count);
            if (*cex) {
                goto out;
            }

            /* However, we still need to test to make sure that the
             * accumulator is less than the shifted divisor after
             * subtracting the product.  This is by far the most
             * common scenario.  If the accumulator is not less than
             * the shifted divisor, it means "digit" could be (at
             * least) one unit larger. */
            compare_rv = cgul_bcd__compare_magnitudes(cex,
                                                      accumulator,
                                                      divisor,
                                                      shift_count);
            if (compare_rv < 0) {
                /* We found the next digit. */
                break;
            }

            /* 
             * There are always going to be values where our guess
             * is way too low.  For example, look at the following
             * quotient:
             *
             *      12341368
             *     ----------
             *       123418
             *
             * In order to guess that the first digit of the
             * quotient is 9, one would have to go out 7 digits in
             * the numerator and 6 in the denominator, and no
             * matter how many digits you use for your
             * approximation, there will always be sets of numbers
             * that can cause our guess to be way too low.
             */

            /* If have entered this section of code once already,
             * throw an exception to avoid an infinite loop. */
            if (is_guess_bad) {
                *cex = cgul_exception__new(
                           "cgul_bcd__divide: "
                           "Internal Error: "
                           "Guess digit will not converge.");
                goto out;
            }
             
            /* Remember that the guess was bad. */
            is_guess_bad = 1;

            /* First, restore the accumulator to the value it had
             * before we subtracted our guess. */
            cgul_bcd__add_magnitudes(cex,
                                     accumulator,
                                     accumulator,
                                     table[digit],
                                     shift_count);

            /* Because the initial guess digit was too small, it is
             * likely that we have hit a combination like the one in
             * the comment above.  So just start over with the largest
             * digit.  We avoid an infinite loop here by setting
             * is_guess_bad above. */
            digit = 9;

            /* Try this digit. */
            continue;
        }

    append_digit:

        /* Add the digit that we found. */
        working_quotient_digits[shift_count] = (char)digit;
    }

    /* NUL-terminate working_quotient and adjust its length. */
    working_quotient_digits[quotient_size] = '\0';
    working_quotient->digits->length = quotient_size;

    /* Swap quotient into place. */
    if (is_quotient_temporary) {
        cgul_bcd__swap(cex, quotient, working_quotient);
    }
    if (cgul_bcd__is_zero(cex, quotient)) {
        quotient->is_positive = 1;
    } else {
        quotient->is_positive = is_quotient_positive;
    }

    /* Swap remainder into place.  */
    cgul_bcd__swap(cex, remainder, accumulator);
    if (cgul_bcd__is_zero(cex, remainder)) {
        remainder->is_positive = 1;
    } else {
        remainder->is_positive = is_remainder_positive;
    }

 out:

    for (i = 0 ; i < 10 ; ++i) {
        if (table[i]) {
            cgul_bcd__delete(table[i]);
        }
    }
    if( is_quotient_temporary ) {
        cgul_bcd__delete(working_quotient);
    }
    if( accumulator ) {
        cgul_bcd__delete(accumulator);
    }

}

/*
 * This function raises "base" to the power of "exponent" using the
 * "exponentiation by squaring" method.  Normally, this is defined
 * recursively, but there is a clever implementation that works by
 * inspecting the bits of "exponent" which is the implementation used
 * below.
 *
 * For example,
 *
 *     pow(5, 37) = pow(5, 32) * pow(5, 4) * pow(5, 1)   [equation #1]
 *
 * It is easy to see the symmetry in how to calculate pow(5, 37) from
 * the base-2 representation of 37:
 *
 *             37 = pow(2, 32) + pow(2, 4) + pow(5, 1)   [equation #2]
 *
 * Note that if you start with the least-significant bit of
 * "exponent", you can just keep doubling "base" in order to get each
 * of the terms (starting with the right-most term) in equation #1
 * above.
 */
extern void
cgul_bcd__power(cgul_exception_t* cex,
                cgul_bcd_t lhs,
                cgul_bcd_t base,
                cgul_bcd_t exponent)
{
    cgul_bcd_t b = NULL;
    cgul_big_integer_t e = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(base);
    assert(exponent);

    /* Make sure "exponent" is positive.  A negative exponent requires
     * taking the inverse which cgul_bcd cannot do. */
    if (cgul_bcd__is_negative(cex, exponent)) {
        *cex = cgul_exception__new("cgul_bcd__power: "
                                   "negative exponent");
        goto out;
    }

    /* Allocate. */
    b = cgul_bcd__new(cex);
    e = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign "base" and "exponent" to temporary values so they are
     * not harmed in the production of the result.  Note that we are
     * using a cgul_big_integer to represent e instead of a cgul_bcd.
     * This implicitly converts e to base-2 making it easy to
     * following the "exponentiation by squaring" algorithm. */
    cgul_bcd__assign(cex, b, base);
    cgul_bcd__as_big_integer(cex, exponent, e);
    if (*cex) {
        goto out;
    }

    /* Set "lhs" to 1 which basically assumes "exponent" is zero.
     * Note that this means pow(0, 0) = 1 which many math applications
     * use, but some leave it undefined. */
    cgul_bcd__set_to_one(cex, lhs);
    if (*cex) {
        goto out;
    }

    /* Iterate while "e" (or exponent) is not zero.  How this works is
     * described (above) in the main comments to this method. */
    while (!cgul_big_integer__is_zero(cex, e)) {

        /* Inspect the least-significant bit of e.  If it is one, we
         * must multiply b into the result. */
        if ((e->value[0] & 0x1) == 0x1) {
            cgul_bcd__multiply(cex, lhs, lhs, b);
            if (*cex) {
                goto out;
            }
        }

        /* Square the base b so that it tracks the bit-wise inspection
         * of e.  (See the main comments (above) for this method.) */
        cgul_bcd__multiply(cex, b, b, b);
        if (*cex) {
            goto out;
        }

        /* Prepare to inspect the next most-significant bit of e. */
        cgul_big_integer__shift_right(cex, e, e, 1);
    }

 out:

    /* Clean up. */
    if (e) {
        cgul_big_integer__delete(e);
    }
    if (b) {
        cgul_bcd__delete(b);
    }

    return;
}

/*
 * This method uses the same "exponentiation by squaring" that
 * cgul_bcd__power() uses, but it takes the modulo at each
 * stage of the calculation to keep the save of the numbers within
 * reason.  See the main comments for cgul_bcd__power() for an
 * explanation of the basic algorithm.
 *
 * The reason modulo operation can be performed at each stage is that,
 * by the basic rules of modular arithmetic, the modulo of the product
 * of two numbers is the same as the product of the modulo of each
 * number.  This can easily be proven by doing the math.
 */
extern void
cgul_bcd__power_modulo(cgul_exception_t* cex,
                               cgul_bcd_t lhs,
                               cgul_bcd_t base,
                               cgul_bcd_t exponent,
                               cgul_bcd_t modulus)
{
    cgul_bcd_t b = NULL;
    cgul_big_integer_t e = NULL;
    cgul_bcd_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(base);
    assert(exponent);
    assert(modulus);

    /* Make sure "exponent" is positive.  A negative exponent requires
     * taking the inverse which cgul_bcd cannot do. */
    if (cgul_bcd__is_negative(cex, exponent)) {
        *cex = cgul_exception__new("cgul_bcd__power: "
                                   "negative exponent");
        goto out;
    }

    /* Allocate. */
    b = cgul_bcd__new(cex);
    e = cgul_big_integer__new(cex);
    tmp = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign "base" and "exponent" to temporary values so they are
     * not harmed in the production of the result. */
    cgul_bcd__assign(cex, b, base);
    cgul_bcd__as_big_integer(cex, exponent, e);
    if (*cex) {
        goto out;
    }

    /* Set "lhs" to 1 which basically assumes "exponent" is zero.
     * Note that this means pow(0, 0) = 1 which many math applications
     * use, but some leave it undefined. */
    cgul_bcd__set_to_one(cex, lhs);
    if (*cex) {
        goto out;
    }

    /* Iterate while "e" (or exponent) is not zero.  How this works is
     * described (above) in the main comments to this method. */
    while (!cgul_big_integer__is_zero(cex, e)) {

        /* Inspect the least-significant bit of e.  If it is one, we
         * must multiply b into the result. */
        if ((e->value[0] & 0x1) == 0x1) {
            /* Multiply. */
            cgul_bcd__multiply(cex, lhs, lhs, b);
            /* Use division to calculate the modulo of lhs. */
            cgul_bcd__divide(cex, tmp, lhs, lhs, modulus);
            if (*cex) {
                goto out;
            }
        }

        /* Square the base b so that it tracks the bit-wise inspection
         * of e.  (See the main comments (above) for this method.) */
        cgul_bcd__multiply(cex, b, b, b);
        /* Use division to calculate the modulo of b. */
        cgul_bcd__divide(cex, tmp, b, b, modulus);
        if (*cex) {
            goto out;
        }

        /* Prepare to inspect the next most-significant bit of e. */
        cgul_big_integer__shift_right(cex, e, e, 1);
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_bcd__delete(tmp);
    }
    if (e) {
        cgul_big_integer__delete(e);
    }
    if (b) {
        cgul_bcd__delete(b);
    }

    return;
}

/*
 * This method is a simple wrapper around
 * cgul_big_integer__is_probably_prime().
 */
int
cgul_bcd__is_probably_prime(cgul_exception_t* cex,
                            cgul_bcd_t bcd,
                            unsigned int count)
{
    int rv = 0;
    cgul_big_integer_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Allocate. */
    tmp = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Convert bcd to a cgul_big_integer. */
    cgul_bcd__as_big_integer(cex, bcd, tmp);
    if (*cex) {
        goto out;
    }

    /* Test for primality. */
    rv = cgul_big_integer__is_probably_prime(cex, tmp, count);

 out:

    /* Clean up. */
    if (tmp) {
        cgul_big_integer__delete(tmp);
    }

    return rv;
}

void
cgul_bcd__shift_left(cgul_exception_t* cex,
                     cgul_bcd_t lhs,
                     cgul_bcd_t rhs,
                     size_t digit_count)
{
    size_t i = 0;
    size_t extra_size = 0;
    size_t total_size = 0;
    size_t lhs_digits_len = 0;
    size_t rhs_digits_len = 0;
    char* lhs_digits = NULL;
    char* rhs_digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* If digit_count is 0, this is effectively an assignment. */
    if (digit_count == 0) {
        cgul_bcd__assign(cex, lhs, rhs);
        goto out;
    }

    /* Get the number of digits in lhs and rhs. */
    lhs_digits_len = cgul_string__get_length(cex, lhs->digits);
    rhs_digits_len = cgul_string__get_length(cex, rhs->digits);

    assert(lhs_digits_len);
    assert(rhs_digits_len);

    /* Calculate the extra number of bytes that lhs needs in order to
     * accommodate rhs shifted by digit_count.  Because all the values
     * are unsigned, we must be careful when subtracting. */
    total_size =
        cgul_defensive__add_size_type(cex, rhs_digits_len, digit_count);
    if (*cex) { goto out; }
    if (total_size > lhs_digits_len) {
        extra_size = total_size - lhs_digits_len;
        cgul_string__increase_capacity_if_necessary(cex,
                                                    lhs->digits,
                                                    extra_size);
        if (*cex) {
            goto out;
        }
    }

    /* Now that everything has been resized, get the arrays of digits.
     * You must do this after resizing lhs because lhs is allowed to
     * be the same object as rhs. */
    lhs_digits = cgul_string__get_value(cex, lhs->digits);
    rhs_digits = cgul_string__get_value(cex, rhs->digits);

    /* Shift and copy the digits from rhs to lhs.  We have to do this
     * in a nondestructive way in case lhs and rhs point to the same
     * object. */
    for (i = rhs_digits_len ; i > 0 ; --i) {
        lhs_digits[i - 1 + digit_count] = rhs_digits[i - 1];
    }

    /* Now, shift in zeros for the remaining digits. */
    for (i = digit_count ; i > 0 ; --i) {
        lhs_digits[i - 1] = 0;
    }

    /* Adjust the length. */
    cgul_string__set_length(cex, lhs->digits, total_size);

    /* Set the sign. */
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive;
    }

 out:
    return;
}

void
cgul_bcd__shift_right(cgul_exception_t* cex,
                      cgul_bcd_t lhs,
                      cgul_bcd_t rhs,
                      size_t digit_count)
{
    size_t i = 0;
    size_t extra_size = 0;
    size_t total_size = 0;
    size_t lhs_digits_len = 0;
    size_t rhs_digits_len = 0;
    char* lhs_digits = NULL;
    char* rhs_digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* If digit_count is 0, this is effectively an assignment. */
    if (digit_count == 0) {
        cgul_bcd__assign(cex, lhs, rhs);
        goto out;
    }

    /* Get the number of digits in lhs and rhs. */
    lhs_digits_len = cgul_string__get_length(cex, lhs->digits);
    rhs_digits_len = cgul_string__get_length(cex, rhs->digits);

    assert(lhs_digits_len);
    assert(rhs_digits_len);

    /* If rhs_digits_len is less than or equal to digit_count, this
     * effectively sets lhs to zero. */
    if (rhs_digits_len <= digit_count) {
        cgul_bcd__set_to_zero(cex, lhs);
        goto out;
    }

    /* Calculate the extra number of bytes that lhs needs in order to
     * accommodate rhs shifted by digit_count.  Because all the values
     * are unsigned, we must be careful when subtracting.  Note that
     * we have already checked above if rhs_digits_len <= digit_count. */
    total_size = rhs_digits_len - digit_count;
    if (total_size > lhs_digits_len) {
        extra_size = total_size - lhs_digits_len;
        cgul_string__increase_capacity_if_necessary(cex,
                                                    lhs->digits,
                                                    extra_size);
        if (*cex) {
            goto out;
        }
    }

    /* Now that everything has been resized, get the arrays of digits.
     * You must do this after resizing lhs because lhs is allowed to
     * be the same object as rhs. */
    lhs_digits = cgul_string__get_value(cex, lhs->digits);
    rhs_digits = cgul_string__get_value(cex, rhs->digits);

    /* Shift and copy the digits from rhs to lhs.  We have to do this
     * in a nondestructive way in case lhs and rhs point to the same
     * object. */
    for (i = 0 ; i < total_size ; ++i) {
        lhs_digits[i] = rhs_digits[i + digit_count];
    }

    /* Zero out the extra high-order bytes. */
    for ( ; i < lhs_digits_len ; ++i) {
        lhs_digits[i] = 0;
    }

    /* Adjust the length. */
    cgul_string__set_length(cex, lhs->digits, total_size);

    /* Set the sign.  Note that this also removes leading zeros in lhs. */
    if (cgul_bcd__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive;
    }

 out:
    return;
}

int
cgul_bcd__compare(cgul_exception_t* cex,
                  cgul_bcd_t lhs,
                  cgul_bcd_t rhs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* Because we are using a signed magnitude representation, there
     * are both positive and negative zeros.  Call is_zero() to help
     * out. */
    if(    cgul_bcd__is_zero(cex, lhs)
        && cgul_bcd__is_zero(cex, rhs) )
    {
        rv = 0;
        goto out;
    }

    /* Check to see if the signs are different. */
    if( lhs->is_positive != rhs->is_positive ) {
        if( lhs->is_positive ) {
            rv = 1;
        } else {
            rv = -1;
        }
        goto out;
    }

    /* The signs are the same.  Check the magnitudes. */
    rv = cgul_bcd__compare_magnitudes(cex, lhs, rhs, 0 /* shift_count */);

    /* We now know which number has the larger magnitude which is
     * determinative.  For two positive numbers, the larger number is
     * the one with the larger magnitude, but for two negative
     * numbers, the larger number is the one with the smaller
     * magnitude. */
    if( lhs->is_positive == 0 ) {
        /* The two numbers are negative. */
        if( rv < 0 ) {
            rv = 1;
        } else if( rv > 0 ) {
            rv = -1;
        }
    }

 out:
    return rv;
}

/*
 * See comments above cgul_big_integer_gcd() for how this works.
 */
void
cgul_bcd__gcd(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs1,
              cgul_bcd_t rhs2)
{
    cgul_bcd_t tmp1 = NULL;
    cgul_bcd_t tmp2 = NULL;
    cgul_bcd_t tmp3 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Allocate. */
    tmp1 = cgul_bcd__new(cex);
    tmp2 = cgul_bcd__new(cex);
    tmp3 = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Initialize tmp1 to be the larger of rhs1 and rhs2; tmp2 is the
     * smaller. */
    if (cgul_bcd__compare(cex, rhs1, rhs2) >= 0) {
        cgul_bcd__assign(cex, tmp1, rhs1);
        cgul_bcd__assign(cex, tmp2, rhs2);
    } else {
        cgul_bcd__assign(cex, tmp1, rhs2);
        cgul_bcd__assign(cex, tmp2, rhs1);
    }

    /* Divide until we get a zero remainder. */
    while (!cgul_bcd__is_zero(cex, tmp2)) {

        /* Divide the larger value by the smaller value further
         * isolating the GCD in the remainder as explained in the main
         * comments to th cgul_big_integer__gcd() method. */
        cgul_bcd__divide(cex, tmp3, tmp1, tmp1, tmp2);

        /* After division, tmp1 holds the remainder which is always
         * smaller than the divisor.  So, we need to swap these values
         * because we always want to use the smaller value as the
         * divisor. */
        cgul_bcd__swap(cex, tmp1, tmp2);

        /* Check for exceptions. */
        if (*cex) {
            goto out;
        }

    }

    /* Success. */
    cgul_bcd__assign(cex, lhs, tmp1);
    lhs->is_positive = 1;

 out:

    /* Clean up. */
    if (tmp3) {
        cgul_bcd__delete(tmp3);
    }
    if (tmp2) {
        cgul_bcd__delete(tmp2);
    }
    if (tmp1) {
        cgul_bcd__delete(tmp1);
    }

    return;
}

/*
 * See comments above cgul_big_integer_lcm() for how this works.
 */
void
cgul_bcd__lcm(cgul_exception_t* cex,
              cgul_bcd_t lhs,
              cgul_bcd_t rhs1,
              cgul_bcd_t rhs2)
{
    cgul_bcd_t gcd = NULL;
    cgul_bcd_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Allocate. */
    gcd = cgul_bcd__new(cex);
    tmp = cgul_bcd__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assume there is no LCM. */
    cgul_bcd__set_to_zero(cex, lhs);

    /* Find the greatest common divisor (GCD). */
    cgul_bcd__gcd(cex, gcd, rhs1, rhs2);
    if (*cex) {
        goto out;
    }

    /* Do not divide by zero. */
    if (cgul_bcd__is_zero(cex, gcd)) {
        goto out;
    }

    /* Find the LCM as explained in the main comments to
     * cgul_big_integer__lcm(). */
    cgul_bcd__multiply(cex, lhs, rhs1, rhs2);
    cgul_bcd__divide(cex, lhs, tmp, lhs, gcd);

 out:

    /* Clean up. */
    if (tmp) {
        cgul_bcd__delete(tmp);
    }
    if (gcd) {
        cgul_bcd__delete(gcd);
    }

    return;
}

void
cgul_bcd__swap(cgul_exception_t* cex,
               cgul_bcd_t bcd_1,
               cgul_bcd_t bcd_2)
{
    struct cgul_bcd tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd_1);
    assert(bcd_2);

    /* Nothing to swap if they point to the same objects. */
    if (bcd_1 == bcd_2) {
        goto out;
    }

    memcpy((void*)&tmp, (void*)bcd_1,   sizeof(tmp));
    memcpy((void*)bcd_1,   (void*)bcd_2,   sizeof(tmp));
    memcpy((void*)bcd_2,   (void*)&tmp,  sizeof(tmp));

 out:
    return;
}

int
cgul_bcd__as_long(cgul_exception_t* cex,
                  cgul_bcd_t bcd,
                  long int* value)
{
    int rv = 0;
    long int tmp = 0;
    size_t i = 0;
    size_t digits_len = 0;
    char* digits = NULL;
    cgul_bcd_t long_min = NULL;
    cgul_bcd_t long_max = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(value);

    /* Get cgul_bcd objects for LONG_MIN and LONG_MAX. */
    long_min = cgul_bcd__new_from_long(cex, LONG_MIN);
    long_max = cgul_bcd__new_from_long(cex, LONG_MAX);

    /* Make sure "bcd" is in range. */
    if (   (cgul_bcd__compare(cex, bcd, long_min) < 0)
        || (cgul_bcd__compare(cex, bcd, long_max) > 0))
    {
        /* Out of range. */
        goto out;
    }

    /* Get the array of digits and the length of the array. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    /* Remove leading zeros from bcd. */
    cgul_bcd__remove_leading_zeros(cex, bcd);

    /* Manually convert to a long. */
    for (i = digits_len ; i > 0 ; --i) {
        tmp = 10 * tmp + digits[i - 1];
    }

    /* Make negative if necessary. */
    if (cgul_bcd__is_negative(cex, bcd)) {
        tmp = -tmp;
    }

    /* Success. */
    *value = tmp;
    rv = 1;

 out:

    /* Clean up. */
    if (long_max) {
        cgul_bcd__delete(long_max);
    }
    if (long_min) {
        cgul_bcd__delete(long_min);
    }

    return rv;
}

int
cgul_bcd__as_unsigned_long(cgul_exception_t* cex,
                           cgul_bcd_t bcd,
                           unsigned long int* value)
{
    int rv = 0;
    unsigned long int tmp = 0;
    size_t i = 0;
    size_t digits_len = 0;
    char* digits = NULL;
    cgul_bcd_t ulong_max = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(value);

    /* Get cgul_bcd objects for ULONG_MAX. */
    ulong_max = cgul_bcd__new_from_unsigned_long(cex, ULONG_MAX);

    /* Make sure "bcd" is in range. */
    if (   cgul_bcd__is_negative(cex, bcd)
        || (cgul_bcd__compare(cex, bcd, ulong_max) > 0))
    {
        /* Out of range. */
        goto out;
    }

    /* Get the array of digits and the length of the array. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    /* Remove leading zeros from bcd. */
    cgul_bcd__remove_leading_zeros(cex, bcd);

    /* Manually convert to an unsigned long. */
    for (i = digits_len ; i > 0 ; --i) {
        tmp = 10 * tmp + digits[i - 1];
    }

    /* Success. */
    *value = tmp;
    rv = 1;

 out:

    /* Clean up. */
    if (ulong_max) {
        cgul_bcd__delete(ulong_max);
    }

    return rv;
}

void
cgul_bcd__as_big_integer(cgul_exception_t* cex,
                         cgul_bcd_t bcd,
                         cgul_big_integer_t big_integer)
{
    size_t i = 0;
    size_t digits_len = 0;
    char* digits = NULL;
    cgul_big_integer_t big_ten = NULL;
    cgul_big_integer_t big_digit = NULL;
    cgul_big_integer_t accumulator = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Allocate temporary cgul_big_integer objects. */
    big_ten = cgul_big_integer__new_from_long(cex, 10);
    big_digit = cgul_big_integer__new(cex);
    accumulator = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the array of digits and the array's length. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    /* Convert "digits" from base-10 to base-16.  Remember that the
     * most-significant digit in "digits" is at "digits[digits_len -
     * 1]" so we have to iterate over "digits" in reverse order.  */
    for (i = digits_len ; i > 0 ; --i) {
        cgul_big_integer__assign_from_long(cex, big_digit, digits[i - 1]);
        cgul_big_integer__multiply(cex, accumulator, accumulator, big_ten);
        cgul_big_integer__add(cex, accumulator, accumulator, big_digit);
    }

    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_big_integer__swap(cex, big_integer, accumulator);
    if (cgul_bcd__is_zero(cex, bcd)) {
        big_integer->is_positive = 1;
    } else {
        big_integer->is_positive = bcd->is_positive;
    }

 out:

    /* Clean up. */
    if (big_ten) {
        cgul_big_integer__delete(big_ten);
    }
    if (accumulator) {
        cgul_big_integer__delete(accumulator);
    }
    if (big_digit) {
        cgul_big_integer__delete(big_digit);
    }

    return;
}

char*
cgul_bcd__as_binary(cgul_exception_t* cex,
                    cgul_bcd_t bcd)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Append the string to s. */
    cgul_bcd__append_as_binary(cex, bcd, s);
    if (*cex) {
        goto out;
    }

    /* Take the value of s and return it. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

void
cgul_bcd__append_as_binary(cgul_exception_t* cex,
                           cgul_bcd_t bcd,
                           cgul_string_t s)
{
    cgul_big_integer_t big_integer = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * The idea is that it is faster for cgul_big_integer to multiply
     * and add than it is for cgul_bcd to divide.  So rather than have
     * cgul_bcd repeatedly divide to peel off each binary digit, we
     * feed each decimal digit to cgul_big_integer and let it generate
     * the binary version of the number.  It can then convert to a
     * binary string by inspection.
     */

    /* Allocate a cgul_big_integer object. */
    big_integer = cgul_big_integer__new(cex);

    /* Convert bcd to a cgul_big_integer.  This is where the
     * multiplication and addition occurs. */
    cgul_bcd__as_big_integer(cex, bcd, big_integer);

    /* Append as binary. */
    cgul_big_integer__append_as_binary(cex, big_integer, s);

 out:

    /* Clean up. */
    if (big_integer) {
        cgul_big_integer__delete(big_integer);
    }

    return;
}

char*
cgul_bcd__as_octal(cgul_exception_t* cex,
                   cgul_bcd_t bcd)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Append the string to s. */
    cgul_bcd__append_as_octal(cex, bcd, s);
    if (*cex) {
        goto out;
    }

    /* Take the value of s and return it. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

void
cgul_bcd__append_as_octal(cgul_exception_t* cex,
                          cgul_bcd_t bcd,
                          cgul_string_t s)
{
    cgul_big_integer_t big_integer = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * The idea is that it is faster for cgul_big_integer to multiply
     * and add than it is for cgul_bcd to divide.  So rather than have
     * cgul_bcd repeatedly divide to peel off each octal digit, we
     * feed each decimal digit to cgul_big_integer and let it generate
     * the binary version of the number.  It can then convert to an
     * octal string by inspection.
     */

    /* Allocate a cgul_big_integer object. */
    big_integer = cgul_big_integer__new(cex);

    /* Convert bcd to a cgul_big_integer.  This is where the
     * multiplication and addition occurs. */
    cgul_bcd__as_big_integer(cex, bcd, big_integer);

    /* Append as octal. */
    cgul_big_integer__append_as_octal(cex, big_integer, s);

 out:

    /* Clean up. */
    if (big_integer) {
        cgul_big_integer__delete(big_integer);
    }

    return;
}

char*
cgul_bcd__as_decimal(cgul_exception_t* cex,
                     cgul_bcd_t bcd)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Append the string to s. */
    cgul_bcd__append_as_decimal(cex, bcd, s);
    if (*cex) {
        goto out;
    }

    /* Take the value of s and return it. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

void
cgul_bcd__append_as_decimal(cgul_exception_t* cex,
                            cgul_bcd_t bcd,
                            cgul_string_t s)
{
    static const char cgul_bcd__digit_as_char[] = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
    };
    int digit = 0;
    size_t i = 0;
    size_t digits_len = 0;
    char* digits = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(s);

    /* Remove any leading zeros. */
    cgul_bcd__remove_leading_zeros(cex, bcd);

    /* Get the array of digits and the array's length. */
    digits = cgul_string__get_value(cex, bcd->digits);
    digits_len = cgul_string__get_length(cex, bcd->digits);

    /* First, append the leading sign. */
    if (cgul_bcd__is_negative(cex, bcd)) {
        cgul_string__append_char(cex, s, '-');
        if (*cex) {
            goto out;
        }
    }

    /* Copy digits from "digits" to "s".  Remember that the values in
     * "digits" are the raw binary values of the decimal digits.  They
     * have to be converted to the OS's code page (typically ASCII).
     * Also remember that the most-significant digit in "digits" is at
     * "digits[digits_len - 1]" so we have to iterate over "digits" in
     * reverse order.  */
    for (i = digits_len ; i > 0 ; --i) {
        digit = digits[i - 1];
        assert((digit >= 0) && (digit <= 9));
        cgul_string__append_char(cex, s, cgul_bcd__digit_as_char[digit]);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

char*
cgul_bcd__as_hexadecimal(cgul_exception_t* cex,
                         cgul_bcd_t bcd)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Append the string to s. */
    cgul_bcd__append_as_hexadecimal(cex, bcd, s);
    if (*cex) {
        goto out;
    }

    /* Take the value of s and return it. */
    rv = cgul_string__take_value(cex, s);

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

#if 1
void
cgul_bcd__append_as_hexadecimal(cgul_exception_t* cex,
                                cgul_bcd_t bcd,
                                cgul_string_t s)
{
    cgul_big_integer_t big_integer = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * The idea is that it is faster for cgul_big_integer to multiply
     * and add than it is for cgul_bcd to divide.  So rather than have
     * cgul_bcd repeatedly divide to peel off each hexadecimal digit,
     * we feed each decimal digit to cgul_big_integer and let it
     * generate the binary version of the number.  It can then convert
     * to a hexadecimal string by inspection.
     */

    /* Allocate a cgul_big_integer object. */
    big_integer = cgul_big_integer__new(cex);

    /* Convert bcd to a cgul_big_integer.  This is where the
     * multiplication and addition occurs. */
    cgul_bcd__as_big_integer(cex, bcd, big_integer);

    /* Append as hexadecimal. */
    cgul_big_integer__append_as_hexadecimal(cex, big_integer, s);

 out:

    /* Clean up. */
    if (big_integer) {
        cgul_big_integer__delete(big_integer);
    }

    return;
}

#else
void
cgul_bcd__append_as_hexadecimal(cgul_exception_t* cex,
                                cgul_bcd_t bcd,
                                cgul_string_t s)
{
    static char cgul_bcd__hexadecimal_digits[] = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
    };

    long int digit = 0;
    size_t digits_length = 0;
    cgul_bcd_t accumulator = NULL;
    cgul_bcd_t remainder = NULL;
    cgul_bcd_t sixteen = NULL;
    cgul_string_t tmp = NULL;

    /*
     *                         WARNING
     *
     * This method is commented out!  I'm keeping it around though
     * because I want to be able to compare different implementations
     * of this method.  Also, I want to see how much a better division
     * algorithm can help speed up this method.
     *
     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(s);

    /* Check for zero up front. */
    if (cgul_bcd__is_zero(cex, bcd)) {
        cgul_string__append(cex, s, "0x0");
        goto out;
    }

    /* Allocate temporary cgul_bcd objects. */
    accumulator = cgul_bcd__new(cex);
    remainder = cgul_bcd__new(cex);
    sixteen = cgul_bcd__new_from_unsigned_long(cex, 16);

    /* Allocate a temporary string for storing the result. */
    tmp = cgul_string__new(cex);

    /* accumulator = bcd */
    cgul_bcd__assign(cex, accumulator, bcd);

    /* Iterate over each hexadecimal digit in the accumulator.  If bcd
     * is zero to start off with, we already checked for this above
     * and short circuited process so that we do not reach here. */
    while (!cgul_bcd__is_zero(cex, accumulator)) {

        /* Repeatedly divide the accumulator by sixteen.  This stores
         * the next hexadecimal digit in the remainder. */
        cgul_bcd__divide(cex, accumulator, remainder, accumulator, sixteen);
        if (*cex) {
            goto out;
        }

        /* Get the hexadecimal digit as a long. */
        digits_length = cgul_string__get_length(cex, remainder->digits);
        switch (digits_length) {
        case 1:
            digit = cgul_string__get_value_at(cex, remainder->digits, 0);
            break;
        case 2:
            digit = cgul_string__get_value_at(cex, remainder->digits, 0);
            digit += 10 * cgul_string__get_value_at(cex, remainder->digits, 1);
            break;
        default:
            *cex = cgul_exception__new(
                       "cgul_bcd__append_as_hexadecimal: "
                       "Internal Error: Unable to convert digit to index.");
            goto out;
        }

        /* The digit should always be less than 16. */
        assert((digit >= 0) && (digit <= 15));

        /* Add the digit to the string. */
        cgul_string__append_char(
            cex,
            tmp,
            cgul_bcd__hexadecimal_digits[digit]);

        if (*cex) {
            goto out;
        }

    }

    /* Append the leading "0x".  Note that everything is still going
     * on the string in reverse order. */
    cgul_string__append_char(cex, tmp, 'x');
    cgul_string__append_char(cex, tmp, '0');

    /* Append the sign if necessary and if it is not a negative zero. */
    if (   !bcd->is_positive
        && !cgul_bcd__is_zero(cex, bcd))
    {
        cgul_string__append_char(cex, tmp, '-');
    }

    /* Above, we pushed all the characters onto the string in reverse order. */
    cgul_string__reverse(cex, tmp);

    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_string__swap(cex, s, tmp);

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }
    if (sixteen) {
        cgul_bcd__delete(sixteen);
    }
    if (remainder) {
        cgul_bcd__delete(remainder);
    }
    if (accumulator) {
        cgul_bcd__delete(accumulator);
    }

    return;
}
#endif

void
cgul_bcd__print_as_binary(cgul_exception_t* cex,
                          cgul_bcd_t bcd,
                          FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(f);

    s = cgul_bcd__as_binary(cex, bcd);
    cgul_libc__fprintf(cex, f, "%s", s);

 out:
    if (s) {
        free(s);
    }
}

void
cgul_bcd__print_as_octal(cgul_exception_t* cex,
                         cgul_bcd_t bcd,
                         FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(f);

    s = cgul_bcd__as_octal(cex, bcd);
    cgul_libc__fprintf(cex, f, "%s", s);

 out:
    if (s) {
        free(s);
    }
}

void
cgul_bcd__print_as_decimal(cgul_exception_t* cex,
                           cgul_bcd_t bcd,
                           FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(f);

    s = cgul_bcd__as_decimal(cex, bcd);
    cgul_libc__fprintf(cex, f, "%s", s);

 out:
    if (s) {
        free(s);
    }
}

void
cgul_bcd__print_as_hexadecimal(cgul_exception_t* cex,
                               cgul_bcd_t bcd,
                               FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bcd);
    assert(f);

    s = cgul_bcd__as_hexadecimal(cex, bcd);
    cgul_libc__fprintf(cex, f, "%s", s);

 out:
    if (s) {
        free(s);
    }
}
