#include "cgul_big_integer_p.h"
#include "cgul_bcd_p.h"
#include "cgul_libc.h"
#include "cgul_bcd.h"
#include "cgul_defensive.h"
#include "cgul_random_lcg256.h"
#include "cgul_string_p.h"

static size_t CGUL_BIG_INTEGER__DEFAULT_MINIMUM_CAPACITY_INCREMENT = 16;

/**
 * Set the capacity to <code>capacity</code> for the
 * <code>big_integer</code>.  If <code>capacity</code> is less than
 * the current size, the current size is used instead.  If
 * <code>capacity</code> is zero, the default capacity is used
 * instead.  If memory cannot be allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>cgul_big_integer</code> instance
 * @param[in] capacity new capacity to request
 */
static void
cgul_big_integer__set_capacity(cgul_exception_t* cex,
                               cgul_big_integer_t big_integer,
                               unsigned long int capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* Sanity check. */
    if (capacity == 0) {
        capacity = big_integer->minimum_capacity_increment;
    }

    /* Make sure the capacity is at least large enough to hold the
     * current number of elements. */
    capacity = CGUL_MAX(capacity, big_integer->size);

#if 0
    fprintf(stderr, "cgul_big_integer__set_capacity = %lu\n", capacity);
    fflush(stderr);
#endif

    /* realloc */
    big_integer->value = (cbi_element_t*)cgul_libc__realloc(
                             cex,
                             big_integer->value,
                             sizeof(*big_integer->value) * capacity);
    if (*cex) {
        goto out;
    }

    /* Rememeber the new capacity. */
    big_integer->capacity = capacity;

    /* Clear the extra elements. */
    if (capacity > big_integer->size) {
        memset(big_integer->value + big_integer->size,
               0,
               sizeof(*big_integer->value)
               * (big_integer->capacity - big_integer->size));
    }

 out:
    return;
}

/**
 * Decrease the capacity of the underlying array if necessary such
 * that the capacity is reduced by 1/2 if the size is 1/4th the
 * capacity.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>big_integer</code> instance
 */
static void
cgul_big_integer__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                                 cgul_big_integer_t big_integer)
{
    size_t actual_new_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We only have excess capacity when 1/4 capacity is greater than
     * or equal to the size.  So, do nothing when 1/4 capacity is less
     * than the size. */
    if ((big_integer->capacity >> 2) < big_integer->size) {
        goto out;
    }

    /* Keep halving the capacity to get the best fit.  It should be at
     * least double the size to avoid thrashing the allocator.  We
     * know from the if() above that we can start iterating with 1/2
     * of big_integer->capacity. */
    actual_new_capacity = big_integer->capacity >> 1;
    while (   ((actual_new_capacity >> 1) > 0)
           && ((actual_new_capacity >> 1) >= (big_integer->size << 1)))
    {
        actual_new_capacity >>= 1;
    }

    /* The actual new capacity should never drop below the minimum
     * capacity. */
    actual_new_capacity = CGUL_MAX(actual_new_capacity,
                                   big_integer->minimum_capacity_increment);
    assert(actual_new_capacity > 0);

    /* If the capacity is already less than or equal to the actual new
     * capacity, do nothing. */
    if (big_integer->capacity <= actual_new_capacity) {
        goto out;
    }

    /* Sanity check to make sure we are not going to accidentally
     * reduce the capacity below the necessary size. */
    if (actual_new_capacity <= big_integer->size) {
        goto out;
    }

    /* Set the reduced capacity. */
    cgul_big_integer__set_capacity(cex, big_integer, actual_new_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Increase the capacity of the underlying array if necessary.  If the
 * capacity needs to be adjusted but memory cannot be allocated, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] big_integer <code>big_integer</code> instance
 * @param[in] needed_capacity needed capacity
 */
static void
cgul_big_integer__increase_capacity_if_necessary(cgul_exception_t* cex,
                                                 cgul_big_integer_t big_integer,
                                                 size_t needed_capacity)
{
    size_t target_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* Sanity check. */
    if (needed_capacity == 0) {
        goto out;
    }

    /* If we need more capacity, realloc() it.*/
    if (needed_capacity > big_integer->capacity) {

        /* Try to double the capacity to allow room to grow.  The test
         * against zero is used to detect overflow. */
        target_capacity = big_integer->capacity;
        while (   ((target_capacity << 1) > 0)
               && ((target_capacity << 1) <= (needed_capacity << 1)))
        {
            target_capacity <<= 1;
        }

        /* If there is overflow, just use needed_capacity as-is which
         * doesn't allow for room to grow without reallocating but is
         * large enough to hold the new elements.  Also, we know that
         * needed_capacity did not overflow when calculated above
         * because an exception would have been thrown. */
        if (target_capacity < needed_capacity) {
            target_capacity = needed_capacity;
        }

        /* Set the new capacity. */
        cgul_big_integer__set_capacity(cex, big_integer, target_capacity);
    }

 out:
    return;
}

/**
 * This function converts a binary ascii character into a long and
 * returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_big_integer__get_binary_digit(cgul_exception_t* cex,
                                   char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * This function converts an octal ascii character into a long and
 * returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_big_integer__get_octal_digit(cgul_exception_t* cex,
                                  char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * This function converts a decimal ascii character into a long and
 * returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_big_integer__get_decimal_digit(cgul_exception_t* cex,
                                    char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    case '8':
        rv = 8;
        break;
    case '9':
        rv = 9;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * This function converts a hexadecimal ascii character into a long
 * and returns it.  If a conversion cannot be made, it returns -1.
 *
 * @param[in] cex c-style exception
 * @param[in] c character representation of a digit
 * @return conversion of <code>c</code> or -1 if no conversion is possible
 */
static long
cgul_big_integer__get_hexadecimal_digit(cgul_exception_t* cex,
                                        char c)
{
    long rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch( c ) {
    case '0':
        rv = 0;
        break;
    case '1':
        rv = 1;
        break;
    case '2':
        rv = 2;
        break;
    case '3':
        rv = 3;
        break;
    case '4':
        rv = 4;
        break;
    case '5':
        rv = 5;
        break;
    case '6':
        rv = 6;
        break;
    case '7':
        rv = 7;
        break;
    case '8':
        rv = 8;
        break;
    case '9':
        rv = 9;
        break;
    case 'a':
    case 'A':
        rv = 10;
        break;
    case 'b':
    case 'B':
        rv = 11;
        break;
    case 'c':
    case 'C':
        rv = 12;
        break;
    case 'd':
    case 'D':
        rv = 13;
        break;
    case 'e':
    case 'E':
        rv = 14;
        break;
    case 'f':
    case 'F':
        rv = 15;
        break;
    default:
        break;
    }

 out:
    return rv;
}

/**
 * Pass in the full string in <code>value</code>.  Pass in the length
 * of <code>value</code> in <code>value_length</code>.  Pass in the
 * index in <code>*i</code> of the first character where you want this
 * function to start looking for a prefix that determines the base.
 * The index is to help you scan pass leading white space in
 * <code>value</code>.  This function returns 16 if the string at
 * <code>&value[*i]</code> starts with "0x" or "0X".  Returns 2 if it
 * starts with "0b" or "0B".  Returns 8 if it starts with "0" followed
 * by a digit between 0 and 7.  It otherwise returns 10.  Also, it
 * will increment <code>*i</code> to index one past the last character
 * in the prefix that determines the base.
 *
 * @param[in] cex c-style exception
 * @param[in] value value of big integer as a string
 * @param[in] value_length length of <code>value</code>
 * @param[in,out] index into <code>value</code>
 * @return base to use for <code>value</code>
 */
static long
cgul_big_integer__determine_base(cgul_exception_t* cex,
                                 const char* value,
                                 size_t value_length,
                                 size_t* i)
{
    long rv = 10;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(value);
    assert(i);

    if( value[*i] == '0' ) {
        size_t next = *i + 1;
        if( next < value_length ) {
            if( (value[next] == 'x') || (value[next] == 'X') ) {
                rv = 16;
                *i += 2;
            } else if( (value[next] == 'b') || (value[next] == 'B') ) {
                rv = 2;
                *i += 2;
            } else {
                long digit =
                    cgul_big_integer__get_octal_digit(cex, value[next]);
                if (digit != -1) {
                    rv = 8;
                    ++*i;
                }
            }
        }
    }

 out:
    return rv;
}

/**
 * It is an invariant of this class that the "size" attribute is
 * always the smallest value possible.  Call this function when you
 * need to resize the "value" array.
 *
 * @param[in] cex c-style exception
 * @param[in] big_integer <code>big_integer</code> integer
 */
static void
cgul_big_integer__minimize_size(cgul_exception_t* cex,
                                cgul_big_integer_t big_integer)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* Find the index of the first byte that is non-zero stopping
     * before you reach the byte indexed by 0. */
    for( i = big_integer->size - 1 ; i > 0 ; --i ) {
        if( big_integer->value[i] ) {
            break;
        }
    }

    /* The size is always at least one. */
    big_integer->size = i + 1;

    /* Decrease the capacity if necessary.  The call to
     * cgul_big_integer__decrease_capacity_if_necessary() is wrapped
     * in an if() to avoid making unnecessary calls to the function
     * which is otherwise expensive. */
    if ((big_integer->size << 2) <= big_integer->capacity) {
        cgul_big_integer__decrease_capacity_if_necessary(cex, big_integer);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

/**
 * Skip leading white space.
 *
 * @param[in] cex c-style exception
 * @param[in] s string
 * @param[in] s_length length of <code>s</code>
 * @param[in,out] index into <code>s</code>
 */
static void
cgul_big_integer__skip_white_space(cgul_exception_t* cex,
                                   const char* s,
                                   size_t s_length,
                                   size_t* i)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(i);

    for( ; *i < s_length ; ++*i ) {
        if( !isspace((int)s[*i]) ) {
            break;
        }
    }

 out:
    return;
}

/**
 * Add the magnitudes of rhs1 and rhs2.  Put the value in lhs without
 * changing the sign of lhs.  It is safe for <code>rhs1</code> or
 * <code>rhs2</code> and <code>lhs</code> to point to the same
 * underlying <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
static void
cgul_big_integer__add_magnitudes(cgul_exception_t* cex,
                                 cgul_big_integer_t lhs,
                                 cgul_big_integer_t rhs1,
                                 cgul_big_integer_t rhs2)
{
    cbi_element_double_t sum = 0;
    cbi_element_double_t carry = 0;
    cgul_big_integer_t n_longer = NULL;
    cgul_big_integer_t n_shorter = NULL;
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Determine the relative lengths of rhs1 and rhs2. */
    if( rhs1->size >= rhs2->size ){
        n_longer = rhs1;
        n_shorter = rhs2;
    } else {
        n_longer = rhs2;
        n_shorter = rhs1;
    }

    /* Make sure lhs is large enough for addition plus a possible carry. */
    cgul_big_integer__increase_capacity_if_necessary(
        cex,
        lhs,
        cgul_defensive__add_size_type(cex, n_longer->size, 1));
    if (*cex) {
        goto out;
    }

    /* Clear the bytes in lhs that are no longer significant.
     * (Remember, lhs could have been much larger than n_longer
     * immediately before calling this function.)  You also need to
     * clear the byte that may or may not end up holding the carry. */
    if( lhs->size > n_longer->size ) {
        memset(lhs->value + n_longer->size,
               0,
               (lhs->size - n_longer->size) * sizeof(cbi_element_t));
    }

    /* Add together the "cbi_element_t" elements that align for both
       numbers.  This exhausts the elements of n_shorter. */
    for( i = 0 ; i < n_shorter->size ; ++i ) {
        sum =   (cbi_element_double_t)n_longer->value[i]
              + (cbi_element_double_t)n_shorter->value[i]
              + carry;
        carry = sum >> CBI_ELEMENT_BITS;
        lhs->value[i] = (cbi_element_t)sum;
    }

    /* Propagate the carry through the remaining elements of n_longer. */
    for( ; i < n_longer->size ; ++i ) {
        sum = (cbi_element_double_t)n_longer->value[i] + carry;
        carry = sum >> CBI_ELEMENT_BITS;
        lhs->value[i] = (cbi_element_t)sum;
    }

    /* Add the final carry. */
    lhs->value[i] = (cbi_element_t)carry;

    /* Tighten up the size. */
    lhs->size = n_longer->size + 1;
    cgul_big_integer__minimize_size(cex, lhs);

 out:
    return;
}

/**
 * Subtract the magnitude of <code>rhs2</code> from <code>rhs1</code>.
 * Put the value in <code>lhs</code> without changing the sign of
 * <code>lhs</code>.  It is IMPORTANT that the caller arrange for
 * <code>rhs1</code> to ALWAYS have a larger magnitude than
 * <code>rhs2</code>.  Remember, a larger magnitude does not
 * necessarily mean <code>rhs1</code> is greater than
 * <code>rhs2</code>.  (Think about two negative numbers.)  You can
 * determine which has the greater magnitude by calling
 * <code>cgul_big_integer__compare_magnitudes()</code>.  It is safe
 * for <code>rhs1</code> or <code>rhs2</code> and <code>lhs</code> to
 * point to the same underlying <code>cgul_big_integer</code> object.
 *
 * @param[in,out] cex c-style exception
 * @param[out] lhs left-hand side
 * @param[in] rhs1 first right-hand side
 * @param[in] rhs2 second right-hand side
 */
static void
cgul_big_integer__subtract_magnitudes(cgul_exception_t* cex,
                                      cgul_big_integer_t lhs,
                                      cgul_big_integer_t rhs1,
                                      cgul_big_integer_t rhs2)
{
    size_t i = 0;
    cbi_element_double_t d = 0;
    cbi_element_double_t base = 0;
    int borrow = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Find the base so that when we borrow, we know how much to add
     * to the digit that does the borrowing. */
    base = 1;
    base <<= CBI_ELEMENT_BITS;

    /* Make sure lhs has enough capacity. */
    if (lhs->capacity < rhs1->capacity) {
        cgul_big_integer__increase_capacity_if_necessary(cex,
                                                         lhs,
                                                         rhs1->capacity);
        if (*cex) {
            goto out;
        }
    }

    /* Subtract the "cbi_element_t" elements where both numbers align.
     * This exhausts the elements of rhs2. */
    for (i = 0 ; i < rhs2->size ; ++i) {

        /* Get the next element. */
        d = rhs1->value[i];

        /* If we have a pending borrow ... */
        if (borrow) {
            /* ... and this digit is not zero, we can satisfy and
             * clear the borrow by subtracting one from d. */
            if (d > 0) {
                --d;
                borrow = 0;
            } else {
                /* ... and this digit is zero, the borrow passes
                 * through causing this digit to be all 1's. */
                d = ~0;
            }
        }

        /* Determine if a borrow is needed. */
        if (d < rhs2->value[i]) {
            /* This is the same as "d += base" given that "base" has
             * no low-order bits set. */
            d |= base;
            borrow = 1;
        }

        /*
         * Subtract.  If you ignore the casting, you have the
         * following:
         *
         *  lhs->value[i] = d - rhs2->value[i].
         */
        lhs->value[i] =
            (cbi_element_t)(d - (cbi_element_double_t)rhs2->value[i]);
    }

    /* Propagate any borrow that remains. */
    for ( ; i < rhs1->size ; ++i) {

        /* With borrow == 0, this loop just copies digits from rhs1 to
         * lhs, but if lhs == rhs1, the digits are already present. */
        if ((borrow == 0) && (lhs == rhs1)) {
            /* Update i so we know the length of the string. */
            i = rhs1->size;
            break;
        }

        /* Get the next digit from rhs1. */
        d = rhs1->value[i];

        /* If we have a pending borrow ... */
        if (borrow) {
            /* ... and this digit is not zero, we can satisfy and
             * clear the borrow by subtracting one from d. */
            if (d > 0) {
                --d;
                borrow = 0;
            } else {
                /* ... and this digit is zero, the borrow passes
                 * through causing this digit to be all 1's. */
                d = ~0;
            }
        }

        /* Subtract (d - 0 = d). */
        lhs->value[i] = (cbi_element_t)d;
        
    }

    /* We have finished subtracting.  If we still have a borrow, it
     * means rhs1 had a smaller magnitude than rhs2 which is not
     * allowed. */
    assert(borrow == 0);

    /* Tighten up the size. */
    lhs->size = rhs1->size;
    cgul_big_integer__minimize_size(cex, lhs);

 out:
    return;
}

/**
 * Create a new <code>cgul_big_integer</code> instance initialized to
 * positive 0 and a starting capacity of <code>capacity</code> and a
 * minimum capacity increment of
 * <code>minimum_capacity_increment</code>.  The caller is responsible
 * for calling <code>cgul_big_integer__delete()</code> on the value
 * returned.  If memory allocation fails, <code>NULL</code> is
 * returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] capacity initial capacity
 * @param[in] minimum_capacity_increment minimum capacity increment
 * @return new <code>cgul_big_integer</code> instance
 */
static cgul_big_integer_t
cgul_big_integer__new_with_capacity(cgul_exception_t* cex,
                                    size_t capacity,
                                    size_t minimum_capacity_increment)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Bounds check "capacity". */
    if (capacity == 0) {
        *cex = cgul_exception__new(
                   "cgul_big_integer__new_with_capacity: "
                   "capacity cannot be zero");
        goto out;
    }

    /* Bounds check "minimum_capacity_increment". */
    if (minimum_capacity_increment == 0) {
        *cex = cgul_exception__new(
                   "cgul_big_integer__new_with_capacity: "
                   "minimum capacity increment cannot be zero");
        goto out;
    }

    /* Allocate. */
    rv = (cgul_big_integer_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Initialize. */
    rv->value =
        (cbi_element_t*)cgul_libc__calloc(cex, capacity, sizeof(*rv->value));
    if (*cex) {
        goto out;
    }

    rv->capacity = capacity;
    rv->minimum_capacity_increment = minimum_capacity_increment;
    rv->is_positive = 1;
    rv->size = 1;

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_big_integer_t
cgul_big_integer__new(cgul_exception_t* cex)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = cgul_big_integer__new_with_capacity(
                     cex,
                     CGUL_BIG_INTEGER__DEFAULT_MINIMUM_CAPACITY_INCREMENT,
                     CGUL_BIG_INTEGER__DEFAULT_MINIMUM_CAPACITY_INCREMENT);

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_big_integer_t
cgul_big_integer__new_from_big_integer(cgul_exception_t* cex,
                                       cgul_big_integer_t value)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign the value. */
    cgul_big_integer__assign(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_big_integer_t
cgul_big_integer__new_from_long(cgul_exception_t* cex,
                                long value)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign the value. */
    cgul_big_integer__assign_from_long(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_big_integer_t
cgul_big_integer__new_from_unsigned_long(cgul_exception_t* cex,
                                         unsigned long value)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign the value. */
    cgul_big_integer__assign_from_unsigned_long(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_big_integer_t
cgul_big_integer__new_from_size_type(cgul_exception_t* cex,
                                     size_t value)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign the value. */
    cgul_big_integer__assign_from_size_type(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_big_integer_t
cgul_big_integer__new_from_string(cgul_exception_t* cex,
                                  const char* s)
{
    cgul_big_integer_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Create a default object. */
    rv = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign the value. */
    cgul_big_integer__assign_from_string(cex, rv, s);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_big_integer__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_big_integer__delete(cgul_big_integer_t big_integer)
{
    if (big_integer) {
        if (big_integer->value) {
            free(big_integer->value);
        }
        free(big_integer);
    }
}

int
cgul_big_integer__is_positive(cgul_exception_t* cex,
                              cgul_big_integer_t big_integer)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    rv =    big_integer->is_positive
        && !cgul_big_integer__is_zero(cex, big_integer);

 out:
    return rv;
}

int
cgul_big_integer__is_negative(cgul_exception_t* cex,
                              cgul_big_integer_t big_integer)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    rv =   !big_integer->is_positive
        && !cgul_big_integer__is_zero(cex, big_integer);

 out:
    return rv;
}

int
cgul_big_integer__is_zero(cgul_exception_t* cex,
                          cgul_big_integer_t big_integer)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* Normalize the size so we can check the length of big_integer. */
    cgul_big_integer__minimize_size(cex, big_integer);

    /* There is a positive zero and a negative zero.  Both are "zero".
     * Notice how the sign is disregarded. */
    rv =   (big_integer->size == 1)
        && (big_integer->value[0] == (cbi_element_t)0);

 out:
    return rv;
}

void
cgul_big_integer__set_to_zero(cgul_exception_t* cex,
                              cgul_big_integer_t big_integer)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* You only need to clear the part that you were using.  The rest
     * should already be cleared. */
    memset(big_integer->value, 0, big_integer->size * sizeof(cbi_element_t));
    big_integer->is_positive = 1;
    big_integer->size = 1;

 out:
    return;
}

int
cgul_big_integer__is_one(cgul_exception_t* cex,
                         cgul_big_integer_t big_integer)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    rv =   big_integer->is_positive
        && (big_integer->size == 1)
        && (big_integer->value[0] == (cbi_element_t)1);

 out:
    return rv;
}

void
cgul_big_integer__set_to_one(cgul_exception_t* cex,
                             cgul_big_integer_t big_integer)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* You only need to clear the part that you were using.  The rest
     * should already be cleared. */
    memset(big_integer->value, 0, big_integer->size * sizeof(cbi_element_t));
    big_integer->is_positive = 1;
    big_integer->size = 1;
    big_integer->value[0] = (cbi_element_t)1;

 out:
    return;
}

int
cgul_big_integer__is_even(cgul_exception_t* cex,
                          cgul_big_integer_t big_integer)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(big_integer->value);
    assert(big_integer->size);

    /* The number is even if the least-significant bit is not set. */
    if ((big_integer->value[0] & 0x1) == 0) {
        rv = 1;
    }

 out:
    return rv;
}

int
cgul_big_integer__is_odd(cgul_exception_t* cex,
                         cgul_big_integer_t big_integer)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(big_integer->value);
    assert(big_integer->size);

    /* The number is odd if the least-significant bit is set. */
    if ((big_integer->value[0] & 0x1) == 1) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_big_integer__assign(cgul_exception_t* cex,
                         cgul_big_integer_t lhs,
                         cgul_big_integer_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    if( lhs != rhs ) {

        /* Minimize the size of rhs. */
        cgul_big_integer__minimize_size(cex, rhs);

        /* Add capacity if necessary. */
        cgul_big_integer__increase_capacity_if_necessary(cex, lhs, rhs->size);
        if (*cex) {
            goto out;
        }

        /* Copy the value. */
        memcpy(lhs->value, rhs->value, rhs->size * sizeof(cbi_element_t));

        /* Clear the bytes that are no longer in use (if any). */
        if( lhs->size > rhs->size ) {
            memset(lhs->value + rhs->size,
                   0,
                   (lhs->size - rhs->size) * sizeof(cbi_element_t));
        }

        /* Set the size. */
        lhs->size = rhs->size;

        /* Set the sign. */
        if (cgul_big_integer__is_zero(cex, rhs)) {
            lhs->is_positive = 1;
        } else {
            lhs->is_positive = rhs->is_positive;
        }
    }

 out:
    return;
}

void
cgul_big_integer__assign_from_long(cgul_exception_t* cex,
                                   cgul_big_integer_t lhs,
                                   long rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    if( rhs < 0 ) {
        cgul_big_integer__assign_from_unsigned_long(cex, lhs, -rhs);
        lhs->is_positive = 0;
    } else {
        cgul_big_integer__assign_from_unsigned_long(cex, lhs, rhs);
    }

 out:
    return;
}

void
cgul_big_integer__assign_from_unsigned_long(cgul_exception_t* cex,
                                            cgul_big_integer_t lhs,
                                            unsigned long rhs)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    /* Add capacity if necessary. */
    cgul_big_integer__increase_capacity_if_necessary(cex,
                                                     lhs,
                                                     sizeof(unsigned long));

    /* Set lhs to zero. */
    cgul_big_integer__set_to_zero(cex, lhs);
    if (*cex) {
        goto out;
    }

    /* By definition, the "unsigned long" is too large to be contained
     * by a single "cbi_element_t".  So break it into multiple
     * "cbi_element_t" chunks.  Note that the code below also avoids
     * byte-swapping and compile-time size issues while assigning. */
    for( i = 0 ; rhs != 0 ; ++i, rhs >>= CBI_ELEMENT_BITS ) {
        lhs->value[i] = (cbi_element_t)rhs;
    }

    /* Tighten up the size.  Note that the size is always at least one. */
    lhs->size = CGUL_MAX(i, 1);
    cgul_big_integer__minimize_size(cex, lhs);

    /* Make sure lhs is positive. */
    lhs->is_positive = 1;

 out:
    return;
}

void
cgul_big_integer__assign_from_size_type(cgul_exception_t* cex,
                                        cgul_big_integer_t lhs,
                                        size_t rhs)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);

    /* Add capacity if necessary. */
    cgul_big_integer__increase_capacity_if_necessary(cex,
                                                     lhs,
                                                     sizeof(size_t));

    /* Set lhs to zero. */
    cgul_big_integer__set_to_zero(cex, lhs);
    if (*cex) {
        goto out;
    }

    /* By definition, the "size_t" is too large to be contained
     * by a single "cbi_element_t".  So break it into multiple
     * "cbi_element_t" chunks.  Note that the code below also avoids
     * byte-swapping and compile-time size issues while assigning. */
    for( i = 0 ; rhs != 0 ; ++i, rhs >>= CBI_ELEMENT_BITS ) {
        lhs->value[i] = (cbi_element_t)rhs;
    }

    /* Tighten up the size.  Note that the size is always at least one. */
    lhs->size = CGUL_MAX(i, 1);
    cgul_big_integer__minimize_size(cex, lhs);

    /* Make sure lhs is positive. */
    lhs->is_positive = 1;

 out:
    return;
}

void
cgul_big_integer__assign_from_string(cgul_exception_t* cex,
                                     cgul_big_integer_t lhs,
                                     const char* rhs)
{
    /* Error message if conversion fails. */
    static const char* errmsg = "cgul_big_integer__assign_from_string: "
                                "invalid conversion";

    /* We need to know what the current locale considers to be the
     * thousands separator. */
    struct lconv* locale_info = NULL;

    /* thousands separator */
    char thousands_separator = '\0';

    /* Determine the base for the digits in the "rhs" string.  This is
     * either 16 (hexadecimal), 10 (decimal), 8 (octal), or 2
     * (binary). */
    long rhs_base = 10;

    /* Function used to determine if the digit in "rhs" is valid for
     * the base of "rhs". */
    long (*local_get_digit)(cgul_exception_t*, char) = NULL;

    /* Iterator for traversing "rhs". */
    size_t i = 0;

    /* String length of "rhs". */
    size_t rhs_length = 0;

    /* Save the sign.  You need to set this last because multiplying
     * by zero will set the sign back to positive. */
    int is_positive_saved = 1;

    /* Next digit to convert. */
    long digit = 0;

    /* "digit" as a big_integer. */
    cgul_big_integer_t big_digit = NULL;

    /* The base as a big_integer. */
    cgul_big_integer_t big_rhs_base =  NULL;

    /* Store the converstion in "tmp" and swap with "lhs" only if the
     * conversion completes successfully. */
    cgul_big_integer_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    rhs_length = strlen(rhs);
    locale_info = localeconv();
    thousands_separator = *locale_info->thousands_sep;

    /*
     * Search for leading positive or negative sign.
     */

    cgul_big_integer__skip_white_space(cex, rhs, rhs_length, &i);

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    /* See if there is a leading positive or negative sign.  I believe
     * all locales uses '+' and '-' for LC_NUMERICAL positive and
     * negative signs (as opposed to LC_MONETARY which accounts for
     * special monetary versions of positive and negative). */
    if( rhs[i] == '+' ) {
        is_positive_saved = 1;
        ++i;
    } else if( rhs[i] == '-' ) {
        is_positive_saved = 0;
        ++i;
    }

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    cgul_big_integer__skip_white_space(cex, rhs, rhs_length, &i);

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    /* Determine the base of "rhs". */
    rhs_base = cgul_big_integer__determine_base(cex, rhs, rhs_length, &i);
 
    /* Set up function that determines whether the digit in "rhs" is
     * valid for the particular base.  */
    if( rhs_base == 10 ) {
        local_get_digit = &cgul_big_integer__get_decimal_digit;
    } else if( rhs_base == 16 ) {
        local_get_digit = &cgul_big_integer__get_hexadecimal_digit;
    } else if( rhs_base == 8 ) {
        local_get_digit = &cgul_big_integer__get_octal_digit;
    } else if( rhs_base == 2 ) {
        local_get_digit = &cgul_big_integer__get_binary_digit;
    }

    if( i >= rhs_length ) {
        *cex = cgul_exception__new("%s", errmsg);
        goto out;
    }

    /* Prepare to do the arithmetic. */
    big_digit = cgul_big_integer__new(cex);
    big_rhs_base = cgul_big_integer__new_from_long(cex, rhs_base);
    /* We will swap tmp with lhs.  So, it should have the same capacity. */
    tmp = cgul_big_integer__new_with_capacity(cex,
                                              lhs->capacity,
                                              lhs->minimum_capacity_increment);
    if (*cex) {
        goto out;
    }

    /* Iterate over the digits in "rhs". */
    for( ; i < rhs_length ; ++i ) {
        if( rhs[i] == thousands_separator ) {
            continue;
        }
        if( (digit = local_get_digit(cex, rhs[i])) == -1 ) {
            *cex = cgul_exception__new("%s", errmsg);
            goto out;
        }
        cgul_big_integer__multiply(cex, tmp, tmp, big_rhs_base);
        cgul_big_integer__assign_from_long(cex, big_digit, digit);
        cgul_big_integer__add_magnitudes(cex, tmp, tmp, big_digit);
        if (*cex) {
            goto out;
        }
    }

    /* Success! */
    cgul_big_integer__swap(cex, lhs, tmp);
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = is_positive_saved;
    }

 out:

    if (tmp) {
        cgul_big_integer__delete(tmp);
    }
    if (big_rhs_base) {
        cgul_big_integer__delete(big_rhs_base);
    }
    if (big_digit) {
        cgul_big_integer__delete(big_digit);
    }

    return;
}

void
cgul_big_integer__negate(cgul_exception_t* cex,
                         cgul_big_integer_t lhs,
                         cgul_big_integer_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    cgul_big_integer__assign(cex, lhs, rhs);
    if (*cex) {
        goto out;
    }

    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive ? 0 : 1;
    }

 out:
    return;
}

void
cgul_big_integer__abs(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    cgul_big_integer__assign(cex, lhs, rhs);
    lhs->is_positive = 1;

 out:
    return;
}

/*
 * =============================
 * Inversion Using Newton Method:
 * =============================
 *
 * General:
 * =======
 *
 * This method uses Newton's Method to calculate the scaled
 * multiplicative inverse of rhs:
 *
 *            1        scale
 *     lhs = ---  *  2
 *           rhs
 *
 *
 * Newton's Method:
 * ===============
 *
 * Briefly, Newton's Method is an iterative approach where an initial
 * value for x at x0 is specified.  Then, y = f(x0) is calculated which
 * (of course) determines a point that is "y" units above the x-axis.
 *
 * It is then possible to f'(x) to create a linear approximation for
 * f(x) at f(x0).  This is just a line with slope m equal dy/dx.
 * Because we know the change along the y-axis (y = f(x0)) and because
 * we know the slope (m), it is possible to determine the change along
 * the x-axis:
 *
 *                      dx        f(x0)
 *     x1 - x0  =  y * ----  =  --------
 *                      dy       f'(x0)
 *
 * Thus, the point at which the line described by m and f(x0)
 * intersects the x-axis is:
 * 
 *                   f(x0)
 *     x1 = x0  -  --------
 *                  f'(x0)
 *
 * If the process is repeated and if f(x) is well-behaved, each
 * successive point along the x-axis will converge to a root of f(x):
 *
 *                    f(x)
 *     x     =  x -  ------
 *      n+1      n    f'(x)
 *
 *
 * Controlling Equation:
 * ====================
 *
 * Let D be the divisor, the following equation is used with Newton's
 * Method because it has a root at 1/D:
 *
 *               1            1 - Dx
 *     f(x)  =  ---  -  D  = -------
 *               x              x
 *
 * The first derivative of f(x) is
 *
 *               -2
 *     f'(x) = -x
 *
 *
 * Thus, each iteration of Newton's Method is given by the following
 * equation:
 *
 *                    f(x)
 *     x     =  x -  ------
 *      n+1      n    f'(x)
 *
 *
 *                       2
 *                      x  * (1 - Dx)     
 *     x      =  x  +  ---------------
 *      n+1                  x
 *
 *
 *     x      =  x + x * (1 - Dx)
 *      n+1
 *
 *                         2
 *     x      =  x + x - Dx
 *      n+1
 *
 *                      2
 *     x      =  2x - Dx
 *      n+1
 * 
 *
 *     x      =  x * (2 - Dx)
 *      n+1
 * 
 *
 * Scaling the Controlling Equation:
 * ================================
 *
 * Unfortunately, the controlling equation "x = x * (2 - Dx)" cannot
 * be used as-is because, as explained below in the "Convergence"
 * section, D needs to be scaled such that 1 <= D_scaled < 2 which
 * requires real numbers, but cgul_big_integer only works with
 * integers.
 *
 * The trick here is to scale the controlling equation so, instead of
 * calculating 1/D, it calculates
 *
 *      1     scale
 *     --- * 2
 *      D
 *
 * This is a good thing because multiplying by a power of 2 is the
 * same as shifting bits to the left.  Thus, the new controlling
 * equation will generate "scale" bits of precision because "scale"
 * bits will be shifted left from the fractional side to the
 * whole-number side of the radix point.
 *
 * It will then be possible to perform division by multiplying by the
 * scaled reciprocal followed by a right shift to remove the scale
 * factor:
 *
 *                               1        scale     -scale
 *    quotient  =  dividend * -------- * 2       * 2
 *                             divisor
 *
 *
 * For many equations that need to be scaled in this manner, it is
 * possible to just scale the inputs without adjusting the equation.
 * For example, let
 *
 *     f(x) = 3x
 *
 * Then
 *
 *     f(sx) = s * 3x
 *
 *     f(sx) = s * f(x)
 *
 * You can see that for some functions, the function will
 * automatically produce output that is scaled by the same
 * proportions.
 *
 * However, "f(x) = x * (2 - Dx)" is a little tricky to scale.  I
 * assume I'm not the first person to solve this problem, but what
 * follows is the result of independent research.
 *
 * In order to use the analysis of "f(x) = 1/x - D" (given in the
 * "Convergence" section below), it is important to be able to scale D
 * such that
 *
 *     1 <= D_scaled < 2
 *
 * Because D_scaled resides between 1 and 2, we must be able to
 * represent D_scaled as a fraction by shifting bits across the radix
 * point as follows:
 *
 *                 scale
 *     D_scaled * 2
 *
 * The notation above is a little confusing because we end up scaling
 * D twice.  The first time, we scale D so that it maps to values
 * between 1 and 2 (because we have studied "f(x) = 1/x - D" and know
 * it is well-behaved in that range).  We then have to scale D a
 * second time in order to have access to the significant bits of the
 * fractional part.
 *
 * Thus, we must scale the input larger in order to have access to the
 * fractional part.  So starting with the controlling equation for
 * real numbers, we need a way to use scaled inputs to generate scaled
 * outputs so we can use integers to perform the calculation:
 *
 *
 *     x    =  x * (2 - Dx)
 *      n+1
 *
 *
 * After a while, I realized that it is not possible to scale the
 * outputs by the same factor as the inputs, but it is possible to
 * scale the outputs by the cube of the scale factor:
 *
 * Let
 *          scale
 *     s = 2
 *
 * Then
 *
 *             3          3
 *     x    * s   =  x * s * (2 - Dx)
 *      n+1
 *
 * Then distribute part of the s^3 factor to each component part on
 * the right-hand side that needs it:
 *
 *             3              2
 *     x    * s   =  x * s * s * (2 - Dx)
 *      n+1
 *
 *             3                   2        2
 *     x    * s   =  x * s * (2 * s - Dx * s )
 *      n+1
 *
 *             3                   2
 *     x    * s   =  x * s * (2 * s - D * s * x * s)
 *      n+1          ^^^^^            ^^^^^   ^^^^^
 *
 * [I have underlined each component part to help the eye see how
 * things are being scaled.]
 *
 * The first thing to notice about the final equation is that all of
 * the non-constant inputs are scaled by "s = 2^scale".  This allows
 * us to control exactly how many bits are considered significant by
 * shifting "scale" bits across the radix point where they will all
 * participate in the calculation.
 *
 * The second thing to notice about the final equation is that we put
 * in a scaled factor of s but get out a scale factor of s^3.  Thus
 * before each iteration, the output from the previous iteration has
 * to be scaled down from s^3 to s so that it can be used as the input
 * value "x * s".  So before each iteration, the output from the
 * previous iteration needs to divided by "s^2".  Because
 * "s^2 = s^(2*scale)", this operation can efficiently be performed by
 * shifting right by "2*scale" bits.
 *
 * It is no fun calculating "3 * scale" bits only to throw away
 * "2 * scale" bits.  So the implementation below actually shifts
 * right by "scale" after the subtraction.  This means the code only
 * has to calculate "2 * scale" bits in the last multiplication.  It
 * then just shifts the last result right by "scale" bits to prepare
 * for the next iteration.
 *
 * Lastly, when implementating the final equation above, make sure you
 * use "2 * s^2", not "2 * s".
 *
 *
 * Convergence:
 * ===========
 *
 * The equation "1/x - D" converges quickly but only if a very good
 * initial value for x0 is selected.  The best way to convince
 * yourself of this is to use gnuplot to look at the function:
 *
 *     $ gnuplot
 *     gnuplot> plot [-10:10] 1/x - 100
 *     gnuplot> plot [.005:.1] 1/x - 100
 *
 * If you look at the first plot above, it is a wonder "1/x - 100"
 * converges at all, but if you look at just the right range as given
 * in the second plot, you'll see a curve that is almost ideal except
 * that the sweet part of the curve is well below the x-axis.
 *
 * It is hard to calculate a good initial value for x0 without first
 * knowing the value of 1/D.  For example, in the plots listed above,
 * D is 100 so 1/100 = 0.01 making it possible to easy to pick a range
 * that contains 0.01, but this is not so easy to do without a priori
 * knowledge of the value of 1/D.
 *
 * So, the usual approach is to scale D so that "1/x - D" falls within
 * a well-studied range where it is trivial to calculate or pick a
 * good initial value of x0.
 *
 * For cgul_big_integer__invert(), D is scaled such that
 *
 *     1  <=  D_scaled  <  2
 *
 * This means
 *
 *      1          1
 *     ---  <  ---------  <=  1
 *      2       D_scaled
 *
 * The best thing to do now is to look closely at the following plot
 * which graphs the upper and lower bound of 1/D_scaled.  All other
 * values for 1/D_scaled fall between these two lines:
 *
 *     $gnuplot
 *     gnuplot> plot [.15:1.05] 1/x - 2, 1/x - 1
 *
 * As you can see from the smooth graphs, Newton's Method will likely
 * converge for almost any reasonable value X0.  Empirically, I found
 * that an initial value of 1.0 converged fastest on average.
 *
 * Note that it is trivial to scale D such that "1 <= D_scaled < 2"
 * because it is just a simple matter of shifting D until its most
 * significant bit is in the "ones" column.  As explained below, this
 * means shifting D to the left because the entire calculation is
 * scaled using a multiplier of 2^scale where "D < 2^scale".  The
 * scaling is done so that "scale" number of bits from the fractional
 * side are shifted across to the whole-number side of the radix
 * point.
 *
 *
 * Linear Approximation for x0:
 * ===========================
 *
 * However, it is possible to do a little better by using a linear
 * approximation instead of a constant for x0.  The purpose of the
 * linear approximation is to input D_scaled and output x0:
 *
 *     x0 = g(D_scaled)
 *
 * It helps at this point to watch how the following graphs change:
 *
 *     gnuplot> plot [.15:1.05] 1/x - 2, 1/x - 1, 1/x - 1.25
 *     gnuplot> plot [.15:1.05] 1/x - 2, 1/x - 1, 1/x - 1.5 
 *     gnuplot> plot [.15:1.05] 1/x - 2, 1/x - 1, 1/x - 1.75
 *
 * The idea is to create a linear approximation for where each of the
 * graphs above intersect the x-axis.  In theory, it should be
 * possible to find a linear approximation that is a best fit, but I
 * was happy with the following approximation which is easy to
 * implement and is based on f(1) = 1 and f(2) = 1/2:
 *
 *           y1 - y0     1/2 - 1        1
 *     m  =  -------  =  -------  =  - ---
 *           x1 - x0      2 - 1         2
 *
 * Thus, the slope for the line is m = -1/2.  The equation for a line
 * can be derived from the definition of slope (of course):
 *
 *           y1 - y0
 *     m  =  -------
 *           x1 - x0
 *
 *    y1  =  m * (x1 - x0) + y0
 *
 *   g(x) =  m * (x - x0)  + y0
 *
 * Choose (x0, y0) to be (1, 1) (i.e., f(1) = 1):
 *
 *               1
 *     g(x) = - --- * (x - 1) + 1
 *               2
 *
 *             -x + 1 + 2
 *     g(x) = ------------
 *                  2
 *
 *             3 - x
 *     g(x) = ------
 *               2
 *
 * Specifically, this means
 *
 *            3 - D_scaled
 *     x0  = -------------
 *                 2
 *
 * The nice thing about this equation for calculating x0 is that it is
 * easy to implement because it is just a subtraction followed by a
 * shift.  Based on testing, the linear approximation given above
 * reduces the number of iterations by 15% on average when compared
 * with just using a constant initial value of "x0 = 1".
 */
void
cgul_big_integer__invert(cgul_exception_t* cex,
                         cgul_big_integer_t lhs,
                         cgul_big_integer_t rhs,
                         size_t users_scale)
{
    int is_inverse_positive = 0;
    size_t min_bits = 0;
    size_t actual_scale = 0;
    size_t rhs_scale = 0;
    size_t remainder_bit_count = 0;
    cbi_element_double_t mask = 0;
    cbi_element_double_t high_order_bits = 0;
    cgul_big_integer_t prev = NULL;
    cgul_big_integer_t curr = NULL;
    cgul_big_integer_t next = NULL;
    cgul_big_integer_t zero = NULL;
    cgul_big_integer_t one = NULL;
    cgul_big_integer_t rhs_scaled = NULL;
    cgul_big_integer_t two_scaled = NULL;
    cgul_big_integer_t three_scaled = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* Check for division by zero. */
    if (cgul_big_integer__is_zero(cex, rhs)) {
        *cex = cgul_exception__new("cgul_big_integer__invert: "
                                   "division by zero");
        goto out;
    }

    /* Check if scale is 0.  This means no bits are shifted across the
     * radix point. */
    if (users_scale == 0) {
        cgul_big_integer__set_to_zero(cex, lhs);
        goto out;
    }

    /* Allocate. */
    prev = cgul_big_integer__new(cex);
    curr = cgul_big_integer__new(cex);
    next = cgul_big_integer__new(cex);
    zero = cgul_big_integer__new(cex);
    one = cgul_big_integer__new_from_long(cex, 1);
    rhs_scaled = cgul_big_integer__new(cex);
    two_scaled = cgul_big_integer__new(cex);
    three_scaled = cgul_big_integer__new(cex);

    /* Determine the sign. */
    is_inverse_positive = rhs->is_positive;

    /* Check if rhs is 1 or -1.  The inversion of 1 or -1 is itself,
     * but it still needs to be scaled. */
    if (cgul_big_integer__compare_magnitudes(cex, rhs, one) == 0) {
        cgul_big_integer__shift_left(cex, lhs, rhs, users_scale);
        goto out;
    }

    /* Make sure we have a minimum number of bits and make sure
     * actual_scale is an exact multiple of CBI_ELEMENT_BITS.  By
     * constraining actual_size in this way, it makes it possible to
     * quickly shift rhs such that 1 <= rhs < 2. */
    min_bits = (rhs->size + 1) * CBI_ELEMENT_BITS;
    actual_scale = CGUL_MAX(min_bits, users_scale);
    remainder_bit_count = (actual_scale % CBI_ELEMENT_BITS);
    if (remainder_bit_count) {
        actual_scale += (CBI_ELEMENT_BITS - remainder_bit_count);
    }

    /* Calculate the number of bits that rhs needs to be shifted left
     * such that 1 <= rhs < 2.  This calculation depends on
     * actual_scale being an exact multiple of CBI_ELEMENT_BITS (which
     * is guaranteed above).  Also, actual_scale is guaranteed to be
     * larger than "rhs->size * CBI_ELEMENT_BITS" so the subtraction
     * below should not underflow.
     *
     * The line that starts with "rhs_scale = ..." calculates the
     * number of bits needed to shift rhs such that rhs_shifted would
     * be within one cbi_element_t of where it needs to be.
     *
     * We then use mask and high_order_bits (both
     * cbi_element_double_t) to shift just the high order bits until
     * the most significant high-order bit of rhs_shifted would cross
     * into the next cbi_element_t which is where the "ones" column is
     * located.  When we shift rhs by this amount, it will by
     * definition mean 1 <= rhs_shifted < 2.
     *
     * Also note that high_order_bits cannot be zero because this is
     * checked above when we check for division by zero.  Thus, the
     * while() loop will always terminate.
     */
    mask = ((cbi_element_double_t)1 << CBI_ELEMENT_BITS);
    high_order_bits = rhs->value[rhs->size - 1];
    rhs_scale = actual_scale - (rhs->size * CBI_ELEMENT_BITS);
    while ((high_order_bits & mask) == 0) {
        high_order_bits <<= 1;
        ++rhs_scale;
    }

    /* Scale rhs such that 1 <= rhs_scaled < 2. */
    cgul_big_integer__shift_left(cex, rhs_scaled, rhs, rhs_scale);
    rhs_scaled->is_positive = 1;

    /* Scale 2 (carefully) because the equations below need a scaled
     * value of 2^(2*actual_scale), not 2^actual_scale. */
    cgul_big_integer__assign_from_long(cex, two_scaled, 2);
    cgul_big_integer__shift_left(cex,
                                 two_scaled,
                                 two_scaled,
                                 actual_scale << 1);

    /* Scale 3 for use in the equation that calculates the initial seed. */
    cgul_big_integer__assign_from_long(cex, three_scaled, 3);
    cgul_big_integer__shift_left(cex, three_scaled, three_scaled, actual_scale);

    /* Calculate the initial seed as "curr = 1/2 * (3 - D_scaled)" as
     * explained above in the main comments for this method. */
    cgul_big_integer__subtract(cex, curr, three_scaled, rhs_scaled);
    cgul_big_integer__shift_right(cex, curr, curr, 1);

    if (*cex) {
        goto out;
    }

    /*
     * Newton's Method
     */

    /* Iterate until Newton's Method converges. */
    for (;;) {

        /* Remember the result from the last iteration so the while()
         * loop can check for early convergence. */
        cgul_big_integer__assign(cex, prev, curr);

        /* Perform the next iteration of Newton's Method. */
        cgul_big_integer__multiply(cex, next, rhs_scaled, curr);
        cgul_big_integer__subtract(cex, next, two_scaled, next);
        cgul_big_integer__shift_right(cex, next, next, actual_scale);
        cgul_big_integer__multiply(cex, next, next, curr);

        /* For each iteration of Newton's Method according to the
         * Controlling Formula discussed above in the main comments
         * for this method, we feed in "curr^(actual_scale)" but get
         * out "next^(3*actual_scale)", but notice that there is
         * already one strategically placed shift_right() in the
         * calculation above (to avoid having to calculate bits that
         * are just going to be shifted out anyway).  Because one set
         * of actual_scale bits have already been shifted out, we only
         * need to shift out the remaining set. */
        cgul_big_integer__shift_right(cex, curr, next, actual_scale);

        /* Check for exceptions. */
        if (*cex) {
            goto out;
        }

        /* Break if the loop converged early. */
        if (cgul_big_integer__compare(cex, prev, curr) == 0) {
            break;
        }
    }

    /*
     * Success.
     */

    /* Return "curr" (which is the result of Newton's Method) to the
     * user, but we have to first shift out the extra scale factors we
     * inserted to make the math work. */
    cgul_big_integer__shift_right(cex,
                                  lhs,
                                  curr,
                                  (actual_scale - rhs_scale)
                                  + (actual_scale - users_scale));

    /* Set the sign of lhs. */
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = is_inverse_positive;
    }

 out:

    /* Clean up. */
    if (three_scaled) {
        cgul_big_integer__delete(three_scaled);
    }
    if (two_scaled) {
        cgul_big_integer__delete(two_scaled);
    }
    if (rhs_scaled) {
        cgul_big_integer__delete(rhs_scaled);
    }
    if (one) {
        cgul_big_integer__delete(one);
    }
    if (zero) {
        cgul_big_integer__delete(zero);
    }
    if (next) {
        cgul_big_integer__delete(next);
    }
    if (curr) {
        cgul_big_integer__delete(curr);
    }
    if (prev) {
        cgul_big_integer__delete(prev);
    }

    return;
}

size_t
cgul_big_integer__get_inversion_scale(cgul_exception_t* cex,
                                      cgul_big_integer_t dividend,
                                      cgul_big_integer_t divisor)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(dividend);
    assert(divisor);

    /* Calculate the number of bits needed to invert the divisor with
     * the accuracy required for integer division.  Note that this
     * depends only on the size of the dividend.  In fact, "divisor"
     * is not used by this method, but we require the user to pass it
     * in to help avoid problems of accidentally getting the wrong
     * inversion scale. */
    rv = (dividend->size + 1) * CBI_ELEMENT_BITS;

 out:
  return rv;
}


void
cgul_big_integer__add(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    if( rhs1->is_positive == rhs2->is_positive ) {
        /* rhs1 and rhs2 have the same sign.  Adjust lhs's sign and
         * add the magnitudes. */
        cgul_big_integer__add_magnitudes(cex, lhs, rhs1, rhs2);
        lhs->is_positive = rhs1->is_positive;
    } else {
        /* rhs1 and rhs2 have different signs.  Adjust lhs's sign and
         * subtract the magnitudes. */
        int cmp = cgul_big_integer__compare_magnitudes(cex, rhs1, rhs2);
        if( cmp < 0 ) {
            cgul_big_integer__subtract_magnitudes(cex, lhs, rhs2, rhs1);
            lhs->is_positive = rhs2->is_positive;
        } else if( cmp > 0 ) {
            cgul_big_integer__subtract_magnitudes(cex, lhs, rhs1, rhs2);
            lhs->is_positive = rhs1->is_positive;
        } else {
            /* The numbers have opposite signs and the same magnitude
             * so an addition will result in a value of zero. */
            cgul_big_integer__set_to_zero(cex, lhs);
        }
    }
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    }

 out:
    return;
}

void
cgul_big_integer__subtract(cgul_exception_t* cex,
                           cgul_big_integer_t lhs,
                           cgul_big_integer_t rhs1,
                           cgul_big_integer_t rhs2)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    if( rhs1->is_positive != rhs2->is_positive ) {
        /* rhs1 and rhs2 have opposite signs.  Adjust lhs's sign and
         * add the magnitudes. */
        cgul_big_integer__add_magnitudes(cex, lhs, rhs1, rhs2);
        lhs->is_positive = rhs1->is_positive;
    } else {
        /* rhs1 and rhs2 have the same signs.  Adjust lhs's sign and
         * subtract the magnitudes. */
        int cmp = cgul_big_integer__compare_magnitudes(cex, rhs1, rhs2);
        if( cmp < 0 ) {
            /* The answer has the opposite sign. */
            cgul_big_integer__subtract_magnitudes(cex, lhs, rhs2, rhs1);
            lhs->is_positive = rhs1->is_positive ? 0 : 1;
        } else if( cmp > 0 ) {
            /* The answer has the same sign. */
            cgul_big_integer__subtract_magnitudes(cex, lhs, rhs1, rhs2);
            lhs->is_positive = rhs1->is_positive;
        } else {
            /* The numbers have the same sign and the same magnitude
             * so a subtraction will result in a value of zero. */
            cgul_big_integer__set_to_zero(cex, lhs);
        }
    }
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    }

 out:
    return;
}

/*
 * This method performs "long multiplication" which is O(n^2).  Do not
 * let the O(n^2) fool you.  On a machine with a fast integer
 * multiplier, this method actually runs faster than the other
 * multiplication methods up to around 512 bits.  All the
 * divide-and-conquer algorithms eventually produce factors that are
 * small enough for this method to handle efficiently.
 */
void
cgul_big_integer__multiply_long(cgul_exception_t* cex,
                                cgul_big_integer_t lhs,
                                cgul_big_integer_t rhs1,
                                cgul_big_integer_t rhs2)
{
    cbi_element_double_t carry = 0;
    cbi_element_double_t product = 0;
    cbi_element_double_t sum = 0;
    size_t acc_index = 0;
    int is_result_positive = 1;
    int is_accumulator_temporary = 0;
    cgul_big_integer_t accumulator = NULL;
    size_t i = 0;
    size_t j = 0;
    size_t capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Check for multiplication by zero. */
    if (   cgul_big_integer__is_zero(cex, rhs1)
        || cgul_big_integer__is_zero(cex, rhs2))
    {
        cgul_big_integer__set_to_zero(cex, lhs);
        goto out;
    }

    /* Determine the sign of the result.  You need to do this early to
     * avoid accidentally overwriting part of "lhs" when "lhs" == "rhs1"
     * or when "lhs" == "rhs2". */
    if( rhs1->is_positive != rhs2->is_positive ) {
        is_result_positive = 0;
    }

    /* capacity = rhs1->size + rhs2->size + 1 */
    capacity =  cgul_defensive__add_size_type(cex, rhs1->size, rhs2->size);
    capacity = cgul_defensive__add_size_type(cex, capacity, 1);
    if (*cex) {
        goto out;
    }

    /* Determine whether you can directly write to lhs or whether you
     * need a temporary accumulator. */
    if( (lhs == rhs1) || (lhs == rhs2) ) {
        /* lhs is the same as either rhs1 or rhs2.  You do not want to
         * corrupt either of these during the multiplication.  So, a
         * temporary accumulator is needed. */
        accumulator = cgul_big_integer__new_with_capacity(
                          cex,
                          lhs->capacity,
                          lhs->minimum_capacity_increment);
        is_accumulator_temporary = 1;
    } else {
        /* We can get a slight increase in speed by using the
         * independent lhs object. */
        accumulator = lhs;
    }

    /* Get the accumulator ready. */
    cgul_big_integer__increase_capacity_if_necessary(cex,
                                                     accumulator,
                                                     capacity);
    cgul_big_integer__set_to_zero(cex, accumulator);
    if (*cex) {
        goto out;
    }

    /* Multiply */
    for( i = 0 ; i < rhs2->size ; ++i ) {
        carry = 0;
        for( j = 0 ; j < rhs1->size ; ++j ) {
            acc_index = i + j;
            /* Multiply bytes. */
            product = (cbi_element_double_t)rhs1->value[j]
                      * (cbi_element_double_t)rhs2->value[i];
            /* Add the low order byte to the accumulator. "sum" is
             * just big enough to hold a carry caused by the addition
             * after factoring in the high-order bits that could be
             * set in product. */
            sum = product + (cbi_element_double_t)accumulator->value[acc_index];
            /* Store low order. */
            accumulator->value[acc_index] = (cbi_element_t)sum;
            /* Add the high order byte to the accumulator. */
            ++acc_index;
            /* Store high order. */
            sum = (cbi_element_double_t)accumulator->value[acc_index]
                    /* High-order word from multiply plus the carry
                     * from the low-order sum. */
                  + (sum >> CBI_ELEMENT_BITS)
                    /* carry from last iteration of loop */
                  + carry;
            accumulator->value[acc_index] = (cbi_element_t)sum;
            carry = sum >> CBI_ELEMENT_BITS;
        }
    }

    /* Assign the size. */
    accumulator->size = rhs1->size + rhs2->size;
    cgul_big_integer__minimize_size(cex, accumulator);

    /* Swap the answer into place. */
    if( is_accumulator_temporary ) {
        cgul_big_integer__swap(cex, lhs, accumulator);
    }
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = is_result_positive;
    }

 out:

    /* Clean up. */
    if( is_accumulator_temporary ) {
        cgul_big_integer__delete(accumulator);
    }

    return;
}

/*
 * This method implements multiplication using the Karatsuba algorithm
 * which is roughly O(n^1.6) which is big improvement over long
 * multiplication in theory, but there is a good deal of overhead
 * which makes the "long division" algorithm faster for numbers up to
 * around 512 bits.
 *
 * It works by breaking the factors rhs1 and rhs2 into two parts and
 * uses clever algebra to reduce the number of multiplies from 4 to 3
 * at the expense of two extra subtractions.
 *
 * Let x = rhs1 and y = rhs2.  The Karatsuba algorithm generates x1,
 * x2, y1, and y2 as follows:
 *
 *     rhs1 = x = x1 * 2^m + x2
 *     rhs2 = y = y1 * 2^m + y2
 *
 * It then combines these into intermediate values using the following
 * 3 multiplications:
 *
 *    a = x1 * y1
 *    b = x2 * y2
 *    c = (x1 + x2)(y1 + y2) - a - b
 *
 * Finally, it calculates the product as follows:
 *
 *     x * y = (a * 2^2m) + (c * 2^m) + b
 *
 * The important thing to note when calculating c is that 2^m does not
 * appear in the calculation.  Instead, you add x1 + x2 and y1 + y2
 * without any shifts.  Specifically, "x1 + x2" is not the same as x
 * which is "x1 * 2^m + x2".  The same holds for the y parts.
 *
 * Also, note that this method is really just a front-end to the "long
 * division" algorithm.  Without the "long division" algorithm to stop
 * the recursion, the stack would never be able to unwind.
 */
void
cgul_big_integer__multiply_karatsuba(cgul_exception_t* cex,
                                     cgul_big_integer_t lhs,
                                     cgul_big_integer_t rhs1,
                                     cgul_big_integer_t rhs2)
{
    size_t m = 0;
    cgul_big_integer_t x1 = NULL;
    cgul_big_integer_t x2 = NULL;
    cgul_big_integer_t y1 = NULL;
    cgul_big_integer_t y2 = NULL;
    cgul_big_integer_t a = NULL;
    cgul_big_integer_t b = NULL;
    cgul_big_integer_t c = NULL;
    cgul_big_integer_t one = NULL;
    cgul_big_integer_t mask = NULL;
    cgul_big_integer_t tmp1 = NULL;
    cgul_big_integer_t tmp2 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Sanity check.  If the smaller factor is too small to split use
     * the "long multiplication" algorithm instead. */
    m = CGUL_MIN(rhs1->size, rhs2->size);
    if (m <= 1) {
        cgul_big_integer__multiply_long(cex, lhs, rhs1, rhs2);
        goto out;
    }

    /* Calculate the number of bits in the lower-order parts.  Note
     * that "size" is the number of cbi_element_t slots in the array.
     * So, we have to convert from "size" to bits.  For Karatsuba to
     * work, both factors have to be split the same.  We determine
     * this simply by halving the size of the smaller factor. */
    m *= CBI_ELEMENT_BITS;
    m >>= 1;

    /* Allocate. */
    x1 = cgul_big_integer__new(cex);
    x2 = cgul_big_integer__new(cex);
    y1 = cgul_big_integer__new(cex);
    y2 = cgul_big_integer__new(cex);
    a = cgul_big_integer__new(cex);
    b = cgul_big_integer__new(cex);
    c = cgul_big_integer__new(cex);
    one = cgul_big_integer__new_from_long(cex, 1);
    mask = cgul_big_integer__new(cex);
    tmp1 = cgul_big_integer__new(cex);
    tmp2 = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Generate the mask used to extract the low-order bytes from rhs1
     * and rhs2. */
    cgul_big_integer__shift_left(cex, mask, one, m);
    cgul_big_integer__subtract(cex, mask, mask, one);

    /* Initialize x1 and x2. */
    cgul_big_integer__shift_right(cex, x1, rhs1, m);
    cgul_big_integer__and(cex, x2, rhs1, mask);

    /* Initialize y1 and y2. */
    cgul_big_integer__shift_right(cex, y1, rhs2, m);
    cgul_big_integer__and(cex, y2, rhs2, mask);

    /* Initialize a and b. */
    cgul_big_integer__multiply(cex, a, x1, y1);
    cgul_big_integer__multiply(cex, b, x2, y2);

    /* Initialize c. */
    cgul_big_integer__add(cex, tmp1, x1, x2);
    cgul_big_integer__add(cex, tmp2, y1, y2);
    cgul_big_integer__multiply(cex, c, tmp1, tmp2);
    cgul_big_integer__subtract(cex, c, c, a);
    cgul_big_integer__subtract(cex, c, c, b);

    /* Calculate the final product. */
    cgul_big_integer__shift_left(cex, lhs, a, (m << 1));
    cgul_big_integer__shift_left(cex, c, c, m);
    cgul_big_integer__add(cex, lhs, lhs, c);
    cgul_big_integer__add(cex, lhs, lhs, b);

 out:

    /* Clean up. */
    if (tmp1) {
        cgul_big_integer__delete(tmp1);
    }
    if (tmp2) {
        cgul_big_integer__delete(tmp2);
    }
    if (mask) {
        cgul_big_integer__delete(mask);
    }
    if (one) {
        cgul_big_integer__delete(one);
    }
    if (c) {
        cgul_big_integer__delete(c);
    }
    if (b) {
        cgul_big_integer__delete(b);
    }
    if (a) {
        cgul_big_integer__delete(a);
    }
    if (y2) {
        cgul_big_integer__delete(y2);
    }
    if (y1) {
        cgul_big_integer__delete(y1);
    }
    if (x2) {
        cgul_big_integer__delete(x2);
    }
    if (x1) {
        cgul_big_integer__delete(x1);
    }

    return;
}

/*
 * This method just redirects the multiplication operation to the
 * correct algorithm based on the size of the factors rhs1 and rhs2.
 */
void
cgul_big_integer__multiply(cgul_exception_t* cex,
                           cgul_big_integer_t lhs,
                           cgul_big_integer_t rhs1,
                           cgul_big_integer_t rhs2)
{
    size_t min_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Get the size of the smaller factor. */
    min_size = CGUL_MIN(rhs1->size, rhs2->size);

    /* Pick a multiplication algorithm based on min_size. */
    if (min_size <= 64) {
        cgul_big_integer__multiply_long(cex, lhs, rhs1, rhs2);
    } else {
        cgul_big_integer__multiply_karatsuba(cex, lhs, rhs1, rhs2);
    }

 out:
    return;
}

/*
 * This method implements "long division" which is O(n^2).
 */
void
cgul_big_integer__divide_long(cgul_exception_t* cex,
                              cgul_big_integer_t quotient,
                              cgul_big_integer_t remainder,
                              cgul_big_integer_t dividend,
                              cgul_big_integer_t divisor)
{
    int is_guess_bad = 0;
    int is_quotient_positive = 0;
    int is_remainder_positive = 0;
    int is_quotient_temporary = 0;
    int compare_rv = 0;
    size_t align_bit_count = 0;
    size_t shift_count = 0;
    size_t shift_count_index = 0;
    size_t quotient_size = 0;
    cbi_element_double_t digit = 0;
    cbi_element_double_t guess_dividend = 0;
    cbi_element_t guess_divisor = 0;
    cbi_element_t mask = 0;
    cgul_big_integer_t one = NULL;
    cgul_big_integer_t accumulator = NULL;
    cgul_big_integer_t aligned_divisor = NULL;
    cgul_big_integer_t working_quotient = NULL;
    cgul_big_integer_t shifted_divisor = NULL;
    cgul_big_integer_t product = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(quotient);
    assert(remainder);
    assert(dividend);
    assert(divisor);

    /* Check for division by zero. */
    if (cgul_big_integer__is_zero(cex, divisor)) {
        *cex = cgul_exception__new("cgul_big_integer__divide_long: "
                                   "division by zero");
        goto out;
    }

    /* Determine the sign of the quotient. */
    is_quotient_positive =
        (dividend->is_positive == divisor->is_positive) ? 1 : 0;

    /* Determine the sign of the remainder. */
    is_remainder_positive = dividend->is_positive;

    /* Now check if the dividend is zero. */
    if (cgul_big_integer__is_zero(cex, dividend)) {
        cgul_big_integer__set_to_zero(cex, quotient);
        cgul_big_integer__set_to_zero(cex, remainder);
        goto out;
    }

    /* Check if the divisor is 1 or -1. */
    one = cgul_big_integer__new_from_long(cex, 1);
    if (cgul_big_integer__compare_magnitudes(cex, divisor, one) == 0) {
        /* Assign the dividend to the quotient before setting the
         * remainder to zero.  This is necessary because the remainder
         * is allowed to point to the same object as the dividend. */
        cgul_big_integer__assign(cex, quotient, dividend);
        quotient->is_positive = is_quotient_positive;
        cgul_big_integer__set_to_zero(cex, remainder);
        goto out;
    }

    /* If the magnitude of the dividend is less than the magnitude of
     * the divisor, the quotient is zero and the dividend is the
     * remainder.  By performing this test, we guarantee that the
     * quotient will always have at least one digit if we continue
     * into the main loop. */
    if (cgul_big_integer__compare_magnitudes(cex, dividend, divisor) < 0) {
        /* Assign the dividend to the remainder before setting the
         * quotient to zero.  This is necessary because the quotient
         * is allowed to point to the same object as the dividend. */
        cgul_big_integer__assign(cex, remainder, dividend);
        remainder->is_positive = is_remainder_positive;
        cgul_big_integer__set_to_zero(cex, quotient);
        goto out;
    }

    /* Determine the size of the quotient when we are done.  Note that
     * we just finished comparing the magnitudes of the dividend and
     * divisor.  So we know that the size of the dividend is at least
     * as large as the size of the divisor.  So, it is safe to
     * subtract these unsigned values here. */
    quotient_size = dividend->size - divisor->size + 1;

    /* Determine whether you can directly write to the quotient or
     * whether you need a temporary quotient.  Because later we are
     * going to swap the "working_quotient" with "quotient", it needs
     * to have the same minimum capacity increment as "quotient". */
    if (quotient == dividend) {
        working_quotient =
            cgul_big_integer__new_with_capacity(
                cex,
                quotient_size,
                dividend->minimum_capacity_increment);
        is_quotient_temporary = 1;
    } else if (quotient == divisor) {
        working_quotient =
            cgul_big_integer__new_with_capacity(
                cex,
                quotient_size,
                divisor->minimum_capacity_increment);
        is_quotient_temporary = 1;
    } else {
        /* We can gain a little speed by overwriting "quotient" directly. */
        working_quotient = quotient;
        cgul_big_integer__set_to_zero(cex, working_quotient);
        cgul_big_integer__increase_capacity_if_necessary(cex,
                                                         working_quotient,
                                                         quotient_size);
    }

    /* We always need an accumulator. */
    accumulator = cgul_big_integer__new(cex);
    cgul_big_integer__assign(cex, accumulator, dividend);

    /* We always need an aligned_divisor which is used to hold the
     * divisor after shifting it to the left before the main loop.  We
     * shift the accumulator to the left by the same amount to avoid
     * changing the result of the division.  We align the divisor and
     * accumulator to make the guess_divisor as large as possible so
     * its guess can be as accurate as possible which greatly improves
     * the speed in certain circumstances. */
    aligned_divisor = cgul_big_integer__new_with_capacity(
                          cex,
                          divisor->capacity + 1,
                          divisor->minimum_capacity_increment);
    cgul_big_integer__assign(cex, aligned_divisor, divisor);

    /* We need a cgul_big_integer object to hold the shifted divisor
     * inside the main loop. */
    shifted_divisor = cgul_big_integer__new_with_capacity(
                          cex,
                          dividend->capacity,
                          dividend->minimum_capacity_increment);

    /* We need a cgul_big_integer objects to hold the products of the
     * shifted divisor and our guesses. */
    product = cgul_big_integer__new_with_capacity(
                  cex,
                  dividend->capacity,
                  dividend->minimum_capacity_increment);
    
    if (*cex) {
        goto out;
    }

    /* Shift both the accumulator and the aligned_divisor bit-wise to
     * the left until the most-significant bit of the most-significant
     * byte of the aligned divisor is set.  This makes the
     * guess_divisor as large as possible which makes its guesses as
     * accurate as possible.  By shifting both values, the quotient
     * will not be altered, but we will need to shift the remainder
     * back when we are finished doing the division. */
    assert(aligned_divisor->size > 0);
    mask = (1 << (CBI_ELEMENT_BITS - 1));
    while ((aligned_divisor->value[aligned_divisor->size - 1] & mask) == 0) {
        /* Keep track of the bit count so we can adjust the remainder. */
        ++align_bit_count;
        /* Left shift aligned_divisor. */
        cgul_big_integer__shift_left(cex, aligned_divisor, aligned_divisor, 1);
        /* Left shift accumulator. */
        cgul_big_integer__shift_left(cex, accumulator, accumulator, 1);
    }

    /* Get the aligned divisor we will use to generate our guess for
     * what to use as the next digit in the quotient. */
    assert(aligned_divisor->size > 0);
    guess_divisor = aligned_divisor->value[aligned_divisor->size - 1];

    /* We divide by guess_divisor so make sure it is not zero. */
    if (guess_divisor == 0) {
        *cex = cgul_exception__new(
                   "cgul_big_integer__divide_long: "
                   "Internal Error: "
                   "guess_divisor is zero!");
        goto out;
    }

    /*
     * Chinese Remainder Theorem.
     *
     * The main for() loop is controlled by "shift_count".  The
     * "shift_count" variable is used to control how far over we are
     * shifting "aligned_divisor" before multiplying it by the next
     * digit in the quotient and subtracting the multiple from the
     * accumulator.  This necessarily means that "shift_count" also
     * tells us exactly where in the quotient the next digit belongs.
     */
    for (shift_count_index = quotient_size ;
         shift_count_index > 0 ;
         --shift_count_index)
    {

        /* shift_count_index is "one beyond" the real shift count. */
        shift_count = shift_count_index - 1;

        /* Reset the digit.  Note that we are not working in base-2 or
         * base-16.  Rather we are working in base-2^32 or base-2^16
         * depending on the size of big_element_t.  So a "digit" in
         * this base can be rather large. */
        digit = 0;

        /* This value is used to prevent an infinite loop below when
         * the guess is bad. */
        is_guess_bad = 0;

        /* Get the value of the aligned_divisor shifted into place.
         * Note that shift_count is in terms of how many "digits" to
         * shift where each "digit" occupies an entire cbi_element_t,
         * but the cgul_big_integer__shift_left() method is in terms
         * of bits.  We use "aligned_divisor" below instead of
         * "divisor" because we need to be able to the accumulator and
         * the divisor by the same amount to the left (yes, left)
         * before entiring the loop.  This is done so that
         * guess_divisor can be as large as possible in order to be as
         * accurate as possible. */
        cgul_big_integer__shift_left(cex,
                                     shifted_divisor,
                                     aligned_divisor,
                                     shift_count * CBI_ELEMENT_BITS);
        if (*cex) {
            goto out;
        }

        /* If the accumulator is already less than the divisor, this
         * digit is a zero so we can skip the subtraction and
         * immediately append a zero. */
        compare_rv = cgul_big_integer__compare_magnitudes(cex,
                                                          accumulator,
                                                          shifted_divisor);
        if (compare_rv < 0) {
            goto append_digit;
        }

        /* Get the two most-significant digits in the accumulator and
         * save them as the dividend we are going to use to guess the
         * next digit. */
        guess_dividend = 0;
        if (accumulator->size >= 1) {
            guess_dividend = accumulator->value[accumulator->size - 1];
        }
        if (accumulator->size >= 2) {
            guess_dividend = (guess_dividend << CBI_ELEMENT_BITS)
                                | accumulator->value[accumulator->size - 2];
        }

        /* Guess the next digit in the quotient.  First, we try to
         * divide the guess_dividend by the guess_divisor.  Our first
         * guess might be way off, for example, if guess_divisor is 1.
         * In this case, you could easily end up with a "digit" that
         * requires a cbi_element_double_t to hold.  This is obviously
         * too big because big_integer->value is an array of just
         * cbi_element_t values.  In this case, the right thing to do
         * is to just use the most significant digit of the
         * accumulator instead of the two most significant bits.  This
         * is what we are testing for and what we are doing inside the
         * while() immediately below.
         */
        digit = guess_dividend / guess_divisor;
        while (digit >> CBI_ELEMENT_BITS) {
            guess_dividend >>= CBI_ELEMENT_BITS;
            digit = guess_dividend / guess_divisor;
        }

        /* This guessing business is not an exact science.  Actually,
         * I suppose it is an exact science, but it is hard to predict
         * how good the guess is.  So, we iterate _and_ test to make
         * sure we have found the correct digit. */
        for (;;) {

            /* Generate the shifted product by multiplying the shifted
             * divisor by our guess digit. */
            cgul_big_integer__assign_from_unsigned_long(cex, product, digit);
            cgul_big_integer__multiply(cex, product, shifted_divisor, product);

            /* Compare the magnitudes of the accumulator and the
             * shifted product. */
            compare_rv = cgul_big_integer__compare_magnitudes(cex,
                                                              accumulator,
                                                              product);

            /* If the magnitude of the accumulator is less than the
             * magnitude of the shifted product, the digit is too
             * large.  Carefully reduce it by one. */    
            if (compare_rv < 0) {
                /* First, make sure "digit" is not going to underflow. */
                if (digit == 0) {
                    *cex = cgul_exception__new(
                               "cgul_big_integer__divide_long: "
                               "Internal Error: "
                               "Underflow detected when guessing digit!");
                    goto out;
                }

                /* Reduce our guess by one to account for inaccuracies
                 * in our guess (e.g., "81000 / 901") */
                --digit;

                /* Try this digit. */
                continue;
            }

            /* If you reach here, the magnitude of the accumulator is
             * equal to or greater than the magnitude of the shifted
             * product.  So we try "digit" by subtracting the shifted
             * product from the accumulator. */
            cgul_big_integer__subtract_magnitudes(cex,
                                          accumulator,
                                          accumulator,
                                          product);
            if (*cex) {
                goto out;
            }

            /* However, we still need to test to make sure that the
             * accumulator is less than the shifted divisor after
             * subtracting the product.  This is by far the most
             * common scenario.  If the accumulator is not less than
             * the shifted divisor, it means "digit" could be (at
             * least) one unit larger. */
            compare_rv = cgul_big_integer__compare_magnitudes(cex,
                                                              accumulator,
                                                              shifted_divisor);
            if (compare_rv < 0) {
                /* We found the next digit. */
                break;
            }

            /* 
             * There are always going to be values where our guess
             * is way too low.  For example, look at the following
             * quotient:
             *
             *      12341368
             *     ----------
             *       123418
             *
             * In order to guess that the first digit of the
             * quotient is 9, one would have to go out 7 digits in
             * the numerator and 6 in the denominator, and no
             * matter how many digits you use for your
             * approximation, there will always be sets of numbers
             * that can cause our guess to be way too low.
             */

            /* If have entered this section of code once already,
             * throw an exception to avoid an infinite loop. */
            if (is_guess_bad) {
                *cex = cgul_exception__new(
                           "cgul_big_integer__divide_long: "
                           "Internal Error: "
                           "Guess digit will not converge.");
                goto out;
            }
             
            /* Remember that the guess was bad. */
            is_guess_bad = 1;

            /* First, restore the accumulator to the value it had
             * before we subtracted our guess. */
            cgul_big_integer__add_magnitudes(cex,
                                     accumulator,
                                     accumulator,
                                     product);

            /* Because the initial guess digit was too small, it is
             * likely that we have hit a combination like the one in
             * the comment above.  So just start over with the largest
             * digit.  We avoid an infinite loop here by setting
             * is_guess_bad above. */
            digit = ~(cbi_element_t)0;

            /* Try this digit. */
            continue;
        }

    append_digit:

        /* Add the digit that we found. */
        working_quotient->value[shift_count] = (cbi_element_t)digit;
    }

    /* Shift the accumulator back.  This is to undo the effect of
     * shifting both the accumulator and the aligned_divisor to the
     * left before entering the main loop which was done so that
     * guess_divisor could be as large as possible so its guesses
     * could be as accurate as possible.  This left shifting did not
     * affect the quotient because, effectively, both the dividend and
     * the divisor were shift; however, it does affect the remainder
     * if it is not undone, and the remainder is currently stored in
     * the accumulator. */
    cgul_big_integer__shift_right(
        cex, accumulator, accumulator, align_bit_count);

    /* Set the size of the working quotient. */
    working_quotient->size = quotient_size;
    cgul_big_integer__minimize_size(cex, working_quotient);

    /* Swap quotient into place. */
    if (is_quotient_temporary) {
        cgul_big_integer__swap(cex, quotient, working_quotient);
    }
    if (cgul_big_integer__is_zero(cex, quotient)) {
        quotient->is_positive = 1;
    } else {
        quotient->is_positive = is_quotient_positive;
    }

    /* Swap remainder into place.  */
    cgul_big_integer__swap(cex, remainder, accumulator);
    if (cgul_big_integer__is_zero(cex, remainder)) {
        remainder->is_positive = 1;
    } else {
        remainder->is_positive = is_remainder_positive;
    }

 out:

    /* Clean up. */
    if (product) {
        cgul_big_integer__delete(product);
    }
    if (shifted_divisor) {
        cgul_big_integer__delete(shifted_divisor);
    }
    if (aligned_divisor) {
        cgul_big_integer__delete(aligned_divisor);
    }
    if (accumulator) {
        cgul_big_integer__delete(accumulator);
    }
    if (is_quotient_temporary) {
        cgul_big_integer__delete(working_quotient);
    }
    if (one) {
        cgul_big_integer__delete(one);
    }

    return;
}

/*
 * This method performs division by multiplying by the pre-calculated,
 * scaled reciprocal of the divisor.  It gets its name from the fact
 * that Newton's Method is used to determine the inverse of the
 * divisor.
 */
void
cgul_big_integer__divide_newton(cgul_exception_t* cex,
                                cgul_big_integer_t quotient,
                                cgul_big_integer_t remainder,
                                cgul_big_integer_t dividend,
                                cgul_big_integer_t divisor,
                                cgul_big_integer_t reciprocal,
                                size_t scale)
{
    int is_quotient_positive = 0;
    int is_remainder_positive = 0;
    cgul_big_integer_t one = NULL;
    cgul_big_integer_t tmp_quotient = NULL;
    cgul_big_integer_t tmp_remainder = NULL;
    cgul_big_integer_t tmp_product = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(quotient);
    assert(remainder);
    assert(dividend);
    assert(divisor);

    /*
     * Allocate.
     */
    one = cgul_big_integer__new_from_long(cex, 1);
    tmp_quotient = cgul_big_integer__new(cex);
    tmp_remainder = cgul_big_integer__new(cex);
    tmp_product = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Determine the sign of the quotient. */
    is_quotient_positive =
        (dividend->is_positive == reciprocal->is_positive) ? 1 : 0;

    /* Determine the sign of the remainder. */
    is_remainder_positive = dividend->is_positive;

    /*
     * Multiply by the reciprocal.
     *
     * The "reciprocal" is multiplied by "2^scale".  So after
     * multiplying the dividend by the reciprocal, it has to be
     * shifted right to remove the scale factor.
     *
     * Both "quotient" and "remainder" can point to the same object as
     * "dividend" or "divisor" so we have to avoid accidentally
     * overwriting "dividend" or "divisor" before we are through with
     * them.
     *
     * If there is round-off error tmp_quotient might be off by a
     * little.  We adjust tmp_quotient below as the remainder is being
     * calculated.
     *
     * We set tmp_quotient to be positive to make the math easier.
     * The correct sign will be set before returning.
     */
    cgul_big_integer__multiply(cex, tmp_quotient, dividend, reciprocal);
    cgul_big_integer__shift_right(cex, tmp_quotient, tmp_quotient, scale);
    tmp_quotient->is_positive = 1;
    if (*cex) {
        goto out;
    }

    /*
     * Check if the estimate of the reciprocal was too large.  If so,
     * decrease tmp_quotient and tmp_product.  We have to calculate
     * the product any way in order to calculate the remainder.
     */
    cgul_big_integer__multiply(cex, tmp_product, tmp_quotient, divisor);
    while (cgul_big_integer__compare_magnitudes(
               cex, tmp_product, dividend) > 0)
    {
        cgul_big_integer__subtract(cex, tmp_quotient, tmp_quotient, one);
        cgul_big_integer__subtract_magnitudes(cex,
                                              tmp_product,
                                              tmp_product,
                                              divisor);
        if (*cex) {
            goto out;
        }
    }

    /* Calculate the remainder. */
    cgul_big_integer__subtract_magnitudes(cex,
                                          tmp_remainder,
                                          dividend,
                                          tmp_product);
    if (*cex) {
        goto out;
    }

    /* Check if the estimate of the reciprocal was too small.  If so,
     * increase tmp_quotient and decrease the remainder. */
    while (cgul_big_integer__compare_magnitudes(
               cex, tmp_remainder, divisor) >= 0)
    {
        cgul_big_integer__add(cex, tmp_quotient, tmp_quotient, one);
        cgul_big_integer__subtract_magnitudes(cex,
                                              tmp_remainder,
                                              tmp_remainder,
                                              divisor);
        if (*cex) {
            goto out;
        }
    }

    /* Set the quotient and its sign. */
    cgul_big_integer__swap(cex, quotient, tmp_quotient);
    if (cgul_big_integer__is_zero(cex, quotient)) {
        quotient->is_positive = 1;
    } else {
        quotient->is_positive = is_quotient_positive;
    }

    /* Set the remainder and its sign. */
    cgul_big_integer__swap(cex, remainder, tmp_remainder);
    if (cgul_big_integer__is_zero(cex, remainder)) {
        remainder->is_positive = 1;
    } else {
        remainder->is_positive = is_remainder_positive;
    }

 out:

    /* Clean up. */
    if (tmp_product) {
        cgul_big_integer__delete(tmp_product);
    }
    if (tmp_remainder) {
        cgul_big_integer__delete(tmp_remainder);
    }
    if (tmp_quotient) {
        cgul_big_integer__delete(tmp_quotient);
    }
    if (one) {
        cgul_big_integer__delete(one);
    }

    return;
}

/*
 * This method just redirects the division operation to the correct
 * algorithm based on the size of the divisor.
 */
void
cgul_big_integer__divide(cgul_exception_t* cex,
                         cgul_big_integer_t quotient,
                         cgul_big_integer_t remainder,
                         cgul_big_integer_t dividend,
                         cgul_big_integer_t divisor)
{
    size_t scale = 0;
    cgul_big_integer_t reciprocal = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(quotient);
    assert(remainder);
    assert(dividend);
    assert(divisor);

#if 1
    /**
     * @todo profile to see when long division is faster
     */
#endif

#if 1
    cgul_big_integer__divide_long(cex,
                                  quotient,
                                  remainder,
                                  dividend,
                                  divisor);
    goto out;
#endif

    /* Pick a multiplication algorithm based on the size of the divisor. */
    if (divisor->size <= 0) {
        cgul_big_integer__divide_long(cex,
                                      quotient,
                                      remainder,
                                      dividend,
                                      divisor);
    } else {
        /* Use Newton's Method to calculate the scaled reciprocal. */
        reciprocal = cgul_big_integer__new(cex);
        scale = cgul_big_integer__get_inversion_scale(cex, dividend, divisor);
        cgul_big_integer__invert(cex, reciprocal, divisor, scale);
        /* Divide by multiplying by the scaled reciprocal. */
        cgul_big_integer__divide_newton(cex,
                                        quotient,
                                        remainder,
                                        dividend,
                                        divisor,
                                        reciprocal,
                                        scale);
    }

 out:

    /* Clean up. */
    if (reciprocal) {
        cgul_big_integer__delete(reciprocal);
    }

    return;
}

/*
 * This function raises "base" to the power of "exponent" using the
 * "exponentiation by squaring" method.  Normally, this is defined
 * recursively, but there is a clever implementation that works by
 * inspecting the bits of "exponent" which is the implementation used
 * below.
 *
 * For example,
 *
 *     pow(5, 37) = pow(5, 32) * pow(5, 4) * pow(5, 1)   [equation #1]
 *
 * It is easy to see the symmetry in how to calculate pow(5, 37) from
 * the base-2 representation of 37:
 *
 *             37 = pow(2, 32) + pow(2, 4) + pow(5, 1)   [equation #2]
 *
 * Note that if you start with the least-significant bit of
 * "exponent", you can just keep doubling "base" in order to get each
 * of the terms (starting with the right-most term) in equation #1
 * above.
 */
extern void
cgul_big_integer__power(cgul_exception_t* cex,
                        cgul_big_integer_t lhs,
                        cgul_big_integer_t base,
                        cgul_big_integer_t exponent)
{
    cgul_big_integer_t b = NULL;
    cgul_big_integer_t e = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(base);
    assert(exponent);

    /* Make sure "exponent" is positive.  A negative exponent requires
     * taking the inverse which cgul_big_integer cannot do. */
    if (cgul_big_integer__is_negative(cex, exponent)) {
        *cex = cgul_exception__new("cgul_big_integer__power: "
                                   "negative exponent");
        goto out;
    }

    /* Allocate. */
    b = cgul_big_integer__new(cex);
    e = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign "base" and "exponent" to temporary values so they are
     * not harmed in the production of the result. */
    cgul_big_integer__assign(cex, b, base);
    cgul_big_integer__assign(cex, e, exponent);
    if (*cex) {
        goto out;
    }

    /* Set "lhs" to 1 which basically assumes "exponent" is zero.
     * Note that this means pow(0, 0) = 1 which many math applications
     * use, but some leave it undefined. */
    cgul_big_integer__set_to_one(cex, lhs);
    if (*cex) {
        goto out;
    }

    /* Iterate while "e" (or exponent) is not zero.  How this works is
     * described (above) in the main comments to this method. */
    while (!cgul_big_integer__is_zero(cex, e)) {

        /* Inspect the least-significant bit of e.  If it is one, we
         * must multiply b into the result. */
        if ((e->value[0] & 0x1) == 0x1) {
            cgul_big_integer__multiply(cex, lhs, lhs, b);
            if (*cex) {
                goto out;
            }
        }

        /* Square the base b so that it tracks the bit-wise inspection
         * of e.  (See the main comments (above) for this method.) */
        cgul_big_integer__multiply(cex, b, b, b);
        if (*cex) {
            goto out;
        }

        /* Prepare to inspect the next most-significant bit of e. */
        cgul_big_integer__shift_right(cex, e, e, 1);
    }

 out:

    /* Clean up. */
    if (e) {
        cgul_big_integer__delete(e);
    }
    if (b) {
        cgul_big_integer__delete(b);
    }

    return;
}

/*
 * This method uses the same "exponentiation by squaring" that
 * cgul_big_integer__power() uses, but it takes the modulo at each
 * stage of the calculation to keep the save of the numbers within
 * reason.  See the main comments for cgul_big_integer__power() for an
 * explanation of the basic algorithm.
 *
 * The reason modulo operation can be performed at each stage is that,
 * by the basic rules of modular arithmetic, the modulo of the product
 * of two numbers is the same as the product of the modulo of each
 * number.  This can easily be proven by doing the math.
 */
extern void
cgul_big_integer__power_modulo(cgul_exception_t* cex,
                               cgul_big_integer_t lhs,
                               cgul_big_integer_t base,
                               cgul_big_integer_t exponent,
                               cgul_big_integer_t modulus)
{
    cgul_big_integer_t b = NULL;
    cgul_big_integer_t e = NULL;
    cgul_big_integer_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(base);
    assert(exponent);
    assert(modulus);

    /* Make sure "exponent" is positive.  A negative exponent requires
     * taking the inverse which cgul_big_integer cannot do. */
    if (cgul_big_integer__is_negative(cex, exponent)) {
        *cex = cgul_exception__new("cgul_big_integer__power: "
                                   "negative exponent");
        goto out;
    }

    /* Allocate. */
    b = cgul_big_integer__new(cex);
    e = cgul_big_integer__new(cex);
    tmp = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assign "base" and "exponent" to temporary values so they are
     * not harmed in the production of the result. */
    cgul_big_integer__assign(cex, b, base);
    cgul_big_integer__assign(cex, e, exponent);
    if (*cex) {
        goto out;
    }

    /* Set "lhs" to 1 which basically assumes "exponent" is zero.
     * Note that this means pow(0, 0) = 1 which many math applications
     * use, but some leave it undefined. */
    cgul_big_integer__set_to_one(cex, lhs);
    if (*cex) {
        goto out;
    }

    /* Iterate while "e" (or exponent) is not zero.  How this works is
     * described (above) in the main comments to this method. */
    while (!cgul_big_integer__is_zero(cex, e)) {

        /* Inspect the least-significant bit of e.  If it is one, we
         * must multiply b into the result. */
        if ((e->value[0] & 0x1) == 0x1) {
            /* Multiply. */
            cgul_big_integer__multiply(cex, lhs, lhs, b);
            /* Use division to calculate the modulo of lhs. */
            cgul_big_integer__divide(cex, tmp, lhs, lhs, modulus);
            if (*cex) {
                goto out;
            }
        }

        /* Square the base b so that it tracks the bit-wise inspection
         * of e.  (See the main comments (above) for this method.) */
        cgul_big_integer__multiply(cex, b, b, b);
        /* Use division to calculate the modulo of b. */
        cgul_big_integer__divide(cex, tmp, b, b, modulus);
        if (*cex) {
            goto out;
        }

        /* Prepare to inspect the next most-significant bit of e. */
        cgul_big_integer__shift_right(cex, e, e, 1);
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_big_integer__delete(tmp);
    }
    if (e) {
        cgul_big_integer__delete(e);
    }
    if (b) {
        cgul_big_integer__delete(b);
    }

    return;
}

/*
 * ===========================
 * Miller-Rabin Primality Test:
 * ===========================
 *
 * This method uses the Miller-Rabin Primality Test to determine if
 * big_integer is prime.  There are two keys to understanding
 * Miller-Rabin.
 *
 *
 * Fermat's Little Theorem:
 * =======================
 *
 * First, Fermat's Little Theorem requires that a^(p-1) is congruent
 * to 1 (mod p) (provided a is not a multiple of p).  This theorem
 * more or less falls out of the fact that all the products (mod p)
 * for (a, 2a, 3a, ..., (p-1)a) map to unique values on the range
 * [1, p-1] (provided a does not have any factors in common with p).
 *
 * In fact, this seems to be the essence of being prime.  The reason
 * for this is that if a and p had factors in common, p could be
 * written as "i * j" where i is the product of the common factors and
 * j is product of the remaining factors.  It is important to see that
 * both i and j must be less less than p (because i * j = p).
 *
 * So, if you reconsider the series of products "a, 2a, ... , (p-1)a",
 * at some point one of the elements of the 1, 2, 3, ..., (p-1) part
 * of the series is guaranteed to be equal to j (because we know j is
 * less than p so the series must pass through j).  If a has "i"
 * factors in common with p, "a * j" = "k * i * j" = kp which is
 * obviously a multiple of p.  Thus, "a * j" is congruent to 0 (mod
 * p), and the series start to repeat.
 *
 * If a and p have no factors in common, this cannot happen.  Thus,
 * the series will not start to repeat until "a * p" (which is a full
 * cycle):
 *
 *               [ a (mod p),
 *                2a (mod p),
 *                ...,
 *                (p-1)a (mod p)] = [1, 2, ..., p-1] (in no particular order)
 *
 * Thus,
 *
 *     a * 2a * 3a * ... * (p-1)a = 1 * 2 * 3 * ... * (p-1) (mod p)
 *               (p-1)! * a^(p-1) = (p-1)!  (mod p)
 *                        a^(p-1) = 1  (mod p)
 *                        ^^^^^^^^^^^^^^^^^^^^
 *
 * Square Roots of 1 (mod p) where p is prime:
 * ==========================================
 *
 * Second, if p is prime, it has no square roots of 1 (mod p) other
 * than 1 (mod p) or -1 (mod p).  Square roots are a little more
 * complicated for modular arithmetic than for basic arithmethic.  For
 * example, 6 is a square root of 1 (mod 5) because 6^2 = 36 which is
 * congruent to 1 (mod 5).  Because 5 is prime, our theorem requires
 * that 6 must be congruent to 1 (mod 5) or congruent to -1 (mod 5).
 * The "proof" requires a little math and a good plausibility
 * argument:
 *
 *                a^2 = 1  (mod n)
 *                a^2 = kn + 1
 *            a^2 - 1 = kn
 *     (a + 1)(a - 1) = kn
 *     (a + 1)(a - 1) = 0  (mod n)
 *
 * Thus, if a^2 is congruent to 1 (mod n), (a+1)(a-1) must be a
 * multiple of n (regardless of whether n is prime).
 *
 * For example consider n = 16 and a = 7.  7^2 = 49 which is congruent
 * to 1 (mod 16).  Thus, it follows that (7+1)(7-1) = 8 * 6 = 48 is a
 * multiple of 16.  It is important to see that 7 is a square root of
 * 1 (mod 16), but 7 itself is not congruent to 1 (mod 16).  Thus,
 * according to what we are trying to prove, 16 must be composite.
 *
 * It is not a big deal to multiply (a+1)(a-1) and get a multiple of n
 * when (a+1) or (a-1) is already a multiple of n.  In fact, we will
 * show that this is the only way a prime number p can have a square
 * root of 1 (mod p).
 *
 * It is a big deal, however, to multiply (a+1)(a-1) and get a
 * multiple of n when both (a+1) and (a-1) are not already a multiple
 * of n because it means (a+1) and (a-1) -- in combination -- have all
 * the same factors as n.  Thus, n has factors.  Thus, n must be
 * composite in this case.
 *
 * In the same case where (a+1)(a-1) is not a multiple of n and n is
 * prime, (a+1)(a-1) will necessarily be missing the factors that
 * makes this magic possible.
 *
 * When trying to determine if n is prime, it is important to
 * understand that we must closely scrutinize (a+1) and (a-1) to see
 * whether either is a multiple of n.  This is easy to do.  Just check
 * to see if a is congruent to either 1 (mod n) or -1 (mod n).  If so,
 * a is permitted to be the square root of 1 (mod p) where p is prime.
 * If not, a can only be a square root of 1 (mod n) where n is
 * composite.
 *
 * Thus, if a^2 is congruent to 1 (mod n) but a is not congruent to
 * either 1 (mod n) or -1 (mod n), n is composite.
 *
 *
 * How This Works in Practice:
 * ==========================
 *
 * So the idea behind the Miller-Rabin Primality Test is to combine
 * Fermat's Little Theorem with the observation that (n-1) is even
 * because all primes (other than 2) are odd.  Thus, it is always
 * possible to take the square root of a^(n-1) because taking the
 * square root is the same thing as dividing the exponent by 2.
 *
 * After taking the square root, Miller-Rabin checks to make sure it
 * is 1 (mod n) or -1 (mod n).  If the square root is 1 (mod n) and if
 * (n-1)/2 is itself even, Miller-Rabin repeats the test to make sure
 * the square root is again either 1 (mod n) or -1 (mod n).  This
 * repeats until the exponent for the square root is odd (meaning it
 * cannot be divided evenly by 2) or the square root is -1 (mod n).
 *
 * Note that if the square root is -1 (mod n), the test does not
 * repeat because the theorem above only addresses the square roots of
 * 1 (mod n), not -1 (mod n).  So we test each square root in turn.
 * If the square root is 1 (mod n), we continue with the next square
 * root.  If the square root is -1 (mod n), it is the end of the road,
 * and we return "probably prime" because we don't know anything about
 * what the value of the square root of -1 (mod n) should be if p is
 * prime.
 *
 * The main optimization is to work from the first odd exponent up to
 * a^((n-1)/2) instead of the more natural approach of working from
 * a^((n-1)/2) down to the first odd exponent.  This is more efficient
 * because, instead of calculating the largest value first and then
 * making it smaller in order to inspect each square root, we inspect
 * each square root as we are in the process of generating the largest
 * value.
 *
 * Now we need a smart way to know when to compare each square root
 * against 1 (mod n) or -1 (mod n).  If we were working down from
 * a^((n-1)/2), this would be easy.  Each square root would have to be
 * either 1 (mod n) or -1 (mod n), and we stop when we reach the first
 * square root that is -1 (mod n) (because we have no knowledge of
 * what the square root -1 (mod n)) or when we reach the first square
 * root with an odd exponent (because it cannot be evenly divided by 2
 * meaning it has no integral square root).
 *
 * However, we are working up from the first odd exponent, and we do
 * not know whether we should check it against 1 (mod n) because there
 * may be a -1 (mod n) higher in the sequence that short-circuits the
 * lower part of the sequence.
 *
 * The solution is to search up the sequence for -1 (mod n).  If you
 * find it, you know that the next square in the sequence has to be 1
 * (mod n) (regardless of whether the modulus is prime) as shown
 * below:
 *
 *        a = -1 (mod n)
 *    a * a = -1 * -1 (mod n)
 *      a^2 = 1 (mod n)
 *
 * Furthermore, we know that once the square becomes 1 (mod n), each
 * subsequent square stays 1 (mod n) (regardless of whether the
 * modulus is prime) as shown below:
 *
 *        a = 1 (mod n)
 *    a * a = 1 * 1 (mod n)
 *      a^2 = 1 (mod n)
 *
 * So once you find a square root that is -1 (mod n), you know all the
 * higher-order square roots are 1 (mod n).  So just like when working
 * down from the highest exponent, you can once again short-circuit
 * your evaluation when you find a square root that is -1 (mod n).
 *
 * The question remains though what happens if you search up the
 * entire sequence and do not find a -1 (mod n).  The necessary
 * implication of this is that all of the square roots in the entire
 * sequence have to be 1 (mod n) because -1 (mod n) is the only
 * jumping off point (when working down from the largest exponent),
 * and no -1 (mod n) square roots were found.
 *
 * Furthermore, there is an easy test for this.  As explained above,
 * if a square root is 1 (mod n), each subsequent square must be 1
 * (mod n) (as explained above with simple math).  Thus, it is
 * sufficient to test just the square root with the smallest exponent
 * to make sure it is 1 (mod n) because, if it is 1 (mod n), all
 * subsequenct squares are guaranteed to be 1 (mod n) (regardless of
 * whether n is prime).
 *
 *
 * Selecting s and d:
 * =================
 *
 * When your read about Miller-Rabin, they talk about s and d.  It
 * should be obvious, but it is worth stating explicitly.  When you
 * divide the exponent of the witness by 2, you are effectively taking
 * the square root of the witness, and each time you do this, you need
 * to check to make sure the square root is either 1 (mod p) or -1
 * (mod p).
 *
 * Because dividing by 2 is the same as shifting right, s is just the
 * count of the number of bits the exponent is shifted right until it
 * is no longer possible to shift the exponent right because it has
 * become odd and can longer be evenly divided by two.  This odd
 * exponent is d.  You then carefully start working your way _up_ the
 * sequence of squares (as explained above) starting with a^d.  You
 * then check a^(2d), a^(4d), ... a^((n-1)/2).
 */
int
cgul_big_integer__is_probably_prime(cgul_exception_t* cex,
                                    cgul_big_integer_t big_integer,
                                    unsigned int count)
{
    int rv = 0;
    unsigned int i = 0;
    unsigned long j = 0;
    unsigned long s = 0;
    cgul_random_lcg256_t prng = NULL;
    cgul_big_integer_t one = NULL;
    cgul_big_integer_t two = NULL;
    cgul_big_integer_t d = NULL;
    cgul_big_integer_t r = NULL;
    cgul_big_integer_t witness = NULL;
    cgul_big_integer_t n_minus_one = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* Throw an exception if big_integer is negative. */
    if (cgul_big_integer__is_negative(cex, big_integer)) {
        *cex = cgul_exception__new("cgul_big_integer__is_probably_prime: "
                                   "attempt to test negative integer");
        goto out;
    }

    /* Allocate. */
    prng = cgul_random_lcg256__new(cex);
    one = cgul_big_integer__new_from_long(cex, 1);
    two = cgul_big_integer__new_from_long(cex, 2);
    d = cgul_big_integer__new(cex);
    r = cgul_big_integer__new(cex);
    witness = cgul_big_integer__new(cex);
    n_minus_one = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* 0 is not prime. */
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        goto out;
    }

    /* 1 is not prime. */
    if (cgul_big_integer__is_one(cex, big_integer)) {
        goto out;
    }

    /* 2 is prime. */
    if (cgul_big_integer__compare(cex, big_integer, two) == 0) {
        rv = 1;
        goto out;
    }

    /* All other even numbers are not prime. */
    if (cgul_big_integer__is_even(cex, big_integer)) {
        goto out;
    }

    /* n_minus_one = big_integer - 1 */
    cgul_big_integer__subtract(cex, n_minus_one, big_integer, one);
    if (*cex) {
        goto out;
    }


    /*
     * See main comments (above) for a description of how all this works.
     */


    /*
     * Find d and s such that n_minus_one = d * 2^s.  The idea behind
     * this is explained in the main comments (above) for this method.
     *
     * We know n_minus_one is even because big_integer (the number we
     * are testing for primality) is odd, and n_minus_one is just
     * big_integer minus one.
     *
     * Because big_integer is always at least 3, n_minus_one is always
     * at least two.  Thus, if we keep shifting n_minus_one to the
     * right by one bit, it will eventually become odd.  When
     * n_minus_one becomes odd, this is d, and s is the number of bits
     * n_minus_one had to be shifted.
     */
    cgul_big_integer__shift_right(cex, d, n_minus_one, 1);
    for (s = 1 ; cgul_big_integer__is_even(cex, d) ; ++s) {
        cgul_big_integer__shift_right(cex, d, d, 1);
        if (*cex) {
            goto out;
        }
    }

    /* Perform "count" iterations. */
    for (i = 0 ; i < count ; ++i) {

        /* Get the next potential witness.  Note that next_in_range()
         * returns values in the range [0, big_integer), but we need
         * the witness in the range [1, big_integer).  So, if we get a
         * zero, we just try again. */
        do {
            cgul_random_lcg256__next_in_range(cex, prng, witness, big_integer);
            if (*cex) {
                goto out;
            }
        } while (cgul_big_integer__is_zero(cex, witness));

        /* r = witness^d % big_integer */
        cgul_big_integer__power_modulo(cex, r, witness, d, big_integer);
        if (*cex) {
            goto out;
        }

        /* If r is one, big_integer is probably prime. */
        if (cgul_big_integer__is_one(cex, r)) {
            continue;
        }

        /* If r is -1, big_integer is probably prime.  Note that, for
         * modulo arithmetic, numbers wrap around at zero so -1 is the
         * same as (big_integer - 1). */
        if (cgul_big_integer__compare(cex, r, n_minus_one) == 0) {
            continue;
        }

        /* Now test the higher-order square roots up to but not
         * including witness^(n-1) where (n-1) = (d * 2^s).  There is
         * no point in testing witness^(n-1) because Fermat's Little
         * Theorem requires the modulo to be 1 for that power which is
         * implicitly covered by the other tests.  Also note that we
         * just finished testing the case where s is 0.  So, we start
         * the loop with 1. */
        for (j = 1 ; j < s ; ++j) {

            /*
             * r = (r^2) % big_integer
             *
             * This is just a more efficient way of calculating the
             * remainder of the next square of "witness."  Because r
             * already holds the remainder of the current square of
             * "witness", the remainder of the next square of witness
             * is just the square of the current remainder (mod
             * big_integer).
             *
             * This is true because of basic principles of modular
             * arithmetic that say that the congruence of "a * b" (mod
             * n) is just the "congruence_of_a * congruence_of_b" (mod
             * n).  This is easily proved by converting all the
             * congruences to normal math equations, doing the math,
             * and then converting back to congruences.
             */
            cgul_big_integer__power_modulo(cex, r, r, two, big_integer);
            if (*cex) {
                goto out;
            }

            /* If r is -1, big_integer is probably prime.  Note that,
             * for modular arithmetic, numbers wrap around at zero so
             * -1 is the same as (big_integer - 1). */
            if (cgul_big_integer__compare(cex, r, n_minus_one) == 0) {
                break;
            }
        }

        /* If the loop ran to completion it means none of the square
         * roots that were tested are 1 or -1 (mod big_integer).
         * Thus, big_integer is definitely composite. */
        if (j >= s) {
            goto out;
        }
    }

    /* Success. */
    rv = 1;

 out:

    /* Clean up. */
    if (n_minus_one) {
        cgul_big_integer__delete(n_minus_one);
    }
    if (witness) {
        cgul_big_integer__delete(witness);
    }
    if (r) {
        cgul_big_integer__delete(r);
    }
    if (d) {
        cgul_big_integer__delete(d);
    }
    if (two) {
        cgul_big_integer__delete(two);
    }
    if (one) {
        cgul_big_integer__delete(one);
    }
    if (prng) {
        cgul_random_lcg256__delete(prng);
    }

    return rv;
}

void
cgul_big_integer__shift_left(cgul_exception_t* cex,
                             cgul_big_integer_t lhs,
                             cgul_big_integer_t rhs,
                             size_t bit_count)
{
    size_t lhs_element = 0;
    size_t rhs_element_high = 0;
    size_t rhs_element_low = 0;
    cbi_element_t rhs_bits_high = 0;
    cbi_element_t rhs_bits_low  = 0;
    size_t new_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* A zero shift is equivalent to an assignment. */
    if( bit_count == 0 ) {
        cgul_big_integer__assign(cex, lhs, rhs);
        goto out;
    }

    /* Make sure lhs is large enough to hold the shifted version of
     * rhs.  Unlike with fixed-precision shift lefts, we never shift
     * bits off the left end.  Instead, we allocate enough space to
     * accommodate all the old bits and all the new bits.  The
     * following is a bit complicated because I'm trying to detect
     * overflows in the addition, but it boils down to the following:
     *
     * new_size = rhs->size
     *            + ((bit_count + CBI_ELEMENT_BITS - 1) / CBI_ELEMENT_BITS)
     */
    new_size =
        cgul_defensive__add_size_type(cex, bit_count, CBI_ELEMENT_BITS - 1);
    new_size /= CBI_ELEMENT_BITS;
    new_size = cgul_defensive__add_size_type(cex, rhs->size, new_size);
    if (*cex) {
        goto out;
    }

    cgul_big_integer__increase_capacity_if_necessary(cex, lhs, new_size);
    if (*cex) {
        goto out;
    }

    /* Clear the bytes in lhs that are no longer significant. */
    if( lhs->size > new_size ) {
        memset(lhs->value + new_size,
               0,
               (lhs->size - new_size) * sizeof(cbi_element_t));
    }

    /* The tricky part here is that, in order to allow the user to
     * specify the number of _bits_ to shift, we have to allow for
     * shifting two parts of adjacent source elements to make up a
     * single destination element. */
    lhs_element = new_size;        /* One beyond the end of lhs! */
    rhs_element_high = rhs->size;  /* One beyond the end of rhs! */
    rhs_element_low = rhs_element_high - 1;
    rhs_bits_low = (cbi_element_t)(bit_count % CBI_ELEMENT_BITS);
    rhs_bits_high  = CBI_ELEMENT_BITS - rhs_bits_low;

    if( rhs_bits_low == 0 ) {
        /* There are no low order bits that span across elements.  We
         * optimize the shift a little. */
        for( ; rhs_element_high > 0 ; --rhs_element_high, --lhs_element) {
            lhs->value[lhs_element - 1] = rhs->value[rhs_element_high - 1];
        }
        /* You must zero all the bits that are shifted in on the right. */
        memset(lhs->value, 0, lhs_element * sizeof(cbi_element_t));
        goto assign_is_positive;
    }

    /* 
     * No more short cuts.  We have to perform a completely general
     * shift.  It helps to know that the high order bits that are
     * going into lhs are in the low order part of rhs_element_high, and
     * the low order bits are in the high order part of
     * rhs_element_low.
     */

    /* The first element you copy has its high order bits beyond the
     * end of the array.  Just treat those bits as all zeros. */
    lhs->value[lhs_element - 1] =
        rhs->value[rhs_element_high - 1] >> rhs_bits_high;

    /* The rest of the elements (except for the last one) have both
     * the high order and low order bits squarely in the array. */
    for( --lhs_element ;
         rhs_element_low > 0 ;
         --rhs_element_low, --rhs_element_high, --lhs_element)
    {
        lhs->value[lhs_element - 1]
            = (rhs->value[rhs_element_high - 1] << rhs_bits_low)
            | (rhs->value[rhs_element_low - 1] >> rhs_bits_high);
    }

    /* The last element has its low order bits beyond the start of the array. */
    lhs->value[lhs_element - 1] =
        rhs->value[rhs_element_high - 1] << rhs_bits_low;

    if( lhs_element - 1 > 0 ) {
        /* You must zero all the other bits that are shifted in on the right. */
        memset(lhs->value, 0, (lhs_element - 1) * sizeof(cbi_element_t));
    }

 assign_is_positive:

    /* Tighten up the size. */
    lhs->size = new_size;
    cgul_big_integer__minimize_size(cex, lhs);

    /* Assign the sign. */
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive;
    }

 out:
    return;
}

void
cgul_big_integer__shift_right(cgul_exception_t* cex,
                              cgul_big_integer_t lhs,
                              cgul_big_integer_t rhs,
                              size_t bit_count)
{
    size_t lhs_element = 0;
    size_t rhs_element_high = 0;
    size_t rhs_element_low = 0;
    cbi_element_t rhs_bits_high = 0;
    cbi_element_t rhs_bits_low  = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* A zero shift is equivalent to an assignment. */
    if( bit_count == 0 ) {
        cgul_big_integer__assign(cex, lhs, rhs);
        goto out;
    }

    /* A shift that is larger than rhs->size is equivalent to setting
     * the value to zero. */
    if( bit_count > rhs->size * CBI_ELEMENT_BITS )    {
        cgul_big_integer__set_to_zero(cex, lhs);
        goto out;
    }

    /* Add capacity if necessary. */
    cgul_big_integer__increase_capacity_if_necessary(cex, lhs, rhs->size);
    if (*cex) {
        goto out;
    }

    /* Clear the bytes in lhs that are no longer significant. */
    if( lhs->size > rhs->size ) {
        memset(lhs->value + rhs->size,
               0,
               (lhs->size - rhs->size) * sizeof(cbi_element_t));
    }

    /* The tricky part here is that, in order to allow the user to
     * specify the number of _bits_ to shift, we have to allow for
     * shifting two parts of adjacent source elements to make up a
     * single destination element. */
    lhs_element = 0;
    rhs_element_low = bit_count / CBI_ELEMENT_BITS;
    rhs_element_high = rhs_element_low + 1;
    rhs_bits_high = (cbi_element_t)(bit_count % CBI_ELEMENT_BITS);
    rhs_bits_low = CBI_ELEMENT_BITS - rhs_bits_high;

    if( rhs_bits_high == 0 ) {
        /* There are no low order bits that span across elements.  We
         * optimize the shift a little. */
        for( ; rhs_element_low < rhs->size ; ++rhs_element_low, ++lhs_element) {
            lhs->value[lhs_element] = rhs->value[rhs_element_low];
        }
        /* You must zero all the bits that are shifted in on the left. */
        memset(lhs->value + lhs_element,
               0,
               (rhs->size - lhs_element) * sizeof(cbi_element_t));
        goto assign_is_positive;
    }

    /* No more short cuts.  We have to perform a completely general
     * shift.  It helps to know that the high order bits that are
     * going into lhs are in the low order part of rhs_element_high, and
     * the low order bits are in the high order part of
     * rhs_element_low. */
    for( ;
         rhs_element_high < rhs->size ;
         ++rhs_element_low, ++rhs_element_high, ++lhs_element)
    {
        lhs->value[lhs_element]
            = (rhs->value[rhs_element_high] << rhs_bits_low)
            | (rhs->value[rhs_element_low] >> rhs_bits_high);
    }
    /* You must be careful with the last partial shift because
     * there might not be a "rhs->value[rhs->size]". */
    lhs->value[lhs_element] =
        rhs->value[rhs_element_low] >> rhs_bits_high;

    ++lhs_element;

    /* You must zero all the other bits that are shifted in on the left. */
    memset(lhs->value + lhs_element,
           0,
           (rhs->size - lhs_element) * sizeof(cbi_element_t));

 assign_is_positive:

    /* Tighten up the size. */
    lhs->size = rhs->size;
    cgul_big_integer__minimize_size(cex, lhs);

    /* Assign the sign. */
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs->is_positive;
    }

 out:
    return;
}

void
cgul_big_integer__and(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2)
{
    size_t i = 0;
    cgul_big_integer_t n_longer = NULL;
    cgul_big_integer_t n_shorter = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Determine the relative lengths of rhs1 and rhs2. */
    if( rhs1->size >= rhs2->size ){
        n_longer = rhs1;
        n_shorter = rhs2;
    } else {
        n_longer = rhs2;
        n_shorter = rhs1;
    }

    /*
     * The "and" operation effectively cuts off the most significant
     * bits of the longer number because the shorter number implicitly
     * has zeros for all bits in bytes n_shorter->size and above.
     */

    /* Make sure lhs is large enough. */
    cgul_big_integer__increase_capacity_if_necessary(cex, lhs, n_shorter->size);
    if (*cex) {
        goto out;
    }

    /* Clear the bytes in lhs that are no longer significant. */
    if( lhs->size > n_shorter->size ) {
        memset(lhs->value + n_shorter->size,
               0,
               (lhs->size - n_shorter->size) * sizeof(cbi_element_t));
    }

    /* Set the new size of lhs. */
    lhs->size = n_shorter->size;

    /* Bit-wise AND the two numbers where their bytes align. */
    for( i = 0 ; i < n_shorter->size ; ++i ) {
        lhs->value[i] = n_longer->value[i] & n_shorter->value[i];
    }

    /* Set the sign of lhs. */
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs1->is_positive;
    }

 out:
    return;
}

void
cgul_big_integer__or(cgul_exception_t* cex,
                     cgul_big_integer_t lhs,
                     cgul_big_integer_t rhs1,
                     cgul_big_integer_t rhs2)
{
    size_t i = 0;
    cgul_big_integer_t n_longer = NULL;
    cgul_big_integer_t n_shorter = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Determine the relative lengths of rhs1 and rhs2. */
    if( rhs1->size >= rhs2->size ){
        n_longer = rhs1;
        n_shorter = rhs2;
    } else {
        n_longer = rhs2;
        n_shorter = rhs1;
    }

    /*
     * The "or" operation effectively copies the most significant bits
     * of the longer number because the shorter number implicitly has
     * zeros for all bits in bytes n_shorter->size and above.
     */

    /* Make sure lhs is large enough. */
    cgul_big_integer__increase_capacity_if_necessary(cex, lhs, n_longer->size);
    if (*cex) {
        goto out;
    }

    /* Clear the bytes in lhs that are no longer significant. */
    if( lhs->size > n_longer->size ) {
        memset(lhs->value + n_longer->size,
               0,
               (lhs->size - n_longer->size) * sizeof(cbi_element_t));
    }

    /* Set the new size of lhs. */
    lhs->size = n_longer->size;

    /* Bit-wise OR the two numbers where their bytes align. */
    for( i = 0 ; i < n_shorter->size ; ++i ) {
        lhs->value[i] = n_longer->value[i] | n_shorter->value[i];
    }

    /* You can just copy the rest of the bytes (if necessary). */
    if( lhs != n_longer ) {
        memcpy(lhs->value + n_shorter->size,
               n_longer->value + n_shorter->size,
               (n_longer->size - n_shorter->size) * sizeof(cbi_element_t));
    }

    /* Set the sign of lhs. */
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs1->is_positive;
    }

out:
    return;
}

void
cgul_big_integer__xor(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2)
{
    size_t i = 0;
    cgul_big_integer_t n_longer = NULL;
    cgul_big_integer_t n_shorter = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Determine the relative lengths of rhs1 and rhs2. */
    if( rhs1->size >= rhs2->size ){
        n_longer = rhs1;
        n_shorter = rhs2;
    } else {
        n_longer = rhs2;
        n_shorter = rhs1;
    }

    /*
     * The "xor" operation effectively copies the most significant
     * bits of the longer number because the shorter number implicitly
     * has zeros for all bits in bytes n_shorter->size and above.
     */

    /* Make sure lhs is large enough. */
    cgul_big_integer__increase_capacity_if_necessary(cex, lhs, n_longer->size);
    if (*cex) {
        goto out;
    }

    /* Clear the bytes in lhs that are no longer significant. */
    if( lhs->size > n_longer->size ) {
        memset(lhs->value + n_longer->size,
               0,
               (lhs->size - n_longer->size) * sizeof(cbi_element_t));
    }

    /* Set the new size of lhs. */
    lhs->size = n_longer->size;

    /* Bit-wise XOR the two numbers where their bytes align. */
    for( i = 0 ; i < n_shorter->size ; ++i ) {
        lhs->value[i] = n_longer->value[i] ^ n_shorter->value[i];
    }

    /* You can just copy the rest of the bytes (if necessary). */
    if( lhs != n_longer ) {
        memcpy(lhs->value + n_shorter->size,
               n_longer->value + n_shorter->size,
               (n_longer->size - n_shorter->size) * sizeof(cbi_element_t));
    }

    /* Set the sign of lhs. */
    if (cgul_big_integer__is_zero(cex, lhs)) {
        lhs->is_positive = 1;
    } else {
        lhs->is_positive = rhs1->is_positive;
    }

 out:
    return;
}

int
cgul_big_integer__compare(cgul_exception_t* cex,
                          cgul_big_integer_t lhs,
                          cgul_big_integer_t rhs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* Because we are using a signed magnitude representation, there
     * are both positive and negative zeros.  Call is_zero() to help
     * out. */
    if(    cgul_big_integer__is_zero(cex, lhs)
        && cgul_big_integer__is_zero(cex, rhs) )
    {
        rv = 0;
        goto out;
    }

    /* Check to see if the signs are different. */
    if( lhs->is_positive != rhs->is_positive ) {
        if( lhs->is_positive ) {
            rv = 1;
        } else {
            rv = -1;
        }
        goto out;
    }

    /* The signs are the same.  Check the magnitudes. */
    rv = cgul_big_integer__compare_magnitudes(cex, lhs, rhs);

    /* We now know which number has the larger magnitude which is
     * determinative.  For two positive numbers, the larger number is
     * the one with the larger magnitude, but for two negative
     * numbers, the larger number is the one with the smaller
     * magnitude. */
    if( lhs->is_positive == 0 ) {
        /* The two numbers are negative. */
        if( rv < 0 ) {
            rv = 1;
        } else if( rv > 0 ) {
            rv = -1;
        }
    }

 out:
    return rv;
}

int
cgul_big_integer__compare_magnitudes(cgul_exception_t* cex,
                                     cgul_big_integer_t lhs,
                                     cgul_big_integer_t rhs)
{
    int rv = 0;
    size_t i = 0;
    size_t index = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /* Normalize the size so you can make the comparison. */
    cgul_big_integer__minimize_size(cex, lhs);
    cgul_big_integer__minimize_size(cex, rhs);

    /* If one side has more digits, it has the greater magnitude. */
    if( lhs->size != rhs->size ) {
        if( lhs->size > rhs->size ) {
            rv = 1;
        } else {
            rv = -1;
        }
        goto out;
    }

    /* The magnitudes are roughly the same (lhs->size == rhs->size).
     * We have to inspect each byte individually at this point
     * starting with the most-significant byte. */
    for( i = lhs->size ; i > 0 ; --i ) {
        index = i - 1;
        if( lhs->value[index] != rhs->value[index] ) {
            if( lhs->value[index] > rhs->value[index] ) {
                rv = 1;
            } else {
                rv = -1;
            }
            break;
        }
    }

 out:
    return rv;
}

/*
 * Euclidean Algorithm (Perspective #1):
 * ------------------------------------
 *
 * This method calculates the greatest common divisor (GCD) using the
 * Euclidean Algorithm.  This algorithm takes advantage of the fact
 * that when the numerator and denominator are scaled by the same
 * value, the quotient remains the same, but the remainder is also
 * scaled.  For example,
 *
 *     17                  170
 *    ---- = 1 r. 7,  but ----- = 1 r. 70
 *     10                  100
 *
 * The first thing to understand is that this scale value, by
 * definition, is the GCD if it is used to scale two co-primes (i.e.,
 * two numbers with no common factors).  This is true because you can
 * break any integer into the product of factors.  Let a and b be any
 * two integers, you can always write them as
 *
 *     a = s * cp1
 *     b = s * cp2
 *
 * where s is the product of the factors that a and b have in common
 * and cp1 and cp2 are the product of the factors that a and b do not
 * have in common.  Because s holds _all_ the common factors, it is by
 * definition the GCD.  cp1 and cp2 are co-primes because, by
 * definition, they have no factors in common.
 *
 * So the idea in the Euclidean Algorithm is that when two integers
 * are divided, the scale factor cancels out in the quotient but
 * continues to exist in the remainder (as demonstrated above).  Thus,
 * it is possible to write the remainder as
 *
 *     r = s * cp3
 *
 * It is important to see that finding the GCD of a and b is no
 * different than finding the GCD of b and r because they all have the
 * same form:
 *
 *     a = s * cp1
 *     b = s * cp2
 *     r = s * cp3
 *
 * So, if r is the remainder of a divided by b, we know that r is less
 * than b because the remainder is always less than the divisor.
 * Thus, by repeatedly dividing, we are dealing with successively
 * smaller remainders.  The GCD is found when the remainder takes the
 * form
 *
 *     r = s * 1
 *
 * But is is not obvious when the remainder takes the form above
 * unless we repeat the division one last time.  In this case, it will
 * look something like
 *
 *     s * cp
 *    ------- = cp r. 0
 *       s
 *
 * Because the s values cancels out, the remainder must be zero.
 * Thus, the GCD is the first divisor that results in a remainder of
 * zero.
 *
 *
 *
 * Euclidean Algorithm (Perspective #2):
 * ------------------------------------
 *
 * Let s = gcd(a, b):
 *
 *     a = s * cp1
 *     b = s * cp2
 *
 * First, because s divides a and b, it must also divide (a - b):
 *
 *     (a - b) = s * cp1 - s * cp2
 *             = s * (cp1 - cp2)
 *
 * Second, if a > b, by the Chinese Remainder Theorom we know
 *
 *     a = k * b + r,  where k is an integer
 *
 * or solving for r
 *
 *     r = a - k * b
 *     r = s * cp1 - k * s * cp2
 *     r = s * (cp1 - k * cp2)
 *
 * Thus, gcd(a, b) also divides the remainder of a / b.  This makes it
 * possible to find the gcd(a, b) by iterating through the remainders
 * allowing us to use a smaller divisor on each iteration.  Note that
 * it is also required to use a smaller dividend each iteration too
 * which is too complicated to try to explain here, but suffice it to
 * say that the gcd(a, b) = gcd(b, a % b), but this is not true with
 * respect to gcd(a, a%b).  For example:
 *
 *     gcd(80, 36) = gcd(36, 80 % 36) = 4
 *
 * but
 *
 *     gcd(80, 80 % 36) = 8
 *
 * So, the correct way, to find gcd(80, 36) using the Euclidean
 * algorithm is as follows:
 *
 *     a = 80,  b = 36,  a % b = 8
 *     a = 36,  b =  8,  a % b = 4
 *     a =  8,  b =  4,  a % b = 0
 *
 * The divisor that generates the first remainder of zero is the GCD.
 */
void
cgul_big_integer__gcd(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2)
{
    cgul_big_integer_t tmp1 = NULL;
    cgul_big_integer_t tmp2 = NULL;
    cgul_big_integer_t tmp3 = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Allocate. */
    tmp1 = cgul_big_integer__new(cex);
    tmp2 = cgul_big_integer__new(cex);
    tmp3 = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Initialize tmp1 to be the larger of rhs1 and rhs2; tmp2 is the
     * smaller. */
    if (cgul_big_integer__compare(cex, rhs1, rhs2) >= 0) {
        cgul_big_integer__assign(cex, tmp1, rhs1);
        cgul_big_integer__assign(cex, tmp2, rhs2);
    } else {
        cgul_big_integer__assign(cex, tmp1, rhs2);
        cgul_big_integer__assign(cex, tmp2, rhs1);
    }

    /* Divide until we get a zero remainder. */
    while (!cgul_big_integer__is_zero(cex, tmp2)) {

        /* Divide the larger value by the smaller value further
         * isolating the GCD in the remainder as explained in the main
         * comments (above) to this method. */
        cgul_big_integer__divide(cex, tmp3, tmp1, tmp1, tmp2);

        /* After division, tmp1 holds the remainder which is always
         * smaller than the divisor.  So, we need to swap these values
         * because we always want to use the smaller value as the
         * divisor. */
        cgul_big_integer__swap(cex, tmp1, tmp2);

        /* Check for exceptions. */
        if (*cex) {
            goto out;
        }

    }

    /* Success. */
    cgul_big_integer__assign(cex, lhs, tmp1);
    lhs->is_positive = 1;

 out:

    /* Clean up. */
    if (tmp3) {
        cgul_big_integer__delete(tmp3);
    }
    if (tmp2) {
        cgul_big_integer__delete(tmp2);
    }
    if (tmp1) {
        cgul_big_integer__delete(tmp1);
    }

    return;
}

/*
 * As explained above in the main comments for
 * cgul_big_integer__gcd(), any two integers can be broken down into
 * common factors and uncommon factors
 *
 *     a = s * cp1
 *     b = s * cp2
 *
 * where s is the product of the common factors and scales the
 * products of the uncommon factors cp1 and cp2.  cp1 and cp2 are by
 * definition co-prime because co-primes have no factors in common.
 *
 * Because s holds _all_ the common factors, it is by
 * definition the GCD.
 *
 * Because cp1 and cp2 have no factors in common, it should be clear
 * that the least common multiple (LCM) of a and b is
 *
 *     lcm = s * cp1 * cp2
 *
 * This can be rewritten in terms of a as
 *
 *     lcm = (s * cp1) * cp2 = a * cp2
 *
 * This can be rewritten in terms of b as
 *
 *     lcm = (s * cp2) * cp1 = b * cp1
 *
 * The easiest way to calculate LCM is as follows:
 *
 *            a * b    s * cp1 * s * cp2
 *     lcm = ------- = ----------------- = s * cp1 * cp2
 *              s              s
 *
 * So we use
 *
 *            a * b
 *     lcm = -------
 *              s
 */
void
cgul_big_integer__lcm(cgul_exception_t* cex,
                      cgul_big_integer_t lhs,
                      cgul_big_integer_t rhs1,
                      cgul_big_integer_t rhs2)
{
    cgul_big_integer_t gcd = NULL;
    cgul_big_integer_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs1);
    assert(rhs2);

    /* Allocate. */
    gcd = cgul_big_integer__new(cex);
    tmp = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Assume there is no LCM. */
    cgul_big_integer__set_to_zero(cex, lhs);

    /* Find the greatest common divisor (GCD). */
    cgul_big_integer__gcd(cex, gcd, rhs1, rhs2);
    if (*cex) {
        goto out;
    }

    /* Do not divide by zero. */
    if (cgul_big_integer__is_zero(cex, gcd)) {
        goto out;
    }

    /* Find the LCM as explained in the main comments (above). */
    cgul_big_integer__multiply(cex, lhs, rhs1, rhs2);
    cgul_big_integer__divide(cex, lhs, tmp, lhs, gcd);

 out:

    /* Clean up. */
    if (tmp) {
        cgul_big_integer__delete(tmp);
    }
    if (gcd) {
        cgul_big_integer__delete(gcd);
    }

    return;
}

void
cgul_big_integer__swap(cgul_exception_t* cex,
                       cgul_big_integer_t big_integer_1,
                       cgul_big_integer_t big_integer_2)
{
    struct cgul_big_integer tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer_1);
    assert(big_integer_2);

    /* Nothing to swap if they point to the same objects. */
    if (big_integer_1 == big_integer_2) {
        goto out;
    }

    memcpy((void*)&tmp, (void*)big_integer_1,   sizeof(tmp));
    memcpy((void*)big_integer_1,   (void*)big_integer_2,   sizeof(tmp));
    memcpy((void*)big_integer_2,   (void*)&tmp,  sizeof(tmp));

 out:
    return;
}

const unsigned char*
cgul_big_integer__get_array(cgul_exception_t* cex,
                            cgul_big_integer_t big_integer)
{
    const unsigned char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    rv = (const unsigned char*)big_integer->value;

 out:
    return rv;
}

size_t
cgul_big_integer__get_size(cgul_exception_t* cex,
                           cgul_big_integer_t big_integer)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    /* Return the byte count. */
    rv = big_integer->size * sizeof(cbi_element_t);

 out:
    return rv;
}

int
cgul_big_integer__as_long(cgul_exception_t* cex,
                          cgul_big_integer_t big_integer,
                          long* value)
{
    int rv = 0;
    long int tmp = 0;
    size_t i = 0;
    cgul_big_integer_t long_min = NULL;
    cgul_big_integer_t long_max = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(value);

    /* Get cgul_big_integer objects for LONG_MIN and LONG_MAX. */
    long_min = cgul_big_integer__new_from_long(cex, LONG_MIN);
    long_max = cgul_big_integer__new_from_long(cex, LONG_MAX);

    /* Make sure "big_integer" is in range. */
    if (   (cgul_big_integer__compare(cex, big_integer, long_min) < 0)
        || (cgul_big_integer__compare(cex, big_integer, long_max) > 0))
    {
        /* Out of range. */
        goto out;
    }

    /* Minimize big_integer->size. */
    cgul_big_integer__minimize_size(cex, big_integer);

    /* Manually convert to a long. */
    for (i = big_integer->size ; i > 0 ; --i) {
        tmp = (tmp << CBI_ELEMENT_BITS) | big_integer->value[i - 1];
    }

    /* Make negative if necessary. */
    if (cgul_big_integer__is_negative(cex, big_integer)) {
        tmp = -tmp;
    }

    /* Success. */
    *value = tmp;
    rv = 1;

 out:

    /* Clean up. */
    if (long_max) {
        cgul_big_integer__delete(long_max);
    }
    if (long_min) {
        cgul_big_integer__delete(long_min);
    }

    return rv;
}

int
cgul_big_integer__as_unsigned_long(cgul_exception_t* cex,
                                   cgul_big_integer_t big_integer,
                                   unsigned long int* value)
{
    int rv = 0;
    unsigned long int tmp = 0;
    size_t i = 0;
    cgul_big_integer_t ulong_max = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(value);

    /* Get cgul_big_integer objects for ULONG_MAX. */
    ulong_max = cgul_big_integer__new_from_unsigned_long(cex, ULONG_MAX);

    /* Make sure "big_integer" is in range. */
    if (   cgul_big_integer__is_negative(cex, big_integer)
        || (cgul_big_integer__compare(cex, big_integer, ulong_max) > 0))
    {
        /* Out of range. */
        goto out;
    }

    /* Minimize big_integer->size. */
    cgul_big_integer__minimize_size(cex, big_integer);

    /* Manually convert to an unsigned long. */
    for (i = big_integer->size ; i > 0 ; --i) {
        tmp = (tmp << CBI_ELEMENT_BITS) | big_integer->value[i - 1];
    }

    /* Success. */
    *value = tmp;
    rv = 1;

 out:

    /* Clean up. */
    if (ulong_max) {
        cgul_big_integer__delete(ulong_max);
    }

    return rv;
}

void
cgul_big_integer__as_bcd(cgul_exception_t* cex,
                         cgul_big_integer_t big_integer,
                         cgul_bcd_t bcd)
{
    size_t i = 0;
    cgul_bcd_t base = NULL;
    cgul_bcd_t next = NULL;
    cgul_bcd_t two = NULL;
    cgul_bcd_t accumulator = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(bcd);

    /*
     * The usual approach of peeling off the least-significant digit
     * by dividing by ten and taking the remainder is awfully slow for
     * large numbers.  So, we simply do the multiplication and
     * addition as it is encoded in the bits of big_integer except
     * that we store the results in binary-coded decimal (BCD) rather
     * than in binary.  This allows use to find the conversion to
     * decimal using a single pass and using only addition and
     * multiplication because we can get the decimal digits from
     * cgul_bcd by inspection.
     *
     * In order to chew on as many bits as possible at a time, it
     * helps to realize that it doesn't matter whether you add one
     * bit, shift one bit, and then add the next etc. or whether you
     * add, say, 32 bits, shift 32 bits, and then add the next 32 bits
     * etc.  It's all the same (unless you're a mathematician in which
     * case you are working in base 2 ^ 32 rather than base 2).
     *
     * It is worth noting that you really don't get a speed boost by
     * chewing on more bits because it just increases the complexity
     * of the multiplication.  The main reason for chewing on one
     * cbi_element_t at a time is that it makes the main loop easier
     * to write.
     */

    /* Allocate the accumulator. */
    accumulator = cgul_bcd__new(cex);

    /* Allocate two as a cgul_bcd.  We need this to scale "base". */
    two = cgul_bcd__new_from_long(cex, 2);

    /* Allocate the cgul_bcd object that holds the base of your
     * number system.  (See general comment above.)  Because we are
     * going to chew on one unsigned long at a time.  The base needs
     * to be scaled appropriately. */
    base = cgul_bcd__new_from_unsigned_long(cex, 1);
    for (i = 0 ; i < CBI_ELEMENT_BITS ; ++i) {
        cgul_bcd__multiply(cex, base, base, two);
    }
    
    /* Allocate the cgul_bcd object that holds the next chunk of bits
     * to add to your result. */
    next = cgul_bcd__new(cex);

    if (*cex) {
        goto out;
    }

    /* For each cbi_element_t in big_integer, multiply the current
     * value of the accumulator by "base".  Then add in the
     * cbi_element_t. */
    for (i = big_integer->size ; i > 0 ; --i) {
        cgul_bcd__assign_from_cbi_element(cex, next, big_integer->value[i - 1]);
        cgul_bcd__multiply(cex, accumulator, accumulator, base);
        cgul_bcd__add(cex, accumulator, accumulator, next);
    }

    if (*cex) {
        goto out;
    }
    
    /* Success. */
    cgul_bcd__swap(cex, bcd, accumulator);
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        bcd->is_positive = 1;
    } else {
        bcd->is_positive = big_integer->is_positive;
    }

 out:

    if (next) {
        cgul_bcd__delete(next);
    }
    if (base) {
        cgul_bcd__delete(base);
    }
    if (two) {
        cgul_bcd__delete(two);
    }
    if (accumulator) {
        cgul_bcd__delete(accumulator);
    }

    return;
}

char*
cgul_big_integer__as_binary(cgul_exception_t* cex,
                            cgul_big_integer_t big_integer)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    s = cgul_string__new(cex);
    cgul_big_integer__append_as_binary(cex, big_integer, s);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

void
cgul_big_integer__append_as_binary(cgul_exception_t* cex,
                                   cgul_big_integer_t big_integer,
                                   cgul_string_t s)
{
    static const char* cgul_big_integer__four_bits_as_string[] = {
        "0000", "0001", "0010", "0011", "0100", "0101", "0110", "0111",
        "1000", "1001", "1010", "1011", "1100", "1101", "1110", "1111",
    };

    cbi_element_t lsb = 0;
    cbi_element_t v = 0;
    int still_no_msb = 1;
    size_t i = 0;
    size_t len = 0;
    cbi_element_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(s);

    /* Handle zero as a special case. */
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        cgul_string__append(cex, s, "0b0");
        goto out;
    }

    /* Get the length of s. */
    len = cgul_string__get_length(cex, s);

    /* Increase the capacity of s. */
    cgul_string__set_capacity(
        cex,
        s,
        len + (big_integer->size * CBI_ELEMENT_BITS) + 4);
    if (*cex) {
        goto out;
    }

    /* Insert the prefix including a possible sign. */
    if( cgul_big_integer__is_negative(cex, big_integer) ) {
        cgul_string__append(cex, s, "-0b");
    } else {
        cgul_string__append(cex, s, "0b");
    }

    /* For the most-significant byte, do not insert leading zeros. */
    i = big_integer->size;
    v = big_integer->value[i - 1];
    for( j = CBI_ELEMENT_BITS ; j > 0 ; --j ) {
        lsb = (v >> (j - 1)) & 1;
        if( still_no_msb && (lsb == 0) ) {
            continue;
        }
        if( lsb ) {
            cgul_string__append_char(cex, s, '1');
        } else {
            cgul_string__append_char(cex, s, '0');
        }
        still_no_msb = 0;
    }

    /* For the remaining bytes, use a table lookup. */
    for( --i ; i > 0 ; --i ) {
        v = big_integer->value[i - 1];
        for( j = CBI_ELEMENT_BITS ; j > 0 ; j -= 4 ) {
            cgul_string__append(
                cex,
                s,
                cgul_big_integer__four_bits_as_string[(v >> (j - 4)) & 0x0f]);
        }
    }

 out:

    return;
}

char*
cgul_big_integer__as_octal(cgul_exception_t* cex,
                           cgul_big_integer_t big_integer)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);

    s = cgul_string__new(cex);
    cgul_big_integer__append_as_octal(cex, big_integer, s);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

void
cgul_big_integer__append_as_octal(cgul_exception_t* cex,
                                  cgul_big_integer_t big_integer,
                                  cgul_string_t s)
{
    static const char three_bits_as_char[] = {
        '0', '1', '2', '3', '4', '5', '6', '7'
    };

    int bit_count = 0;
    size_t i = 0;
    size_t tmp_s_buf_len = 0;
    cbi_element_t j = 0;
    cbi_element_t bit = 0;
    cbi_element_t accumulator = 0;
    char* tmp_s_buf = NULL;
    cgul_string_t tmp_s = NULL;
    

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(s);

    /*
     * Handle zero specially.
     */
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        cgul_string__append_char(cex, s, '0');
        goto out;
    }

    /*
     * The idea here is to peel off the last three bits of big_integer
     * and directly convert them to an octal digit.
     */

    /* Allocate the string that will hold the octal conversion. */
    tmp_s = cgul_string__new(cex);

    /* Iterate over the value elements of big_integer inspecting each
     * bit in turn starting with the least-significant bit. */
    for (i = 0 ; i < big_integer->size ; ++i) {
        for (j = 0 ; j < CBI_ELEMENT_BITS ; ++j) {

            if (*cex) {
                goto out;
            }

            /* Get the next bit. */
            bit = (big_integer->value[i] >> j) & 0x1;

            /* Increment the bit count.  The idea is to use bit_count
             * to splice together bits from adjacent bytes. */
            ++bit_count;

            /* Do the right thing based on the bit_count. */
            switch (bit_count) {
            case 1:
                accumulator = bit;
                break;
            case 2:
                if (bit) {
                    accumulator |= 0x2;
                }
                break;
            case 3:
                if (bit) {
                    accumulator |= 0x4;
                }
                cgul_string__append_char(cex,
                                         tmp_s,
                                         three_bits_as_char[accumulator]);
                bit_count = 0;
                break;
            default:
                assert(0);
            }

        }
    }

    /* Handle any left over bits. */
    if (bit_count) {
        cgul_string__append_char(cex,
                                 tmp_s,
                                 three_bits_as_char[accumulator]);
        bit_count = 0;
    }

    if (*cex) {
        goto out;
    }

    /*
     * The string we need is in tmp_s, but it is in reverse order.
     */

    /* Unpack tmp_s. */
    tmp_s_buf = cgul_string__get_value(cex, tmp_s);
    tmp_s_buf_len = cgul_string__get_length(cex, tmp_s);

    /* Remove the extra zeros at the front of the number which is the
     * back of the string at this point. */
    for (i = tmp_s_buf_len ; i > 0 ; --i) {
        if (tmp_s_buf[i - 1] != '0') {
            tmp_s_buf[i] = '\0';
            cgul_string__externally_adjusted(cex, tmp_s);
            break;
        }
    }

    /* All the leading zeros have been removed.  We still need exactly
     * on leading zero however to indicate that the string is
     * octal. */
    cgul_string__append_char(cex, tmp_s, '0');

    /* Append the leading sign if necessary. */
    if (cgul_big_integer__is_negative(cex, big_integer)) {
        cgul_string__append_char(cex, tmp_s, '-');
        if (*cex) {
            goto out;
        }
    }

    /* Reverse the string. */
    cgul_string__reverse(cex, tmp_s);

    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_string__append_block(cex,
                              s,
                              cgul_string__get_value(cex, tmp_s),
                              cgul_string__get_length(cex, tmp_s));

 out:

    if (tmp_s) {
        cgul_string__delete(tmp_s);
    }

    return;
}

char*
cgul_big_integer__as_decimal(cgul_exception_t* cex,
                             cgul_big_integer_t big_integer)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    s = cgul_string__new(cex);
    cgul_big_integer__append_as_decimal(cex, big_integer, s);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

#if 1
void
cgul_big_integer__append_as_decimal(cgul_exception_t* cex,
                                    cgul_big_integer_t big_integer,
                                    cgul_string_t s)
{
    size_t i = 0;
    size_t j = 0;
    size_t k = 0;
    size_t power10 = 0;
    size_t buf_size = 0;
    size_t decimal_digits_max = 0;
    cbi_element_double_t n1 = 0;
    cbi_element_double_t n2 = 0;
    cbi_element_double_t base = 0;
    cbi_element_double_t* buf = NULL;
    cgul_string_t tmp = NULL;
    static char cgul_big_integer__decimal_digits[] = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
    };

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(s);

    /* Check for zero up front. */
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        cgul_string__append_char(cex, s, '0');
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /*
     * The idea is that many small hardware divisions are faster than
     * cgul_bcd's addition and multiplication or cgul_big_integer's
     * division.  This routine is similar to how cgul_bcd does the
     * conversion in base 10.  The main difference are that this
     * routine does the conversion using a larger power of 10 but
     * still small enough to allow for hardware division.  After
     * performing the conversion, the coefficients are relative to the
     * larger power of 10.  Fortunately, you can just convert each
     * coefficient individually (using many small hardware divisions)
     * and concatenate the results.  For example, in base 10^3
     *
     *     (123 * 10^6) + (456 * 10^3) + (789 * 10^0) = 123456789
     *      ^^^            ^^^            ^^^           ^^^^^^^^^
     */

    /* Determine an upper bound on the total number of decimal digits.
     * The value 2.5 comes from rounding (2.0 * 1.205) up.  2.0 comes
     * from there being two hex digits per byte, and 1.205 comes from
     * ln(16)/ln(10) which gives you the ratio of decimal digits to
     * hex digits.  This is probably easier to see with hex and octal
     * where 12 bits can be represented by either 4 octal digits or 3
     * hex digits, and ln(16)/ln(8) = 1.333... = 4/3. */
    decimal_digits_max =
        (size_t)(2.5 * (big_integer->size * sizeof(cbi_element_t))) + 1;

    /* Determine the power of 10 to use depending on the size of
     * cbi_element_double_t.  The values below refect the largest
     * power of ten where the coefficients do not overflow.  Note that
     * the main loop (below) chews on 8-bits (one byte) at a time so
     * the coefficients need to leave enough room to temporarily hold
     * the carry produced by multiplying by 256 and adding 255. */
    if (sizeof(*buf) >= 16) {
        power10 = 36;
    } else if (sizeof(*buf) >= 8) {
        power10 = 16;
    } else if (sizeof(*buf) >= 4) {
        power10 = 7;
    } else if (sizeof(*buf) >= 2) {
        power10 = 2;
    } else {
        /* Not reached. */
        assert(0);
    }

    /* Calculate the larger power of 10 to use as the base.  We
     * calculate this instead of just hard coding it to avoid problems
     * with older compilers warning about 10^16 or even 10^7 being too
     * large. */
    base = 10;
    for (i = 1 ; i < power10 ; ++i) {
        base *= 10;
    }

    /* Allocate a buffer that is large enough to hold the coefficients
     * of the larger power of 10.  Each element in the "buf" array
     * represents one coefficient of the larger power of 10, and each
     * coefficient is capable of holding "power10" decimal digits. */
    buf_size = decimal_digits_max / power10 + 1;
    buf = (cbi_element_double_t*)cgul_libc__calloc(cex, buf_size, sizeof(*buf));
    if (*cex) {
        goto out;
    }

    /* Iterate over each element in the big_integer value starting
     * with the most-significant element.  This is just your typical
     * loop that converts from one base to another using
     * multiplication and addition in the target base's arithmetic.
     * Of course, computers don't directly support base (10 ^ power10)
     * arithmetic so we have to implement it ourselves.  This boils
     * down to lots of hardware divisions. */
    for (i = big_integer->size ; i > 0 ; --i) {

        /* Iterate over each byte in each element. */
        for (j = sizeof(cbi_element_t) ; j > 0 ; --j) {

            /* Repeated preform the multiplication and addition
             * typical of converting from one base to another.  For
             * efficiency, the multiplication by 256 (distributing
             * across all coefficients) and the "addition and carry"
             * of the next byte are done in a single loop.  */
            for (k = 0 ; k < buf_size ; ++k) {

                /* Multiply. */
                buf[k] = buf[k] << 8;

                /* Add. */
                if (k == 0) {
                    buf[0] += (big_integer->value[i - 1] >> ((j - 1) << 3))
                              & 0xff;

                /* Carry. */
                } else if (buf[k - 1] >= base) {
                    /* Do the division just once.  Note that the third
                     * line below just sets buf[k - 1] to the remainder
                     * without having to do a modulo operation. */
                    n1 = buf[k - 1] / base;
                    buf[k] += n1;
                    buf[k - 1] -= n1 * base;
                }
            }
        }
    }

    /* Find the most significant coefficient. */
    for (k = buf_size ; k > 0 ; --k) {
        if (buf[k - 1]) {
            break;
        }
    }

    /* Convert each coefficient to decimal appending the digits to
     * "tmp" in reverse order. */
    for (i = 0 ; i < k ; ++i) {
        /* Set n1 equal to the next coefficient. */
        n1 = buf[i];
        for (j = 0 ; j < power10 ; ++j) {
            /* Do the division just once. */
            n2 = n1 / 10;
            /* Append the next digit.  Note that (n1 - n2 * 10) is
             * just the remainder of dividing n1 by 10. */
            cgul_string__append_char(
                cex, tmp, cgul_big_integer__decimal_digits[n1 - n2 * 10]);
            if (*cex) {
                goto out;
            }
            /* Avoid appending leading zeros if this is the
             * most-significant coefficient. */
            if ((n2 == 0) && (i >= k - 1)) {
                break;
            }
            /* Reduce. */
            n1 = n2;
        }
    }

    /* Append the sign if necessary and if it is not a negative zero. */
    if (   !big_integer->is_positive
        && !cgul_big_integer__is_zero(cex, big_integer))
    {
        cgul_string__append_char(cex, tmp, '-');
        if (*cex) {
            goto out;
        }
    }

    /* Above, we pushed all the characters onto the string in reverse order. */
    cgul_string__reverse(cex, tmp);

    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_string__append_block(cex,
                              s,
                              cgul_string__get_value(cex, tmp),
                              cgul_string__get_length(cex, tmp));

 out:

    /* Clean up. */
    if (buf) {
        free(buf);
    }
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}
#endif

#if 0
void
cgul_big_integer__append_as_decimal(cgul_exception_t* cex,
                                    cgul_big_integer_t big_integer,
                                    cgul_string_t s)
{
    cgul_bcd_t bcd = NULL;

    /*
     *                         WARNING
     *
     * This method is commented out!  I'm keeping it around though
     * because I want to be able to compare different implementations.
     *
     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * The idea is that it is faster for cgul_bcd to multiply and add
     * than it is for cgul_big_integer to divide.  So rather than have
     * cgul_big_integer repeatedly divide to peel off each decimal
     * digit, we feed each binary digit to cgul_bcd and let it
     * generate the decimal version of the number.  It can then
     * convert to a decimal string by inspection.
     */

    /* Allocate a cgul_big_integer object. */
    bcd = cgul_bcd__new(cex);

    /* Convert big_integer to a cgul_bcd.  This is where the
     * multiplication and addition occurs. */
    cgul_big_integer__as_bcd(cex, big_integer, bcd);

    /* Append as decimal. */
    cgul_bcd__append_as_decimal(cex, bcd, s);

 out:

    /* Clean up. */
    if (bcd) {
        cgul_bcd__delete(bcd);
    }

    return;
}
#endif

#if 0
void
cgul_big_integer__append_as_decimal(cgul_exception_t* cex,
                                    cgul_big_integer_t big_integer,
                                    cgul_string_t s)
{
    static char cgul_big_integer__decimal_digits[] = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
    };

    long int digit = 0;
    size_t scale = 0;
    cgul_big_integer_t reciprocal = NULL;
    cgul_big_integer_t accumulator = NULL;
    cgul_big_integer_t remainder = NULL;
    cgul_big_integer_t ten = NULL;
    cgul_string_t tmp = NULL;

    /*
     *                         WARNING
     *
     * This method is commented out!  I'm keeping it around though
     * because I want to be able to compare different implementations.
     * Also, I want to see how much a better division algorithm can
     * improve performance.
     *
     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(s);

    /* Check for zero up front. */
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        cgul_string__append_char(cex, s, '0');
        goto out;
    }

    /* Allocate. */
    reciprocal = cgul_big_integer__new(cex);
    accumulator = cgul_big_integer__new(cex);
    remainder = cgul_big_integer__new(cex);
    ten = cgul_big_integer__new_from_unsigned_long(cex, 10);

    /* Allocate a temporary string for storing the result. */
    tmp = cgul_string__new(cex);

    /* Use Newton's Method to calculate the scaled reciprocal of 10. */
    scale = cgul_big_integer__get_inversion_scale(cex, big_integer, ten);
    cgul_big_integer__invert(cex, reciprocal, ten, scale);
    if (*cex) {
        goto out;
    }

    /* accumulator = big_integer */
    cgul_big_integer__assign(cex, accumulator, big_integer);

    /* Iterate over each decimal digit in the accumulator.  We don't
     * have to worry about the case where big_integer is zero because
     * we have already checked for this above. */
    while (!cgul_big_integer__is_zero(cex, accumulator)) {

        /* Reduce the size of the reciprocal as the accumulator is reduced. */
        while (reciprocal->size > accumulator->size + 2) {
            cgul_big_integer__shift_right(cex,
                                          reciprocal,
                                          reciprocal,
                                          CBI_ELEMENT_BITS);
            scale -= CBI_ELEMENT_BITS;
        }

        /* Repeatedly divide the accumulator by multiplying by the
         * reciprocal of ten.  This stores the next decimal digit in
         * the remainder. */
        cgul_big_integer__divide_newton(cex,
                                        accumulator,
                                        remainder,
                                        accumulator,
                                        ten,
                                        reciprocal,
                                        scale);
        if (*cex) {
            goto out;
        }

        /* Get the decimal digit as a long. */
        if (!cgul_big_integer__as_long(cex, remainder, &digit)) {
            *cex = cgul_exception__new(
                       "cgul_big_integer__append_as_decimal:"
                       "Internal Error: Failed to convert digit to index.");
            goto out;
        }

        /* Handle negative remainders.  We just need the magnitude. */
        if (digit < 0) {
            digit = -digit;
        }

        /* The digit should always be less than 10. */
        assert((digit >= 0) && (digit <= 9));

        /* Add the digit to the string. */
        cgul_string__append_char(
            cex,
            tmp,
            cgul_big_integer__decimal_digits[digit]);

        if (*cex) {
            goto out;
        }

    }

    /* Append the sign if necessary and if it is not a negative zero. */
    if (   !big_integer->is_positive
        && !cgul_big_integer__is_zero(cex, big_integer))
    {
        cgul_string__append_char(cex, tmp, '-');
        if (*cex) {
            goto out;
        }
    }

    /* Above, we pushed all the characters onto the string in reverse order. */
    cgul_string__reverse(cex, tmp);

    if (*cex) {
        goto out;
    }

    /* Success. */
    cgul_string__append_block(cex,
                              s,
                              cgul_string__get_value(cex, tmp),
                              cgul_string__get_length(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }
    if (ten) {
        cgul_big_integer__delete(ten);
    }
    if (remainder) {
        cgul_big_integer__delete(remainder);
    }
    if (accumulator) {
        cgul_big_integer__delete(accumulator);
    }
    if (reciprocal) {
        cgul_big_integer__delete(reciprocal);
    }

    return;
}
#endif

char*
cgul_big_integer__as_hexadecimal(cgul_exception_t* cex,
                                 cgul_big_integer_t big_integer)
{
    char* rv = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    s = cgul_string__new(cex);
    cgul_big_integer__append_as_hexadecimal(cex, big_integer, s);
    if (*cex) {
        goto out;
    }

    /* Success. */
    rv = cgul_string__take_value(cex, s);

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

void
cgul_big_integer__append_as_hexadecimal(cgul_exception_t* cex,
                                        cgul_big_integer_t big_integer,
                                        cgul_string_t s)
{
    static const char cgul_big_integer__four_bits_as_char[] = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
    };

    cbi_element_t v = 0;
    int still_no_msb = 1;
    size_t i = 0;
    size_t len = 0;
    cbi_element_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Handle zero as a special case. */
    if (cgul_big_integer__is_zero(cex, big_integer)) {
        cgul_string__append(cex, s, "0x0");
        goto out;
    }

    /* Get the length of s. */
    len = cgul_string__get_length(cex, s);

    /* Increase the capacity of s. */
    cgul_string__set_capacity(
        cex,
        s,
        len + (big_integer->size * CBI_ELEMENT_BITS) / 4 + 4);
    if (*cex) {
        goto out;
    }

    /* Insert the prefix including a possible sign. */
    if( cgul_big_integer__is_negative(cex, big_integer) ) {
        cgul_string__append(cex, s, "-0x");
    } else {
        cgul_string__append(cex, s, "0x");
    }

    /* For the most-significant byte, do not insert leading zeros. */
    i = big_integer->size;
    v = big_integer->value[i - 1];
    for( j = CBI_ELEMENT_BITS ; j > 0 ; j -= 4 ) {
        char c = cgul_big_integer__four_bits_as_char[(v >> (j - 4)) & 0xf];
        if( still_no_msb && (c == '0') ) {
            continue;
        }
        cgul_string__append_char(cex, s, c);
        still_no_msb = 0;
    }

    /* For the remaining bytes, use a table lookup. */
    for( --i ; i > 0 ; --i ) {
        v = big_integer->value[i - 1];
        for( j = CBI_ELEMENT_BITS ; j > 0 ; j -= 4 ) {
            cgul_string__append_char(
                cex,
                s,
                cgul_big_integer__four_bits_as_char[(v >> (j - 4)) & 0xf]);
        }
    }

 out:

    return;
}

void
cgul_big_integer__print_as_binary(cgul_exception_t* cex,
                                  cgul_big_integer_t big_integer,
                                  FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(f);

    /* Convert. */
    s = cgul_big_integer__as_binary(cex, big_integer);
    if (*cex) {
        goto out;
    }

    /* Print. */
    if (fprintf(f, "%s", s) < 0) {
        *cex = cgul_exception__new("cgul_big_integer__print_as_binary: %m");
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        free(s);
    }
}

void
cgul_big_integer__print_as_octal(cgul_exception_t* cex,
                                 cgul_big_integer_t big_integer,
                                 FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(f);

    /* Convert. */
    s = cgul_big_integer__as_octal(cex, big_integer);
    if (*cex) {
        goto out;
    }

    /* Print. */
    if (fprintf(f, "%s", s) < 0) {
        *cex = cgul_exception__new("cgul_big_integer__print_as_octal: %m");
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        free(s);
    }
}

void
cgul_big_integer__print_as_decimal(cgul_exception_t* cex,
                                   cgul_big_integer_t big_integer,
                                   FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(f);

    /* Convert. */
    s = cgul_big_integer__as_decimal(cex, big_integer);
    if (*cex) {
        goto out;
    }

    /* Print. */
    if (fprintf(f, "%s", s) < 0) {
        *cex = cgul_exception__new("cgul_big_integer__print_as_decimal: %m");
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        free(s);
    }
}

void
cgul_big_integer__print_as_hexadecimal(cgul_exception_t* cex,
                                       cgul_big_integer_t big_integer,
                                       FILE* f)
{
    char* s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(big_integer);
    assert(f);

    /* Convert. */
    s = cgul_big_integer__as_hexadecimal(cex, big_integer);
    if (*cex) {
        goto out;
    }

    /* Print. */
    if (fprintf(f, "%s", s) < 0) {
        *cex = cgul_exception__new(
                   "cgul_big_integer__print_as_hexadecimal: %m");
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        free(s);
    }
}
