#include "cgul_binary_search.h"

void*
cgul_binary_search(cgul_exception_t* cex,
                   const void* key,
                   const void* v,
                   size_t v_count,
                   size_t element_size,
                   cgul_binary_search__compare_t f)
{
    void* rv = NULL;
    char* tmp = (char*)v;
    size_t next = 0;
    size_t low = 0;
    size_t high = v_count;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(v);

    if( (v_count == 0) || (element_size == 0) ) {
        goto out;
    }

    while (low < high) {
        next = (high + low) >> 1;
        f_rv = (*f)(key, &tmp[next * element_size]);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = (void*)&tmp[next * element_size];
            break;
        }
    }

 out:
    return rv;
}

void*
cgul_binary_search_pointers(cgul_exception_t* cex,
                            const void* key,
                            const void** v,
                            size_t v_count,
                            cgul_binary_search__compare_t f)
{
    void* rv = NULL;
    size_t next = 0;
    size_t low = 0;
    size_t high = v_count;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(v);

    if( v_count == 0 ) {
        return NULL;
    }

    while (low < high) {
        next = (high + low) >> 1;
        f_rv = (*f)(key, v[next]);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = (void*)v[next];
            break;
        }
    }

 out:
    return rv;
}

void*
cgul_binary_search_ceiling(cgul_exception_t* cex,
                           const void* key,
                           const void* v,
                           size_t v_count,
                           size_t element_size,
                           cgul_binary_search__compare_t f)
{
    void* rv = NULL;
    char* tmp = (char*)v;
    size_t next = 0;
    size_t low = 0;
    size_t high = v_count;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(v);

    if( (v_count == 0) || (element_size == 0) ) {
        goto out;
    }

    /* Perform a stock binary search. */
    while (low < high) {
        next = (high + low) >> 1;
        f_rv = (*f)(key, &tmp[next * element_size]);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = (void*)&tmp[next * element_size];
            break;
        }
    }

    /* If the element wasn't found, return the smallest element
     * greater than the search key. */
    if (!rv) {
        if (low < v_count) {
            rv = (void*)&tmp[low * element_size];
        }
    }

 out:
    return rv;
}

void*
cgul_binary_search_ceiling_pointers(cgul_exception_t* cex,
                                    const void* key,
                                    const void** v,
                                    size_t v_count,
                                    cgul_binary_search__compare_t f)
{
    void* rv = NULL;
    size_t next = 0;
    size_t low = 0;
    size_t high = v_count;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(v);

    if( v_count == 0 ) {
        return NULL;
    }

    /* Perform a stock binary search. */
    while (low < high) {
        next = (high + low) >> 1;
        f_rv = (*f)(key, v[next]);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = (void*)v[next];
            break;
        }
    }

    /* If the element wasn't found, return the smallest element
     * greater than the search key. */
    if (!rv) {
        if (low < v_count) {
            rv = (void*)v[low];
        }
    }

 out:
    return rv;
}

void*
cgul_binary_search_floor(cgul_exception_t* cex,
                         const void* key,
                         const void* v,
                         size_t v_count,
                         size_t element_size,
                         cgul_binary_search__compare_t f)
{
    void* rv = NULL;
    char* tmp = (char*)v;
    size_t next = 0;
    size_t low = 0;
    size_t high = v_count;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(v);

    if( (v_count == 0) || (element_size == 0) ) {
        goto out;
    }

    /* Perform a stock binary search. */
    while (low < high) {
        next = (high + low) >> 1;
        f_rv = (*f)(key, &tmp[next * element_size]);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = (void*)&tmp[next * element_size];
            break;
        }
    }

    /* If the element wasn't found, return the largest element less
     * than the search key. */
    if (!rv) {
        if (low > 0) {
            rv = (void*)&tmp[(low - 1) * element_size];
        }
    }

 out:
    return rv;
}

void*
cgul_binary_search_floor_pointers(cgul_exception_t* cex,
                                  const void* key,
                                  const void** v,
                                  size_t v_count,
                                  cgul_binary_search__compare_t f)
{
    void* rv = NULL;
    size_t next = 0;
    size_t low = 0;
    size_t high = v_count;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);
    assert(v);

    if( v_count == 0 ) {
        return NULL;
    }

    /* Perform a stock binary search. */
    while (low < high) {
        next = (high + low) >> 1;
        f_rv = (*f)(key, v[next]);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = (void*)v[next];
            break;
        }
    }

    /* If the element wasn't found, return the largest element less
     * than the search key. */
    if (!rv) {
        if (low > 0) {
            rv = (void*)v[low - 1];
        }
    }

 out:
    return rv;
}

int
cgul_binary_search_irange(cgul_exception_t* cex,
                          int low,
                          int high,
                          int (*f)(int, void*),
                          void* data)
{
    int rv = 0;
    int next = 0;
    int f_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    while (low < high) {
        next = (high + low) >> 1;
        f_rv = f(next, data);
        if (f_rv < 0) {
            high = next;  /* "high" is still the _exclusive_ upper bound. */
        } else if (f_rv > 0) {
            low = next + 1;  /* "low" is still the _inclusive_ lower bound. */
        } else if (f_rv == 0) {
            rv = 1;
            break;
        }
    }

 out:
    return rv;
}
