#include "cgul_cache_p.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_vector_p.h"

/**
 * Free all the objects currently on the cache.  Then free the cache
 * itself.
 *
 * @param[in] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 */
static void
cgul_cache__free_cache(cgul_exception_t* cex,
                       cgul_cache_t cache)
{
    void* obj = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    /* Check to see if the cache is already free. */
    if (cache->v == NULL) {
        goto out;
    }

    /* Free the cached objects. */
    for (;;) {
        obj = cgul_vector__pop_back(cex, cache->v);
        if (obj == NULL) {
            break;
        }
        if (cache->destructor) {
            (*cache->destructor)(obj);
        }
    }

    assert(cgul_vector__get_size(cex, cache->v) == 0);

    /* Free the vector itself. */
    cgul_vector__delete(cache->v);

    cache->v = NULL;
    cache->size = 0;
    cache->reserve = 0;

 out:
    return;
}

/**
 * This method resizes the cache to accommodate
 * <code>total_size</code> number of objects. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] cache <code>cgul_cache</code> instance
 * @param[in] total_size total size
 */
static void
cgul_cache__resize(cgul_exception_t* cex,
                   cgul_cache_t cache,
                   unsigned long int total_size)
{
    void* obj = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    /* A total_size of 0 means the user wants us to stop caching and
     * clean up. */
    if (total_size == 0) {
        cgul_cache__free_cache(cex, cache);
        goto out;
    }

    /* Lazily create the cache. */
    if (cache->v == NULL) {
        /* Create a new vector with enough capacity. */
        cache->v = cgul_vector__new_with_capacity(cex, total_size);
        if (*cex) {
            goto out;
        }
    }

    /* If the user is making the cache smaller, delete any excess
     * objects. */
    while (cgul_vector__get_size(cex, cache->v) > total_size) {
        obj = cgul_vector__pop_back(cex, cache->v);
        assert(obj);
        if (cache->destructor) {
            (*cache->destructor)(obj);
        }
    }

 out:
    return;
}

static struct cgul_cache
cgul_cache__the_freer = {
    0,                                /* size */
    0,                                /* reserve */
    NULL,                             /* v */
    NULL,                             /* constructor */
    (cgul_cache__destructor_t)free,   /* destructor */
    NULL                              /* restructor */
};

cgul_cache_t
cgul_cache__get_freer(cgul_exception_t* cex)
{
    return &cgul_cache__the_freer;
}

cgul_cache_t
cgul_cache__new(cgul_exception_t* cex,
                cgul_cache__constructor_t constructor,
                cgul_cache__destructor_t destructor,
                cgul_cache__restructor_t restructor)
{
    cgul_cache_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* NOTE: The "restructor" is allowed to be NULL.  The client is
     * then responsible for resetting the object before putting it
     * back on the cgul_cache. */

    /* Sanity checks. */
    if (constructor == NULL) {
        *cex = cgul_exception__new(
            "cgul_cache__new: Invalid parameter: constructor == NULL");
        goto out;
    }
    if (destructor == NULL) {
        *cex = cgul_exception__new(
            "cgul_cache__new: Invalid parameter: destructor == NULL");
        goto out;
    }

    /* Allocate the cache object. */
    rv = (cgul_cache_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Initialize non-zero members. */
    rv->constructor = constructor;
    rv->destructor = destructor;
    rv->restructor = restructor;

 out:

    /* Clean up partially allocated objects. */
    if (*cex && rv) {
        cgul_cache__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_cache__delete(cgul_cache_t cache)
{
    cgul_exception_t local = NULL;

    /* Make sure the user isn't trying to delete the statically
     * allocate "freer" instance.  (See cgul_cache__get_freer().) */
    if (cache && (cache != &cgul_cache__the_freer)) {
        cgul_cache__free_cache(&local, cache);
        free(cache);
    }

}

unsigned long int
cgul_cache__get_size(cgul_exception_t* cex,
                     cgul_cache_t cache)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    rv = cache->size;

 out:
    return rv;
}

void
cgul_cache__set_size(cgul_exception_t* cex,
                     cgul_cache_t cache,
                     unsigned long int size)
{
    unsigned long int total_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    /* Ignore resize requests that are identical to the current size. */
    if (size == cache->size) {
        goto out;
    }

    /* Carefully calculate the new size. */
    total_size = cgul_defensive__add_unsigned_long(cex, size, cache->reserve);
    if (*cex) {
        goto out;
    }

    /* Resize the cache.  Notice that any objects that have recently
     * "moved" from the reserve to the main cache could be freed if
     * "size" (above) is less than the number of those objects.  This
     * is documented in the doxygen comments, and I think this is the
     * right thing to do based on the principle of least surprise. */
    cgul_cache__resize(cex, cache, total_size);
    if (*cex) {
        goto out;
    }

    /* Remember the new size. */
    cache->size = size;

 out:
    return;
}

unsigned long int
cgul_cache__get_reserve(cgul_exception_t* cex,
                        cgul_cache_t cache)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    rv = cache->reserve;

 out:
    return rv;
}

void
cgul_cache__set_reserve(cgul_exception_t* cex,
                        cgul_cache_t cache,
                        unsigned long int reserve)
{
    void* obj = NULL;
    unsigned long int total_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    /* Ignore reserve requests that are identical to the current reserve. */
    if (reserve == cache->reserve) {
        goto out;
    }

    /* If the new reserve is less than the old reserve, the user wants
     * us to make those reserved objects available.  Thus, we just
     * leave the objects in the vector and reduce cache->reserve so
     * that those objects are now available.  (We do not call
     * cgul_cache__resize() which might free those objects.) */
    if (reserve < cache->reserve) {
        cache->reserve = reserve;
        goto out;
    }

    /*
     * If you reach here, you are increasing the size of the reserve.
     */

    /* Carefully calculate the new total size. */
    total_size = cgul_defensive__add_unsigned_long(cex, cache->size, reserve);
    if (*cex) {
        goto out;
    }

    /* Resize the cache to accommodate the reserve plus any extra
     * objects that have been "moved" from the reserve to the main
     * cache. */
    cgul_cache__resize(cex, cache, total_size);
    if (*cex) {
        goto out;
    }

    /* Remember the new reserve. */
    cache->reserve = reserve;

    /* Allocate the reserve objects.  We do this by repeatedly getting
     * new cache objects and immediately putting them back on the
     * cache.  This works because cgul_cache__get() and
     * cgul_cache__put() are coded to do the right thing when the
     * size of cache->v is below the reserve. */
    while (cgul_vector__get_size(cex, cache->v) < reserve) {
        obj = cgul_cache__get(cex, cache);
        cgul_cache__put(cex, cache, obj);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

void*
cgul_cache__get(cgul_exception_t* cex,
                cgul_cache_t cache)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);

    /* Try to get an object from the cache. */
    if (cache->v) {
        /* We can only return unreserved objects. */
        if (cgul_vector__get_size(cex, cache->v) > cache->reserve) {
            rv = cgul_vector__pop_back(cex, cache->v);
            if (rv) {
                goto out;
            }
        }
    }

    /* Try to create a new object.  */
    if (cache->constructor) {
        rv = (*cache->constructor)(cex);
        if (*cex) {
            goto out;
        }
    }

 out:
    return rv;
}

void
cgul_cache__put(cgul_exception_t* cex,
                cgul_cache_t cache,
                void* obj)
{
    cgul_exception_t local = NULL;
    unsigned long int total_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cache);
    assert(obj);

    /* If the user set the restructor, reset the object to minimize
     * its memory footprint; otherwise, the user is responsible for
     * reseting the object before calling this method. */
    if (cache->restructor) {
        (*cache->restructor)(obj);
    }

    /* If the cache is not full, add the object.  One thing to notice
     * is that, because we have the option of adding the object to the
     * cache or deleting it, a failure in cgul_vector__push_back() is
     * not fatal to the operation of this method because we can always
     * just delete the object instead. */
    if (cache->v) {
        total_size = cache->size + cache->reserve;
        if (cgul_vector__get_size(&local, cache->v) < total_size) {
            cgul_vector__push_back(&local, cache->v, obj);
            if (local == NULL) {
                /* Success. */
                goto out;
            }
        }
    }

    /* Either the cache was full or an error occurred when trying to
     * add the object to the cache.  Either way, just delete it. */
    if (cache->destructor) {
        (*cache->destructor)(obj);
    }

 out:

    /* Clean up. */
    if (local) {
        cgul_exception__delete(local);
    }

    return;
}
