#include "cgul_convspec_p.h"
#include "cgul_libc.h"

/**
 * Reset <code>cs</code> so that it is in the default state.
 *
 * @param[in] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 */
static void
cgul_convspec__reset(cgul_exception_t* cex,
                     cgul_convspec_t cs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    /*
     * NOTE: If you are modifying this method, you should probably
     * also modify the cgul_convspec__copy() method.
     */

    /* Flags */
    cs->flag_hash = 0;
    cs->flag_zero = 0;
    cs->flag_dash = 0;
    cs->flag_space = 0;
    cs->flag_plus = 0;

    /* Field Width */
    cs->is_field_width_present = 0;
    cs->is_field_width_on_stack = 0;
    cs->field_width = 0;

    /* Precision */
    cs->is_precision_present = 0;
    cs->is_precision_on_stack = 0;
    cs->precision = 0;

    /* Length modifier */
    cs->is_length_modifier_present = 0;
    cgul_string__set_value(cex, cs->length_modifier, "");

    /* Conversion specifier */
    cs->cspec = '\0';

 out:
    return;
}

/**
 * This method returns whether <code>c</code> is a conversion
 * specifier character.  A "conversion specifier" is a character like
 * d in "%d".
 *
 * @param[in] cex c-style exception
 * @param[in] c character
 */
static int
cgul_convspec__is_cspec(cgul_exception_t* cex,
                        char c)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch (c) {
    case '%':
    case 'a':
    case 'A':
    case 'c':
    case 'C':
    case 'd':
    case 'e':
    case 'E':
    case 'f':
    case 'F':
    case 'g':
    case 'G':
    case 'i':
    case 'm':
    case 'n':
    case 'o':
    case 'p':
    case 's':
    case 'S':
    case 'u':
    case 'x':
    case 'X':
        rv = 1;
        break;
    default:
        rv = 0;
        break;
    }

 out:
    return rv;
}

/**
 * This method deserializes just the flags section of the conversion
 * specification which should begin at <code>format[*index]</code>.
 * This method increments <code>*index</code> so that it points one
 * beyond the end of the flags section of the conversion
 * specification.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] format format string
 * @param[in,out] index index into <code>format</code>
 */
static void
cgul_convspec__deserialize_flags(cgul_exception_t* cex,
                                 cgul_convspec_t cs,
                                 const char* format,
                                 unsigned long* index)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(format);
    assert(index);

    /* Reset the flags. */
    cs->flag_hash = 0;
    cs->flag_zero = 0;
    cs->flag_dash = 0;
    cs->flag_space = 0;
    cs->flag_plus = 0;

    /* Loop over all the flags. */
    for ( ; format[*index] != '\0' ; ++(*index)) {
        /* Set flags based on the current character. */
        switch (format[*index]) {
        case '#':
            cs->flag_hash = 1;
            break;
        case '0':
            cs->flag_zero = 1;
            break;
        case '-':
            cs->flag_dash = 1;
            break;
        case ' ':
            cs->flag_space = 1;
            break;
        case '+':
            cs->flag_plus = 1;
            break;
        default:
            /* Must not be a flag. */
            goto end_of_flags_loop;
        }
    }

 end_of_flags_loop:

    /* Check for an unexpected end of string. */
    if (format[*index] == '\0') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "unexpected end of conversion specification: %s",
                   format);
        goto out;
    }

 out:
    return;
}

/**
 * This method deserializes just the field width section of the
 * conversion specification which should begin at
 * <code>format[*index]</code>.  This method increments
 * <code>*index</code> so that it points one beyond the end of the
 * field width section of the conversion specification.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] format format string
 * @param[in,out] index index into <code>format</code>
 */
static void
cgul_convspec__deserialize_field_width(cgul_exception_t* cex,
                                       cgul_convspec_t cs,
                                       const char* format,
                                       unsigned long* index)
{
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(format);
    assert(index);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Reset the field width. */
    cs->is_field_width_present = 0;
    cs->is_field_width_on_stack = 0;
    cs->field_width = 0;

    /* Process the field width if it is present. */
    for ( ; format[*index] != '\0' ; ++(*index)) {
        /* The field width is always a positive decimal string or the
         * special '*' character that indicates the field width is on
         * the stack. */
        switch (format[*index]) {
        case '*':
            cs->is_field_width_present = 1;
            cs->is_field_width_on_stack = 1;
            break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            cs->is_field_width_present = 1;
            cgul_string__append_char(cex, s, format[*index]);
            break;
        default:
            goto end_of_field_width_loop;
        }
    }

 end_of_field_width_loop:

    /* Convert the field width from a string to an int. */
    if (cgul_string__get_length(cex, s) > 0) {
        /* Sanity check.  The user is not allowed to specify both a
         * field width and an '*' character. */
        if (cs->is_field_width_on_stack) {
            *cex = cgul_exception__new(
                       "cgul_convspec__deserialize: "
                       "dual specification of field width: %s",
                       format);
            goto out;
        }
        /* Convert the field width from a string to an int. */
        cs->field_width = cgul_string__to_int(cex, s, 10 /* base */);
        if (*cex) {
            goto out;
        }
    }

    /* Check for an unexpected end of string. */
    if (format[*index] == '\0') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "unexpected end of conversion specification: %s",
                   format);
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * This method deserializes just the precision section of the
 * conversion specification which should begin at
 * <code>format[*index]</code>.  This method increments
 * <code>*index</code> so that it points one beyond the end of the
 * precision section of the conversion specification.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] format format string
 * @param[in,out] index index into <code>format</code>
 */
static void
cgul_convspec__deserialize_precision(cgul_exception_t* cex,
                                     cgul_convspec_t cs,
                                     const char* format,
                                     unsigned long* index)
{
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(format);
    assert(index);

    /* Allocate a temporary string. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Reset the precision. */
    cs->is_precision_present = 0;
    cs->is_precision_on_stack = 0;
    cs->precision = 0;

    /* A precision is indicated by a leading '.' character; otherwise,
       the precision is not present. */
    if (format[*index] != '.') {
        goto out;
    }

    /* Skip the leading '.' character. */
    ++(*index);

    /* Remember that a precision is present.  Note that an empty
     * precision is interpreted the same as though the user specified
     * a precision of "0" which is how we are currently set up.  */
    cs->is_precision_present = 1;

    /* Process the precision if it is present. */
    for ( ; format[*index] != '\0' ; ++(*index)) {
        /* The precision is always a decimal string or the special '*'
         * character that indicates the precision is on the stack. */
        switch (format[*index]) {
        case '*':
            cs->is_precision_on_stack = 1;
            break;
        case '-':
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            cgul_string__append_char(cex, s, format[*index]);
            break;
        default:
            goto end_of_precision_loop;
        }
    }

 end_of_precision_loop:

    /* Convert the precision from a string to an int. */
    if (cgul_string__get_length(cex, s) > 0) {
        /* Sanity check.  The user is not allowed to specify both a
         * precision and an '*' character. */
        if (cs->is_precision_on_stack) {
            *cex = cgul_exception__new(
                       "cgul_convspec__deserialize: "
                       "dual specification of precision: %s",
                       format);
            goto out;
        }
        /* Convert the precision from a string to an int. */
        cs->precision = cgul_string__to_int(cex, s, 10 /* base */);
        if (*cex) {
            goto out;
        }
        /* A negative precision should be interpreted the same as a
         * precision of zero. */
        if (cs->precision < 0) {
            cs->precision = 0;
        }
    }

    /* Check for an unexpected end of string. */
    if (format[*index] == '\0') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "unexpected end of conversion specification: %s",
                   format);
        goto out;
    }

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

/**
 * This method deserializes just the length modifier section of the
 * conversion specification which should begin at
 * <code>format[*index]</code>.  This method increments
 * <code>*index</code> so that it points one beyond the end of the
 * length modifier section of the conversion specification.  If an
 * error occurs, an exception is thrown. <p>
 *
 * This method does not pass judgment on whether the length modifier
 * is supported.  It simply remembers what it is.  It is up to the
 * caller to do something with it. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] format format string
 * @param[in,out] index index into <code>format</code>
 */
static void
cgul_convspec__deserialize_length_modifier(cgul_exception_t* cex,
                                           cgul_convspec_t cs,
                                           const char* format,
                                           unsigned long* index)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(format);
    assert(index);

    /* Reset the length modifier. */
    cs->is_length_modifier_present = 0;
    cgul_string__set_value(cex, cs->length_modifier, "");

    /* The length modifier is everything up to the conversion
     * specifier. */
    for ( ; format[*index] != '\0' ; ++(*index)) {
        /* If this is a conversion specifier, we are done. */
        if (cgul_convspec__is_cspec(cex, format[*index])) {
            break;
        }
        /* This character must be part of the length modifier. */
        cs->is_length_modifier_present = 1;
        cgul_string__append_char(cex, cs->length_modifier, format[*index]);
        if (*cex) {
            goto out;
        }
    }

    /* Check for an unexpected end of string. */
    if (format[*index] == '\0') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "unexpected end of conversion specification: %s",
                   format);
        goto out;
    }

 out:
    return;
}

/**
 * This method deserializes just the conversion specifier section of
 * the conversion specification which should begin at
 * <code>format[*index]</code>.  This method increments
 * <code>*index</code> so that it points one beyond the end of the
 * conversion specifier section of the conversion specification.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] cs <code>cgul_convspec</code> instance
 * @param[in] format format string
 * @param[in,out] index index into <code>format</code>
 */
static void
cgul_convspec__deserialize_cspec(cgul_exception_t* cex,
                                 cgul_convspec_t cs,
                                 const char* format,
                                 unsigned long* index)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(format);
    assert(index);

    /* Reset the conversion specifier. */
    cs->cspec = '\0';

    /* Check for an unexpected end of string. */
    if (format[*index] == '\0') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "unexpected end of conversion specification: %s",
                   format);
        goto out;
    }

    /* Test to make sure the character is a conversion specifier. */
    if (cgul_convspec__is_cspec(cex, format[*index])) {
        /* Remember the conversion specifier. */
        cs->cspec = format[*index];
    } else {
        /* Otherwise, throw an exception. */
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "unexpected conversion specification: %s",
                   format);
        goto out;
    }

    /* Increment index. */
    ++(*index);

 out:
    return;
}

cgul_convspec_t
cgul_convspec__new(cgul_exception_t* cex)
{
    cgul_convspec_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate space for the new object. */
    rv = (cgul_convspec_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the length modifier string. */
    rv->length_modifier = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Reset the new object so that it is in the default state. */
    cgul_convspec__reset(cex, rv);

 out:

    /* Clean up partially instantiated objects. */
    if (*cex && rv) {
        cgul_convspec__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_convspec__delete(cgul_convspec_t cs)
{
    if (cs) {

        /* length_modifier */
        if (cs->length_modifier) {
            cgul_string__delete(cs->length_modifier);
        }

        /* Delete cs itself. */
        free(cs);
    }
}

void
cgul_convspec__copy(cgul_exception_t* cex,
                    cgul_convspec_t lhs,
                    cgul_convspec_t rhs)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(lhs);
    assert(rhs);

    /*
     * NOTE: If you are modifying this method, you should probably
     * also modify the cgul_convspec__reset() method.
     */

    /* Do not try to copy if lhs and rhs are the same object. */
    if (lhs == rhs) {
        goto out;
    }

    /* Flags */
    lhs->flag_hash = rhs->flag_hash;
    lhs->flag_zero = rhs->flag_zero;
    lhs->flag_dash = rhs->flag_dash;
    lhs->flag_space = rhs->flag_space;
    lhs->flag_plus = rhs->flag_plus;

    /* Field Width */
    lhs->is_field_width_present = rhs->is_field_width_present;
    lhs->is_field_width_on_stack = rhs->is_field_width_on_stack;
    lhs->field_width = rhs->field_width;

    /* Precision */
    lhs->is_precision_present = rhs->is_precision_present;
    lhs->is_precision_on_stack = rhs->is_precision_on_stack;
    lhs->precision = rhs->precision;

    /* Length modifier */
    lhs->is_length_modifier_present = rhs->is_length_modifier_present;
    cgul_string__set_value(cex,
                           lhs->length_modifier,
                           cgul_string__get_value(cex, rhs->length_modifier));

    /* Conversion specifier */
    lhs->cspec = rhs->cspec;

 out:
    return;
}

void
cgul_convspec__deserialize(cgul_exception_t* cex,
                           cgul_convspec_t cs,
                           const char* format,
                           unsigned long* index)
{
    unsigned long int index_tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(format);

    /* Reset this object. */
    cgul_convspec__reset(cex, cs);

    /* Manipulate a temporary index so we do not disturb "index"
     * unless the deserialization is a success. */
    index_tmp = *index;

    /* The conversion specification must not be empty. */
    if (format[index_tmp] == '\0') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "conversion specification is empty");
        goto out;
    }

    /* The conversion specification must start with a '%' character. */
    if (format[index_tmp] != '%') {
        *cex = cgul_exception__new(
                   "cgul_convspec__deserialize: "
                   "conversion specification does not begin with a %%: %s",
                   format);
        goto out;
    }

    /* Skip the leading '%' character. */
    ++index_tmp;

    /* Process the flags section. */
    cgul_convspec__deserialize_flags(
        cex, cs, format, &index_tmp);

    /* Process the field width section. */
    cgul_convspec__deserialize_field_width(
        cex, cs, format, &index_tmp);

    /* Process the precision section. */
    cgul_convspec__deserialize_precision(
        cex, cs, format, &index_tmp);

    /* Process the length modifier section. */
    cgul_convspec__deserialize_length_modifier(
        cex, cs, format, &index_tmp);

    /* Process the conversion specifier. */
    cgul_convspec__deserialize_cspec(
        cex, cs, format, &index_tmp);

    if (*cex) {
        goto out;
    }

    /* Success. */
    *index = index_tmp;

 out:

    return;
}

void
cgul_convspec__serialize(cgul_exception_t* cex,
                         cgul_convspec_t cs,
                         cgul_string_t s)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);
    assert(s);

    /* We always start with a '%' character. */
    cgul_string__append_char(cex, s, '%');

    /* The + flag overrides the space flag. */
    if (cs->flag_plus) {
        cgul_string__append_char(cex, s, '+');
    } else if (cs->flag_space) {
        cgul_string__append_char(cex, s, ' ');
    }

    /* Insert the # flag if necessary. */
    if (cs->flag_hash) {
        cgul_string__append_char(cex, s, '#');
    }

    /* The - flag overrides the 0 flag. */
    if (cs->flag_dash) {
        cgul_string__append_char(cex, s, '-');
    } else if (cs->flag_zero) {
        cgul_string__append_char(cex, s, '0');
    }

    /* Field width */
    if (cs->is_field_width_present) {
        if (cs->is_field_width_on_stack) {
            cgul_string__append_char(cex, s, '*');
        } else {
            cgul_string__append_decimal(cex, s, cs->field_width);
        }
    }

    /* Precision */
    if (cs->is_precision_present) {
        cgul_string__append_char(cex, s, '.');
        if (cs->is_precision_on_stack) {
            cgul_string__append_char(cex, s, '*');
        } else {
            cgul_string__append_decimal(cex, s, cs->precision);
        }
    }

    /* Length modifier */
    if (cs->is_length_modifier_present) {
        const char* lm = cgul_string__get_value(cex, cs->length_modifier);
        cgul_string__append(cex, s, lm);
    }

    /* Conversion specifier */
    assert(cs->cspec != '\0');
    cgul_string__append_char(cex, s, cs->cspec);

 out:
    return;
}

int
cgul_convspec__get_flag_hash(cgul_exception_t* cex,
                             cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->flag_hash) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_flag_hash(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_hash)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (flag_hash) {
        cs->flag_hash = 1;
    } else {
        cs->flag_hash = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_flag_zero(cgul_exception_t* cex,
                             cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->flag_zero) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_flag_zero(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_zero)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (flag_zero) {
        cs->flag_zero = 1;
    } else {
        cs->flag_zero = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_flag_dash(cgul_exception_t* cex,
                             cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->flag_dash) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_flag_dash(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_dash)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (flag_dash) {
        cs->flag_dash = 1;
    } else {
        cs->flag_dash = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_flag_space(cgul_exception_t* cex,
                              cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->flag_space) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_flag_space(cgul_exception_t* cex,
                              cgul_convspec_t cs,
                              int flag_space)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (flag_space) {
        cs->flag_space = 1;
    } else {
        cs->flag_space = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_flag_plus(cgul_exception_t* cex,
                             cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->flag_plus) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_flag_plus(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int flag_plus)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (flag_plus) {
        cs->flag_plus = 1;
    } else {
        cs->flag_plus = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_is_field_width_present(cgul_exception_t* cex,
                                          cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->is_field_width_present) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_is_field_width_present(cgul_exception_t* cex,
                                          cgul_convspec_t cs,
                                          int is_field_width_present)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (is_field_width_present) {
        cs->is_field_width_present = 1;
    } else {
        cs->is_field_width_present = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_is_field_width_on_stack(cgul_exception_t* cex,
                                           cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->is_field_width_on_stack) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_is_field_width_on_stack(cgul_exception_t* cex,
                                           cgul_convspec_t cs,
                                           int is_field_width_on_stack)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (is_field_width_on_stack) {
        cs->is_field_width_on_stack = 1;
    } else {
        cs->is_field_width_on_stack = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_field_width(cgul_exception_t* cex,
                               cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    rv = cs->field_width;

 out:
    return rv;
}

void
cgul_convspec__set_field_width(cgul_exception_t* cex,
                               cgul_convspec_t cs,
                               int field_width)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    cs->field_width = field_width;

 out:
    return;
}

int
cgul_convspec__get_is_precision_present(cgul_exception_t* cex,
                                        cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->is_precision_present) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_is_precision_present(cgul_exception_t* cex,
                                        cgul_convspec_t cs,
                                        int is_precision_present)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (is_precision_present) {
        cs->is_precision_present = 1;
    } else {
        cs->is_precision_present = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_is_precision_on_stack(cgul_exception_t* cex,
                                         cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->is_precision_on_stack) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_is_precision_on_stack(cgul_exception_t* cex,
                                         cgul_convspec_t cs,
                                         int is_precision_on_stack)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (is_precision_on_stack) {
        cs->is_precision_on_stack = 1;
    } else {
        cs->is_precision_on_stack = 0;
    }

 out:
    return;
}

int
cgul_convspec__get_precision(cgul_exception_t* cex,
                             cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    rv = cs->precision;

 out:
    return rv;
}

void
cgul_convspec__set_precision(cgul_exception_t* cex,
                             cgul_convspec_t cs,
                             int precision)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    cs->precision = precision;

 out:
    return;
}

int
cgul_convspec__get_is_length_modifier_present(cgul_exception_t* cex,
                                              cgul_convspec_t cs)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (cs->is_length_modifier_present) {
        rv = 1;
    }

 out:
    return rv;
}

void
cgul_convspec__set_is_length_modifier_present(cgul_exception_t* cex,
                                              cgul_convspec_t cs,
                                              int is_length_modifier_present)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    if (is_length_modifier_present) {
        cs->is_length_modifier_present = 1;
    } else {
        cs->is_length_modifier_present = 0;
    }

 out:
    return;
}

const char*
cgul_convspec__get_length_modifier(cgul_exception_t* cex,
                                   cgul_convspec_t cs)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    rv = cgul_string__get_value(cex, cs->length_modifier);

 out:
    return rv;
}

void
cgul_convspec__set_length_modifier(cgul_exception_t* cex,
                                   cgul_convspec_t cs,
                                   const char* length_modifier)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    cgul_string__set_value(cex, cs->length_modifier, length_modifier);

 out:
    return;
}

char
cgul_convspec__get_cspec(cgul_exception_t* cex,
                         cgul_convspec_t cs)
{
    char rv = '\0';

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    rv = cs->cspec;

 out:
    return rv;
}

void
cgul_convspec__set_cspec(cgul_exception_t* cex,
                         cgul_convspec_t cs,
                         char cspec)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cs);

    cs->cspec = cspec;

 out:
    return;
}
