#include "cgul_heap_p.h"
#include "cgul_libc.h"

cgul_heap_t
cgul_heap__new(cgul_exception_t* cex,
               cgul_heap_primitives__compare_t cf)
{
    cgul_heap_t rv = NULL;
    cgul_vector_t v = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cf);

    /* Allocate. */
    v = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Create a new cgul_heap instance that uses our vector. */
    rv = cgul_heap__new_from_vector(cex, v, cf);
    if (*cex) {
        goto out;
    }

    /* The cgul_heap instance now owns the vector. */
    v = NULL;

 out:

    /* Clean up. */
    if (v) {
        cgul_vector__delete(v);
    }

    return rv;
}

cgul_heap_t
cgul_heap__new_from_vector(cgul_exception_t* cex,
                           cgul_vector_t v,
                           cgul_heap_primitives__compare_t cf)
{
    cgul_heap_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(cf);

    /* Allocate. */
    rv = (cgul_heap_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Initialize. */
    rv->v = v;
    rv->cf = cf;

    /* Heapify. */
    cgul_heap_primitives__heapify(cex,
                                  cgul_vector__get_array(cex, rv->v),
                                  cgul_vector__get_size(cex, rv->v),
                                  cf);

 out:

    /* Clean up. */
    if (*cex && rv) {
        cgul_heap__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_heap__delete(cgul_heap_t heap)
{
    if (heap) {

        /* v */
        if (heap->v) {
            cgul_vector__delete(heap->v);
        }

        /* Free the heap itself. */
        free(heap);
    }
}

void
cgul_heap__free_values(cgul_heap_t heap)
{
    assert(heap);
    assert(heap->v);

    cgul_vector__free_values(heap->v);
}

int
cgul_heap__is_empty(cgul_exception_t* cex,
                    cgul_heap_t heap)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(heap);

    rv = cgul_vector__is_empty(cex, heap->v);

 out:
    return rv;
}

void
cgul_heap__push(cgul_exception_t* cex,
                cgul_heap_t heap,
                void* value)
{
    unsigned long int heap_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(heap);

    /* Add the new value to the back of the underlying array. */
    cgul_vector__push_back(cex, heap->v, value);
    if (*cex) {
        goto out;
    }

    /* Shift the value up to maintain heap order. */
    heap_size = cgul_vector__get_size(cex, heap->v);
    cgul_heap_primitives__shift_up(cex,
                                   cgul_vector__get_array(cex, heap->v),
                                   heap_size,      /* size  */
                                   heap_size - 1,  /* index */
                                   heap->cf);

 out:
    return;
}

void*
cgul_heap__pop(cgul_exception_t* cex,
               cgul_heap_t heap)
{
    void* rv = NULL;
    void** array = NULL;
    size_t array_size = 0;
    void* back_value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(heap);

    /* Do not attempt to pop an empty heap. */
    array_size = cgul_vector__get_size(cex, heap->v);
    if (array_size <= 0) {
        goto out;
    }

    /* Get the top value. */
    rv = cgul_vector__at(cex, heap->v, 0);
    if (*cex) {
        goto out;
    }

    /* Remove the back value from the heap. */
    back_value = cgul_vector__pop_back(cex, heap->v);
    if (*cex) {
        goto out;
    }
    --array_size;

    /* If the heap is now empty, there is no replacement value because
     * there was only one value on the heap to begin with. */
    if (array_size <= 0) {
        assert(rv == back_value);
        goto out;
    }

    /* Replace the top value with the back value. */
    array = cgul_vector__get_array(cex, heap->v);
    array[0] = back_value;

    /* Shift down the back value to maintain heap order. */
    cgul_heap_primitives__shift_down(cex,
                                     array,
                                     array_size,
                                     0,  /* index */
                                     1,  /* use_floyds_method */
                                     heap->cf);

 out:
    return rv;
}

void*
cgul_heap__top(cgul_exception_t* cex,
               cgul_heap_t heap)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(heap);

    /* Do not attempt to pop an empty heap. */
    if (cgul_vector__is_empty(cex, heap->v)) {
        goto out;
    }

    /* Get the top value. */
    rv = cgul_vector__at(cex, heap->v, 0);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

unsigned long int
cgul_heap__get_size(cgul_exception_t* cex,
                    cgul_heap_t heap)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(heap);

    rv = cgul_vector__get_size(cex, heap->v);

 out:
    return rv;
}

void
cgul_heap__swap(cgul_exception_t* cex,
                cgul_heap_t t1,
                cgul_heap_t t2)
{
    struct cgul_heap tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t1);
    assert(t2);

    /* Nothing to swap if they point to the same objects. */
    if (t1 == t2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *t1;
    *t1 = *t2;
    *t2 = tmp;

 out:
    return;
}
