#include "cgul_stream_reader_p.h"
#include "cgul_defensive.h"
#include "cgul_endian.h"
#include "cgul_libc.h"
#include "cgul_unicode.h"

/**
 * Default block size to use with <code>cgul_block_reader</code>.
 */
static size_t
CGUL_STREAM_READER__DEFAULT_BLOCK_SIZE = 16384;

/**
 * Unicode code point for "\n".
 */
static cgul_wchar_t
CGUL_STREAM_READER__LINE_FEED = (cgul_wchar_t)0x0a;

/**
 * Unicode code point for "\r".
 */
static cgul_wchar_t
CGUL_STREAM_READER__CARRIAGE_RETURN = (cgul_wchar_t)0x0d;

/**
 * Unicode byte-order mark (BOM).
 */
static cgul_wchar_t
CGUL_STREAM_READER__BOM = (cgul_wchar_t)0xfeff;

/**
 * Byte swap each 16-bit value in the block <code>block16</code>
 * having <code>block_size</code> count of 16-bit values.
 *
 * @param[in] cex c-style exception
 * @param[in,out] block16 block of 16-bit values
 * @param[in] block_size block size in 16-bit values
 */
static void
cgul_stream_reader__byte_swap_block16(cgul_exception_t* cex,
                                      cgul_hchar_t* block16,
                                      size_t block_size)
{
    size_t i = 0;
    unsigned char tmp = '\0';
    unsigned char* value8 = NULL;

    for (i = 0 ; i < block_size ; ++i) {
        value8 = (unsigned char*)&block16[i];
        tmp = value8[0];
        value8[0] = value8[1];
        value8[1] = tmp;
    }
}

/**
 * Byte swap each 32-bit value in the block <code>block32</code>
 * having <code>block_size</code> count of 32-bit values.
 *
 * @param[in] cex c-style exception
 * @param[in,out] block32 block of 32-bit values
 * @param[in] block_size block size in 32-bit values
 */
static void
cgul_stream_reader__byte_swap_block32(cgul_exception_t* cex,
                                      cgul_wchar_t* block32,
                                      size_t block_size)
{
    size_t i = 0;
    unsigned char tmp = '\0';
    unsigned char* value8 = NULL;

    for (i = 0 ; i < block_size ; ++i) {
        value8 = (unsigned char*)&block32[i];
        tmp = value8[0];
        value8[0] = value8[3];
        value8[3] = tmp;
        tmp = value8[1];
        value8[1] = value8[2];
        value8[2] = tmp;
    }
}

/**
 * This function holds the common code for
 * <code>cgul_stream_reader__decode_utf16be()</code> and
 * <code>cgul_stream_reader__decode_utf16le()</code>.  It is not a
 * public function because UTF-16 usually means a Unicode Byte-Order
 * Mark (BOM) is present which should be automatically detected.  This
 * function however assumes the UTF-16 character stream is in the
 * native byte order for the host machine.  This allows it to be used
 * by either of the calling functions after possibly byte swapping the
 * <code>block_encoded</code> buffer. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
static void
cgul_stream_reader__decode_utf16(cgul_exception_t* cex,
                                 unsigned char* block_encoded,
                                 size_t size_encoded,
                                 size_t* index_encoded,
                                 cgul_wchar_t* block_decoded,
                                 size_t size_decoded,
                                 size_t* count_decoded)
{
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block_encoded);
    assert(index_encoded);
    assert(block_decoded);
    assert(count_decoded);

    /* Sanity check. */
    if (size_encoded < 2) {
        *cex = cgul_exception__new(
            "cgul_stream_reader__decode_utf16: "
            "block size (%zu) is too small to decode UTF-16.",
            size_encoded);
        goto out;
    }

    /* Iterate over each UTF-16 character.  Because "size_encoded" is
     * in (8-bit) bytes but "i" is in (16-bit) cgul_wchar characters,
     * we have to multiply "i" by 2 (using i << 1) to properly compare
     * them.  Add 1 to make sure all 16-bits are present in the block. */
    for (i = 0, j = 0;
         ((i << 1) + 1 < size_encoded) && (j < size_decoded);
         ++j)
    {

        /* Decode the next UTF-16 character.  Note that this updates
         * "i" to point at the start of the next 16-bit character. */
        block_decoded[j] = cgul_unicode__hctowc(
            cex, (cgul_hchar_t*)block_encoded, size_encoded, &i);
        if (*cex) {
            if (i == 0) {
                /* Error. */
                goto out;
            } else {
                /* Quash the exception because it could be caused by
                 * having an encoded character that spans blocks.  If
                 * it really is an error, it will be detected the next
                 * time this function is called because the offending
                 * bytes will be at the front of the block and (i == 0)
                 * will be true. */
                cgul_exception__delete(*cex);
                *cex = NULL;
                break;
            }
        }

    }

    /* Return new index and count.  Note that we double "i" to convert
     * it from an index for (16-bit) cgul_hchar characters to an index
     * for (8-bit) bytes. */
    *index_encoded = (i << 1);
    *count_decoded = j;

 out:
    return;
}

/**
 * This function holds the common code for
 * <code>cgul_stream_reader__decode_utf32be()</code> and
 * <code>cgul_stream_reader__decode_utf32le()</code>.  It is not a
 * public function because UTF-32 usually means a Unicode Byte-Order
 * Mark (BOM) is present which should be automatically detected.  This
 * function however assumes the UTF-32 character stream is in the
 * native byte order for the host machine.  This allows it to be used
 * by either of the calling functions after possibly byte swapping the
 * <code>block_encoded</code> buffer. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] block_encoded block of encoded bytes
 * @param[in] size_encoded total size of <code>block_encoded</code> in bytes
 * @param[in,out] index_encoded index into <code>block_encoded</code>
 * @param[out] block_decoded block of decoded wide characters
 * @param[in] size_decoded total size of <code>block_decoded</code> in chars
 * @param[out] count_decoded count of decoded characters
 * @see cgul_stream_reader__decode_t
 */
static void
cgul_stream_reader__decode_utf32(cgul_exception_t* cex,
                                 unsigned char* block_encoded,
                                 size_t size_encoded,
                                 size_t* index_encoded,
                                 cgul_wchar_t* block_decoded,
                                 size_t size_decoded,
                                 size_t* count_decoded)
{
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block_encoded);
    assert(index_encoded);
    assert(block_decoded);
    assert(count_decoded);

    /* Sanity check. */
    if (size_encoded < 4) {
        *cex = cgul_exception__new(
            "cgul_stream_reader__decode_utf32: "
            "block size (%zu) is too small to decode UTF-32.",
            size_encoded);
        goto out;
    }

    /* Iterate over each UTF-32 character.  Because "size_encoded" is
     * in (8-bit) bytes but "i" is in (32-bit) cgul_wchar characters,
     * we have to multiply "i" by 4 (using i << 2) to properly compare
     * them.  Add 3 to make sure all 32-bits are present in the block. */
    for (i = 0, j = 0;
         ((i << 2) + 3 < size_encoded) && (j < size_decoded);
         ++i, ++j)
    {

        /* Nothing to do except copy the UTF-32 character from the
         * encoded block to the decoded block. */
        block_decoded[j] = ((cgul_wchar_t*)block_encoded)[i];

    }

    /* Return new index and count.  Note that we quadruple "i" to
     * convert it from an index for (32-bit) cgul_hchar characters to
     * an index for (8-bit) bytes. */
    *index_encoded = (i << 2);
    *count_decoded = j;

 out:
    return;
}

/**
 * Clear the queue of block readers <code>block_readers</code> without
 * deleting the queue itself.
 *
 * @param[in] cex c-style exception
 * @param[in] block_readers block readers
 */
static void
cgul_stream_reader__clear_block_readers(cgul_exception_t* cex,
                                        cgul_list_t block_readers)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block_readers);

    /* We do not own the block readers so all we have to do is clear
     * the queue itself. */
    cgul_list__clear(cex, block_readers, NULL);

 out:
    return;
}

/**
 * Try to automatically detect the correct decoder by looking for the
 * Unicode Byte-Order Mark (BOM).  Failing that, use the UTF-8
 * decoder.  This method assumes the first block of data has already
 * been read and that it has a size of at least 4 bytes.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 */
static void
cgul_stream_reader__autodetect_decoder(cgul_exception_t* cex,
                                       cgul_stream_reader_t sr)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Sanity check. */
    if (sr->block_size < 4) {
        *cex = cgul_exception__new(
            "cgul_stream_reader__autodetect_decoder: "
            "block size (%zu) must be at least 4 to automatically "
            "detect the correct decoder",
            sr->block_size);
        goto out;
    }

    /* Sanity check. */
    if (sr->index_encoded != 0) {
        *cex = cgul_exception__new(
            "cgul_stream_reader__autodetect_decoder: internal error: "
            "attempt to automatically detect decoder after initial read");
        goto out;
    }

    /* Sanity check. */
    if (sr->decode) {
        *cex = cgul_exception__new(
            "cgul_stream_reader__autodetect_decoder: internal error: "
            "attempt to automatically detect decoder "
            "after decoder has been set");
        goto out;
    }

    /* UTF-32BE */
    if ((sr->block_encoded[0] == 0x00) &&
        (sr->block_encoded[1] == 0x00) &&
        (sr->block_encoded[2] == 0xFE) &&
        (sr->block_encoded[3] == 0xFF))
    {
        sr->decode = cgul_stream_reader__decode_utf32be;

    /* UTF-32LE */
    } else if ((sr->block_encoded[0] == 0xFF) &&
               (sr->block_encoded[1] == 0xFE) &&
               (sr->block_encoded[2] == 0x00) &&
               (sr->block_encoded[3] == 0x00))
    {
        sr->decode = cgul_stream_reader__decode_utf32le;

    /* UTF-16BE */
    } else if ((sr->block_encoded[0] == 0xFE) &&
               (sr->block_encoded[1] == 0xFF))
    {
        sr->decode = cgul_stream_reader__decode_utf16be;

    /* UTF-16LE */
    } else if ((sr->block_encoded[0] == 0xFF) &&
               (sr->block_encoded[1] == 0xFE))
    {
        sr->decode = cgul_stream_reader__decode_utf16le;

    /* UTF-8 */
    } else if ((sr->block_encoded[0] == 0xEF) &&
               (sr->block_encoded[1] == 0xBB) &&
               (sr->block_encoded[2] == 0xBF))
    {
        sr->decode = cgul_stream_reader__decode_utf8;

    /* Default. */
    } else {
        sr->decode = cgul_stream_reader__decode_utf8;
    }

 out:
    return;
}

/**
 * Decode the next block of characters.  Unused characters in
 * <code>sr->block_encoded</code> will be stitched together with new
 * characters read from the current block reader or, if the current
 * block reader is exhausted, from the next block reader in the
 * queue.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 */
static void
cgul_stream_reader__decode_next_block(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr)
{
    size_t i = 0;
    size_t rcount = 0;
    cgul_block_reader_t br_next = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Effectively clear sr->block_decoded. */
    sr->index_decoded = 0;
    sr->rcount_decoded = 0;

    /* Move the unused bytes in sr->block_encoded to the front. */
    if (sr->index_encoded > 0) {
        for (i = 0 ; i + sr->index_encoded < sr->rcount_encoded ; ++i) {
            sr->block_encoded[i] = sr->block_encoded[i + sr->index_encoded];
        }
        /* Only the bytes that were copied to the front are now valid. */
        sr->index_encoded = 0;
        sr->rcount_encoded = i;
    }

    /* Try to fill the remainder of the block using multiple block
     * readers from the queue if necessary. */
    while (sr->rcount_encoded < sr->block_size) {

        /* Read the next block (taking into account the bytes that
         * are already in block). */
        rcount = cgul_block_reader__read_next_block(
                     cex,
                     sr->br,
                     &sr->block_encoded[sr->rcount_encoded],
                     sr->block_size - sr->rcount_encoded);
        if (*cex) {
            goto out;
        }

        if (rcount > 0) {

            /* Increment the count of valid bytes in sr->rcount_encoded. */
            sr->rcount_encoded += rcount;

        } else {

            /* This is the end of file for the current block
             * reader.  Try to replace the current block reader
             * with the one at the front of the block_readers
             * queue. */
            br_next = (cgul_block_reader_t)
                      cgul_list__pop_front(cex, sr->block_readers);
            if (!br_next) {
                /* EOF */
                break;
            }

            /* Replace the exhausted block reader with the block
             * reader at the front of the queue. */
            sr->br = br_next;
        }

    }

    /* If the decoder is not set, try to automatically detect the
     * correct decoder by looking for the Unicode Byte-Order Mark
     * (BOM).  Failing that, use the UTF-8 decoder. */
    if (!sr->decode) {
        cgul_stream_reader__autodetect_decoder(cex, sr);
        if (*cex) {
            goto out;
        }
    }

    /* Decode. */
    if (sr->rcount_encoded > 0) {
        sr->decode(cex,
                   sr->block_encoded,
                   sr->rcount_encoded,
                   &sr->index_encoded,
                   sr->block_decoded,
                   sr->block_size,
                   &sr->rcount_decoded);
        if (*cex) {
            /* Quash the exception. */
            cgul_exception__delete(*cex);
            *cex = NULL;
            /* Replace the exception with one having a better message. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: invalid encoding",
                cgul_stream_reader__get_presentation_name(cex, sr),
                sr->line_current,
                sr->column_current);
            goto out;
        }
    }

 out:
    return;
}

/**
 * Return the next decoded character as <code>next_char</code>.  If
 * successful, <code>result</code> is set to <code>1</code>.  If the
 * end of file (EOF) is reached, <code>result</code> is set to
 * <code>0</code>, and <code>next_char</code> is set to
 * <code>CGUL_WCHAR__NUL</code>.  If an error occurs,
 * <code>result</code> is set to <code>0</code>,
 * <code>next_char</code> is set to <code>CGUL_WCHAR__NUL</code>, and
 * an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[out] next_char next decoded character
 * @param[out] result whether a character was successfully decoded
 */
#define CGUL_STREAM_READER__DECODE(cex, sr, next_char, result)          \
do {                                                                    \
    assert(cex);                                                        \
    if (*(cex)) {                                                       \
        break;                                                          \
    };                                                                  \
    assert(sr);                                                         \
    /* Try to get the character from our block buffer. */               \
    if ((sr)->index_decoded < (sr)->rcount_decoded) {                   \
        (next_char) = (sr)->block_decoded[(sr)->index_decoded++];       \
        (result) = 1;                                                   \
    } else {                                                            \
        /* Assume failure. */                                           \
        (next_char) = CGUL_WCHAR__NUL;                                  \
        (result) = 0;                                                   \
        /* The block buffer is exhausted.  Decode the next block .*/    \
        cgul_stream_reader__decode_next_block((cex), (sr));             \
        if (*(cex)) {                                                   \
            break;                                                      \
        }                                                               \
        /* Try again to get the character from our block buffer. */     \
        if ((sr)->index_decoded < (sr)->rcount_decoded) {               \
            /* Success. */                                              \
            (next_char) = (sr)->block_decoded[(sr)->index_decoded++];   \
            (result) = 1;                                               \
        }                                                               \
    }                                                                   \
} while (0)

/**
 * This is a helper macro for <code>cgul_stream_reader__read()</code>
 * that advances the line and column numbers based on the character
 * <code>wc</code> that was read.  If <code>wc</code> is a '\r' or a
 * '\n' character, advance to the next line; otherwise, advance to the
 * next column.  This macro assumes that <code>wc</code> has already
 * had EOL normalization performed if it is enabled. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 * @param[in] wc wide character that was read
 */
#define CGUL_STREAM_READER__ADVANCE_LINE_AND_COLUMN_NUMBERS(cex, sr, wc) \
do {                                                                    \
    assert((cex));                                                      \
    if (*(cex)) {                                                       \
        break;                                                          \
    }                                                                   \
    assert(sr);                                                         \
    /* Do not advance if the character is the Unicode */                \
    /* byte-order mark (BOM). */                                        \
    if ((wc) == CGUL_STREAM_READER__BOM) {                              \
        break;                                                          \
    }                                                                   \
    /* Remember the previous line and column numbers. */                \
    (sr)->line_previous = (sr)->line_current;                           \
    (sr)->column_previous = (sr)->column_current;                       \
    /* Determine if we are moving to the next line or else just the */  \
    /* next column. */                                                  \
    if (((wc) == CGUL_STREAM_READER__LINE_FEED) ||                      \
        ((wc) == CGUL_STREAM_READER__CARRIAGE_RETURN))                  \
    {                                                                   \
        /* Advance to the first column on the next line. */             \
        (sr)->line_current =                                            \
            cgul_defensive__add_size_type((cex), (sr)->line_current, 1); \
        (sr)->column_current = 1;                                       \
        if (*(cex)) {                                                   \
            break;                                                      \
        }                                                               \
    } else {                                                            \
        /* Advance to the next column on the same line. */              \
        (sr)->column_current =                                          \
            cgul_defensive__add_size_type((cex), (sr)->column_current, 1); \
        if (*(cex)) {                                                   \
            break;                                                      \
        }                                                               \
    }                                                                   \
} while (0)

/**
 * Reinitialize the <code>cgul_stream_reader</code> instance.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 */
static void
cgul_stream_reader__reinitialize(cgul_exception_t* cex,
                                 cgul_stream_reader_t sr)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* decode -- re-enable auto-detection */
    sr->decode = NULL;

    /* br -- We do not own. */
    sr->br = NULL;

    /* block_readers */
    if (sr->block_readers) {
        cgul_stream_reader__clear_block_readers(cex, sr->block_readers);
    } else {
        sr->block_readers = cgul_list__new(cex);
        if (*cex) {
            goto out;
        }
    }

    /* block_encoded */
    if (!sr->block_encoded) {
        sr->block_encoded = (unsigned char*)
                            cgul_libc__calloc(cex,
                                              sr->block_size,
                                              sizeof(*sr->block_encoded));
        if (*cex) {
            goto out;
        }
    }

    /* block_decoded */
    if (!sr->block_decoded) {
        sr->block_decoded = (cgul_wchar_t*)
                            cgul_libc__calloc(cex,
                                              sr->block_size,
                                              sizeof(*sr->block_decoded));
        if (*cex) {
            goto out;
        }
    }

    /* block_size -- initialized in constructor */

    /* index_encoded */
    sr->index_encoded = 0;

    /* rcount_encoded */
    sr->rcount_encoded = 0;

    /* index_decoded */
    sr->index_decoded = 0;

    /* rcount_decoded */
    sr->rcount_decoded = 0;

    /* eof */
    sr->eof = 0;

    /* track_line_and_column -- initialized in constructor */

    /* line_current */
    sr->line_current = 1;

    /* column_current */
    sr->column_current = 1;

    /* line_previous */
    sr->line_previous = 0;

    /* column_previous */
    sr->column_previous = 0;

    /* normalize_eol -- initialized in constructor */

    /* actual_previous_char */
    sr->actual_previous_char = CGUL_WCHAR__NUL;

    /* presentation_name */
    if (sr->presentation_name) {
        free(sr->presentation_name);
        sr->presentation_name = NULL;
    }

 out:
    return;
}

/* Public callback function that implements cgul_stream_reader__decode_t. */
void
cgul_stream_reader__decode_ascii(cgul_exception_t* cex,
                                 unsigned char* block_encoded,
                                 size_t size_encoded,
                                 size_t* index_encoded,
                                 cgul_wchar_t* block_decoded,
                                 size_t size_decoded,
                                 size_t* count_decoded)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block_encoded);
    assert(index_encoded);
    assert(block_decoded);
    assert(count_decoded);

    /* Iterate over each ASCII character. */
    for (i = 0 ; (i < size_encoded) && (i < size_decoded) ; ++i) {

        /* Copy the next ASCII character to the wide-character array. */
        block_decoded[i] = (cgul_wchar_t)block_encoded[i];

    }

    /* Return new index and count. */
    *index_encoded = i;
    *count_decoded = i;

 out:
    return;
}

/* Public callback function that implements cgul_stream_reader__decode_t. */
void
cgul_stream_reader__decode_utf8(cgul_exception_t* cex,
                                unsigned char* block_encoded,
                                size_t size_encoded,
                                size_t* index_encoded,
                                cgul_wchar_t* block_decoded,
                                size_t size_decoded,
                                size_t* count_decoded)
{
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block_encoded);
    assert(index_encoded);
    assert(block_decoded);
    assert(count_decoded);

    /* Iterate over each UTF-8 character. */
    for (i = 0, j = 0 ; (i < size_encoded) && (j < size_decoded) ; ++j) {

        /* Special Case: 7-bit ASCII can be copied directly. */
        if (block_encoded[i] < 0x80) {
            block_decoded[j] = (cgul_wchar_t)block_encoded[i++];
            continue;
        }

        /* General Case: UTF-8 characters need to be decoded.  Note
         * that this updates "i" to point at the start of the next
         * character. */
        block_decoded[j] = cgul_unicode__mbtowc(
            cex, (char*)block_encoded, size_encoded, &i);
        if (*cex) {
            if (i == 0) {
                /* Error. */
                goto out;
            } else {
                /* Quash the exception because it could be caused by
                 * having an encoded character that spans blocks.  If
                 * it really is an error, it will be detected the next
                 * time this function is called because the offending
                 * bytes will be at the front of the block and (i == 0)
                 * will be true. */
                cgul_exception__delete(*cex);
                *cex = NULL;
                break;
            }
        }

    }

    /* Return new index and count. */
    *index_encoded = i;
    *count_decoded = j;

 out:
    return;
}

/* Public callback function that implements cgul_stream_reader__decode_t. */
void
cgul_stream_reader__decode_utf16be(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* If the machine architecture is little endian, byte swap the
     * big-endian block so it is in the native byte order.  Note that
     * "size_encoded" is the size in (8-bit) bytes, but
     * byte_swap_block16() needs the number of (16-bit) cgul_hchar
     * characters.  It is ok if "size_encoded" is odd. */
    if (cgul_endian__is_little_endian(cex)) {
        cgul_stream_reader__byte_swap_block16(cex,
                                              (cgul_hchar_t*)block_encoded,
                                              size_encoded >> 1);
    }

    /* Decode the block now that it is in the native byte order. */
    cgul_stream_reader__decode_utf16(cex,
        block_encoded, size_encoded, index_encoded,
        block_decoded, size_decoded, count_decoded);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/* Public callback function that implements cgul_stream_reader__decode_t. */
void
cgul_stream_reader__decode_utf16le(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* If the machine architecture is big endian, byte swap the
     * little-endian block so it is in the native byte order.  Note
     * that "size_encoded" is the size in (8-bit) bytes, but
     * byte_swap_block16() needs the number of (16-bit) cgul_hchar
     * characters.  It is ok if "size_encoded" is odd. */
    if (cgul_endian__is_big_endian(cex)) {
        cgul_stream_reader__byte_swap_block16(cex,
                                              (cgul_hchar_t*)block_encoded,
                                              size_encoded >> 1);
    }

    /* Decode the block now that it is in the native byte order. */
    cgul_stream_reader__decode_utf16(cex,
        block_encoded, size_encoded, index_encoded,
        block_decoded, size_decoded, count_decoded);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/* Public callback function that implements cgul_stream_reader__decode_t. */
void
cgul_stream_reader__decode_utf32be(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(block_encoded);
    assert(index_encoded);
    assert(block_decoded);
    assert(count_decoded);

    /* If the machine architecture is little endian, byte swap the
     * big-endian block so it is in the native byte order.  Note that
     * "size_encoded" is the size in (8-bit) bytes, but
     * byte_swap_block32() needs the number of (32-bit) cgul_hchar
     * characters.  It is ok if "size_encoded" is odd. */
    if (cgul_endian__is_little_endian(cex)) {
        cgul_stream_reader__byte_swap_block32(cex,
                                              (cgul_wchar_t*)block_encoded,
                                              size_encoded >> 2);
    }

    /* Decode the block now that it is in the native byte order. */
    cgul_stream_reader__decode_utf32(cex,
        block_encoded, size_encoded, index_encoded,
        block_decoded, size_decoded, count_decoded);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/* Public callback function that implements cgul_stream_reader__decode_t. */
void
cgul_stream_reader__decode_utf32le(cgul_exception_t* cex,
                                   unsigned char* block_encoded,
                                   size_t size_encoded,
                                   size_t* index_encoded,
                                   cgul_wchar_t* block_decoded,
                                   size_t size_decoded,
                                   size_t* count_decoded)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* If the machine architecture is big endian, byte swap the
     * little-endian block so it is in the native byte order.  Note
     * that "size_encoded" is the size in (8-bit) bytes, but
     * byte_swap_block32() needs the number of (32-bit) cgul_hchar
     * characters.  It is ok if "size_encoded" is odd. */
    if (cgul_endian__is_big_endian(cex)) {
        cgul_stream_reader__byte_swap_block32(cex,
                                              (cgul_wchar_t*)block_encoded,
                                              size_encoded >> 2);
    }

    /* Decode the block now that it is in the native byte order. */
    cgul_stream_reader__decode_utf32(cex,
        block_encoded, size_encoded, index_encoded,
        block_decoded, size_decoded, count_decoded);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

cgul_stream_reader_t
cgul_stream_reader__new(cgul_exception_t* cex)
{
    cgul_stream_reader_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = (cgul_stream_reader_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* block_size */
    rv->block_size = CGUL_STREAM_READER__DEFAULT_BLOCK_SIZE;

    /* track_line_and_column */
    rv->track_line_and_column = 1;

    /* normalize_eol */
    rv->normalize_eol = 1;

    /* Initialize. */
    cgul_stream_reader__reinitialize(cex, rv);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_stream_reader__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_stream_reader_t
cgul_stream_reader__new_from_block_reader(cgul_exception_t* cex,
                                          cgul_block_reader_t br)
{
    cgul_stream_reader_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_stream_reader__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the block reader. */
    cgul_stream_reader__add_block_reader(cex, rv, br);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_stream_reader__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_stream_reader__delete(cgul_stream_reader_t sr)
{
    cgul_exception_t local = NULL;

    if (sr) {

        /* presentation_name */
        if (sr->presentation_name) {
            free(sr->presentation_name);
            sr->presentation_name = NULL;
        }

        /* block_decoded */
        if (sr->block_decoded) {
            free(sr->block_decoded);
        }

        /* block_encoded */
        if (sr->block_encoded) {
            free(sr->block_encoded);
        }

        /* block_readers */
        if (sr->block_readers) {
            cgul_stream_reader__clear_block_readers(&local, sr->block_readers);
            cgul_list__delete(sr->block_readers);
        }

        /* br -- We do not own. */
        
        /* Free the object itself. */
        free(sr);

    }
}

void
cgul_stream_reader__reset(cgul_exception_t* cex,
                          cgul_stream_reader_t sr)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Reinitialize. */
    cgul_stream_reader__reinitialize(cex, sr);

 out:
    return;
}

int
cgul_stream_reader__read(cgul_exception_t* cex,
                         cgul_stream_reader_t sr,
                         cgul_wchar_t* next_char)
{
    int rv = 0;
    int decode_rv = 0;
    cgul_wchar_t tmp = CGUL_WCHAR__NUL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Sanity check. */
    if (!sr->br) {
        *cex = cgul_exception__new("cgul_stream_reader__read: "
                                   "no block reader selected");
        goto out;
    }

    /* Assume failure. */
    if (next_char) {
        *next_char = CGUL_WCHAR__NUL;
    }

    /* Decode the next character. */
    CGUL_STREAM_READER__DECODE(cex, sr, tmp, decode_rv);
    if (*cex) {
        goto out;
    }
    if (!decode_rv) {
        /* EOF */
        sr->eof = 1;
        goto out;
    }

    /* EOL normalization is split across two reads.  This if()
     * statement handles the second read where the '\n' character (if
     * any) that is part of a "\r\n" EOL sequence is skipped. */
    if (sr->normalize_eol &&
        (sr->actual_previous_char == CGUL_STREAM_READER__CARRIAGE_RETURN) &&
        (tmp == CGUL_STREAM_READER__LINE_FEED))
    {
        /* Remember the actual character read. */
        sr->actual_previous_char = tmp;

        /* Read the next character thereby skipping the '\n' character. */
        CGUL_STREAM_READER__DECODE(cex, sr, tmp, decode_rv);
        if (*cex) {
            goto out;
        }
        if (!decode_rv) {
            /* EOF */
            sr->eof = 1;
            goto out;
        }
    }

    /* Remember the actual character read.  Do this before EOL
     * normalization which can change "tmp". */
    sr->actual_previous_char = tmp;

    /* EOL normalization is split across two reads.  This if()
     * statement handles the first read.  It simply converts any '\r'
     * character to '\n' if EOL normalization is enabled.  On the next
     * read, a '\n' character that follows (if any) is skipped. */
    if (sr->normalize_eol) {
        /* Normalize '\r' and "\r\n" to '\n'. */
        if (tmp == CGUL_STREAM_READER__CARRIAGE_RETURN) {
            /* Return '\n' instead of '\r'. */
            tmp = CGUL_STREAM_READER__LINE_FEED;
        }
    }

    /* Advance the line and column numbers. */
    if (sr->track_line_and_column) {
        CGUL_STREAM_READER__ADVANCE_LINE_AND_COLUMN_NUMBERS(cex, sr, tmp);
        if (*cex) {
            goto out;
        }
    }

    /* Return the result to the user. */
    if (next_char) {
        *next_char = tmp;
    }

    /* Success. */
    rv = 1;

 out:
    return rv;
}

int
cgul_stream_reader__eof(cgul_exception_t* cex,
                        cgul_stream_reader_t sr)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return whether EOF has been reached. */
    rv = sr->eof;

 out:
    return rv;
}

int
cgul_stream_reader__get_line_and_column_tracking(cgul_exception_t* cex,
                                                 cgul_stream_reader_t sr)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return whether line and column number tracking is enabled. */
    rv = sr->track_line_and_column;

 out:
    return rv;
}

void
cgul_stream_reader__set_line_and_column_tracking(cgul_exception_t* cex,
                                                 cgul_stream_reader_t sr,
                                                 int is_tracking_enabled)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Set whether line and column number tracking is enabled. */
    if (is_tracking_enabled) {
        sr->track_line_and_column = 1;
    } else {
        sr->track_line_and_column = 0;
    }

 out:
    return;
}

size_t
cgul_stream_reader__get_line_number_current(cgul_exception_t* cex,
                                            cgul_stream_reader_t sr)
{
    size_t rv = 0;

    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return the current line number. */
    rv = sr->line_current;

 out:
    return rv;
}

size_t
cgul_stream_reader__get_column_number_current(cgul_exception_t* cex,
                                              cgul_stream_reader_t sr)
{
    size_t rv = 0;

    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return the current column number. */
    rv = sr->column_current;

 out:
    return rv;
}

size_t
cgul_stream_reader__get_line_number_previous(cgul_exception_t* cex,
                                             cgul_stream_reader_t sr)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return the previous line number to the user. */
    rv = sr->line_previous;

 out:
    return rv;
}

size_t
cgul_stream_reader__get_column_number_previous(cgul_exception_t* cex,
                                               cgul_stream_reader_t sr)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return the previous column number to the user. */
    rv = sr->column_previous;

 out:
    return rv;
}

int
cgul_stream_reader__get_normalize_eol(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr)
{
    int rv = 0;

    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return whether EOLs are normalized. */
    rv = sr->normalize_eol;

 out:
    return rv;
}

void
cgul_stream_reader__set_normalize_eol(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr,
                                      int normalize)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Set whether to normalize EOLs. */
    if (normalize) {
        sr->normalize_eol = 1;
    } else {
        sr->normalize_eol = 0;
    }

 out:
    return;
}

void
cgul_stream_reader__add_block_reader(cgul_exception_t* cex,
                                     cgul_stream_reader_t sr,
                                     cgul_block_reader_t br)
{
    size_t block_reader_count = 0;
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* If a block reader has not yet been set, set it. */
    if (!sr->br) {
        sr->br = br;
        goto out;
    }

    /*
     * Otherwise, we need to enqueue the block reader so it can
     * eventually replace sr->br when all the other block readers
     * ahead of it in the queue have been exhausted.
     */

    /* Get the number of block readers. */
    block_reader_count = cgul_list__get_size(cex, sr->block_readers);

    /* Pre-allocate the "MULTIPART" presentation name if we are about
     * to have more than one block reader and the user has not
     * explicitly set the presentation name. */
    if ((block_reader_count == 0) && !sr->presentation_name) {
        tmp = cgul_libc__strdup(cex, "MULTIPART");
        if (*cex) {
            goto out;
        }
    }

    /* Enqueue the block reader. */
    cgul_list__push_back(cex, sr->block_readers, br);
    if (*cex) {
        goto out;
    }

    /* Check if we need to set the default "MULTIPART" presentation name. */
    if (tmp) {
        /* We checked above to make sure the presentation name is not set. */
        assert(!sr->presentation_name);
        /* Update the presentation name. */
        sr->presentation_name = tmp;
        /* Take ownership of tmp. */
        tmp = NULL;
    }

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}

size_t
cgul_stream_reader__get_block_size(cgul_exception_t* cex,
                                   cgul_stream_reader_t sr)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return the block size to the user. */
    rv = sr->block_size;

 out:
    return rv;
}

int
cgul_stream_reader__set_block_size(cgul_exception_t* cex,
                                   cgul_stream_reader_t sr,
                                   size_t block_size)
{
    int rv = 0;
    cgul_wchar_t* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Sanity check. */
    if (block_size == 0) {
        goto out;
    }

    /* Sanity check. */
    if (block_size == sr->block_size) {
        /* Success */
        rv = 1;
        goto out;
    }

    /* It is always safe to try to increase the block size, but it is
     * only safe to try to decrease the block size if a block reader
     * is not currently selected. */
    if ((block_size > sr->block_size) || !sr->br) {

        /* Allocate a new block for decoding.  We do not yet assign
         * the new decoding block to sr->block_decoded in case the
         * realloc() of the *encoding* block fails. */
        tmp = (cgul_wchar_t*)cgul_libc__calloc(cex, block_size, sizeof(*tmp));
        if (*cex) {
            goto out;
        }

        /* Try to realloc the encoded block. */
        sr->block_encoded = (unsigned char*)
                            cgul_libc__realloc(
                                cex,
                                sr->block_encoded,
                                block_size * sizeof(*sr->block_encoded));
        if (*cex) {
            goto out;
        }

        /* Assign the new decoding block and take ownership. */
        if (sr->block_decoded) {
            free(sr->block_decoded);
        }
        sr->block_decoded = tmp;
        tmp = NULL;

        /* Remember the new block size. */
        sr->block_size = block_size;

        /* Success */
        rv = 1;
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return rv;
}

cgul_stream_reader__decode_t
cgul_stream_reader__get_decoder(cgul_exception_t* cex,
                                cgul_stream_reader_t sr)
{
    cgul_stream_reader__decode_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Return the decode callback to the user. */
    rv = sr->decode;

 out:
    return rv;
}

void
cgul_stream_reader__set_decoder(cgul_exception_t* cex,
                                cgul_stream_reader_t sr,
                                cgul_stream_reader__decode_t decoder)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Set the decode callback. */
    if (sr->decode) {
        sr->decode = decoder;
    }

 out:
    return;
}

const char*
cgul_stream_reader__get_presentation_name(cgul_exception_t* cex,
                                          cgul_stream_reader_t sr)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* An explicitly set presentation name takes precedence.  Note
     * that this includes the case where more than one block reader
     * has been set and we are returning the default "MULTIPART"
     * presentation name. */
    if (sr->presentation_name) {
        /* Success. */
        rv = sr->presentation_name;
        goto out;
    }

    /* Otherwise, if there is only one block reader, use its
     * presentation name.  We do not need to explicitly check the size
     * of sr->block_readers for a multipart stream because those are
     * handled above.  All that is left now is either one or zero
     * block readers. */
    if (sr->br) {
        /* Success. */
        rv = cgul_block_reader__get_presentation_name(cex, sr->br);
        goto out;
    }

    /* Error. */
    *cex = cgul_exception__new("cgul_stream_reader__get_presentation_name: "
                               "no input source selected");
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

void
cgul_stream_reader__set_presentation_name(cgul_exception_t* cex,
                                          cgul_stream_reader_t sr,
                                          const char* presentation_name)
{
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Sanity check. */
    if (!presentation_name) {
        goto out;
    }

    /* Pre-allocate the new presentation name. */
    tmp = cgul_libc__strdup(cex, presentation_name);
    if (*cex) {
        goto out;
    }

    /* Free the old presentation name (if any). */
    if (sr->presentation_name) {
        free(sr->presentation_name);
        sr->presentation_name = NULL;
    }

    /* Set the new presentation name. */
    sr->presentation_name = tmp;

    /* Take ownership of "tmp". */
    tmp = NULL;

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}
