#ifndef CGUL_TRIE_NODE_P_H
#define CGUL_TRIE_NODE_P_H

#include "cgul_common.h"
#include "cgul_trie_node.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_trie_node</code> class.
 */
struct cgul_trie_node {

    /**
     * <code>ch</code> is the character for this node.  Its main
     * purpose is to control tree traversal.  It is one of the
     * characters in the <code>key</code> string.  It's exact value
     * depends on where in the trie the node was inserted. <p>
     *
     * @note <code>ch</code> is ignored for the root node because the
     * root node maps to an empty string. <p>
     */
    char ch;

    /**
     * Pointer to parent node.  <code>parent</code> is
     * <code>NULL</code> for the root node.
     */
    cgul_trie_node_t parent;

    /** Pointer to left node. */
    cgul_trie_node_t left;

    /** Pointer to middle node. */
    cgul_trie_node_t middle;

    /** Pointer to right node. */
    cgul_trie_node_t right;

    /**
     * Pointer to older node.  Together with <code>younger</code>,
     * <code>older</code> forms a doubly linked list that has nothing
     * to do with sort order; rather, the <code>older</code> node
     * points to the next older node.  If there is no older node,
     * <code>older</code> is <code>NULL</code>.  By calling
     * <code>cgul_trie_node__get_older()</code>, the user can
     * iterate over the nodes in reverse chronological order. <p>
     *
     * If this node is not a terminating character, <code>older</code>
     * will be <code>NULL</code> because there is no key associated
     * with this node so it doesn't make sense to try to point to an
     * older node. <p>
     */
    cgul_trie_node_t older;

    /**
     * Pointer to younger node.  Together with <code>older</code>,
     * <code>younger</code> forms a doubly linked list that has
     * nothing to do with sort order; rather, the <code>younger</code>
     * node points to the next younger node.  If there is no younger
     * node, <code>younger</code> is <code>NULL</code>.  By calling
     * <code>cgul_trie_node__get_younger()</code>, the user can
     * iterate over the nodes in chronological order. <p>
     *
     * If this node is not a terminating character,
     * <code>younger</code> will be <code>NULL</code> because there is
     * no key associated with this node so it doesn't make sense to
     * try to point to a younger node. <p>
     */
    cgul_trie_node_t younger;

    /**
     * Key used to lookup the value.  If <code>key</code> is NULL, it
     * means <code>ch</code> is not a terminating character for any
     * string that has been inserted into the trie.
     */
    const char* key;

    /**
     * Value returned by a lookup based on a key.
     */
    void* value;

};

/*
 * Private Methods
 */

/**
 * Create a new <code>cgul_trie_node</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_trie_node__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_trie_node</code> instance
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the <code>key</code> or <code>value</code> pairs.  The
 * value for <code>n</code> should have been instantiated by
 * <code>cgul_trie_node__new()</code>.  Do not try to access
 * <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_trie_node</code> instance
 */
extern CGUL_EXPORT void
cgul_trie_node__delete(cgul_trie_node_t n);

/**
 * This method is shared between <code>cgul_trie_node__new()</code>
 * and the <code>restructor</code> for the <code>cgul_cache()</code>.
 * It initializes new nodes and re-initializes existing nodes.
 *
 * @param[in] n <code>cgul_trie_node</code> instance to initialize
 */
extern CGUL_EXPORT void
cgul_trie_node__initialize(cgul_trie_node_t n);

/**
 * Set the key used to lookup this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_trie_node</code> instance
 * @param[in] key key
 */
extern CGUL_EXPORT void
cgul_trie_node__set_key(cgul_exception_t* cex,
                        cgul_trie_node_t n,
                        const char* key);

/**
 * Return the smallest node in the subtree rooted at <code>n</code>.
 * This function returns <code>NULL</code> if the subtree is
 * empty. <p>
 *
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return the smallest node in the subtree rooted at <code>n</code>
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__get_smallest_node_in_subtree(cgul_exception_t* cex,
                                             cgul_trie_node_t n);

/**
 * Return the largest node in the subtree rooted at <code>n</code>.
 * This function returns <code>NULL</code> if the subtree is
 * empty. <p>
 *
 * @param[in] n <code>cgul_trie_node</code> instance
 * @return the largest node in the subtree rooted at <code>n</code>
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie_node__get_largest_node_in_subtree(cgul_exception_t* cex,
                                            cgul_trie_node_t n);

CGUL_END_C

#endif  /*  CGUL_TRIE_NODE_P_H  */
