#include "cgul_varsub.h"

/**
 * This structure defines the client data that gets passed into
 * <code>cgul_varsub_formatted__lookup_cb()</code>.  It is used to
 * remember the data passed into <code>cgul_varsub_formatted()</code>.
 */
typedef struct cgul_varsub_formatted__lookup_cb_data {

    /** Destination string. */
    cgul_string_t dst;

    /** Client callback. */
    cgul_varsub_formatted__lookup_t f;

    /** Client data. */
    void* data;

} *cgul_varsub_formatted__lookup_cb_data_t;

/**
 * This function is the local callback
 * <code>cgul_varsub_formatted()</code> registers with
 * <code>cgul_varsub()</code>.  The idea is to let
 * <code>cgul_varsub()</code> parse the source string and extract all
 * the variable names including variable names like "${foo:%32s}" that
 * contain embedded printf()-style conversion specifications. <p>
 *
 * <code>cgul_varsub()</code> will then feed the entire variable name
 * to this function as <code>varname</code>.  It is the job of this
 * function to split <code>varname</code> into the real variable name
 * and the optional printf()-style conversion specification. <p>
 *
 * After splitting <code>varname</code>, this function than calls the
 * client's callback function with the real variable name and optional
 * printf()-style conversion specification. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] varname variable name
 * @param[in,out] dst destination string
 * @param[in] cb_data callback data
 * @return whether the lookup was successful
 */
static int
cgul_varsub_formatted__lookup_cb(
    cgul_exception_t* cex,
    const char* varname,
    cgul_string_t dst,
    cgul_varsub_formatted__lookup_cb_data_t cb_data)
{
    char c_saved = '\0';
    int i_saved = -1;
    int i = 0;
    int is_cs_ok = 0;
    int rv = 0;
    unsigned long int format_index = 0;
    const char* format = NULL;
    cgul_convspec_t cs = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the convspec object. */
    cs = cgul_convspec__new(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Split varname.
     *
     * This is done in-place for efficiency.  Note that "varname" can
     * potentially be used when this function returns, so it must be
     * "unsplit" before returning.
     */
    for (i = 0 ; varname[i] != '\0' ; ++i) {
        if (varname[i] == ':') {
            /* Remember the original index and character. */
            c_saved = varname[i];
            i_saved = i;
            /* Split. */
            ((char*)varname)[i] = '\0';
            /* Remember where the format string supposedly starts. */
            format = &varname[i+1];
            break;
        }
    }

    /*
     * Validate the format specification (to the extent possible).
     */

    /* If no format string, skip the validation. */
    if (format == NULL) {
        goto callback;
    }

    /* Parse the format string. */
    format_index = 0;
    cgul_convspec__deserialize(cex, cs, format, &format_index);
    if (*cex) {
        /* Ignore the exception. */
        cgul_exception__delete(*cex);
        *cex = NULL;
        goto callback;
    }

    /* All characters in the format string should be consumed. */
    if (format[format_index] != '\0') {
        goto callback;
    }

    /* Avoid security issues related to trying to dereferencing
     * stack variables that we don't have. */
    if (   cgul_convspec__get_is_field_width_on_stack(cex, cs)
        || cgul_convspec__get_is_precision_on_stack(cex, cs))
    {
        goto callback;
    }

    /* The conversion specification is ok. */
    is_cs_ok = 1;

 callback:

    /* Call the client's callback so it can perform the substitution. */
    rv = (*cb_data->f)(cex,
                       varname,
                       (is_cs_ok ? cs : NULL),
                       dst,
                       cb_data->data);
    if (*cex) {
        goto out;
    }

 out:

    /* Unsplit varname. */
    if (i_saved >= 0) {
        ((char*)varname)[i_saved] = c_saved;
    }

    /* Clean up. */
    if (cs) {
        cgul_convspec__delete(cs);
    }

    return rv;
}

void
cgul_varsub(cgul_exception_t* cex,
            const char* src,
            cgul_string_t dst,
            cgul_varsub__lookup_t f,
            void* data)
{
    int brace_level = 0;
    size_t i = 0;
    const char* varname_c = NULL;
    cgul_string_t varname = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(src);
    assert(dst);
    assert(f);

    /* Allocate varname. */
    varname = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over src checking for end-of-string (EOS). */
    for (i = 0 ; src[i] != '\0' ; ++i) {

        /* Just copy characters directly to dst until we hit the next
         * variable name being mindful of escape sequences. */
        if (src[i] != '$') {

            /* Check for an escape sequence before appending. */
            if (src[i] == '\\') {

                /* If not EOS, move to the next character so it gets
                 * appended verbatim; otherwise, the backslash will be
                 * appended verbatim. */
                if (src[i+1] != '\0') {
                    ++i;
                }

                /* Append the escaped character being mindful of the
                 * conventional escape sequences like '\t' and '\n' */
                switch (src[i]) {
                case 'a':
                    cgul_string__append_char(cex, dst, '\a');
                    break;
                case 'b':
                    cgul_string__append_char(cex, dst, '\b');
                    break;
                case 'f':
                    cgul_string__append_char(cex, dst, '\f');
                    break;
                case 'n':
                    cgul_string__append_char(cex, dst, '\n');
                    break;
                case 'r':
                    cgul_string__append_char(cex, dst, '\r');
                    break;
                case 't':
                    cgul_string__append_char(cex, dst, '\t');
                    break;
                case 'v':
                    cgul_string__append_char(cex, dst, '\v');
                    break;
                default:
                    /* It is not a conventional escape sequence so
                     * just append it verbatim. */
                    cgul_string__append_char(cex, dst, src[i]);
                    break;
                }

                /* Check for exceptions. */
                if (*cex) {
                    goto out;
                }

                continue;
            }

            /* The character wasn't escaped.  Append it. */
            cgul_string__append_char(cex, dst, src[i]);
            continue;
        }

        /* If EOS, just append the '$' character; otherwise, skip
         * passed the '$' character. */
        if (src[i+1] == '\0') {
            cgul_string__append_char(cex, dst, src[i]);
            goto out;
        } else {
            /* Skip passed the '$' character. */
            ++i;
        }

        /* Remember if braces are being used.  If so, move to the next
         * character. */
        if (src[i] == '{') {
            brace_level = 1;
            ++i;
        } else {
            brace_level = 0;
        }

        /* Iterate over the variable name.  The point of this inner
         * loop is to build up "varname". */
        for ( ; src[i] != '\0' ; ++i) {

            /* Check if braces are being used. */
            if (brace_level > 0) {

                /* Break if the end of varname is reached. */
                if ((brace_level == 1) && (src[i] == '}')) {
                    break;
                }

                /* Check for an escape sequence. */
                if (src[i] == '\\') {

                    /* If not EOS, move to the next character so it
                     * gets appended verbatim; otherwise, the
                     * backslash will be appended verbatim. */
                    if (src[i+1] != '\0') {
                        ++i;
                    }

                    /* Append the escaped character being mindful of the
                     * conventional escape sequences like '\t' and '\n' */
                    switch (src[i]) {
                    case 'a':
                        cgul_string__append_char(cex, varname, '\a');
                        break;
                    case 'b':
                        cgul_string__append_char(cex, varname, '\b');
                        break;
                    case 'f':
                        cgul_string__append_char(cex, varname, '\f');
                        break;
                    case 'n':
                        cgul_string__append_char(cex, varname, '\n');
                        break;
                    case 'r':
                        cgul_string__append_char(cex, varname, '\r');
                        break;
                    case 't':
                        cgul_string__append_char(cex, varname, '\t');
                        break;
                    case 'v':
                        cgul_string__append_char(cex, varname, '\v');
                        break;
                    default:
                        /* It is not a conventional escape sequence so
                         * just append it verbatim. */
                        cgul_string__append_char(cex, varname, src[i]);
                        break;
                    }

                    /* Check for exceptions. */
                    if (*cex) {
                        goto out;
                    }

                    continue;
                }

                /* Check if the brace level is increasing. */
                if (src[i] == '{') {
                    ++brace_level;
                }

                /* Check if the brace level is decreasing. */
                if (src[i] == '}') {
                    --brace_level;
                }

                /* Append to varname. */
                cgul_string__append_char(cex, varname, src[i]);
                if (*cex) {
                    goto out;
                }

                continue;
            }

            /* Braces are not being used so the end of varname is
             * indicated by the first non-alphanumeric, non-underscore
             * character. */
            if (!isalnum((unsigned char)src[i]) && (src[i] != '_')) {
                /* The end of varname has been reached.  Move back one
                 * character to avoid consuming it. */
                --i;
                break;
            }

            /* Append the alphanumeric character to varname. */
            cgul_string__append_char(cex, varname, src[i]);
            if (*cex) {
                goto out;
            }
        }

        /*
         * NOTE: Any code after this point must fall through to the
         * end of the loop so that "varname" gets cleared before the
         * next iteration.
         */

        /* Check for EOS. */
        if (src[i] == '\0') {
            /* Check if braces are being used because reaching EOS
             * before reaching the closing brace means that "varname"
             * is not complete. */
            if (brace_level > 0) {
                /* Do not try to perform variable substitution.
                 * Instead, just append as much of the variable name
                 * as we were able to read.  (Because we are using
                 * braces, we know that "varname" should be preceded
                 * by "${" when appending it to dst.) */
                varname_c = cgul_string__get_value(cex, varname);
                cgul_string__append(cex, dst, "${");
                cgul_string__append(cex, dst, varname_c);
                goto out;
            }
            /* Move back one character to avoid consuming EOS. */
            --i;
        }

        /* Convert "varname" to a C-style string for convenience. */
        varname_c = cgul_string__get_value(cex, varname);

        /* Call the callback "f" to perform the lookup and substitution. */
        if (!f(cex, varname_c, dst, data)) {
            /* The lookup was not successful so just copy the variable
             * name directly to dst instead.  Don't forget to add
             * braces if necessary. */
            cgul_string__append_char(cex, dst, '$');
            if (brace_level > 0) {
                cgul_string__append_char(cex, dst, '{');
            }
            cgul_string__append(cex, dst, varname_c);
            if (brace_level > 0) {
                cgul_string__append_char(cex, dst, '}');
            }
        }

        /* Reset varname before the next iteration. */
        cgul_string__set_value(cex, varname, "");

        /* Check for exceptions. */
        if (*cex) {
            goto out;
        }
    }

 out:

    /* Clean up. */
    if (varname) {
        cgul_string__delete(varname);
    }

    return;
}

extern void
cgul_varsub_formatted(cgul_exception_t* cex,
                      const char* src,
                      cgul_string_t dst,
                      cgul_varsub_formatted__lookup_t f,
                      void* data)
{
    struct cgul_varsub_formatted__lookup_cb_data cb_data;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Remember the client's parameters. */
    cb_data.dst = dst;
    cb_data.f = f;
    cb_data.data = data;

    /* Call cgul_varsub() so it can find the literal variable names,
     * but register cgul_varsub_formatted__lookup_cb() as the callback
     * for cgul_varsub() so it can split the literal variable names
     * into the real variable names and optional printf()-style
     * conversion specifications.  cgul_varsub_formatted__lookup_cb()
     * will then pass the real variable names and options
     * printf()-style conversion specifications to "f". */
    cgul_varsub(cex,
                src,
                dst,
                (cgul_varsub__lookup_t)cgul_varsub_formatted__lookup_cb,
                &cb_data);

 out:
    return;
}
