/*
 * This program reads binary data from stdin, converts it to base64,
 * and prints the resulting conversion to stdout.
 */

#include "cgul.h"

#ifdef WIN32
#include <fcntl.h>
#include <io.h>
#endif

int main(int argc, char* argv[])
{
    int rv = 0;
    size_t rcount = (size_t)-1;
    cgul_exception_t cex = NULL;
    cgul_base64_t b64 = NULL;
    char* wbuf = NULL;
    char rbuf[4096];

#ifdef WIN32
    /* Convert stdin and stdout to binary mode. */
    if (_setmode(_fileno(stdin), _O_BINARY) < 0) {
        cex = cgul_exception__new("_setmode(stdin): %m");
        goto out;
    }
    if (_setmode(_fileno(stdout), _O_BINARY) < 0) {
        cex = cgul_exception__new("_setmode(stdout): %m");
        goto out;
    }
#endif

    /* Get the base64 object ready. */
    b64 = cgul_base64__new(&cex);
    cgul_base64__set_line_length_max(&cex, b64, 76);
    if (cex) {
        goto out;
    }

    while (cex == NULL) {

        /* Read a block from stdin. */
        rcount = cgul_libc__fread(&cex, rbuf, 1, sizeof(rbuf), stdin);
        if (cex) {
            goto out;
        }

        /* EOF */
        if (rcount == 0) {
            break;
        }

        /* Base64 encode the block. */
        cgul_base64__encode(&cex, b64, rbuf, rcount);
        if (cex) {
            goto out;
        }

        /* Print the encoding as it exists so far. */
        wbuf = cgul_base64__get_encode_conversion(&cex, b64, NULL);
        cgul_libc__printf(&cex, "%s", wbuf);
        if (cex) {
            goto out;
        }
    }

    /* Flush and print any remaining bytes. */
    cgul_base64__encode_finish(&cex, b64);
    wbuf = cgul_base64__get_encode_conversion(&cex, b64, NULL);
    cgul_libc__printf(&cex, "%s\n", wbuf);
    if (cex) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (cex) {
        const char* msg = cgul_exception__get_message(cex);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(cex);
        cex = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (b64) {
        cgul_base64__delete(b64);
    }

    return rv;
}
