/**
 * This program uses <code>cgul_csv</code> to convert a CSV file to
 * <code>cgul_stanza</code> format.  It assumes the first row of the
 * CSV file contains the column headers.  It then produces one stanza
 * per row where the column headers map to the stanza keys and the
 * column values map to stanza values. <p>
 *
 * One use for this program is to convert CSV files into a format that
 * is easy to diff. <p>
 *
 * Use the <code>main_stanza_to_csv</code> program to convert from
 * <code>cgul_stanza</code> format back into CSV format. <p>
 *
 * @file
 * @author Paul Serice
 * @brief csv to stanza converter
 */

#include <assert.h>
#include "cgul.h"

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned long int i = 0;
    unsigned long int headers_count = 0;
    unsigned long int fields_count = 0;
    unsigned long int data_row_count = 0;
    const char* header = NULL;
    const char* value = NULL;
    cgul_csv_t csv = NULL;
    cgul_vector_t headers = NULL;
    cgul_exception_t local = NULL;

    /* Verify command line. */
    if (argc != 2) {
        fprintf(stderr, "Usage: main_csv_to_stanza <file.csv>\n");
        goto out;
    }

    /* Allocate the "headers" vector. */
    headers = cgul_vector__new(&local);
    if (local) {
        goto out;
    }

    /* Open the cvs file. */
    csv = cgul_csv__new_from_fname(&local, argv[1], 1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Read the headers. */
    cgul_csv__load_next(&local, csv);
    if(local) {
        goto out;
    }

    /* Remember the headers and their exact order so that the order of
     * the resulting stanza matches the order of the columns in the
     * csv file */
    headers_count = cgul_csv__get_number_of_fields(&local, csv);
    for (i = 0 ; i < headers_count ; ++i) {
        header = cgul_csv__take_value(&local, csv, i);
        cgul_vector__push_back(&local, headers, header);
        if (local) {
            goto out;
        }
    }

    /* Iterate over each data row. */
    while (cgul_csv__load_next(&local, csv)) {

        /* Check for exceptions. */
        if (local) {
            goto out;
        }

        /* Increment the count of data rows. */
        ++data_row_count;

        /* Make sure this row has the expected number of columns. */
        fields_count = cgul_csv__get_number_of_fields(&local, csv);
        if (fields_count != headers_count) {
            local = cgul_exception__new(
                        "Data Row %lu: Expecting %lu columns but found %lu.",
                        data_row_count,
                        headers_count,
                        fields_count);
            goto out;
        }

        /* Print the stanza separator if necessary. */
        if (data_row_count > 1) {
            cgul_libc__puts(&local, "");
            if (local) {
                goto out;
            }
        }

        /* Iterate over each column in the row. */
        for (i = 0 ; i < headers_count ; ++i) {

            /* Get the header for this column. */
            header = (const char*)cgul_vector__at(&local, headers, i);
            if (local) {
                goto out;
            }

            /* Get the value for this column. */
            value = (const char*)cgul_csv__lookup(&local, csv, i);
            if (local) {
                goto out;
            }

            /* Print the stanza line which could be multi-line text. */
            cgul_stanza__print(&local, stdout, header, value);
            if (local) {
                goto out;
            }
        }
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /*
     * Clean up.
     */

    /* headers */
    if (headers) {
        /* Free each individual header. */
        for (;;) {
            header = (const char*)cgul_vector__pop_back(&local, headers);
            if (!header) {
                break;
            }
            free((void*)header);
        }
        /* Free the "headers" vector. */
        cgul_vector__delete(headers);
    }

    /* csv */
    if (csv) {
        cgul_csv__delete(csv);
    }

    return rv;
}
