/*
 * General:
 * =======
 *
 * This program demonstrates how to add Managed .NET classes as event
 * listeners for cgul_event_router_cxx.
 *
 *
 * Problem:
 * =======
 *
 * The main thing you need to know is that you cannot directly convert
 * a .NET tracking handle to a "void" pointer.  This means it is not
 * directly possible to pass "this" into
 * cgul_event_router_cxx::add_listner().
 *
 * The main reason for this seems to be that tracking handles are
 * capable of tracking where the object has been moved when the
 * garbage collector defragments memory.  This is obviously something
 * that a "void" pointer cannot do.
 *
 * One alternative to the approach presented here is to "pin" the
 * object in memory.  This prevents the garbage collector from moving
 * the the object in memory.  The problem with pinning is that it only
 * works for primitive types and arrays.  It is not a general
 * solution.
 *
 *
 * General Solution:
 * ================
 *
 * Instead, you use a GCHandle.  The primary function of a GCHandle is
 * to increase the reference count for a Managed object so that the
 * garbage collector will not finialize it when the only reference to
 * the object is in Native code.
 *
 * So obviously it is possible for Native code to refer to a Managed
 * object, but how?  It seems GCHandle is the solution here too.  You
 * can use the following method to convert "this" into a magic cookie
 * that fits exactly in a "void" pointer that you can pass into your
 * Native code:
 *
 *     MyObj^ obj = gcnew MyObj();
 *     GCHandle gch = GCHandle::Alloc(obj);
 *     void* data = GCHandle::ToIntPtr(gch).ToPointer();
 *
 * Your Native code can then pass the "void" pointer back to the
 * Managed code which can then pass the magic cookie to the following
 * method to get the tracking handle:
 *
 *     GCHandle gch = GCHandle::FromIntPtr(IntPtr((void*)data));
 *     MyObj^ obj = (MyObj^)gch.Target;
 *
 * Don't forget to call Free() on your GCHandle.  These things appear
 * to be owned by the garbage collector and do not appear to have the
 * same scope as the GCHandle variable.  Furthermore, an exception is
 * raised if you call Free() on an unallocated GCHandle variable.  So,
 * you have to carefully check before calling Free():
 *
 *      if (gch.IsAllocated) {
 *          gch.Free();
 *      }
 *
 * The code below shows how to wrap all this into a single class such
 * that virtual methods for individual objects are called whenever the
 * Native cgul_event_router_cxx sends an event.
 */

#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace System;
using namespace System::Runtime::InteropServices;

/* See comments at the top of the file. */
public ref class ManagedClass {

    /* Constructor */
    public: ManagedClass()
    {
        /* Create a GCHandle for "this". */
        gch = GCHandle::Alloc(this);

        /* Convert "this" to a magic cookie that maps to a "void" pointer. */
        const void* clientData = GCHandle::ToIntPtr(gch).ToPointer();

        /* Get the event router. */
        cgul_event_router_cxx* er = cgul_event_router_cxx::get_instance();

        /* Register for the "print" event using the magic cookie. */
        er->add_listener("print", PrintMessageCB, clientData);
    }

    /* Destructor */
    public: ~ManagedClass() {
        CleanUp();
    }

    /* Finializer */
    public: !ManagedClass() {
        CleanUp();
    }

    /* Common clean up code. */
    private: void CleanUp() {
        /* Free the GCHandle if it has been allocated. */
        if (gch.IsAllocated) {
            gch.Free();
        }
    }

    /* Virtual method that is _indirectly_ sent the "print" event. */
    protected: virtual void PrintMessage(const char* rendezvous,
                                         const void* event)
    {
        String^ message = gcnew String((const char*)event);
        Console::WriteLine(message);
    }

    /* Static method that is _directly_ sent the "print" event. */
    private: static int PrintMessageCB(const char* rendezvous,
                                       const void* event,
                                       const void* data)
    {
        GCHandle gch = GCHandle::FromIntPtr(IntPtr((void*)data));
        ManagedClass^ obj = (ManagedClass^)gch.Target;
        obj->PrintMessage(rendezvous, event);
        return 1;
    }

    /* GCHandle to convert "this" to and from a "void" pointer. */
    private: GCHandle gch;
};

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_event_router_cxx* er = NULL;

    try {

        /* Get the event router. */
        er = cgul_event_router_cxx::get_instance();

        /* Create a new ManagedClass instance. */
        ManagedClass^ mc = gcnew ManagedClass();

        /* Send "print" events. */
        er->send_event("print", "Hello, World!");
        er->send_event("print", "Goodbye, Cruel World!");

    } catch (Exception^ e) {
        Console::Error->WriteLine("*** Error: " + e);
        rv = 1;
    }

    return rv;
}
