/**
 * Test the hash function and comparision function when using custom
 * keys in a <code>cgul_hash_cxx</code> object.
 *
 * @file
 * @brief test program for custom hash keys
 */

#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * Data structure to use as custom keys.
 */
typedef struct point {
    int x;
    int y;
} *point_t;

/**
 * Hash function.  This is an adaptor for the
 * <code>cgul_hash__hash_block()</code> function.  The purpose of this
 * function is to provide the size of the block
 * <code>cgul_hash__hash_block()</code> can hash the block.
 *
 * @param[in] key key
 * @param[in] data client data
 * @return hash of <code>key</code>
 */
static unsigned int
hf(void* key,
   void* data)
{
    return cgul_hash__hash_block(key, sizeof(point_t*));
}

/**
 * Comparison function for <code>point_t</code> objects.  This is
 * needed so the hash can find a particular key when the hash for two
 * different keys collide.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with <code>key2</code>
 */
static int
cf(const void* key1,
   const void* key2)
{
    int rv = 0;
    point_t p1 = (point_t)key1;
    point_t p2 = (point_t)key2;
    if (p1->x < p2->x) {
        rv = -1;
    } else if (p1->x > p2->x) {
        rv = 1;
    } else if (p1->y < p2->y) {
        rv = -1;
    } else if (p1->y > p2->y) {
        rv = 1;
    }
    return rv;
}

/**
 * Print a point to <code>f</code>.
 *
 * @param[in] p point
 */
static void
print_point(point_t p)
{
    if (p) {
        printf("(%d, %d)\n", p->x, p->y);
    }
}

/**
 * This function loads keys into the hash <code>h</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] h hash
 */
static void
load_hash(cgul_hash_cxx& h)
{
    int i = 0;
    int j = 0;
    int insert_rv = 0;
    point_t p = NULL;
    cgul_exception_t local = NULL;

    /* Insert points into h. */
    for (i = 0 ; i < 5 ; ++i) {
        for (j = 0 ; j < 5 ; ++j) {

            /* Allocate a new point. */
            p = (point_t)cgul_libc__calloc(&local, 1, sizeof(*p));
            if (local) {
                throw cgul_exception_cxx(local);
            }

            /* Initialize the point. */
            p->x = 5 - i;
            p->y = 5 - j;

            /* Insert the point as a key in h. */
            try {
                insert_rv = h.insert((void*)p, NULL);
            } catch (...) {
                /* Don't leak memory. */
                free(p);
                throw;
            }

            /* Free the point if it is a duplicate of one already in
             * the hash. */
            if (!insert_rv) {
                free(p);
            }

            /* Either way, we no longer own p. */
            p = NULL;
        }
    }

}

/**
 * This function prints the keys the hash <code>h</code> in insertion
 * order.
 *
 * @param[in] h hash
 */
static void
print_in_insertion_order(cgul_hash_cxx& h)
{
    cgul_hash_node_cxx* n = h.get_oldest();
    for ( ; n ; n = n->get_younger()) {
        print_point((point_t)n->get_key());
    }
}

/**
 * This function prints the keys the hash <code>h</code> in sort order
 * from least to greatest.
 *
 * @param[in] h hash
 */
static void
print_in_sort_order(cgul_hash_cxx& h)
{
    size_t i = 0;
    point_t* v = NULL;

    /* Get an array of keys. */
    v = (point_t*)h.get_sorted_keys(NULL);

    /* Iterate over the array. */
    for (i = 0 ; v[i] ; ++i) {
        print_point((point_t)v[i]);
    }

    /* Clean up. */
    if (v) {
        free(v);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    point_t p = NULL;
    cgul_hash_cxx h(hf, NULL, cf);

    try {

        /* Load custom keys into the hash. */
        load_hash(h);

        /* Print the points in insertion order. */
        print_in_insertion_order(h);

        puts("========");

        /* Print the points in sort order from least to greatest. */
        print_in_sort_order(h);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }
    
    /* Free each point in the hash. */
    while (h.remove_back((void**)&p, NULL)) {
        free(p);
    }

    return rv;
}
