/**
 * This program is the same as "main_hash_list_new" except it
 * registers custom hash and comparison function callbacks instead of
 * using the defaults.
 *
 * This program can be used to test the following method:
 *
 *     <code>cgul_hash__hash_block()</code>
 */

#include <assert.h>
#include "cgul.h"

/**
 * This is the custom hash function.  It comes in two flavors
 * depending on what you are trying to accomplish.
 *
 * @param[in] key key
 * @param[in] data client data
 * @return hash of <code>key</code>
 */
unsigned int
hash_string(void* key,
            void* data)
{
#if 0

    /*
     * We calculate the hash.
     */

    unsigned char c;
    unsigned int rv = 2166136261;
    unsigned char* tmp = (unsigned char*)key;
    for (c = *tmp ; c ; c = *++tmp) {
        rv = rv * 31 + c;
    }
    return rv;

#else

    /*
     *  cgul_hash__hash_block() calculates the hash.
     */

    return cgul_hash__hash_block(key,
                                 strlen((const char*)key));

#endif
}

/**
 * This method frees the keys in the hash <code>h</code>.  It removes
 * each node from <code>h</code> and calls <code>free</code> on the
 * key and value associated with the node but only if they are not
 * <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] h hash
 */
static void
free_keys_and_values(cgul_exception_t* cex,
                     cgul_hash_t h)
{
    void* key = NULL;
    void* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Remove each node. */
    while (cgul_hash__remove_youngest(cex, h, &key, &value)) {
        /* Free the key. */
        if (key) {
            free(key);
        }
        /* Free the value. */
        if (value) {
            free(value);
        }
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    char* progname = NULL;
    const char* line = NULL;
    cgul_string_t s = NULL;
    cgul_crlf_file_t crlf = NULL;
    cgul_hash_t entries = NULL;
    cgul_exception_t local = NULL;

    /* Get the program name. */
    progname = cgul_libc__basename(&local, argv[0]);
    if (local) {
        goto out;
    }

    /* Verify the command line. */
    if (argc != 3) {
        fprintf(stderr, "Usage: %s <old> <new>\n", progname);
        rv = 1;
        goto out;
    }

    /* Allocate entries. */
    entries = cgul_hash__new_custom(
                  &local,
                  hash_string,
                  NULL,
                  (cgul_hash__compare_t)strcmp);
    if (local) {
        goto out;
    }

#if 0
    cgul_hash__set_load_factor(&local, entries, 0.75);
#endif

    /* Allocate s. */
    s = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Open the "old" file. */
    crlf = cgul_crlf_file__new_from_fname(&local,
                                          argv[1],
                                          1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Add each line in the "old" file to the "entries" tree. */
    while ((line = cgul_crlf_file__get_line(&local, crlf)) != NULL) {
        /* Copy the line assuming it will be inserted into "entries." */
        cgul_string__set_value(&local, s, line);
        /* Try to insert the copy of "line". */
        if (cgul_hash__insert(&local,
                              entries,
                              cgul_string__get_value(&local, s),
                              NULL)) {
            /* "entries" now owns the string. */
            cgul_string__take_value(&local, s);
        }
        if (local) {
            goto out;
        }
    }

    /* Close the "old" file. */
    cgul_crlf_file__delete(crlf);
    crlf = NULL;

    /* Open the "new" file. */
    crlf = cgul_crlf_file__new_from_fname(&local,
                                          argv[2],
                                          1 /* is_block_buffered */);
    if (local) {
        goto out;
    }

    /* Print each line in the "new" file that is not present in the
     * "old" file. */
    while ((line = cgul_crlf_file__get_line(&local, crlf)) != NULL) {
        if (local) {
            goto out;
        }
        if (!cgul_hash__find(&local, entries, line)) {
            puts(line);
        }
    }

    /* Close the "new" file. */
    cgul_crlf_file__delete(crlf);
    crlf = NULL;

#if 1
    {
        char* s = cgul_hash__get_stats(&local, entries);
        if (local) {
            goto out;
        }
        fprintf(stderr, "%s\n", s);
        free(s);
    }
#endif

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

#if 1
    /* Clean up. */
    if (crlf) {
        cgul_crlf_file__delete(crlf);
    }
    if (entries) {
        free_keys_and_values(&local, entries);
        cgul_hash__delete(entries);
    }
    if (s) {
        cgul_string__delete(s);
    }
    if (progname) {
        free(progname);
    }
#endif

    return rv;
}
