#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * This function prints the value stored in the node <code>n</code>.
 *
 * @param[in] t tree
 * @param[in] n node
 * @param[in] data client data
 */
static int
print_pair(cgul_rbtree_cxx* t,
           cgul_rbtree_node_cxx* n,
           void* data)
{
    printf("key=%s  value=%d\n", (char*)n->get_key(), *(int*)n->get_value());
    return 1;
}

/**
 * This function calls <code>free()</code> on the value stored in the
 * node <code>n</code>.
 *
 * @param[in] t tree
 * @param[in] n node
 * @param[in] data client data
 */
static int
free_pair(cgul_rbtree_cxx* t,
          cgul_rbtree_node_cxx* n,
          void* data)
{
    free((void*)n->get_key());
    free(n->get_value());
    return 1;
}

/**
 * Initialize the rbtree <code>t</code> by loading key/value pairs.
 *
 * @param[in] t tree
 */
static void
init_rbtree(cgul_rbtree_cxx* t)
{
    int i = 0;
    char* key = NULL;
    int* value = NULL;
    cgul_rbtree_node_cxx* n = NULL;

    assert(t);

    try {

        for (i = 0 ; i < 5 ; ++i) {

            /* Allocate. */
            key = (char*)calloc(1, sizeof(char) * 2);
            value = (int*)calloc(1, sizeof(int));
            if (!key || !value) {
                throw runtime_error("out of memory");
            }

            /* Assign values. */
            *key = 'a' + i;
            *value = i;

            /* Insert into rbtree. */
            if (!t->insert(key, &n)) {
                throw cgul_exception_cxx("key already exists: %s", key);
            }
            n->set_value(value);

            /* The rbtree now owns the key and value. */
            key = NULL;
            value = NULL;
        }

    } catch (...) {
        /* Clean up. */
        if (key) {
            free(key);
        }
        if (value) {
            free(value);
        }
        /* Rethrow. */
        throw;
    }
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_rbtree_cxx* t = NULL;
    cgul_rbtree_node_cxx* first = NULL;
    cgul_rbtree_node_cxx* last = NULL;

    try {

        /* Allocate. */
        t = new cgul_rbtree_cxx((cgul_rbtree_cxx::compare_t)strcmp);

        /* Initialize the rbtree. */
        init_rbtree(t);

        /* Print the rbtree. */
        t->traverse(print_pair, NULL);

        cout << "=====" << endl;

        /* Print a range of the rbtree. */
        first = t->get_front();
        if (first) {
            first = first->get_next();
        }
        last = t->get_back();
        if (last) {
            last = last->get_prev();
        }
        t->traverse_range(first, last, print_pair, NULL);

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

    /* Clean up. */
    if (t) {
        t->traverse(free_pair, NULL);
        delete t;
    }

    return rv;
}
