#include "cgul.h"

/**
 * Append key to the <code>keys</code> list and <code>value</code> to
 * the <code>values</code> list.  The key/value pairs that will be
 * inserted into the trie are the result of interleaving the two
 * lists. If an error occurs, an exception is thrown, and the
 * relationship between <code>keys</code> and <code>values</code> is
 * undefined.
 *
 * @param[in,out] cex c-style exception
 * @param[in,out] keys keys list
 * @param[in,out] values values list
 * @param[in] key key to append to keys
 * @param[in] value value to append to values
 */
static void
main__append_kvpair(cgul_exception_t* cex,
                    cgul_vector_t keys,
                    cgul_vector_t values,
                    const char* key,
                    int value)
{
    char* key_ptr = NULL;
    int* value_ptr = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(keys);
    assert(values);

    /* Sanity check. */
    if (!key) {
        goto out;
    }

    /* Duplicate the string. */
    key_ptr = cgul_libc__strdup(cex, key);
    if (*cex) {
        goto out;
    }

    /* Duplicate the value. */
    value_ptr = (int*)cgul_libc__calloc(cex, 1, sizeof(*value_ptr));
    if (*cex) {
        goto out;
    }
    *value_ptr = value;

    /* Add the key/value pair. */
    cgul_vector__push_back(cex, keys, key_ptr);
    if (*cex) {
        goto out;
    }

    /* The "keys" vector now owns "key_ptr". */
    key_ptr = NULL;
    
    cgul_vector__push_back(cex, values, value_ptr);
    if (*cex) {
        goto out;
    }

    /* The "values" vector now owns "value_ptr". */
    value_ptr = NULL;

 out:

    /* Clean up. */
    if (value_ptr) {
        free(value_ptr);
    }
    if (key_ptr) {
        free(key_ptr);
    }

    return;
}

/**
 * Print all the key/value pairs in the trie <code>trie</code> in
 * chronological (i.e., insertion) order.
 *
 * @param[in] cex c-style exception
 * @param]in] trie <code>cgul_trie</code> instance
 */
static void
main__print_chrono_kvpairs(cgul_exception_t* cex,
                           cgul_trie_t trie)
{
    const char* key = NULL;
    int* value = NULL;
    cgul_trie_node_t curr = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!trie) {
        goto out;
    }

    /* Iterate over all the key/value pairs printing each. */
    curr = cgul_trie__get_oldest(cex, trie);
    for ( ; curr ; curr = cgul_trie_node__get_younger(cex, curr)) {
        key = cgul_trie_node__get_key(cex, curr);
        value = (int*)cgul_trie_node__get_value(cex, curr);
        if (value) {
            printf("%s -> %d\n", key, *value);
        } else {
            printf("%s -> %p\n", key, value);
        }
    }

 out:
    return;
}

/**
 * Print all the key/value pairs in the trie <code>trie</code> in sort
 * order.
 *
 * @param[in] cex c-style exception
 * @param]in] trie <code>cgul_trie</code> instance
 */
static void
main__print_sorted_kvpairs(cgul_exception_t* cex,
                           cgul_trie_t trie)
{
    const char* key = NULL;
    int* value = NULL;
    cgul_trie_node_t curr = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!trie) {
        goto out;
    }

    /* Iterate over all the key/value pairs printing each. */
    curr = cgul_trie__get_front(cex, trie);
    for ( ; curr ; curr = cgul_trie_node__get_next(cex, curr)) {
        key = cgul_trie_node__get_key(cex, curr);
        value = (int*)cgul_trie_node__get_value(cex, curr);
        if (value) {
            printf("%s -> %d\n", key, *value);
        } else {
            printf("%s -> %p\n", key, value);
        }
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t i = 0;
    size_t i_max = 0;
    cgul_trie_t trie = NULL;
    cgul_vector_t keys = NULL;
    cgul_vector_t values = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    keys = cgul_vector__new(&local);
    values = cgul_vector__new(&local);
    if (local) {
        goto out;
    }

    /* Append key/value pairs. */
    main__append_kvpair(&local, keys, values, "zero",  0);
    main__append_kvpair(&local, keys, values, "one",   1);
    main__append_kvpair(&local, keys, values, "two",   2);
    main__append_kvpair(&local, keys, values, "three", 3);
    main__append_kvpair(&local, keys, values, "four",  4);
    main__append_kvpair(&local, keys, values, "five",  5);
    main__append_kvpair(&local, keys, values, "six",   6);
    main__append_kvpair(&local, keys, values, "seven", 7);
    main__append_kvpair(&local, keys, values, "eight", 8);
    main__append_kvpair(&local, keys, values, "nine",  9);
    if (local) {
        goto out;
    }

    /* Print the key/value pairs in chronological (i.e., insertion) order. */
    printf("===== Original Chronological Order\n");
    i_max = cgul_vector__get_size(&local, keys);
    for (i = 0 ; i < i_max ; ++i) {
        printf("%s -> %d\n",
               (char*)cgul_vector__at(&local, keys, i),
               *(int*)cgul_vector__at(&local, values, i));
    }

    /* Allocate a new trie initialized from the "keys" and "values" vectors. */
    trie = cgul_trie__new(&local);
    cgul_trie__insert_from_arrays(
        &local,
        trie,
        (const char**)cgul_vector__get_array(&local, keys),
        cgul_vector__get_array(&local, values),
        cgul_vector__get_size(&local, keys),
        0 /* maintain_insertion_order */);
    if (local) {
        goto out;
    }

    /* Print the key/value pairs in chronological (i.e., insertion) order. */
    printf("===== New Chronological Order\n");
    main__print_chrono_kvpairs(&local, trie);
    if (local) {
        goto out;
    }

    /* Print the key/value pairs in sort order. */
    printf("===== Sort Order (by key)\n");
    main__print_sorted_kvpairs(&local, trie);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (values) {
        cgul_vector__delete(values);
    }
    if (keys) {
        cgul_vector__delete(keys);
    }
    if (trie) {
        cgul_trie__free_values(trie);
        cgul_trie__free_keys(trie);
        cgul_trie__delete(trie);
    }

    return rv;
}
