#include "cgul.h"

/**
 * Print all the keys in the trie <code>trie</code> that start with
 * the prefix <code>prefix</code>.  If an error occurs, an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_keys_by_prefix(cgul_exception_t* cex,
                           cgul_trie_t trie,
                           const char* prefix)
{
    unsigned long int i = 0;
    const char** keys = NULL;
    unsigned long int keys_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over all the keys that match the prefix. */
    if (cgul_trie__find_keys_by_prefix(cex, trie, prefix, &keys, &keys_size)) {
        for (i = 0 ; i < keys_size ; ++i) {
            printf("%s -> %s\n", prefix, keys[i]);
        }
    }

 out:

    /* Clean up. */
    if (keys) {
        free(keys);
    }

}

/**
 * Print all the values in the trie <code>trie</code> associated with
 * keys that start with the prefix <code>prefix</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_values_by_prefix(cgul_exception_t* cex,
                             cgul_trie_t trie,
                             const char* prefix)
{
    unsigned long int i = 0;
    void** values = NULL;
    unsigned long int values_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over all the values that match the prefix. */
    if (cgul_trie__find_values_by_prefix(
            cex, trie, prefix, &values, &values_size))
    {
        for (i = 0 ; i < values_size ; ++i) {
            printf("%s -> %p\n", prefix, values[i]);
        }
    }

 out:

    /* Clean up. */
    if (values) {
        free(values);
    }

}

/**
 * Print all the nodes in the trie <code>trie</code> associated with
 * keys that start with the prefix <code>prefix</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_nodes_by_prefix(cgul_exception_t* cex,
                            cgul_trie_t trie,
                            const char* prefix)
{
    unsigned long int i = 0;
    const char* key = NULL;
    void* value = NULL;
    cgul_trie_node_t* nodes = NULL;
    unsigned long int nodes_size = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over all the nodes that match the prefix. */
    if (cgul_trie__find_nodes_by_prefix(
            cex, trie, prefix, &nodes, &nodes_size))
    {
        for (i = 0 ; i < nodes_size ; ++i) {
            key = cgul_trie_node__get_key(cex, nodes[i]);
            value = cgul_trie_node__get_value(cex, nodes[i]);
            printf("%s -> (%s, %p)\n", prefix, key, value);
        }
    }

 out:

    /* Clean up. */
    if (nodes) {
        free(nodes);
    }

}

/**
 * Print the keys in the trie <code>trie</code> that is the longest
 * prefix of <code>prefix</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 */
static void
main__print_longest_prefix_of(cgul_exception_t* cex,
                              cgul_trie_t trie,
                              const char* prefix)
{
    const char* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    key = cgul_trie__find_longest_prefix_of(cex, trie, prefix);
    cgul_libc__printf(cex, "%s -> %s\n", prefix, key);

 out:
    return;
}

static int
main__fold_keys_cb(cgul_exception_t* cex,
                   const char* key,
                   void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(key);

    printf("%s\n", key);

 out:
    /* Continue traversing. */
    return 1;
}

static int
main__fold_values_cb(cgul_exception_t* cex,
                     void* value,
                     void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    printf("%p\n", value);

 out:
    /* Continue traversing. */
    return 1;
}

static int
main__fold_pairs_cb(cgul_exception_t* cex,
                    const char* key,
                    void* value,
                    void* data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(key);

    printf("%s -> %p\n", key, value);

 out:
    /* Continue traversing. */
    return 1;
}

static int
main__traverse_cb(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  cgul_trie_node_t n,
                  void* data)
{
    const char* key = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(trie);

    key = cgul_trie_node__get_key(cex, n);
    printf("%s\n", key);

 out:
    /* Continue traversing. */
    return 1;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    size_t i = 0;
    cgul_trie_t trie = NULL;
    cgul_trie_node_t node = NULL;
    cgul_exception_t local = NULL;
    const char* ss[] = {
        "",
        "Hello",
        "Hello, World!",
        "Goodbye",
        "Goodbye, Cruel World!"
    };
    size_t ss_size = sizeof(ss) / sizeof(ss[0]);

    /* Allocate. */
    trie = cgul_trie__new_with_indexing(&local);
    if (local) {
        goto out;
    }


    /* Test if the trie is empty.  Also, make sure the trie is empty
     * and front, back, oldest, and youngest are all NULL. */
    printf("===== Is Empty? (before)\n");
    printf("    is empty: %d\n", cgul_trie__is_empty(&local, trie));
    printf("       front: %p\n", cgul_trie__get_front(&local, trie));
    printf("        back: %p\n", cgul_trie__get_back(&local, trie));
    printf("      oldest: %p\n", cgul_trie__get_oldest(&local, trie));
    printf("    youngest: %p\n", cgul_trie__get_youngest(&local, trie));

    /* Insert keys into the trie. */
    for (i = 0 ; i < ss_size ; ++i) {
        char* tmp = cgul_libc__strdup(&local, ss[i]);
        int insert_rv = cgul_trie__insert(&local, trie, tmp, NULL);
        if (local) {
            goto out;
        }
        if (!insert_rv) {
            free(tmp);
        }
    }

    printf("===== Is Empty? (after)\n");
    printf("    is empty: %d\n", cgul_trie__is_empty(&local, trie));
    printf("       front: %p\n", cgul_trie__get_front(&local, trie));
    printf("        back: %p\n", cgul_trie__get_back(&local, trie));
    printf("      oldest: %p\n", cgul_trie__get_oldest(&local, trie));
    printf("    youngest: %p\n", cgul_trie__get_youngest(&local, trie));

    /* Iterate over the keys in the trie in chronological order. */
    printf("===== Chronological Order\n");
    node = cgul_trie__get_oldest(&local, trie);
    for ( ; node ; node = cgul_trie_node__get_younger(&local, node)) {
        printf("-->%s<--\n", cgul_trie_node__get_key(&local, node));
    }

    /* Iterate over the keys in the trie in reverse chronological order. */
    printf("===== Reverse Chronological Order\n");
    node = cgul_trie__get_youngest(&local, trie);
    for ( ; node ; node = cgul_trie_node__get_older(&local, node)) {
        printf("-->%s<--\n", cgul_trie_node__get_key(&local, node));
    }

    /* Iterate over the keys in the trie in sort order. */
    printf("===== Sort Order\n");
    node = cgul_trie__get_front(&local, trie);
    for ( ; node ; node = cgul_trie_node__get_next(&local, node)) {
        printf("-->%s<--\n", cgul_trie_node__get_key(&local, node));
    }

    /* Iterate over the keys in the trie in reverse sort order. */
    printf("===== Reverse Sort Order\n");
    node = cgul_trie__get_back(&local, trie);
    for ( ; node ; node = cgul_trie_node__get_prev(&local, node)) {
        printf("-->%s<--\n", cgul_trie_node__get_key(&local, node));
    }

    /* Find or don't find the nodes for individual keys. */
    printf("===== cgul_trie__find() \n");
    for (i = 0 ; i < ss_size ; ++i) {
        node = cgul_trie__find(&local, trie, ss[i]);
        if (node) {
            printf("found: -->%s<--\n", cgul_trie_node__get_key(&local, node));
        }
    }
    node = cgul_trie__find(&local, trie, "not in trie");
    if (node) {
        local = cgul_exception__new("Unexpectedly found key that "
                                    "should not be in trie");
        goto out;
    }

    /* Find keys by prefix. */
    printf("===== cgul_trie__find_keys_by_prefix()\n");
    main__print_keys_by_prefix(&local, trie, "");
    main__print_keys_by_prefix(&local, trie, "Hello");
    main__print_keys_by_prefix(&local, trie, "Good");
    main__print_keys_by_prefix(&local, trie, "foo");

    /* Find values by prefix. */
    printf("===== cgul_trie__find_values_by_prefix()\n");
    main__print_values_by_prefix(&local, trie, "");
    main__print_values_by_prefix(&local, trie, "Hello");
    main__print_values_by_prefix(&local, trie, "Good");
    main__print_values_by_prefix(&local, trie, "foo");

    /* Find nodes by prefix. */
    printf("===== cgul_trie__find_nodes_by_prefix()\n");
    main__print_nodes_by_prefix(&local, trie, "");
    main__print_nodes_by_prefix(&local, trie, "Hello");
    main__print_nodes_by_prefix(&local, trie, "Good");
    main__print_nodes_by_prefix(&local, trie, "foo");

    /* Find key that is the longest prefix of another string. */
    printf("===== cgul_trie__find_longest_prefix_of()\n");
    main__print_longest_prefix_of(&local, trie, "foo");
    main__print_longest_prefix_of(&local, trie, "Hello, Old Chap!");

    /* Left fold over the keys. */
    printf("===== cgul_trie__foldl_keys()\n");
    cgul_trie__foldl_keys(&local, trie, main__fold_keys_cb, NULL);

    /* Right fold over the keys. */
    printf("===== cgul_trie__foldr_keys()\n");
    cgul_trie__foldr_keys(&local, trie, main__fold_keys_cb, NULL);

    /* Left fold over the values. */
    printf("===== cgul_trie__foldl_values()\n");
    cgul_trie__foldl_values(&local, trie, main__fold_values_cb, NULL);

    /* Right fold over the values. */
    printf("===== cgul_trie__foldr_values()\n");
    cgul_trie__foldr_values(&local, trie, main__fold_values_cb, NULL);

    /* Left fold over the pairs. */
    printf("===== cgul_trie__foldl_pairs()\n");
    cgul_trie__foldl_pairs(&local, trie, main__fold_pairs_cb, NULL);

    /* Right fold over the pairs. */
    printf("===== cgul_trie__foldr_pairs()\n");
    cgul_trie__foldr_pairs(&local, trie, main__fold_pairs_cb, NULL);

    /* Traverse the nodes. */
    printf("===== cgul_trie__traverse()\n");
    cgul_trie__traverse(&local, trie, main__traverse_cb, NULL);

    /* Clear the trie.  (Note, you can pass free() in directly to
     * cgul_trie__clear().) */
    printf("===== cgul_trie__clear()\n");
    cgul_trie__clear(&local, trie, cgul_cache__get_freer(), NULL);

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        /* The keys should have been freed above. */
        cgul_trie__delete(trie);
    }

    return rv;
}
